/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.basis;

/**
 * This is a base class of an entity, which is a collection of data that can be processed by some
 * {@link Engine} engine(s). The interface allows getting information about the entity type, which
 * is useful for checking ability of an engine to process a given entity.
 * 
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public abstract class Entity {

  /** The entity type. */
  private final EntityType type;

  /**
   * Constructs an entity with a given type.
   *
   * @param type the entity type.
   */
  public Entity(final EntityType type) {
    this.type = type;
  }

  /**
   * Returns the type of the entity (a unique identifier).
   * 
   * @return {@link EntityType} the type of the entity.
   */
  public final EntityType getEntityType() {
    return type;
  }

  /**
   * Adds a given entity to this one.
   * 
   * @param entity the entity to be added.
   */
  public abstract void add(final Entity entity);

  @Override
  public String toString() {
    return String.format("<%s>", type.toString());
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + ((type == null) ? 0 : type.hashCode());
    return result;
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj == null) {
      return false;
    }
    if (getClass() != obj.getClass()) {
      return false;
    }

    final Entity other = (Entity) obj;

    if (type == null) {
      if (other.type != null) {
        return false;
      }
    } else if (!type.equals(other.type)) {
      return false;
    }

    return true;
  }
}
