/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.basis;

import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.model.basis.EventDrivenModule;
import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.basis.State;
import ru.ispras.retrascope.result.test.Transaction;

import java.util.Collection;
import java.util.Map;

/**
 * The abstract representation of a device model encapsulating a runtime state along with a static
 * structure of an original device.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public abstract class Simulator extends Engine implements EventDrivenModule {
  /**
   * Creates a simulator with the specified ID and I/O entity types.
   *
   * @param id the ID
   * @param inputType the type of an input entity
   * @param outputType the type of an output entity
   */
  public Simulator(String id, EntityType inputType, EntityType outputType) {
    super(id, inputType, outputType);
  }

  /**
   * Sets the specified input of this simulator to the specified value.
   * 
   * @param name a name of the modified input
   * @param value the value to be set
   */
  public abstract void setInputValue(String name, NodeValue value);

  /**
   * Sets the specified inputs of this simulator to the specified values.
   *
   * @param transactions transactions containing the inputs and the corresponding values
   */
  public abstract void setInputValues(Collection<Transaction> transactions);

  /**
   * Returns a value of the specified inner variable of this simulator.
   *
   * @param name a name of the inner variable
   *
   * @return a value of the inner variable
   *
   * @throws IllegalArgumentException if the specified variable is not found among the simulator
   *         inner variables
   */
  public abstract NodeValue getInnerValue(String name);

  /**
   * Returns values of the specified inner variables of this simulator.
   *
   * @param names a collection containing names of the inner variables
   *
   * @return a map containing mappings between names and values of the specified inner variables
   *
   * @throws IllegalArgumentException if at least on of the specified variables is not found among
   *         the simulator inner variables
   */
  public abstract Map<String, NodeValue> getInnerValues(Collection<String> names);

  /**
   * Returns values of all inner variables of this simulator.
   *
   * @return a map containing mappings between names and values of the inner variables
   */
  public abstract Map<String, NodeValue> getInnerValues();

  /**
   * Returns a value of the specified output variable of this simulator.
   *
   * @param name a name of the output
   *
   * @return a value of the output
   *
   * @throws IllegalArgumentException if the specified variable is not found among the simulator
   *         outputs
   */
  public abstract NodeValue getOutputValue(String name);

  /**
   * Returns values of the specified output variables of this simulator.
   *
   * @param names a collection containing names of the outputs
   *
   * @return a map containing mappings between names and values of the specified outputs
   *
   * @throws IllegalArgumentException if at least on of the specified variables is not found among
   *         the simulator outputs
   */
  public abstract Map<String, NodeValue> getOutputValues(Collection<String> names);

  /**
   * Returns values of all output variables of this simulator.
   *
   * @return an output vector containing the outputs
   */
  public abstract Map<String, NodeValue> getOutputValues();

  /**
   * Returns a current state of this simulator.
   *
   * @return a current state of this simulator
   */
  public abstract State getCurrentState();

  /**
   * Returns a static device model related to this simulator.
   *
   * @return a current state of this simulator
   */
  public abstract Model getStaticModel();

  /**
   * Returns a representation of a coverage reached during simulation.
   *
   * @return a test coverage
   */
  public abstract TestCoverage getCoverage();
}
