/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgVisitor;
import ru.ispras.retrascope.model.cfg.CfgWalker;
import ru.ispras.retrascope.model.cfg.CfgWalkerMode;

import java.util.Map;

/**
 * Abstract engine for control flow graph model elaboration.
 *
 * <p>Runs internal walkers on the control flow graph model.</p>
 *
 * <p>Note that while executed, walkers will be used
 * in the same order they appeared in constructor.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class CfgEngine extends Engine {

  private final CfgWalker[] walkers;

  /**
   * Constructs object with the specified identifier, input/output types
   * and visitors.
   *
   * <p>All the defined visitors will be run on the walker with default
   * {@link ru.ispras.retrascope.model.cfg.CfgWalkerMode} walking mode.</p>
   *
   * @param engineId engine identifier
   * @param inputType engine input type
   * @param outputType engine output type
   * @param visitors control flow graph visitors
   *
   * @throws IllegalArgumentException when the specified visitors array is {@code null }.
   */
  public CfgEngine(
      final String engineId,
      final EntityType inputType,
      final EntityType outputType,
      final CfgVisitor ... visitors) {

    super(engineId);

    InvariantChecks.checkNotNull(visitors);

    this.addInputType(inputType);
    this.setOutputType(outputType);
    this.walkers = new CfgWalker[visitors.length];
    for (int i = 0; i < visitors.length; i++) {
      this.walkers[i] = new CfgWalker(visitors[i], CfgWalkerMode.getDefaultMode());
    }
  }

  /**
   * Constructs object with the specified identifier, input/output types,
   * walkers.
   *
   * @param engineId engine identifier
   * @param inputType engine input type
   * @param outputType engine output type
   * @param walkers model walkers
   *
   * @throws IllegalArgumentException when walkers array is {@code null}.
   *
   */
  public CfgEngine(
      final String engineId,
      final EntityType inputType,
      final EntityType outputType,
      final CfgWalker... walkers) {

    super(engineId);

    InvariantChecks.checkNotNull(walkers);

    this.addInputType(inputType);
    this.setOutputType(outputType);

    this.walkers = walkers;
  }

  /**
   * Initializes engine with the specified inputs.
   *
   * @param inputs Engine inputs
   */
  public abstract void initialize(final Map<EntityType, Entity> inputs);

  @Override
  public Entity start(final Map<EntityType, Entity> inputs) {

    initialize(inputs);

    final CfgModel cfgModel = (CfgModel) inputs.get(CfgModel.TYPE);

    for (final CfgWalker walker : this.walkers) {
      walker.start(cfgModel);
    }
    return getOutput();
  }

  /**
   * Returns engine output entity.
   *
   * @return Output entity
   */
  public abstract Entity getOutput();

  /**
   * Returns internal walkers.
   *
   * @return walkers
   */
  public final CfgWalker[] getWalkers() {
    return this.walkers;
  }

  /**
   * Returns initial visitor.
   *
   * @return visitor
   */
  public final CfgVisitor getInitVisitor() {
    return this.walkers[0].getVisitor();
  }
}
