/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Backend;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgVisitor;
import ru.ispras.retrascope.model.cfg.CfgWalker;
import ru.ispras.retrascope.model.cfg.CfgWalkerMode;
import ru.ispras.retrascope.model.cgaa.CgaaModel;

/**
 * Base class for control flow graph model visitor-based back-ends.
 *
 * <p>Base class for {@link Backend} back-ends that are
 * based on {@link CfgVisitor} control flow graph model visitor.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class CfgVisitorBackend extends Backend {

  private final CfgVisitor visitor;
  private final CfgWalkerMode walkerMode;

  /**
   * Creates object with the specified model visitor and walking mode.
   *
   * @param visitor model visitor
   * @param mode model walking mode
   */
  public CfgVisitorBackend(
      final CfgVisitor visitor,
      final CfgWalkerMode mode) {

    this.visitor = visitor == null ? new CfgDefaultVisitor() : visitor;
    this.walkerMode = mode == null ? CfgWalkerMode.getDefaultMode() : mode;
  }

  /**
   * Elaborates the specified entity.
   *
   * @param entity entity to be elaborated
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when argument is {@code null};
   *         (2) when the specified entity is neither
   *         of {@link ru.ispras.retrascope.model.cfg.CfgModel} control flow graph model
   *         nor of {@link ru.ispras.retrascope.model.cgaa.CgaaModel}
   *         clocked guarded atomic actions model kinds.
   */
  @Override
  public void start(final Entity entity) {
    InvariantChecks.checkNotNull(entity);

    CfgModel model;

    if (entity instanceof CfgModel) {
      model = (CfgModel) entity;
    } else if (entity instanceof CgaaModel) {
      model = ((CgaaModel) entity).getModel();
    } else {
      throw new IllegalArgumentException("Incorrect entity type: " + entity.getEntityType() + ".");
    }

    final CfgWalker walker = new CfgWalker(this.visitor, this.walkerMode);
    walker.start(model);
  }
}
