/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.cfg.printer.graphml;

import org.apache.commons.cli.ParseException;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.FileNames;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.engine.cfg.CfgEngine;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgWalker;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.Map;

/**
 * Control flow graph printer to GraphML format.
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */

public class CfgGraphMlPrinter extends CfgEngine {

  /**
   * The engine identifier.
   */
  public static final String ENGINE_ID = "cfg-graphml-printer";

  /**
   * The engine output entity type.
   */
  public static final EntityType OUT_TYPE = EntityType.get("cfg-graphml");

  /**
   * The command line parameters.
   */
  public static final Parameter OUT_FILE_NAME =
      new Parameter("cfg-graphml", true, "Output GraphML file name");

  // The default file name.
  private static final String DEFAULT_FILE_NAME = "cfg-model.graphml";

  // The printer result.
  private FileNames fileNames;

  /**
   * Default constructor.
   */
  public CfgGraphMlPrinter() {
    super(
        CfgGraphMlPrinter.ENGINE_ID,
        CfgModel.TYPE,
        CfgGraphMlPrinter.OUT_TYPE,
        new CfgGraphMlVisitor());
    addParameter(OUT_FILE_NAME);
  }

  @Override
  public void initialize(final Map<EntityType, Entity> inputs) {

    final Arguments arguments = (Arguments)inputs.get(Arguments.TYPE);

    try {
      // Get arguments from command line.
      final Arguments engineArgs = parseCommandLine(arguments.getCommandLine());

      // Create GraphML file.
      fileNames = new FileNames("graphml");
      fileNames.add(engineArgs.getValue(OUT_FILE_NAME, DEFAULT_FILE_NAME));

      initGraphMlVisitor(fileNames.getList().get(0));

    } catch (ParseException e) {
      Log.print(LogLevel.ERROR, e.toString());
    }
  }

  /**
   * Initialize visitor with the graphML file name.
   *
   * @param visitorName the output file name.
   */
  public void initGraphMlVisitor(final String visitorName) {
    ((CfgGraphMlVisitor) getInitVisitor()).initialize(visitorName);
  }

  @Override
  public Entity getOutput() {
    return fileNames;
  }

  /**
   * Print the control flow graph to GraphML file.
   *
   * @param fileName the output file name.
   * @param cfg control flow graph to print.
   */
  public void print(final String fileName, final CfgModel cfg) {

    // Create an output file for the GraphML printer.
    final CfgGraphMlVisitor cfgGraphMlFile = new CfgGraphMlVisitor();
    cfgGraphMlFile.initialize(fileName);

    // Construct a CFG walker for the GraphML printer.
    final CfgWalker cfgWalker = new CfgWalker(cfgGraphMlFile);

    // Run the CFG walker.
    cfgWalker.start(cfg);
  }
}
