/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.cfg.printer.graphml;

import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.FileCreator;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.CfgNodeType;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.cfg.Wait;
import ru.ispras.retrascope.result.graphml.GraphMlUtils;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.nio.charset.Charset;
import java.util.Collection;

/**
 * Control flow graph visitor for GraphML printer.
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */

public class CfgGraphMlVisitor extends CfgDefaultVisitor {

  /* The print-to-GraphMl-file writer. */
  private PrintWriter printWriter;

  /* The expression printer. */
  private ExprTreePrinter exprPrinter;

  private static final Charset DEFAULT_CHARSET = Charset.defaultCharset();

  /**
   * Default constructor.
   */
  public CfgGraphMlVisitor() {
    super();
  }

  /**
   * Initializes object with the specified file name.
   *
   * @param fileName GraphML file name
   */
  public void initialize(final String fileName) {

    try {

      /* Create a GraphML file. */
      final File graphMlFile = FileCreator.newFile(fileName);
      printWriter = createPrintWriter(graphMlFile.getAbsoluteFile());

    } catch (IOException e) {
      throw new RetrascopeRuntimeException(e);
    }

    /* The expression printer initialization. */
    this.exprPrinter = Retrascope.getConfiguration().getStylePrinter();
  }

  private PrintWriter createPrintWriter(final File file) throws FileNotFoundException {
    return new PrintWriter(
        new BufferedWriter(
            new OutputStreamWriter(
                new FileOutputStream(file),
                DEFAULT_CHARSET)));
  }

  @Override
  public Status getStatus() {
    return Status.OK;
  }

  @Override
  public void onRootBegin() {
    GraphMlUtils.onGraphMlBegin(printWriter);
  }

  @Override
  public void onRootEnd() {
    GraphMlUtils.onGraphMlEnd(printWriter);
    GraphMlUtils.closeGraphMlFile(printWriter);
  }

  @Override
  public void onCfgModelBegin() {
    GraphMlUtils.onGraphBegin(printWriter);
  }

  @Override
  public void onCfgModelEnd() {
    GraphMlUtils.onGraphEnd(printWriter);
  }

  /**
   * Prints the node, gets the list of parents for this node and prints the edges.
   *
   * @param child the CFG model node
   */
  private void printNodeBegin(final CfgModelNode child) {
    printNodeBegin(child, GraphMlUtils.FORM_RECTANGLE, GraphMlUtils.COLOR_PURPLE);
  }

  /**
   * Prints the node, gets the list of parents for this node and prints the edges.
   *
   * @param child the CFG model node
   * @param form the node form
   * @param color the node color
   */
  private void printNodeBegin(
      final CfgModelNode child,
      final String form,
      final String color) {

    /* Add the node to the GraphML file. */
    GraphMlUtils.printNode(printWriter, child.getId(), child.getDescription(this.exprPrinter),
        form, color);

    /* Add the edges to the GraphML file. */
    if (child.hasChildren()) {
      final Collection<? extends CfgModelNode> children = child.getChildren();
      for (final CfgModelNode nodeChild : children) {

        /* Case nodes will not be added to GraphML. */
        if (nodeChild.getType() == CfgNodeType.CASE) {
          if (nodeChild.hasOnlyChild()) {
            final CfgModelNode caseChild = nodeChild.getOnlyChild();
            GraphMlUtils.printEdge(printWriter, child.getId(),
                nodeChild.getDescription(this.exprPrinter), caseChild.getId());
          } else {
            Log.getLogger().log(LogLevel.ERROR,
                String.format("Case node must have only one child. Node id: %s.", child.getId()));
          }
        } else {
          GraphMlUtils.printEdge(printWriter, child.getId(), nodeChild.getId());
        }
      }
    }
  }

  @Override
  public void onModuleBegin(final Module module) {
    printNodeBegin(module, GraphMlUtils.FORM_ELLIPSE, GraphMlUtils.COLOR_YELLOW);
  }

  @Override
  public void onProcessBegin(final Process process) {

    /* Add the node to the GraphML file. */
    GraphMlUtils.printNode(printWriter, process.getId(), process.getDescription(this.exprPrinter),
        GraphMlUtils.FORM_DIAMOND, GraphMlUtils.COLOR_BLUE);

    final Source source = (Source) process.getChildren().toArray()[0];
    if (!process.hasOnlyChild()) {
      Log.getLogger()
          .log(LogLevel.ERROR,
              String.format("Process node must " + "have only one child. Node id: %s.",
                  process.getId()));
    }

    final CfgNode node = (CfgNode) source.getChildren().toArray()[0];
    if (!source.hasOnlyChild()) {
      Log.getLogger()
          .log(
              LogLevel.ERROR,
              String.format("Source node must have " + "only one child. Node id: %s.",
                  process.getId()));
    }

    GraphMlUtils.printEdge(printWriter, process.getId(), node.getId());
  }

  @Override
  public void onBasicBlockBegin(final BasicBlock basicBlock) {
    printNodeBegin(basicBlock);
  }

  @Override
  public void onSwitchBegin(final Switch switchNode) {
    printNodeBegin(switchNode, GraphMlUtils.FORM_DIAMOND, GraphMlUtils.COLOR_RED);
  }

  @Override
  public void onMergeBegin(final Merge merge) {
    printNodeBegin(merge, GraphMlUtils.FORM_ELLIPSE, GraphMlUtils.COLOR_WHITE);
  }

  @Override
  public void onSink(final Sink sink) {
    printNodeBegin(sink, GraphMlUtils.FORM_RECTANGLE, GraphMlUtils.COLOR_YELLOW);
  }

  @Override
  public void onWaitBegin(final Wait wait) {
    printNodeBegin(wait, GraphMlUtils.FORM_OCTAGON, GraphMlUtils.COLOR_ORANGE);
  }
}
