/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.cgaa;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.transformer.ReduceOptions;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.engine.basis.TransformUtils;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.CfgNodeType;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Deque;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Clocked guarded atomic actions decision diagram builder.
 *
 * @see CfgCgaaTransformer
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CfgCgaaTransformerBuilder {

  /* The mapping of module to transformations. */
  private final Map<Module, Map<Process, List<CfgTransform>>> moduleMap;

  /**
   * Constructs object with the specified mapping.
   *
   * @param map case -> basic blocks collection
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public CfgCgaaTransformerBuilder(final Map<Module, Map<Process, List<CfgTransform>>> map) {
    InvariantChecks.checkNotNull(map);

    this.moduleMap = map;
  }

  private Map<Module, Map<Process, List<CfgTransform>>> getModuleMap() {
    return this.moduleMap;
  }

  /**
   * Creates clocked guarded atomic action model.
   *
   * <p>Builds clocked guarded atomic actions decision diagram
   * from {@link CfgModel} control flow graph model representation.</p>
   *
   * @return guarded actions decision diagram.
   */
  public CfgModel build() {
    final CfgModel model = new CfgModel();

    for (final Module module : getModuleMap().keySet()) {
      buildModule(model, module);
    }
    return model;
  }

  private void buildModule(final CfgModel model, final Module module) {

    /* The mapping for "old" case nodes and linked sequences of block nodes. */
    final Map<Case, List<BasicBlock>> terminations = new LinkedHashMap<>();

    /* The mapping for old cases and new cases. */
    final Map<Case, List<Case>> caseMap = new LinkedHashMap<>();

    /* Create new module. */
    final Module newModule = new Module(module.getName());
    newModule.setVariablesMapping(module.getVariablesMapping());
    model.addModule(newModule);

    /* Apply transformations of the current module's process. */
    final Map<Process, List<CfgTransform>> ptList = getModuleMap().get(module);
    applyTransformations(terminations, caseMap, newModule, ptList);
  }

  private void applyTransformations(
      final Map<Case, List<BasicBlock>> terminations,
      final Map<Case, List<Case>> caseMap,
      final Module module,
      final Map<Process, List<CfgTransform>> ptMap) {

    for (final Map.Entry<Process, List<CfgTransform>> entry : ptMap.entrySet()) {

      /* Create new process container. */
      final Process oldProcess = entry.getKey();
      final Process process = new Process(oldProcess.isInitial());
      process.setSensitivityList(oldProcess.getSensitivityList());
      process.setVariablesMapping(oldProcess.getVariablesMapping());
      module.addChild(process);

      /* Create new process control flow graph. */
      final List<CfgNode> cfg = new ArrayList<>();
      final Source src = new Source();
      final Sink sink = new Sink();
      src.addChild(sink);
      cfg.add(src);
      cfg.add(sink);
      process.addChild(src);

      /* Build the control flow graph from the list of transformations. */
      buildCfgControlPart(cfg, terminations, caseMap, entry.getValue());

      /* Continue terminal case nodes with sequences of basic blocks. */
      buildCfgDataPart(cfg, terminations);

      process.addNodes(cfg);
    }
  }

  private void buildCfgDataPart(
      final Collection<CfgNode> nodes,
      final Map<Case, List<BasicBlock>> terminations) {

    for (final Map.Entry<Case, List<BasicBlock>> entry : terminations.entrySet()) {

      final Case newCase = entry.getKey();

      if (!newCase.hasChildren()) {
        final List<BasicBlock> termBlocks = entry.getValue();
        addBlocksAfterCase(nodes, newCase, termBlocks);
      }
    }
  }

  private void addBlocksAfterCase(
      final Collection<CfgNode> nodes,
      final Case caseNode,
      final List<BasicBlock> blocks) {

    final Sink sink = (Sink) CfgUtils.getNode(nodes, CfgNodeType.SINK);

    if (blocks.isEmpty()) {

      linkWithSink(nodes, caseNode, sink);
    } else {

      /* Pointer to the currently unlinked (having no children) node. */
      CfgNode unLinked = null;

      for (int i = 0; i < blocks.size(); i++) {

        final BasicBlock block = blocks.get(i);

        final CfgNode blockCopy = (CfgNode) block.deepCopy();

        nodes.add(blockCopy);

        blockCopy.addParent(i == 0 ? caseNode : unLinked);

        unLinked = blockCopy;

        /* If it is the last block. */
        if (i == blocks.size() - 1) {

          linkWithSink(nodes, unLinked, sink);
        }
      }
    }
  }

  private void linkWithSink(
      final Collection<CfgNode> nodes,
      final CfgNode node,
      final Sink sink) {

    if (sink.hasParents()) {

      final CfgModelNode sinkParent = sink.getOnlyParent();

      if (sinkParent instanceof Merge) {

        node.addChild(sinkParent);
      } else {

        sinkParent.removeChild(sink);

        final Merge merge = linkWithNewMerge(nodes, node, sink);

        sinkParent.addChild(merge);
      }
    } else {

      linkWithNewMerge(nodes, node, sink);
    }
  }

  private Merge linkWithNewMerge(
      final Collection<CfgNode> nodes,
      final CfgNode node,
      final Sink sink) {

    final Merge merge = new Merge();
    nodes.add(merge);
    merge.addChild(sink);
    node.addChild(merge);
    return merge;
  }

  private void buildCfgControlPart(
      final List<CfgNode> nodes,
      final Map<Case, List<BasicBlock>> terminals,
      final Map<Case, List<Case>> caseMap,
      final List<CfgTransform> transforms) {

    for (final CfgTransform transform : transforms) {

      if (transform.containsSwitch()) {

        applySwitchTransform(nodes, caseMap, transform);
      } else {

        applyNonSwitchTransform(nodes, transform);
      }
      if (transform.containsMap()) {
        applyMapTransform(terminals, caseMap, transform);
      }
    }
  }

  private void applyMapTransform(
      final Map<Case, List<BasicBlock>> terminals,
      final Map<Case, List<Case>> caseMap,
      final CfgTransform transform) {

    final Map<Case, List<BasicBlock>> transformMap = transform.getMap();

    for (final Map.Entry<Case, List<BasicBlock>> entry : transformMap.entrySet()) {

      final List<Case> newCases = caseMap.get(entry.getKey());

      for (final Case newCase : newCases) {
        if (!newCase.hasChildren() && !terminals.containsKey(newCase)) {
          terminals.put(newCase, entry.getValue());
        }
      }
    }
  }

  private void applyNonSwitchTransform(
      final List<CfgNode> list,
      final CfgTransform transform) {

    final CfgNode src = CfgUtils.getNode(list, CfgNodeType.SOURCE);
    final List<BasicBlock> blocks = transform.getBlocks();

    /* The basic block-only control flow graph - no transformation needed. */

    /* Remove all the connections with source's children. */
    final Collection<CfgModelNode> srcChildren = src.getChildren();
    for (final CfgModelNode child : srcChildren) {
      src.removeChild(child);
    }

    if (blocks.isEmpty()) {
      throw new IllegalStateException("No nodes has been found.");
    }

    CfgNode current = src;

    for (final BasicBlock block : blocks) {
      final CfgNode newBlock = new BasicBlock(block.getAssignments(), block.isConcurrent());
      list.add(newBlock);
      newBlock.addParent(current);
      current = newBlock;
    }
    current.addChild(CfgUtils.getNode(list, CfgNodeType.SINK));
  }

  private void applySwitchTransform(
      final Collection<CfgNode> list,
      final Map<Case, List<Case>> caseMap,
      final CfgTransform transform) {

    final Source src = (Source) CfgUtils.getNode(list, CfgNodeType.SOURCE);

    final Switch oldSwitch = transform.getSwitchNode();
    final List<BasicBlock> blocks = transform.getBlocks();
    final Deque<Case> cases = transform.getCases();

    /* Create new switch with the modified expression. */
    final Event event = oldSwitch.getEvent();
    final Node condition = oldSwitch.getCondition();
    final Switch newSwitch = (event == null)
        ? new Switch(TransformUtils.substituteBackward(blocks, condition))
        : new Switch(oldSwitch.getEvent());
    list.add(newSwitch);

    /* Add all the cases of old switch to terminals and create new case nodes. */
    final Collection<CfgModelNode> children = oldSwitch.getChildren();
    for (final CfgModelNode oldCase : children) {

      /* Create a new child case node & add it to the control flow graph. */
      final Case newCase = new Case(makeValue(oldSwitch, (Case)oldCase));
      list.add(newCase);
      newSwitch.addChild(newCase);

      /* Add new case nodes to old-new mapping. */
      if (caseMap.containsKey(oldCase)) {
        caseMap.get(oldCase).add(newCase);
      } else {
        caseMap.put((Case) oldCase, new ArrayList<>(Collections.singleton(newCase)));
      }
    }

    /* Determine the parent node for the node switch. */
    if (cases.isEmpty()) {

      /* Remove source's children. */
      final Collection<CfgModelNode> srcChildren = src.getChildren();
      for (final CfgModelNode srcChild : srcChildren) {
        srcChild.removeParent(src);
      }

      /* Add new switch as child for source. */
      newSwitch.addParent(src);
    } else {

      /* The new incarnation of old case is already created. */
      final Case tCasesPeek = cases.peek();

      /* Check rather mapping contains old case. */
      if (!caseMap.containsKey(tCasesPeek)) {
        throw new IllegalStateException("Map does not contain item.");
      }

      final List<Case> terminalCases = caseMap.get(tCasesPeek);

      /* Get cases which does not have children. */
      final List<Case> noChildrenCases = new ArrayList<>();
      for (final Case termCase : terminalCases) {
        if (!termCase.hasChildren()) {
          noChildrenCases.add(termCase);
        }
      }

      /* No-child-having cases should be the parents of the new switch node. */
      if (noChildrenCases.size() == 1) {
        newSwitch.addParent(noChildrenCases.get(0));
      } else if (noChildrenCases.size() > 1) {
        final Merge merge = new Merge();
        list.add(merge);
        newSwitch.addParent(merge);
        for (final Case noChildren : noChildrenCases) {
          noChildren.addChild(merge);
        }
      }
    }
  }

  private Collection<NodeValue> makeValue(
      final Switch switchNode,
      final Case caseNode) {

    Node value = null;

    if (caseNode.isDefault()) {
      final Collection<CfgModelNode> children = switchNode.getChildren();

      for (final CfgModelNode child : children) {

        if (!((Case)child).isDefault()) {
          if (value == null) {
            final Collection<NodeValue> childValues = ((Case) child).getValues();
            for (final NodeValue childValue : childValues) {
              if (value == null) {
                value = ExprUtils.getNegation(childValue);
              } else {
                value = ExprUtils.getConjunction(value, ExprUtils.getNegation(childValue));
              }
            }
          } else {
            final Collection<NodeValue> childValues = ((Case) child).getValues();
            for (final NodeValue childValue : childValues) {
              value = ExprUtils.getConjunction(value, ExprUtils.getNegation(childValue));
            }
          }
        }
      }
    } else {
      return caseNode.getValues();
    }

    if (value == null) {
      throw new IllegalStateException(
          String.format(
              "Can't create %s==%s expression.",
              switchNode.getCondition().toString(),
              caseNode.getValues().toString()));
    }

    final Node reduced = Transformer.reduce(ReduceOptions.NEW_INSTANCE, value);

    if (!(reduced instanceof NodeValue)) {
      throw new IllegalStateException("Wrong type: " + reduced.getKind() + ".");
    }

    final Set<NodeValue> set = new LinkedHashSet<>();
    set.add((NodeValue) reduced);
    return set;
  }
}
