/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.cgaa;

import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Deque;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * Visitor component of transformer to clocked guarded atomic actions model.
 *
 * @see CfgCgaaTransformer
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CfgCgaaTransformerVisitor extends CfgDefaultVisitor {

  /* The current module. */
  private final Deque<Module> modules;

  /* The current basic blocks collection. */
  private final List<BasicBlock> blocks;

  /* The current cases stack. */
  private final Deque<Case> cases;

  /* The current process transformation. */
  private List<CfgTransform> transformations;

  /* The mapping of module data to transformations. */
  private final Map<Module, Map<Process, List<CfgTransform>>> moduleMap;

  /**
   * Default constructor.
   */
  public CfgCgaaTransformerVisitor() {
    super();
    this.modules = new ArrayDeque<>();
    this.blocks = new ArrayList<>();
    this.cases = new ArrayDeque<>();
    this.moduleMap = new LinkedHashMap<>();
  }

  /**
   * Returns mapping that keeps the transformed control flow graph model.
   *
   * @return transformed control flow graph model
   */
  public Map<Module, Map<Process, List<CfgTransform>>> getModuleMap() {
    return this.moduleMap;
  }

  private Deque<Module> getModules() {
    return this.modules;
  }

  private List<BasicBlock> getBlocks() {
    return this.blocks;
  }

  private Deque<Case> getCases() {
    return this.cases;
  }

  private List<CfgTransform> getTransformations() {
    return this.transformations;
  }

  @Override
  public void onModuleBegin(final Module module) {
    getModuleMap().put(module, new LinkedHashMap<Process, List<CfgTransform>>());
    getModules().push(module);
  }

  @Override
  public void onModuleEnd(final Module module) {
    getModules().pop();
  }

  @Override
  public void onProcessBegin(final Process process) {
    this.transformations = new ArrayList<>();
  }

  @Override
  public void onProcessEnd(final Process process) {
    getModuleMap().get(getModules().peek()).put(process, getTransformations());
  }

  @Override
  public void onBasicBlockBegin(final BasicBlock basicBlock) {

    if (getBlocks().contains(basicBlock)) {
      Log.getLogger().log(LogLevel.WARNING, "Loop is found");
    }
    getBlocks().add(basicBlock);
  }

  @Override
  public void onBasicBlockEnd(final BasicBlock basicBlock) {

    /* Remove the last block. */
    final int lastBlockNum = getBlocks().size() - 1;
    getBlocks().remove(lastBlockNum);
  }

  @Override
  public void onCaseBegin(final Case caseNode) {
    if (getCases().contains(caseNode)) {
      Log.getLogger().log(LogLevel.WARNING, "Loop is found");
    }
    getCases().push(caseNode);
  }

  @Override
  public void onCaseEnd(final Case caseNode) {
    getCases().pop();
  }

  @Override
  public void onSwitchBegin(final Switch switchNode) {
    addTransformation(getBlocks(), getCases(), switchNode);
  }

  private void addTransformation(
      final List<BasicBlock> blocks,
      final Deque<Case> cases,
      final Switch switchNode) {

    final CfgTransform transformation =
        new CfgTransform(
            cases, blocks, switchNode, new LinkedHashMap<Case, List<BasicBlock>>());

    getTransformations().add(transformation);
  }

  @Override
  public void onSink(final Sink sink) {

    if (getCases().isEmpty()) {

      /* There is no switch\case nodes. */
      getTransformations().add(new CfgTransform(getCases(), getBlocks(), null, null));
    } else {

      /* Get last control flow graph transformation. */
      final int lastProcessTrSize = getTransformations().size();
      final CfgTransform lastTransform = getTransformations().get(lastProcessTrSize - 1);

      /* Create a copy of current blocks list. */
      final List<BasicBlock> terminalBlocks = new ArrayList<>();
      terminalBlocks.addAll(getBlocks());

      lastTransform.addTermination(getCases().peek(), terminalBlocks);
    }
  }
}
