/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.cgaa;

import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Switch;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Deque;
import java.util.List;
import java.util.Map;

/**
 * Control flow graph transformation.
 *
 * <p>The transformation of control flow graph {@link ru.ispras.retrascope.model.cfg.CfgModel}
 * model. The guarded actions decision diagram builder {@link CfgCgaaTransformerBuilder}
 * forms this as intermediate step of it's work.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CfgTransform {
  private final Deque<Case> cases;
  private final List<BasicBlock> blocks;
  private final Switch switchNode;
  private final Map<Case, List<BasicBlock>> map;

  /**
   * Constructs object with the specified parameters.
   *
   * @param cases case nodes
   * @param blocks block nodes
   * @param switchNode switch node
   * @param map mapping
   */
  public CfgTransform(
      final Deque<Case> cases,
      final List<BasicBlock> blocks,
      final Switch switchNode,
      final Map<Case, List<BasicBlock>> map) {

    this.cases = new ArrayDeque<>();
    this.cases.addAll(cases);
    this.blocks = new ArrayList<>();
    this.blocks.addAll(blocks);
    this.switchNode = switchNode;
    this.map = map;
  }

  /**
   * Returns associated case nodes.
   *
   * @return cases stack
   */
  public Deque<Case> getCases() {
    return this.cases;
  }

  /**
   * Returns associated block nodes.
   *
   * @return basic blocks stack
   */
  public List<BasicBlock> getBlocks() {
    return this.blocks;
  }

  /**
   * Returns associated switch node.
   *
   * @return switch node
   */
  public Switch getSwitchNode() {
    return this.switchNode;
  }

  /**
   * Returns mapping of case and basic block nodes.
   *
   * @return cases/basic block nodes mapping
   */
  public Map<Case, List<BasicBlock>> getMap() {
    return this.map;
  }

  /**
   * Adds terminal case and basic block nodes.
   *
   * @param caseNode case node
   * @param blocks basic blocks
   */
  public void addTermination(final Case caseNode, final List<BasicBlock> blocks) {
    this.map.put(caseNode, blocks);
  }

  /**
   * Checks rather object contains a switch node.
   *
   * @return {@code true} if object contains non-null switch node, {@code false} otherwise
   */
  public boolean containsSwitch() {
    return this.switchNode != null;
  }

  /**
   * Checks rather object contains mapping.
   *
   * @return {@code true} if object contains non-null mapping, {@code false} otherwise
   */
  public boolean containsMap() {
    return this.map != null && !this.map.isEmpty();
  }
}
