/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.cgaa;

import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableContainer;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;

import java.util.AbstractMap;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Deque;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Back-end component of clocked guarded atomic action transformer
 * that implements static signal assignment transformation.
 *
 * <p>HDL parser backend that converts clocked guarded atomic actions model
 * {@link ru.ispras.retrascope.model.cgaa.CgaaModel}
 * into Static Signal Assignment (SSA) form.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaSsaBackend extends CfgDefaultVisitor {

  /* Declaration containers & their variables names & their versions. */
  private final Map<VariableContainer, Map<String, Deque<NodeVariable>>> versionMap;

  /* The current declaration container. */
  private VariableContainer containerNode;

  /* The blocks of current action. */
  private final List<BasicBlock> blocks;

  /**
   * Default constructor.
   */
  public CgaaSsaBackend() {
    super();

    this.versionMap = new LinkedHashMap<>();
    this.blocks = new ArrayList<>();
  }

  private Map<VariableContainer, Map<String, Deque<NodeVariable>>> getVersions() {
    return this.versionMap;
  }

  private List<BasicBlock> getBlocks() {
    return this.blocks;
  }

  @Override
  public void onModuleBegin(final Module module) {

    /* Initialize versions mapping with variables. */
    initVersions(module);
  }

  private void initVersions(final VariableContainer containerNode) {

    final List<NodeVariable> vars = containerNode.getVariables();
    final Map<String, Deque<NodeVariable>> variableMap = new LinkedHashMap<>();

    for (final NodeVariable variable : vars) {
      final Deque<NodeVariable> varVersions = new ArrayDeque<>();
      varVersions.push(variable);
      variableMap.put(variable.getName(), varVersions);
    }
    getVersions().put(containerNode, variableMap);
  }

  @Override
  public void onModuleEnd(final Module module) {
    getVersions().clear();
  }

  @Override
  public void onProcessBegin(final Process process) {
    this.containerNode = process;
    initVersions(process);
  }

  @Override
  public void onBasicBlockBegin(final BasicBlock basicBlock) {
    getBlocks().add(basicBlock);
  }

  @Override
  public void onBasicBlockEnd(final BasicBlock basicBlock) {
    getBlocks().remove(basicBlock);
  }

  @Override
  public void onSink(final Sink sink) {

    /* Check rather the current diagram branch contains versionMap. */
    boolean versionsExist = false;

    /* Get block/assignment numbers where variables are defined last time. */
    final Map<RangedVariable, Map.Entry<Integer, Integer>> defNumbers = new LinkedHashMap<>();

    for (int i = 0; i < getBlocks().size(); i++) {

      final List<Assignment> assignments = getBlocks().get(i).getAssignments();

      for (int j = 0; j < assignments.size(); j++) {
        final RangedVariable define = assignments.get(j).getTarget();
        if (defNumbers.containsKey(define)) {
          defNumbers.remove(define);
          versionsExist = true;
        }
        defNumbers.put(define, new AbstractMap.SimpleEntry<>(i, j));
      }
    }

    /* If the current diagram branch contains variable versionMap. */
    if (versionsExist) {

      /* Substitute defined variables by their versionMap except last defines. */
      for (int i = 0; i < getBlocks().size(); i++) {

        final BasicBlock block = getBlocks().get(i);
        final List<Assignment> assignments = block.getAssignments();
        for (int j = 0; j < assignments.size(); j++) {
          final Assignment assign = assignments.get(j);

          /* Assignment define processing. */
          final RangedVariable define = assign.getTarget();
          final NodeVariable defineVariable = define.getVariable();

          final Map.Entry<Integer, Integer> lastDefNumbers = defNumbers.get(define);

          /* If the last definition of this variable goes later. */
          if (!lastDefNumbers.getKey().equals(i) || !lastDefNumbers.getValue().equals(j)) {

            /* Find variable declaration container. */
            VariableContainer container = this.containerNode;

            while (!container.containsVariable(defineVariable)
                && (container instanceof CfgModelNode)) {

              container = (VariableContainer) ((CfgModelNode) container).getOnlyParent();
            }

            /* create and add new version of the defined variable. */
            final NodeVariable newVariable = container.declareVariableVersion(define);

            /* Put version to the stack. */
            final Map<String, Deque<NodeVariable>> containerMap = getVersions().get(container);
            final Deque<NodeVariable> versions = containerMap.get(defineVariable.getName());
            versions.push(newVariable);
            containerMap.remove(defineVariable.getName());
            containerMap.put(defineVariable.getName(), versions);

            assign.setVariable(new RangedVariable(newVariable));
          }

          /* Assignment uses processing. */
          final Set<NodeVariable> uses = assign.getUses();

          if (!uses.isEmpty()) {

            final List<NodeBinding.BoundVariable> bindings = new ArrayList<>();

            for (final NodeVariable use : uses) {

              /* Find variable declaration container. */
              VariableContainer container = this.containerNode;

              while (!container.containsVariable(use)) {
                container = (VariableContainer)((CfgModelNode) container).getOnlyParent();
              }

              final Map<String, Deque<NodeVariable>> containerMap = getVersions().get(container);

              if (containerMap.containsKey(use.getName())) {

                final NodeVariable lastVersion = containerMap.get(use.getName()).peek();
                if (!lastVersion.equals(use)) {
                  bindings.add(NodeBinding.bindVariable(use, lastVersion));
                }
              }
            }
            if (!bindings.isEmpty()) {
              assign.setValue(
                  Transformer.substituteAllBindings(new NodeBinding(assign.getValue(), bindings)));
            }
          }
        }
      }
    }
  }
}
