/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * Component that detects Clock-like variables in the clocked guarded atomic actions model.
 *
 * <p>Visitor of clocked guarded actions model {@link ru.ispras.retrascope.model.cgaa.CgaaModel}
 * for clock-like signals detection.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaClockDetector extends MapCfgVisitor {

  /* The clock-like variables for current module. */
  private List<NodeVariable> currentClockList;

  /**
   * Checks rather the specified variable is a one-bit length signal.
   *
   * @param variable variable
   *
   * @return {@code true} if variable is one bit long signal, {@code false} otherwise
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when any of arguments is {@code null};
   *         (2) when argument's user data is not of {@link VariableData} type
   */
  public static boolean isOneBitSignal(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    final Object variableData = variable.getUserData();
    if (variableData instanceof VariableData) {
      final boolean isSignal =
          VariableType.isSignal(((VariableData) variableData).getVariableType());
      final DataType dataType = variable.getDataType();

      return isSignal
            && (dataType.equals(DataType.BOOLEAN) || dataType.equals(DataType.BIT_VECTOR(1)));
    } else {
      throw new IllegalArgumentException("Incorrect type: " + variableData.getClass() + ".");
    }
  }

  @Override
  public void onRootBegin() {
    this.currentClockList = new ArrayList<>();
  }

  @Override
  public void onModuleBegin(final Module module) {
    this.currentClockList = new ArrayList<>();
  }

  @Override
  public void onModuleEnd(final Module module) {
    final Set<NodeVariable> clocks = new LinkedHashSet<>();
    clocks.addAll(this.currentClockList);
    this.getClockMap().put(module, clocks);
    this.currentClockList.clear();
  }

  @Override
  public void onProcessBegin(final Process process) {

    /* Analyse sensitivity list. */
    final EventList eventList = process.getSensitivityList();
    if (eventList != null && !eventList.isEmpty()) {
      final Collection<Event> events = eventList.getEvents();
      for (final Event event : events) {
        final NodeVariable variable = event.getRangedVariable().getVariable();
        if (isOneBitSignal(variable) && !this.currentClockList.contains(variable)) {
          this.currentClockList.add(variable);
        }
      }
    }
  }

  @Override
  public void onBasicBlockBegin(final BasicBlock basicBlock) {
    final List<Assignment> statements = basicBlock.getAssignments();
    for (final Assignment statement : statements) {

      /* Analyse the signals are used in basic blocks assignments. */
      final Collection<NodeVariable> stmtUses = statement.getUses();
      if (stmtUses != null && !stmtUses.isEmpty()) {
        for (final NodeVariable stmtUse : stmtUses) {
          if (this.currentClockList.contains(stmtUse)) {
            this.currentClockList.remove(stmtUse);
          }
        }
      }

      /* Analyse the signals are defined in basic blocks assignments. */
      final Collection<NodeVariable> stmtDefines = statement.getDefines();
      if (stmtDefines != null && !stmtDefines.isEmpty()) {
        for (final NodeVariable stmtDefine : stmtDefines) {
          if (this.currentClockList.contains(stmtDefine)) {
            this.currentClockList.remove(stmtDefine);
          }
        }
      }
    }
  }
}
