/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
* Component that extracts data for constructing EFSM states
* and transitions from clocked guarded atomic actions model.
*
* @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
*/
final class CgaaEfsmDataExtractor extends MapCfgVisitor {

  /* Processes and corresponding EFSM data containers. */
  private Map<Process, List<EfsmDataContainer>> processDataMap;

  /* Fields that are used throughout walking process. */

  /* The current module. */
  private Module currentModule;

  /* The current process. */
  private Process currentProcess;

  /* The data containers for current process. */
  private List<EfsmDataContainer> dataList;

  /* The current state expression. */
  private List<Node> currentState;

  /* The current list of events. */
  private List<Event> currentEventList;

  /* The current guard. */
  private List<Node> currentGuard;

  /* The current action. */
  private List<BasicBlock> currentAction;

  /* Be set to true when guard processing has been finished. */
  private boolean guardFinished;

  /* Be set to true when guard & state are satisfiable & compatible. */
  private boolean actionReachable;

  /* Cases that used into state-like variables' depending expressions. */
  private List<Case> stateCases;

  /* The current clock. */
  private Event clock;

  /* The current case. */
  private Case clockCase;

  /* Diagram paths number (for statistics keeping). */
  private int pathNum;

  public CgaaEfsmDataExtractor() {
    super();
    this.processDataMap = new LinkedHashMap<>();
    this.currentState = new ArrayList<>();
    this.currentEventList = new ArrayList<>();
    this.currentGuard = new ArrayList<>();
    this.currentAction = new ArrayList<>();
    this.guardFinished = false;
    this.actionReachable = true;
    this.stateCases = new ArrayList<>();

    this.pathNum = 0;
  }

  /**
   * Returns mapping of processes into data containers.
   *
   * @return mapping of processes into data containers
   */
  public Map<Process, List<EfsmDataContainer>> getProcessDataMap() {
    Log.print(LogLevel.INFO, "Number of GADD paths: " + this.pathNum);
    return this.processDataMap;
  }

  @Override
  public void onRootBegin() {
    this.processDataMap = new LinkedHashMap<>();
    this.currentState = new ArrayList<>();
    this.currentEventList = new ArrayList<>();
    this.currentGuard = new ArrayList<>();
    this.currentAction = new ArrayList<>();
    this.guardFinished = false;
    this.actionReachable = true;
    this.stateCases = new ArrayList<>();
    this.pathNum = 0;
  }

  @Override
  public void onModuleBegin(final Module module) {
    this.currentModule = module;
  }

  @Override
  public void onProcessBegin(final Process process) {
    this.currentProcess = process;
    this.dataList = new ArrayList<>();
    this.currentEventList.addAll(process.getSensitivityList().getEvents());
  }

  @Override
  public void onProcessEnd(final Process process) {
    this.processDataMap.put(this.currentProcess, this.dataList);
    this.currentEventList.clear();
  }

  @Override
  public void onCaseBegin(final Case caseNode) {
    Node expr;
    final Switch caseParent = (Switch)caseNode.getOnlyParent();

    /* Construct a condition expression for the current case node. */
    expr = CgaaEfsmTransformerUtils.getExpression(caseNode, caseParent);

    if (isStateVarUsingExpr(expr)) {
      this.currentState.add(expr);
      this.stateCases.add(caseNode);
    } else {
      this.currentGuard.add(expr);
    }

    /* If switch has event, save it as clock. */
    if (caseParent.getEvent() != null) {
      if (this.clock == null) {
        this.clock = caseParent.getEvent();
        this.clockCase = caseNode;
      } else {
        Log.print(LogLevel.WARNING,
            String.format(
                "Unreachable path for: event1=%s, event2=%s.",
                this.clock.toString(),
                caseParent.getEvent().toString()));
      }
    }
  }

  private boolean isStateVarUsingExpr(final Node expr) {
    return !CollectionUtils.intersectSets(
        new LinkedHashSet<>(ExprUtils.getVariables(expr)),
        this.getStateMap().get(this.currentModule)).isEmpty();
  }

  @Override
  public void onCaseEnd(final Case caseNode) {
    if (this.stateCases.contains(caseNode)) {
      this.currentState.remove(this.currentState.size() - 1);
      this.stateCases.remove(caseNode);
    } else {
      this.currentGuard.remove(this.currentGuard.size() - 1);
    }

    /* Visitor is ready again to process guard part elaboration. */
    if (this.guardFinished) {
      this.guardFinished = false;
    }

    if (this.clockCase != null && this.clockCase.equals(caseNode)) {
      this.clockCase = null;
      this.clock = null;
    }
  }

  @Override
  public void onBasicBlockBegin(final BasicBlock block) {

    checkConditionsIfNot();

    if (this.actionReachable) {
      this.currentAction.add(block);
    }
  }

  private void checkConditionsIfNot() {
    if (!this.guardFinished) {
      this.actionReachable =
          (this.currentState.isEmpty()
              || checkInvCompliance(this.currentProcess, this.currentState))
              && (this.currentGuard.isEmpty()
              || checkInvCompliance(this.currentProcess, this.currentGuard));
      this.guardFinished = true;
    }
  }

  /* Check the compliance of nodes from the specified collection to the invariants
  * on used variables at the specified process.*/
  private static boolean checkInvCompliance(final Process process, final List<Node> list) {
    InvariantChecks.checkNotNull(process);
    InvariantChecks.checkNotNull(list);

    boolean isSat;

    Node conjunction = ExprUtils.getConjunction(list.toArray(new Node[list.size()]));

    if (ExprUtils.areCompatible(conjunction)) {
      final Collection<Node> invNodes =
          CgaaEfsmTransformerUtils.getInvariants(process, ExprUtils.getVariables(conjunction));

      for (final Node invNode : invNodes) {
        conjunction = ExprUtils.getConjunction(invNode, conjunction);
      }
      isSat = ExprUtils.areCompatible(conjunction);

    } else {
      isSat = false;
    }
    return isSat;
  }

  @Override
  public void onSink(final Sink sink) {

    this.pathNum++;
    if (this.actionReachable) {

      Action action;

      if (this.currentAction.isEmpty()) {
        action = new Action(new ArrayList<Assignment>());
      } else {
        final List<Assignment> assignments = new ArrayList<>();
        for (final BasicBlock block : this.currentAction) {
          assignments.addAll(block.getAssignments());
        }
        action = new Action(assignments);
      }

      final Set<Node> states = getNodes(this.currentState);
      final Set<Node> guard = getNodes(this.currentGuard);

      if (states.isEmpty()) {
        states.add(NodeValue.newBoolean(true));
      }

      /* Check rather guard is satisfiable. */
      final Guard guardExpr = new Guard(guard);
      if (ExprUtils.isSAT(guardExpr.toNode())) {
        final GuardedAction guardedAction = new GuardedAction(guardExpr, action);
        if (this.clock == null) {
          guardedAction.addClocks(this.currentEventList);
        } else {
          guardedAction.addClock(this.clock.deepCopy());
        }
        final EfsmDataContainer container =
            new EfsmDataContainer(states, guardedAction);

        this.dataList.add(container);
      }
    }
    this.currentAction.clear();
  }

  private Set<Node> getNodes(final List<Node> list) {
    final Set<Node> nodes = new LinkedHashSet<>();
    if (list.isEmpty()) {
      nodes.add(NodeValue.newBoolean(true));
    } else {
      nodes.addAll(list);
    }
    return nodes;
  }
}
