/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import org.apache.commons.cli.ParseException;
import org.apache.commons.collections15.BidiMap;
import org.apache.commons.collections15.bidimap.DualHashBidiMap;
import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.logic.Clause;
import ru.ispras.fortress.logic.NormalForm;
import ru.ispras.fortress.logic.Orthogonalizer;
import ru.ispras.fortress.transformer.ReduceOptions;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.basis.TransformUtils;
import ru.ispras.retrascope.engine.cfg.CfgEngine;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.VariableContainer;
import ru.ispras.retrascope.model.basis.VariableDeclaration;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgVisitor;
import ru.ispras.retrascope.model.cfg.CfgWalker;
import ru.ispras.retrascope.model.cfg.CfgWalkerMode;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cgaa.CgaaModel;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The transformer of clocked guarded actions container
 * {@link ru.ispras.retrascope.model.cgaa.CgaaModel} into
 * extended finite state machine {@link ru.ispras.retrascope.model.efsm.EfsmModel}.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CgaaEfsmTransformer extends CfgEngine {

  /**
   * The engine identifier.
   */
  public static final String ENGINE_ID = "cgaa-efsm-transformer";

  /**
   * State-like variables names command line parameter.
   */
  public static final Parameter STATE_VARIABLES =
      new Parameter("state-vars", true, "Names of state-like variables");

  /* The clock-like variables detector. */
  private final CgaaClockDetector clockDetector;

  /* The state-like variables extractor. */
  private final CgaaStateDetector stateDetector;

  /* The state-like variables statistics keeping visitor. */
  private final CgaaStatisticsKeeper statKeeper;

  /* The depending-on-states switches splitter. */
  private final CgaaSwitchSplitter switchSplitter;

  /* The splitting rules' applicator. */
  private final CgaaRuleApplier ruleApplier;

  /* The states expressions' extractor. */
  private final CgaaEfsmDataExtractor efsmDataExtractor;

  /* Debug mode flag. */
  private boolean debugMode;

  /* File for debug output. */
  private File debugFile;

  /* Debug file default name. */
  private static final String DEFAULT_FILE_NAME = ENGINE_ID + ".log";

  /* The default charset (for printing to debug file). */
  private static final Charset DEFAULT_CHARSET = Charset.defaultCharset();

  /**
   * Default constructor.
   */
  public CgaaEfsmTransformer() {
    super(
      CgaaEfsmTransformer.ENGINE_ID,
      CgaaModel.TYPE,
      EfsmModel.TYPE,
      new CfgWalker(new CgaaClockDetector(), CfgWalkerMode.DFS_NO_REPEAT),
      new CfgWalker(new CgaaStateDetector(), CfgWalkerMode.DFS),
      new CfgWalker(new CgaaStatisticsKeeper(), CfgWalkerMode.DFS),
      new CfgWalker(new CgaaSwitchSplitter(), CfgWalkerMode.DFS),
      new CfgWalker(new CgaaRuleApplier(), CfgWalkerMode.DFS),
      new CfgWalker(new CgaaEfsmDataExtractor(), CfgWalkerMode.DFS));

    /* Engine command line parameters. */
    addParameter(Retrascope.TOOL_DEBUG_PARAM);
    addParameter(STATE_VARIABLES);

    final CfgWalker[] walkers = getWalkers();

    /* Initialize visitor components. */
    CgaaClockDetector clkVisitor = null;
    CgaaStateDetector stateVarVisitor = null;
    CgaaStatisticsKeeper stateStatVisitor = null;
    CgaaSwitchSplitter switchVisitor = null;
    CgaaRuleApplier ruleVisitor = null;
    CgaaEfsmDataExtractor stateExprVisitor = null;

    for (final CfgWalker config : walkers) {

      final CfgVisitor visitor = config.getVisitor();

      if (visitor instanceof CgaaClockDetector) {

        clkVisitor = (CgaaClockDetector) visitor;
      } else if (visitor instanceof CgaaStateDetector) {

        stateVarVisitor = (CgaaStateDetector) visitor;
      } else if (visitor instanceof CgaaStatisticsKeeper) {

        stateStatVisitor = (CgaaStatisticsKeeper) visitor;
      } else if (visitor instanceof CgaaSwitchSplitter) {

        switchVisitor = (CgaaSwitchSplitter) visitor;
      } else if (visitor instanceof CgaaRuleApplier) {

        ruleVisitor = (CgaaRuleApplier) visitor;
      } else if (visitor instanceof CgaaEfsmDataExtractor) {

        stateExprVisitor = (CgaaEfsmDataExtractor) visitor;
      }
    }
    this.clockDetector = clkVisitor;
    this.stateDetector = stateVarVisitor;
    this.statKeeper = stateStatVisitor;
    this.switchSplitter = switchVisitor;
    this.ruleApplier = ruleVisitor;
    this.efsmDataExtractor = stateExprVisitor;
    this.debugMode = false;
  }

  @Override
  public void initialize(final Map<EntityType, Entity> inputs) {

    /* Extract control flow graph from wrapper. */

    final CgaaModel cgaaModel = (CgaaModel) inputs.get(CgaaModel.TYPE);
    inputs.remove(CgaaModel.TYPE);
    inputs.put(CfgModel.TYPE, cgaaModel.getModel());

    /* Initialize the internal visitors. */

    final Map<Module, Set<NodeVariable>> clockMap = new LinkedHashMap<>();
    this.clockDetector.setClockMap(clockMap);
    this.efsmDataExtractor.setClockMap(clockMap);

    final Map<Module, Set<NodeVariable>> stateMap = new LinkedHashMap<>();
    this.stateDetector.setStateMap(stateMap);
    this.statKeeper.setStateMap(stateMap);
    this.switchSplitter.setStateMap(stateMap);
    this.efsmDataExtractor.setStateMap(stateMap);

    final Map<CfgModelNode, List<CfgModelNode>> map = new LinkedHashMap<>();
    this.switchSplitter.setOldNewMap(map);
    this.ruleApplier.setOldNewMap(map);

    /* Elaborate command line parameters. */
    final Arguments arguments = (Arguments)inputs.get(Arguments.TYPE);

    try {
      final Arguments engineArgs = parseCommandLine(arguments.getCommandLine());
      final String value = engineArgs.getValue(STATE_VARIABLES);
      if (value != null) {
        final String[] stateNames = value.split(File.pathSeparator);
        this.stateDetector.setStateNames(stateNames);
      }

      if (engineArgs.contains(Retrascope.TOOL_DEBUG_PARAM)) {
        this.debugMode = true;
        this.debugFile =
            new File(engineArgs.getValue(Retrascope.TOOL_DEBUG_PARAM, DEFAULT_FILE_NAME));
        new PrintWriter(this.debugFile).close();
      }
    } catch (ParseException | FileNotFoundException e) {
      throw new RetrascopeRuntimeException(e.getMessage(), e.getCause());
    }
  }

  @Override
  public Entity getOutput() {

    final EfsmModel model = new EfsmModel();

    final Map<Process, List<EfsmDataContainer>> processDataMap =
        this.efsmDataExtractor.getProcessDataMap();

    Log.print(LogLevel.INFO, "======================================");
    int efsmCount = 0;

    int processCount = 0;
    int stateCount = 0;
    int transitionCount = 0;
    for (final Map.Entry<Process, List<EfsmDataContainer>> entry : processDataMap.entrySet()) {

      final Process keyProcess = entry.getKey();
      final Module module = (Module) keyProcess.getOnlyParent();

      /* Print clock-like variables. */

      final List<String> clockNames = new ArrayList<>();

      final Set<NodeVariable> clockVars = this.clockDetector.getClockMap().get(module);

      for (final NodeVariable var : clockVars) {
        clockNames.add(var.getName());
      }

      Log.print(LogLevel.INFO, "Clock-like variables: " + clockNames.toString() + ".");
      Log.print(LogLevel.INFO, "Transforming the process of module: " + module.getName() + ".");

      final List<EfsmDataContainer> list = entry.getValue();

      final Efsm efsm = new Efsm(module.getName() + "_" + processCount);

      /* Copy invariants from process & parent modules. */
      addInvariants(efsm, keyProcess);

      /* Construct EFSM states. */
      stateCount += addStates(efsm, keyProcess, list);

      if (efsm.containsStates()) {
        transitionCount += addTransitions(efsm, list);
        model.addEfsm(efsm);
        efsmCount++;
      }
      processCount++;
    }
    Log.print(LogLevel.INFO, efsmCount + " EFSM(s) are extracted.");

    /* If debug log mode is enabled, write statistical info to log file. */
    if (this.debugMode) {
      writeToFile(this.debugFile, "states=" + stateCount + " ", false);
      writeToFile(this.debugFile, "transitions=" + transitionCount + " ", false);
      writeToFile(this.debugFile, "efsms=" + efsmCount, true);
    }

    return model;
  }

  private void addInvariants(final Efsm efsm, final Process process) {
    CfgModelNode current = process;
    while (current != null && current instanceof VariableContainer)  {

      final Map<NodeVariable, VariableDeclaration> decMap =
          ((VariableContainer)current).getVariablesMapping();

      for (final Map.Entry<NodeVariable, VariableDeclaration> entry : decMap.entrySet()) {

        final String varName = entry.getKey().getName();
        final Node invariant = entry.getValue().getInvariant();

        if (!efsm.containsInvariant(varName) && invariant != null) {
          efsm.addInvariant(varName, invariant);
        }
      }
      current = current.getOnlyParent();
    }
  }

  private int addStates(
      final Efsm efsm,
      final Process process,
      final Collection<EfsmDataContainer> list) {

    /* Create the convenience collection of pre-states. */
    final Set<Set<Node>> preStates = new LinkedHashSet<>();

    for (final EfsmDataContainer data : list) {
      final Set<Node> dataState = data.getState();
      if (dataState.isEmpty()) {
        throw new IllegalStateException("Empty state.");
      } else {
        preStates.add(dataState);
      }
    }

    final List<Node> states = new ArrayList<>();

    /* Create a bi-directional mapping for literals and expressions. */
    final BidiMap<Integer, Node> literalMap = new DualHashBidiMap<>();

    if (!preStates.isEmpty() && preStates.size() > 1) {

      /* Transform data containers' pre-states into clauses. */
      final List<Clause> clauses = createClauses(preStates, literalMap);

      /* Make clauses orthogonal. */
      final NormalForm normalForm = new NormalForm(NormalForm.Type.DNF, clauses);
      final NormalForm orthogonalForm = Orthogonalizer.orthogonalize(normalForm);
      final List<Clause> ortClauses = orthogonalForm.getClauses();

      /* Transform orthogonal clauses into EFSM states. */
      createStates(process, states, literalMap, ortClauses);
    }

    /* If states are not extracted, a trivial EFSM should be created. */
    if (states.isEmpty()) {
      states.add(NodeValue.newBoolean(true));
    }

    /* Add states to EFSM model. */
    final List<String> stateVarNames = new ArrayList<>();
    int stateCount = 0;
    for (final Node state : states) {

      EfsmState efsmState;

      if (state instanceof NodeOperation) {
        efsmState = new EfsmState(state);

        /* Store state-like variables names (for statistics only). */
        final Collection<NodeVariable> stateVariables = ExprUtils.getVariables(state);

        for (final NodeVariable stateVar : stateVariables) {
          final String name = stateVar.getName();
          if (!stateVarNames.contains(name)) {
            stateVarNames.add(name);
          }
        }
      } else if (state instanceof NodeValue) {
        efsmState = new EfsmState(state);
      } else {
        throw new IllegalArgumentException("Incorrect state: " + state.toString() + ".");
      }

      efsm.addState(efsmState);
      stateCount++;
    }

    Log.print(LogLevel.INFO, stateCount + " states are extracted.");
    Log.print(LogLevel.INFO, "The state-like variables are: " + stateVarNames.toString() + ".");
    return stateCount;
  }

  private void createStates(
      final Process process,
      final Collection<Node> states,
      final BidiMap<Integer, Node> literalMap,
      final Collection<Clause> ortClauses) {

    for (final Clause clause : ortClauses) {

      Node state = null;
      final Set<Integer> clauseLiterals = clause.getVars();
      for (final Integer literal : clauseLiterals) {
        Node node = literalMap.get(literal);

        if (clause.getSign(literal)) {
          node = ExprUtils.getNegation(node);
        }
        state = state == null ? node : ExprUtils.getConjunction(state, node);
        state = Transformer.standardize(state);
      }

      /* Add variable invariants to state constraint. */
      final Collection<NodeVariable> stateVars = ExprUtils.getVariables(state);
      final Collection<Node> stateInvariants =
          CgaaEfsmTransformerUtils.getInvariants(process, stateVars);
      final Node invConjunction = stateInvariants.isEmpty()
          ? new NodeValue(Data.newBoolean(true)) :
          ExprUtils.getConjunction(stateInvariants.toArray(new Node[stateInvariants.size()]));

      /* If state expression is satisfiable, then it is a correct state. */
      if (ExprUtils.isSAT(state) && ExprUtils.areCompatible(state, invConjunction)) {
        states.add(state);
      }
    }
  }

  private List<Clause> createClauses(
      final Collection<Set<Node>> preStates,
      final BidiMap<Integer, Node> literalMap) {

    final List<Clause> clauses = new ArrayList<>();

    for (final Set<Node> preState : preStates) {

      final Clause clause = new Clause();

      for (Node node : preState) {

        boolean negated = false;

        if (node instanceof NodeOperation) {

          final Enum<?> opId = ((NodeOperation) node).getOperationId();

          if (opId == StandardOperation.NOT) {
            node = ((NodeOperation) node).getOperand(0);
            negated = true;
          }

          int literal;

          if (literalMap.values().contains(node)) {
            literal = literalMap.getKey(node);
          } else {
            literal = literalMap.keySet().size();
            literalMap.put(literal, node);
          }
          clause.add(literal, negated);
        }
      }
      clauses.add(clause);
    }
    return clauses;
  }

  /**
   * Adds transitions from the specified collection to the specified {@link Efsm} object.
   *
   * @param efsm model the transitions to be added to
   * @param list collection of transition candidates
   *
   * @return number of transitions that are constructed
   */
  private int addTransitions(
      final Efsm efsm,
      final Collection<EfsmDataContainer> list) {

    /* Collection for candidates on initial model state. */
    final Map<EfsmState, Set<NodeVariable>> initialResetMap = new LinkedHashMap<>();

    /* Variables for EFSM extraction statistics keeping. */
    int transitionCount = 0;

    /* Create EFSM transitions. */
    for (final EfsmDataContainer data : list) {

      final GuardedAction guardedAction = data.getGuardedAction();
      final Action action = guardedAction.getAction();
      final Collection<Event> clocks = guardedAction.getClocks();

      final Guard guard = guardedAction.getGuard();
      final Node guardNode = guard.toNode();

      final List<BasicBlock> blocks = new ArrayList<>();

      final BasicBlock block = new BasicBlock(action.getAssignments());
      blocks.add(block);

      EfsmState srcState;

      final Node expression =
          ExprUtils.getConjunction(data.getState().toArray(new Node[data.getState().size()]));

      final Set<EfsmState> efsmStates = efsm.getStates();
      for (final EfsmState efsmState : efsmStates) {

        final Node stateExpr = efsmState.getExpression();

        if (ExprUtils.areCompatible(expression, stateExpr)) {

          /* Detect source state. */
          srcState = efsmState;

          /* Detect possible destination states. */
          final List<EfsmState> possible = new ArrayList<>();
          for (final EfsmState possibleDst : efsmStates) {

            final Node posExpr = possibleDst.getExpression();
            final Node substitution = posExpr.deepCopy();

            /* Create a weakest precondition and check the compatibility with source state. */

            final Node wPreCond = TransformUtils.substituteBackward(blocks, substitution);
            final Node srcExpr = srcState.getExpression();

            if (ExprUtils.areCompatible(srcExpr, wPreCond)) {
              possible.add(possibleDst);
            }
          }

          /* Add possible destination states. */
          if (possible.isEmpty()) {
            Log.print(
                LogLevel.WARNING,
                String.format(
                    "Lockup found for: src-state=%s, guard=%s, action=%s.",
                    srcState.toString(), guard.toString(), action.toString()));

          } else if (CgaaEfsmTransformerUtils.containsSingleObject(possible)) {

            /* Get the only possible destination state. */
            final EfsmState dst = possible.get(0);

            /* Create transition's guarded action. */
            final GuardedAction transition = new GuardedAction(new Guard(guardNode), action);
            transition.addClocks(clocks);

            efsm.createTransition(srcState, dst, transition);

            /* Check rather transition leads to initial state. */
            if (leadsToInitialState(transition)) {
              initialResetMap.put(dst, getReset(transition));
            }

            /* Increment transition counter. */
            transitionCount++;
          } else {

            /* More than one possible destination state. */
            for (final EfsmState dst : possible) {

              /* Refine the transition guard. */
              final Node dstExprCopy = dst.getExpression().deepCopy();
              final Node refinedGuard = refineGuard(action, guardNode, dstExprCopy);

              final GuardedAction transition = new GuardedAction(new Guard(refinedGuard), action);
              transition.addClocks(clocks);

              efsm.createTransition(srcState, dst, transition);

              /* Check rather transition leads to initial state. */
              if (leadsToInitialState(transition)) {
                initialResetMap.put(dst, getReset(transition));
              }

              /* Increment transition counter. */
              transitionCount++;
            }
          }
        }
        if (!ExprUtils.isSAT(
            ExprUtils.getConjunction(expression, ExprUtils.getNegation(stateExpr)))) {
          break;
        }
      }
    }
    Log.print(LogLevel.INFO, transitionCount + " transitions are extracted.");

    if (initialResetMap.keySet().isEmpty()) {

      /* No initial state has been found. */

      Log.print(LogLevel.WARNING, "No initial state found.");

      final EfsmState first = efsm.getStates().iterator().next();

      efsm.setInitialState(first);

      setResetGuardedAction(efsm, first);

      Log.print(LogLevel.WARNING, "The first is chosen as initial: " + first.toString() + ".");
    } else if (CgaaEfsmTransformerUtils.containsSingleObject(initialResetMap.keySet())) {

      /* Only one candidate is extracted - set is as initial. */

      final Map.Entry<EfsmState, Set<NodeVariable>> initEntry =
          initialResetMap.entrySet().iterator().next();
      final EfsmState initState = initEntry.getKey();

      Log.print(LogLevel.INFO, "The initial state is: " + initState
          + ", reset signal is: " + initEntry.getValue() + ".");

      /* Create pre-initial state & link it with initial state. */
      efsm.addState(Efsm.PREINITIAL_STATE);
      efsm.createTransition(Efsm.PREINITIAL_STATE, initState, getInitTransition());

      efsm.setInitialState(initState);

      setResetGuardedAction(efsm, initState);
    } else {

      /* More than one candidate on initial state has been found. */

      Log.print(LogLevel.WARNING, "The following candidates on initial state are detected: ");

      for (final EfsmState state : initialResetMap.keySet()) {
        Log.print(LogLevel.INFO, state.toString());
      }

      final EfsmState first = initialResetMap.keySet().iterator().next();

      Log.print(LogLevel.WARNING, "The first one is chosen as initial: " + first + ".");

      efsm.setInitialState(first);

      setResetGuardedAction(efsm, first);
    }
    return transitionCount;
  }

  private Node refineGuard(
      final Action action,
      final Node guardNode,
      final Node dstExprCopy) {

    return Transformer.standardize(ExprUtils.getConjunction(
        guardNode,
        Transformer.reduce(ReduceOptions.NEW_INSTANCE,
            TransformUtils.substituteBackward(
                Collections.singletonList(action.toBasicBlock()), dstExprCopy))));
  }

  /* Set as resetting guarded action the transition that leads to the specified state. */
  private void setResetGuardedAction(final Efsm efsm, final EfsmState state) {

    final Set<EfsmTransition> incoming = efsm.getIncomingTransitions(state);

    /* Set the first one. */
    efsm.addResetGuardedAction(incoming.iterator().next().getGuardedAction());
  }

  /**
   * Returns reset signals that are probably used in the specified object.
   *
   * @param guardedAction object that can use reset signals
   *
   * @return reset signals collection
   */
  private Set<NodeVariable> getReset(final GuardedAction guardedAction) {

    final Set<NodeVariable> resets = new LinkedHashSet<>();
    final EventList eventList = guardedAction.getEventList();
    final Collection<Event> clocks = eventList.getEvents();
    final Set<NodeVariable> guardUses = guardedAction.getGuard().getUses();

    if (guardUses.isEmpty()) {

      /* Add all clock-like variables for this guarded action. */
      for (final Event clock : clocks) {
        resets.add(clock.getRangedVariable().getVariable());
      }

    } else if (CgaaEfsmTransformerUtils.containsSingleObject(guardUses)) {
      final NodeVariable guardUse = guardUses.iterator().next();
      if (eventList.contains(guardUse)) {
        resets.add(guardUse);
      }
    }
    return resets;
  }

  /**
   * Check rather this object leads to initial state.
   *
   * @param guardedAction object to be checked
   *
   * @return {@code true} when {@link GuardedAction} guarded action
   *         leads to initial state, {@code false} otherwise.
   */
  private boolean leadsToInitialState(final GuardedAction guardedAction) {

    /* Check rather guard depends only from one clock-like variable. */

    final Collection<Event> clocks = guardedAction.getClocks();
    final Guard guard = guardedAction.getGuard();
    final Set<NodeVariable> guardUses = guard.getUses();

    boolean checkGuard = false;

    if (guardUses.isEmpty()) {
      checkGuard = true;
    } else if (CgaaEfsmTransformerUtils.containsSingleObject(guardUses)) {

      final NodeVariable use = guardUses.iterator().next();

      for (final Event clock : clocks) {
        if (clock.getRangedVariable().getVariable().equals(use)) {
          checkGuard = true;
          break;
        }
      }
    }

    /* Check rather action contains only assignments like var := constant. */

    final Action action = guardedAction.getAction();
    boolean checkAction = false;
    for (final Assignment assignment : action.getAssignments()) {
      if (assignment.getUses().isEmpty()) {
        checkAction = true;
      } else {
        checkAction = false;
        break;
      }
    }
    return checkGuard && checkAction;
  }

  /**
   * Constructs a transition that contains assigning the default values to variables.
   *
   * @return initialization transition
   */
  private GuardedAction getInitTransition() {

    //TODO: action should be filled with default values of process variables

    return new GuardedAction(
        new Guard(NodeValue.newBoolean(true)),
        new Action(new ArrayList<Assignment>()));
  }

  private static void writeToFile(File file, String string, boolean asLine) {
    try {
      BufferedWriter writer =
          new BufferedWriter(
              new OutputStreamWriter(new FileOutputStream(file, true), DEFAULT_CHARSET));
      writer.write(string);
      if (asLine) {
        writer.newLine();
      }
      writer.close();

    } catch (IOException e) {
      throw new RetrascopeRuntimeException(e.getMessage(), e.getCause());
    }
  }
}
