/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.VariableContainer;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Switch;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * Auxiliary methods for extended finite state machine construction.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaEfsmTransformerUtils {

  private CgaaEfsmTransformerUtils() {
    /* Empty. */
  }

  /**
   * Constructs an expression from the expression of the specified switch
   * {@link ru.ispras.retrascope.model.cfg.Switch} node and it's possible values
   * in the specified case {@link ru.ispras.retrascope.model.cfg.Case} node.
   *
   * @param caseNode case node
   * @param switchNode switch node
   *
   * @return expression
   */
  public static Node getExpression(
      final Case caseNode,
      final Switch switchNode) {

    Node expr;

    if (caseNode.isDefault()) {

      /* The 'default' case processing. */
      final Collection<CfgModelNode> children = switchNode.getChildren();
      final List<NodeOperation> list = new ArrayList<>();
      for (final CfgModelNode child : children) {
        if (!child.equals(caseNode)) {
          list.add(constructExpr((Case)child, switchNode));
        }
      }
      expr = ExprUtils.getComplement(list.toArray(new Node[list.size()]));
    } else {

      /* The 'normal' case processing. */
      expr = constructExpr(caseNode, switchNode);
    }

    /* Transform expression into standard form. */
    return Transformer.standardize(expr);
  }

  private static NodeOperation constructExpr(
      final Case child,
      final Switch parent) {

    Node expr = null;

    for (final NodeValue value: child.getValues()) {
      if (expr == null) {
        expr = new NodeOperation(StandardOperation.EQ, parent.getCondition(), value);
      } else {
        expr = ExprUtils.getDisjunction(
            expr,
            new NodeOperation(StandardOperation.EQ, parent.getCondition(), value));
      }
    }
    return (NodeOperation) expr;
  }

  /**
   * Returns invariants of the specified variables that are defined/used in the specified process.
   *
   * @param process process
   * @param variables variables
   *
   * @return collection of variable invariants
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public static Collection<Node> getInvariants(
      final Process process,
      final Collection<NodeVariable> variables) {

    InvariantChecks.checkNotNull(process);
    InvariantChecks.checkNotNull(variables);

    /* The returned collection of invariants. */
    final List<Node> invariants = new ArrayList<>();

    for (final NodeVariable var : variables) {
      final Node invariant = getInvariant(process, var);
      if (invariant != null) {
        invariants.add(invariant);
      }
    }

    return invariants;
  }

  private static Node getInvariant(
      final Process process,
      final NodeVariable var) {
    InvariantChecks.checkNotNull(process);
    InvariantChecks.checkNotNull(var);

    Node invariant = null;

    /* The current variable declarations' container. */
    CfgModelNode current = process;

    do {
      if (current instanceof VariableContainer
          && ((VariableContainer)current).containsVariable(var)) {
        invariant = ((VariableContainer)current).getDeclaration(var).getInvariant();
      }

      current = current.getOnlyParent();
    } while (current != null);

    return invariant;
  }

  /**
   * Checks rather the specified collection contains exactly one element.
   *
   * @param collection collection
   *
   * @return {@code true} when the specified collection contains exactly one element,
   * {@code false} otherwise
   */
  public static boolean containsSingleObject(final Collection<?> collection) {
    return collection.size() == 1;
  }
}
