/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.CfgNodeType;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;

import java.util.Collection;
import java.util.List;

/**
 * Component that applies rules of clocked guarded atomic actions model transformation.
 *
 * <p>The component that applies transformation rules to clocked guarded atomic actions model.
 * Rules are constructed by {@link CgaaSwitchSplitter} visitor
 * as a part of EFSM extraction algorithm.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaRuleApplier extends MapCfgVisitor {

  /* The model to which rules should be applied. */
  private CfgModel cfgModel;

  /* The current process. */
  private Process process;

  private CfgModel getModel() {
    return this.cfgModel;
  }

  private Process getProcess() {
    return this.process;
  }

  @Override
  public void onCfgModelBegin() {
    this.cfgModel = new CfgModel();
  }

  @Override
  public void onModuleBegin(final Module module) {

    final List<CfgModelNode> newModules = getOldNewMap().get(module);
    final Module moduleNode = (Module) newModules.get(0);
    getModel().addModule(moduleNode);

    if (module.hasParents()) {
      final Collection<CfgModelNode> moduleParents = module.getParents();
      for (final CfgModelNode parent : moduleParents) {
        final List<CfgModelNode> newParentModules = getOldNewMap().get(parent);
        moduleNode.addParent(newParentModules.get(0));
      }
    }
  }

  @Override
  public void onProcessBegin(final Process process) {
    final List<CfgModelNode> newProcesses = getOldNewMap().get(process);
    final Process newProcess = (Process) newProcesses.get(0);
    addParentsToNode(process, newProcess);
    this.process = newProcess;
  }

  @Override
  public void onBasicBlockBegin(final BasicBlock basicBlock) {
    final List<CfgModelNode> newBlocks = getOldNewMap().get(basicBlock);
    final BasicBlock blockNode = (BasicBlock) newBlocks.get(0);
    addParentsToNode(basicBlock, blockNode);
  }

  @Override
  public void onCaseBegin(final Case caseNode) {
    final List<CfgModelNode> newCases = getOldNewMap().get(caseNode);
    final CfgModelNode newCase = newCases.get(0);
    addParentsToNode(caseNode, newCase);
  }

  @Override
  public void onSwitchBegin(final Switch switchNode) {
    final List<CfgModelNode> switchCfg = getOldNewMap().get(switchNode);

    if (CgaaEfsmTransformerUtils.containsSingleObject(switchCfg)
        && switchCfg.get(0) instanceof Switch) {

      addParentsToNode(switchNode, switchCfg.get(0));

    } else if (switchCfg.size() > 1) {

      /* This switch was previously transformed into sub-graph. */

      final Source cfgSource = (Source) CfgUtils.getNode(switchCfg, CfgNodeType.SOURCE);

      if (cfgSource == null) {
        throw new IllegalStateException("No " + CfgNodeType.SOURCE + " has been found.");
      }

      /* Get "root" switch in the sub-graph. */
      final CfgModelNode mainSwitch = cfgSource.getOnlyChild();

      /* Re-link "root" switch to main graph parent node. */
      final List<CfgModelNode> newSwitchParents = getOldNewMap().get(switchNode.getOnlyParent());
      mainSwitch.removeParent(cfgSource);
      mainSwitch.addParent(newSwitchParents.get(0));

      /* Get sub-graph nodes. */

      /* Add non-source & non-sink nodes. */
      for (final CfgModelNode node : switchCfg) {
        final CfgNodeType type = node.getType();
        if (!type.equals(CfgNodeType.SOURCE) && !type.equals(CfgNodeType.SINK)) {
          getProcess().addNode((CfgNode) node);
        }
      }
    }
  }

  @Override
  public void onMergeBegin(final Merge merge) {
    final List<CfgModelNode> newMerges = getOldNewMap().get(merge);
    final CfgModelNode newMerge = newMerges.get(0);
    addParentsToNode(merge, newMerge);
  }

  @Override
  public void onSink(final Sink sink) {
    final List<CfgModelNode> newSinks = getOldNewMap().get(sink);
    final CfgModelNode newSink = newSinks.get(0);
    addParentsToNode(sink, newSink);
  }

  private void addParentsToNode(final CfgModelNode oldNode, final CfgModelNode newNode) {
    if (oldNode.hasParents() && !newNode.hasParents()) {
      final Collection<CfgModelNode> parents = oldNode.getParents();
      for (final CfgModelNode parent : parents) {
        final List<CfgModelNode> newParents = getOldNewMap().get(parent);
        final CfgModelNode newParent = newParents.get(0);
        newParent.addChild(newNode);
      }
    }
    if (oldNode instanceof CfgNode) {
      getProcess().addNode((CfgNode) newNode);
    }
  }
}
