/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.VariableContainer;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * Component that detects state-like variables in clocked guarded atomic actions model.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaStateDetector extends MapCfgVisitor {

  /* The control part of current path. */
  private List<Switch> controlPath;

  /* The data part of current path. */
  private List<BasicBlock> dataPath;

  /* The switch nodes expressions. */
  private Set<Node> switchNodes;

  /* The state-like variables. */
  private List<NodeVariable> stateVariables;

  /* Names of user-defined state variables. */
  private String[] predefinedNames;

  private Status status;

  /**
   * The default constructor.
   */
  public CgaaStateDetector() {
    super();
    this.controlPath = new ArrayList<>();
    this.dataPath = new ArrayList<>();
    this.status = Status.OK;
  }

  /**
   * Sets user-defined names of state-like variables.
   *
   * @param names state-like variables names
   */
  public void setStateNames(final String[] names) {
    this.predefinedNames = names;
  }

  @Override
  public Status getStatus() {
    return this.status;
  }

  @Override
  public void onRootBegin() {
    this.controlPath = new ArrayList<>();
    this.dataPath = new ArrayList<>();
    this.stateVariables = new ArrayList<>();
  }

  @Override
  public void onRootEnd() {
    this.status = Status.OK;
  }

  @Override
  public void onModuleBegin(final Module module) {
    if (this.predefinedNames != null && this.predefinedNames.length > 0) {
      final Set<NodeVariable> states = new LinkedHashSet<>();

      // TODO: process can also declare variables
      for (final String name : this.predefinedNames) {
        final VariableContainer variableContainer = CfgUtils.getVariableContainer(module, name);

        if (variableContainer != null) {
          states.add(variableContainer.getVariable(name));
        }
      }
      this.getStateMap().put(module, states);
      this.status = Status.SKIP;

    } else {
      this.switchNodes = new LinkedHashSet<>();
    }
  }

  @Override
  public void onModuleEnd(final Module module) {

    if (this.predefinedNames == null || this.predefinedNames.length == 0) {

      /* Check that all the module switch nodes
       * that use state-like variables does not use non-state-like variables. */
      checkStateVariables();

      final Set<NodeVariable> states = new LinkedHashSet<>();
      states.addAll(this.stateVariables);
      this.getStateMap().put(module, states);
      this.stateVariables.clear();
    }
  }

  private void checkStateVariables() {
    for (final Node node : this.switchNodes) {
      final Collection<NodeVariable> nodeVariables = ExprUtils.getVariables(node);

      boolean hasState = false;
      boolean hasNonState = false;

      for (final NodeVariable variable : nodeVariables) {
        if (this.stateVariables.contains(variable)) {
          hasState = true;
        } else if (!hasNonState) {
          hasNonState = true;
        }
      }
      if (hasState && hasNonState) {
        for (final NodeVariable variable : nodeVariables) {
          if (this.stateVariables.contains(variable)) {
            this.stateVariables.remove(variable);
          }
        }
      }
    }
  }

  @Override
  public void onBasicBlockBegin(final BasicBlock basicBlock) {
    if (this.dataPath.contains(basicBlock)) {
      Log.print(LogLevel.WARNING, "Loop has been found.");
    }
    this.dataPath.add(basicBlock);
  }

  @Override
  public void onBasicBlockEnd(final BasicBlock basicBlock) {
    if (this.dataPath.contains(basicBlock)) {
      this.dataPath.remove(basicBlock);
    } else {
      throw new IllegalStateException("No such node: " + basicBlock.toString() + ".");
    }
  }

  @Override
  public void onSwitchBegin(final Switch switchNode) {
    if (this.controlPath.contains(switchNode)) {
      Log.print(LogLevel.WARNING, "Loop has been found.");
    }
    this.controlPath.add(switchNode);
    this.switchNodes.add(switchNode.getCondition());
  }

  @Override
  public void onSwitchEnd(final Switch switchNode) {
    if (this.controlPath.contains(switchNode)) {
      this.controlPath.remove(switchNode);
    } else {
      throw new IllegalStateException("No such node: " + switchNode.toString() + ".");
    }
  }

  @Override
  public void onSink(final Sink sink) {

    /* Get variables that are defined in data parts of graph paths. */
    final Set<NodeVariable> definedDataVars = getDefinedVariables(this.dataPath);

    /* Get variables that are used in control parts of graph paths. */
    final List<NodeVariable> usedControlVars = getUsedVariables(this.controlPath);

     /* The weak heuristic is applied. */
    for (final NodeVariable var : definedDataVars) {

      if (var.getUserData() != null && var.getUserData() instanceof VariableData) {

        final VariableType variableType = ((VariableData) var.getUserData()).getVariableType();
        if (!VariableType.isSignal(variableType)
            && usedControlVars.contains(var)
            &&  !this.stateVariables.contains(var)) {
          this.stateVariables.add(var);
        }
      }
    }
  }

  private Set<NodeVariable> getDefinedVariables(final List<BasicBlock> dataPath) {

    /* Get those defined variables that do not depend on input signals. */
    final Set<NodeVariable> defined = new LinkedHashSet<>();

    final Set<NodeVariable> forbidden = new LinkedHashSet<>();

    for (final BasicBlock block : dataPath) {
      final List<Assignment> assignments = block.getAssignments();

      for (final Assignment assignment : assignments) {
        final Collection<NodeVariable> stmtUses = assignment.getUses();

        final Set<NodeVariable> defines = assignment.getDefines();
        if (areSignalDependent(stmtUses)) {

          /* Variables that are defined here, cannot be returned. */
          forbidden.addAll(defines);
        } else {
          defined.addAll(CollectionUtils.complementSets(defines, forbidden));
        }
      }

      final Collection<NodeVariable> blockDefines = block.getDefines();
      for (final NodeVariable blockDefined : blockDefines) {
        if (!forbidden.contains(blockDefined)) {
          defined.add(blockDefined);
        }
      }
    }
    return defined;
  }

  private boolean areSignalDependent(final Collection<NodeVariable> stmtUses) {
    boolean dependOnSignal = false;
    for (final NodeVariable use : stmtUses) {
      if (VariableType.isInputSignal(((VariableData) use.getUserData()).getVariableType())) {
        dependOnSignal = true;
        break;
      }
    }
    return dependOnSignal;
  }

  private List<NodeVariable> getUsedVariables(final List<Switch> controlPath) {
    final List<NodeVariable> used = new ArrayList<>();

    for (final Switch node : controlPath) {
      final Collection<NodeVariable> nodeUses = node.getUses();
      for (final NodeVariable nodeUse : nodeUses) {
        if (!used.contains(nodeUse)) {
          used.add(nodeUse);
        }
      }
    }
    return used;
  }
}
