/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Component that keeps some statistics about clocked guarded atomic actions model.
 *
 * <p>The visitor for {@link ru.ispras.retrascope.model.cgaa.CgaaModel}
 * that collects statistical information about state-like variables.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaStatisticsKeeper extends MapCfgVisitor {

  /* The current module. */
  private Module currentModule;

  /* The control part of current CFG path. */
  private Map<Case, Node> controlPath;

  /* The data part of current CFG path. */
  private List<BasicBlock> dataPath;

  /* The mapping for keeping the statistical info about state-like variables. */
  private Map<NodeVariable, VarStatInfo> varStatInfoMap;

  /* Class for keeping the statistical information. */
  static class VarStatInfo {

    /* The number of all process paths. */
    private int allPathNum;

    /* The number of paths in which the variable is used. */
    private int usePathNum;

    /* The number of paths' control parts in which the variable is used. */
    private int controlUsePathNum;

    /* The number of paths' data parts in which the variable is used. */
    private int dataUsePathNum;

    /* The number of paths in which the variable is defined. */
    private int defPathNum;

    /* The number of paths in which the variable is both used and defined. */
    private int useDefPathNum;

    public VarStatInfo() {

      this.allPathNum = 0;
      this.usePathNum = 0;
      this.controlUsePathNum = 0;
      this.dataUsePathNum = 0;
      this.defPathNum = 0;
      this.useDefPathNum = 0;
    }

    public int getAllPathNum() {
      return this.allPathNum;
    }

    public int getUsePathNum() {
      return this.usePathNum;
    }

    public int getControlUsePathNum() {
      return this.controlUsePathNum;
    }

    public int getDataUsePathNum() {
      return this.dataUsePathNum;
    }

    public int getDefPathNum() {
      return this.defPathNum;
    }

    public int getUseDefPathNum() {
      return this.useDefPathNum;
    }

    public void incTotalPathNum() {
      this.allPathNum ++;
    }

    public void incUsePathNum() {
      this.usePathNum ++;
    }

    public void incControlUsePathNum() {
      this.controlUsePathNum++;
    }

    public void incDataUsePathNum() {
      this.dataUsePathNum++;
    }

    public void incDefPathNum() {
      this.defPathNum ++;
    }

    public void incUseDefPathNum() {
      this.useDefPathNum ++;
    }

    @Override
    public String toString() {
      return "{all-path-num=" + getAllPathNum()
          + ", use-path-num=" + getUsePathNum()
          + ", control-use-path-num=" + getControlUsePathNum()
          + ", data-use-path-num=" + getDataUsePathNum()
          + ", def-path-num=" + getDefPathNum()
          + ", def-use-path-num=" + getUseDefPathNum() + "}";
    }
  }

  @Override
  public void onRootBegin() {
    this.controlPath = new LinkedHashMap<>();
    this.dataPath = new ArrayList<>();
  }

  @Override
  public void onModuleBegin(final Module module) {
    this.currentModule = module;
  }

  @Override
  public void onProcessBegin(final Process process) {
    this.varStatInfoMap = new LinkedHashMap<>();
  }

  @Override
  public void onProcessEnd(final Process process) {
    Log.print(LogLevel.DEBUG, this.varStatInfoMap.toString());
  }

  @Override
  public void onBasicBlockBegin(final BasicBlock basicBlock) {
    this.dataPath.add(basicBlock);
  }

  @Override
  public void onBasicBlockEnd(final BasicBlock basicBlock) {
    this.dataPath.remove(basicBlock);
  }

  @Override
  public void onCaseBegin(final Case caseNode) {
    final Switch parent = (Switch) caseNode.getOnlyParent();
    final Node expr = CgaaEfsmTransformerUtils.getExpression(caseNode, parent);
    this.controlPath.put(caseNode, expr);
  }

  @Override
  public void onCaseEnd(final Case caseNode) {
    this.controlPath.remove(caseNode);
  }

  @Override
  public void onSink(final Sink sink) {

    /* Get state-like variables of the current module. */
    final Set<NodeVariable> stateVariables = this.getStateMap().get(this.currentModule);

    /* The collection of variables that are used in the control part of current path. */
    final Set<NodeVariable> controlUsed = new LinkedHashSet<>();

    final Collection<Node> controlNodes = this.controlPath.values();

    for (final Node expr : controlNodes) {
      final Collection<NodeVariable> usedVars = ExprUtils.getVariables(expr);
      controlUsed.addAll(usedVars);
    }

    final Set<NodeVariable> dataUsed = new LinkedHashSet<>();
    final Set<NodeVariable> dataDefined = new LinkedHashSet<>();

    for (final BasicBlock block : this.dataPath) {
      dataUsed.addAll(block.getUses());
      dataDefined.addAll(block.getDefines());
    }

    for (final NodeVariable state : stateVariables) {

      if (!this.varStatInfoMap.containsKey(state)) {
        this.varStatInfoMap.put(state, new VarStatInfo());
      }

      final VarStatInfo statInfo = this.varStatInfoMap.get(state);

      final boolean usedInControl = controlUsed.contains(state);
      final boolean usedInData = dataUsed.contains(state);

      if (usedInControl || usedInData) {
        if (usedInControl) {
          statInfo.incControlUsePathNum();
        }

        if (dataUsed.contains(state)) {
          statInfo.incDataUsePathNum();
        }
        statInfo.incUsePathNum();
      }

      if (dataDefined.contains(state)) {
        statInfo.incDefPathNum();
        if (usedInControl || usedInData) {
          statInfo.incUseDefPathNum();
        }
      }

      statInfo.incTotalPathNum();
    }
  }
}
