/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.expression.ExprTreeVisitor;
import ru.ispras.fortress.expression.ExprTreeVisitorDefault;
import ru.ispras.fortress.expression.ExprTreeWalker;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * Component that creates rules of switch nodes splitting in clocked guarded atomic actions model.
 *
 * <p>Visitor for splitting switch nodes, that are depend on state-like variables
 * and can be split into many-nodes-containing sub-graph. Such splitting
 * can be done when top-level operation identifier in switch node's expression
 * is of Boolean type (like NOT, OR, AND, IML, XOR). If the switch node's expression
 * also contains internal Boolean identifiers, the splitting procedure will be applied
 * to them also only if one of their sub-expressions depends on state-like variable.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaSwitchSplitter extends MapCfgVisitor {

  /* The module to be processed. */
  private Module current;

  @Override
  public void onModuleBegin(final Module module) {

    this.current = module;

    /* Make a copy of module. */
    final Module newModule = new Module(module.getName());
    newModule.setVariablesMapping(module.getVariablesMapping());

    getOldNewMap().put(module, Collections.<CfgModelNode>singletonList(newModule));
  }

  @Override
  public void onProcessBegin(final Process process) {

    /* Make a copy of process. */
    final Process newProcess = new Process(process.isInitial());
    newProcess.setSensitivityList(process.getSensitivityList());

    /* Make copies of SOURCE/SINK internal nodes. */
    final Source newSource = new Source();
    final Sink newSink = new Sink();

    newProcess.addChild(newSource);
    newProcess.addNode(newSource);
    newProcess.addNode(newSink);

    getOldNewMap().put(
        CfgUtils.getSource(process.getNodes()),
        Collections.<CfgModelNode>singletonList(newSource));

    getOldNewMap().put(
        CfgUtils.getSink(process.getNodes()),
        Collections.<CfgModelNode>singletonList(newSink));

    getOldNewMap().put(process, Collections.<CfgModelNode>singletonList(newProcess));
  }

  @Override
  public void onBasicBlockBegin(final BasicBlock basicBlock) {

    final BasicBlock newBlock =
        new BasicBlock(basicBlock.getAssignments(), basicBlock.isConcurrent());
    getOldNewMap().put(basicBlock, Collections.<CfgModelNode>singletonList(newBlock));
  }

  @Override
  public void onSwitchBegin(final Switch switchNode) {

    final List<CfgModelNode> subGraph = splitIntoSubGraph(switchNode);

    if (CgaaEfsmTransformerUtils.containsSingleObject(subGraph)
        && subGraph.get(0) instanceof Switch
        && !subGraph.get(0).hasChildren()) {

      throw new IllegalStateException(
          "Can't create correct sub-graph for " + switchNode.toString() + ".");
    }

    getOldNewMap().put(switchNode, subGraph);
  }

  private List<CfgModelNode> splitIntoSubGraph(final Switch switchNode) {

    List<CfgModelNode> subGraph;
    final Node expr = switchNode.getCondition();
    final Collection<CfgModelNode> sChildren = switchNode.getChildren();

    if (isSplittable(expr)) {

      subGraph = new ArrayList<>();

      /* Every control flow graph contains linked source and sink - delete that link. */
      final Source cfgSource = new Source();
      final Sink cfgSink = new Sink();
      subGraph.add(cfgSource);
      subGraph.add(cfgSink);

      /* Switch expression is boolean
      * so this node has only two case children
      * with 'true'/'false' values.
      * These children are transformed into merge nodes.*/
      Merge trueMerge = null;
      Merge falseMerge = null;
      for (final CfgModelNode child : sChildren) {
        final Collection<NodeValue> childValues = ((Case) child).getValues();
        if (CgaaEfsmTransformerUtils.containsSingleObject(childValues)) {
          final NodeValue cValue = childValues.iterator().next();
          if (cValue.getData().equals(Data.newBoolean(true))) {
            trueMerge = new Merge();
            getOldNewMap().put(child, Collections.<CfgModelNode>singletonList(trueMerge));
          } else if (cValue.getData().equals(Data.newBoolean(false))) {
            falseMerge = new Merge();
            getOldNewMap().put(child, Collections.<CfgModelNode>singletonList(falseMerge));
          }
        } else {
          throw new IllegalStateException(
              "Able-to-be-split switch cannot be processed: multiple values at child case: "
              + childValues.toString());
        }
      }

      /* Check that merge nodes a linked with cases. */
      if (trueMerge == null || falseMerge == null) {
        throw new IllegalStateException("Able-to-be-split switch cannot be processed: " + expr);
      }

      subGraph.add(falseMerge);
      subGraph.add(trueMerge);

      makeSubGraph((NodeOperation) expr, subGraph, cfgSource, trueMerge, falseMerge);
    } else {
      if (switchNode.getEvent() == null) {
        subGraph = new ArrayList<>();
        subGraph.add(new Switch(expr));
      } else {
        subGraph = new ArrayList<>();
        subGraph.add(new Switch(switchNode.getEvent()));
      }

      /* All child nodes are not to be transformed. */
      for (final CfgModelNode child : sChildren) {
        final Case newCase = new Case(((Case) child).getValues());

        if (CgaaEfsmTransformerUtils.containsSingleObject(subGraph)) {
          subGraph.get(0).addChild(newCase);
        } else {
          throw new IllegalStateException("Non-one-element sub-graph.");
        }
        getOldNewMap().put(child, Collections.<CfgModelNode>singletonList(newCase));
      }
    }
    return subGraph;
  }

  private void makeSubGraph(
      final NodeOperation expr,
      final List<CfgModelNode> list,
      final CfgNode root,
      final Merge trueMerge,
      final Merge falseMerge) {

    final int opCount = expr.getOperandCount();

    if (opCount == 1) {

     /* Unary operation. */
      makeUnaryOpSubGraph(expr, list, root, trueMerge, falseMerge);
    } else if (opCount == 2) {

      /* Binary operation. */
      makeBinOpSubGraph(expr, list, root, trueMerge, falseMerge);
    } else {

      /* No additional elaboration is needed here. */
      makeAtomicSubGraph(expr, list, root, trueMerge, falseMerge);
    }
  }

  private void makeUnaryOpSubGraph(
      final NodeOperation expr,
      final List<CfgModelNode> list,
      final CfgNode root,
      final Merge trueMerge,
      final Merge falseMerge) {

    final Enum<?> opId = expr.getOperationId();
    final Node operand = expr.getOperand(0);

    if (isSplittable(operand)) {
      if (opId == StandardOperation.NOT) {
        makeSubGraph((NodeOperation) operand, list, root, falseMerge, trueMerge);
      }
    } else {
      makeAtomicSubGraph(operand, list, root, trueMerge, falseMerge);
    }
  }

  private void makeBinOpSubGraph(
      final NodeOperation expr,
      final List<CfgModelNode> list,
      final CfgNode root,
      final Merge trueMerge,
      final Merge falseMerge) {

    /*
     * The constructed sub-graph has the following structure:
     *
     *                       firstSwitch
     *                      |           |
     *                  tCase         fCase
     *                   |               |
     *         trueSecondSwitch      falseSecondSwitch
     *         |               |     |                |
     *      ttCase           tfCase ftCase          ffCase
     *      |                  |     |                |
     *      <--- link layer that is operation specific>
     *        trueMerge                   falseMerge
     */

    final Enum<?> opId = expr.getOperationId();
    final Node first = expr.getOperand(0);
    final Node second = expr.getOperand(1);

    CfgNode trueSecondRoot;
    CfgNode falseSecondRoot;

    if (isSplittable(first)) {

      /* First operand of binary operation to be split. */
      final Merge firstTrueMerge = new Merge();
      trueSecondRoot = firstTrueMerge;
      list.add(firstTrueMerge);

      final Merge firstFalseMerge = new Merge();
      falseSecondRoot = firstFalseMerge;
      list.add(firstFalseMerge);

      makeSubGraph((NodeOperation) first, list, root, firstTrueMerge, firstFalseMerge);
    } else {

      /* First operand of binary operation cannot be split. */
      final Switch newSwitch = new Switch(first);
      list.add(newSwitch);
      root.addChild(newSwitch);

      final Case tCase = new Case(NodeValue.newBoolean(true));
      trueSecondRoot = tCase;
      newSwitch.addChild(tCase);
      list.add(tCase);

      final Case fCase = new Case(NodeValue.newBoolean(false));
      falseSecondRoot = fCase;
      newSwitch.addChild(fCase);
      list.add(fCase);
    }

    CfgNode ttNode;
    CfgNode tfNode;
    CfgNode ftNode;
    CfgNode ffNode;
    if (isSplittable(second)) {

      /* Second operand of binary operation to be split. */
      final Merge ttMerge = new Merge();
      ttNode = ttMerge;
      list.add(ttMerge);

      final Merge tfMerge = new Merge();
      tfNode = tfMerge;
      list.add(tfMerge);

      final Merge ftMerge = new Merge();
      ftNode = ftMerge;
      list.add(ftMerge);

      final Merge ffMerge = new Merge();
      ffNode = ffMerge;
      list.add(ffMerge);

      makeSubGraph((NodeOperation) second, list, trueSecondRoot, ttMerge, tfMerge);
      makeSubGraph((NodeOperation) second, list, falseSecondRoot, ftMerge, ffMerge);
    } else {

      /* Second operand of binary operation cannot be split. */
      final Switch trueSecondSwitch = new Switch(second);
      list.add(trueSecondSwitch);
      trueSecondSwitch.addParent(trueSecondRoot);

      final Switch falseSecondSwitch = new Switch(second);
      list.add(falseSecondSwitch);
      falseSecondSwitch.addParent(falseSecondRoot);

      final Case ttCase = new Case(NodeValue.newBoolean(true));
      ttNode = ttCase;
      list.add(ttCase);
      trueSecondSwitch.addChild(ttCase);

      final Case tfCase = new Case(NodeValue.newBoolean(false));
      tfNode = tfCase;
      list.add(tfCase);
      trueSecondSwitch.addChild(tfCase);

      final Case ftCase = new Case(NodeValue.newBoolean(true));
      ftNode = ftCase;
      list.add(ftCase);
      falseSecondSwitch.addChild(ftCase);

      final Case ffCase = new Case(NodeValue.newBoolean(false));
      ffNode = ffCase;
      list.add(ffCase);
      falseSecondSwitch.addChild(ffCase);
    }
    /* The operation-id dependent link layer. */
    if (opId == StandardOperation.AND) {
      ttNode.addChild(trueMerge);
      tfNode.addChild(falseMerge);
      ftNode.addChild(falseMerge);
      ffNode.addChild(falseMerge);
    } else if (opId == StandardOperation.OR) {
      ttNode.addChild(trueMerge);
      tfNode.addChild(trueMerge);
      ftNode.addChild(trueMerge);
      ffNode.addChild(falseMerge);
    } else if (opId == StandardOperation.XOR) {
      ttNode.addChild(falseMerge);
      tfNode.addChild(trueMerge);
      ftNode.addChild(trueMerge);
      ffNode.addChild(falseMerge);
    } else if (opId == StandardOperation.IMPL) {
      ttNode.addChild(trueMerge);
      tfNode.addChild(falseMerge);
      ftNode.addChild(trueMerge);
      ffNode.addChild(trueMerge);
    } else {
      throw new IllegalArgumentException("Unexpected operation: " + opId + ".");
    }
  }

  private void makeAtomicSubGraph(
      final Node expr,
      final List<CfgModelNode> list,
      final CfgNode root,
      final Merge trueMerge,
      final Merge falseMerge) {

    final Switch newSwitch = new Switch(expr);
    root.addChild(newSwitch);
    list.add(newSwitch);

    final Case trueCase = new Case(NodeValue.newBoolean(true));
    newSwitch.addChild(trueCase);
    trueCase.addChild(trueMerge);
    list.add(trueCase);

    final Case falseCase = new Case(NodeValue.newBoolean(false));
    newSwitch.addChild(falseCase);
    falseCase.addChild(falseMerge);
    list.add(falseCase);
  }

  private boolean isSplittable(final Node expr) {

    if (expr instanceof NodeOperation) {
      final Set<NodeVariable> stateVars = this.getStateMap().get(this.current);

      final ExprTreeVisitor visitor = new ExprTreeVisitorDefault() {

        @Override
        public void onOperationBegin(final NodeOperation op) {
          final Collection<NodeVariable> variables = ExprUtils.getVariables(op);

          final Set<NodeVariable> varSet = new LinkedHashSet<>(variables);

          if (CgaaEfsmTransformerUtils.containsSingleObject(variables)
              && CollectionUtils.areIntersectedSets(varSet, stateVars)) {

            setStatus(Status.ABORT);
          }
        }
      };

      final ExprTreeWalker walker = new ExprTreeWalker(visitor);
      walker.visit(expr);
      return visitor.getStatus().equals(ExprTreeVisitor.Status.ABORT)
          && isSplittable(((NodeOperation) expr).getOperationId());
    } else {
      return false;
    }
  }

  private boolean isSplittable(final Enum<?> opId) {
    return opId == StandardOperation.NOT
        || opId == StandardOperation.AND
        || opId == StandardOperation.OR
        || opId == StandardOperation.XOR
        || opId == StandardOperation.IMPL;
  }

  @Override
  public void onMergeBegin(final Merge merge) {
    final Merge newMerge = new Merge();
    getOldNewMap().put(merge, Collections.<CfgModelNode>singletonList(newMerge));
  }

  @Override
  public void onSink(final Sink sink) {
    final Sink newSink = new Sink();
    getOldNewMap().put(sink, Collections.<CfgModelNode>singletonList(newSink));
  }
}
