/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.extractor.conflict;

import org.apache.commons.cli.ParseException;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.engine.efsm.extractor.conflict.searcher.EfsmBfs;
import ru.ispras.retrascope.engine.efsm.extractor.conflict.searcher.EfsmDfs;
import ru.ispras.retrascope.engine.efsm.extractor.conflict.searcher.EfsmPathSearcher;
import ru.ispras.retrascope.engine.efsm.extractor.conflict.searcher.SearchCondition;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmWalker;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.result.conflict.Conflict;
import ru.ispras.retrascope.result.conflict.ConflictModel;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * This engine is used to detect possible conflicts among specified EFSMs.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class EfsmConflictExtractor extends Engine {

  /* The engine identifier. */
  public static final String ENGINE_ID = "efsm-conflict-extractor";

  private static final Parameter SEARCH_TYPE = new Parameter("search-type", new String[] {"BFS",
      "DFS"}, "Type of path search algorithm");
  private static final Parameter SEARCH_DEPTH = new Parameter("search-depth", true,
      "Algorithm search depth");

  private EfsmPathSearcher searcher;

  private static final String LOG_HEADER = "EFSM Conflict Extractor: ";

  /**
   * The default constructor.
   */
  public EfsmConflictExtractor() {
    super(ENGINE_ID);

    addInputType(EfsmModel.TYPE);
    setOutputType(ConflictModel.TYPE);

    addParameter(SEARCH_TYPE);
    addParameter(SEARCH_DEPTH);

  }

  private class StopCondition implements SearchCondition {

    public Efsm efsm;
    public int depth;

    @Override
    public boolean conditionMet(Object obj) {
      if ((obj != null) && (obj.getClass() == EfsmState.class)) {
        EfsmState state = (EfsmState) obj;
        return state == efsm.getInitialState();
      } else {
        throw new IllegalArgumentException("The specified object is not a state!");
      }
    }

    @Override
    public boolean limitReached(int depth) {
      return depth >= this.depth;
    }
  }

  @Override
  public Entity start(Map<EntityType, Entity> inputs) {

    final EfsmModel efsmModel = (EfsmModel) inputs.get(EfsmModel.TYPE);

    final ConflictModel conflicts = new ConflictModel();

    // Default value.
    int searchDepth = 10;

    StopCondition condition = new StopCondition();

    Log.print(LogLevel.INFO, LOG_HEADER + "Starting");

    try {

      final Arguments arguments =
          this.parseCommandLine(((Arguments) inputs.get(Arguments.TYPE)).getCommandLine());

      if (arguments.contains(SEARCH_TYPE)) {

        final String value = arguments.getValue(SEARCH_TYPE).toUpperCase();

        switch (value) {
          case "BFS":
            searcher = new EfsmBfs();
            break;
          case "DFS":
            searcher = new EfsmDfs();
            break;
          default:
            Log.print(LogLevel.WARNING, LOG_HEADER + "Using default search type: BFS.");
            break;
        }

      } else {

        Log.print(LogLevel.WARNING, LOG_HEADER + "Using default search type: BFS.");
        searcher = new EfsmBfs();

      }

      if (arguments.contains(SEARCH_DEPTH)) {
        searchDepth = Integer.parseInt(arguments.getValue(SEARCH_DEPTH));
        if (searchDepth <= 0) {
          Log.print(LogLevel.WARNING, LOG_HEADER
              + "Search depth can't be less or equal 0! Using default (10).");
        }
      } else {
        Log.print(LogLevel.WARNING, LOG_HEADER + "Using default search depth: 10.");
      }

    } catch (ParseException | NumberFormatException e1) {
      Log.getLogger().log(LogLevel.SEVERE, LOG_HEADER, e1);
    }

    condition.depth = searchDepth;

    List<Efsm> efsmList = efsmModel.getEfsmList();

    if (efsmList.size() > 1) {
      conflicts.addConflicts(searchConflicts(searcher, condition, efsmList));
    }

    if (conflicts.getConflicts().isEmpty()) {
      Log.print(LogLevel.INFO, LOG_HEADER + "No conflicts found!");
    }

    return conflicts;
  }

  /**
   * Searches for conflicts in the specified Efsm list.
   * 
   * @param searcher - a {@link EfsmPathSearcher} that performs a search;
   * @param condition - a {@link SearchCondition} that is required to stop the search;
   * @param efsmList - a list of EFSMs.
   * @return The list of conflicts.
   */
  private List<Conflict> searchConflicts(EfsmPathSearcher searcher, StopCondition condition,
      List<Efsm> efsmList) {

    List<Conflict> result = new LinkedList<>();

    // Combining EFSMs.
    for (int i = 0; i < efsmList.size() - 1; i++) {
      for (int j = i + 1; j < efsmList.size(); j++) {

        Log.print(LogLevel.INFO, LOG_HEADER + "Examining: " + efsmList.get(i).getName() + ", "
            + efsmList.get(j).getName());

        EfsmConflictVisitor conflictVisitor = new EfsmConflictVisitor();
        EfsmWalker walker = new EfsmWalker(conflictVisitor);
        Map<NodeVariable, List<EfsmTransition>> variables;
        LinkedList<GuardedAction> path;

        // Making product of two EFSMs.
        Efsm product = Efsm.product(efsmList.get(i), efsmList.get(j));

        condition.efsm = product;

        // Searching for variables which are defined more than once in one action block.
        walker.start(product);
        variables = conflictVisitor.getDuplicatingVariables();

        Log.print(LogLevel.INFO, LOG_HEADER + "Duplicating variables: " + variables.size());
        for (NodeVariable var : variables.keySet()) {
          Log.print(LogLevel.INFO, LOG_HEADER + var.getName());
          Log.print(LogLevel.INFO, LOG_HEADER + "Transitions: " + variables.get(var).size());
        }

        // Combining variables and searching paths to the initial state of the product EFSM.
        for (NodeVariable variable : variables.keySet()) {
          for (EfsmTransition transition : variables.get(variable)) {
            searcher.init(product, transition, condition);
            while (searcher.hasNext()) {
              searcher.next();

              // Storing a conflict if the path exists.
              if (searcher.hasValue()) {
                path = (LinkedList<GuardedAction>) searcher.value();
                result.add(new Conflict(efsmList.get(i), efsmList.get(j), variable,
                    Conflict.Types.WAW, path));
              }
            }
          }
        }
      }
    }
    return result;
  }
}
