/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.extractor.conflict.searcher;

import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.GuardedAction;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Stack;

/**
 * This class implements the Depth-First Search algorithm.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class EfsmDfs implements EfsmPathSearcher {

  // The stack of elements to examine.
  private Stack<TreeElement> stack;

  // The stack of visited elements.
  private Stack<TreeElement> visited;

  // Shows whether the search is over.
  private boolean isOver;

  // The result of algorithm execution.
  // Represents the path from the specified transition to the initial state of the EFSM.
  private LinkedList<GuardedAction> path;

  // The examined EFSM.
  private Efsm efsm;

  // The condition which stops and limits the search.
  private SearchCondition condition;

  /**
   * The default constructor.
   */
  public EfsmDfs() {}

  @Override
  public void init(
      final Efsm efsm,
      final EfsmTransition startTransition,
      final SearchCondition condition) {

    isOver = false;
    path = new LinkedList<>();
    stack = new Stack<>();
    visited = new Stack<>();
    this.efsm = efsm;
    this.condition = condition;

    // Adding the first element of the tree.
    stack.push(new TreeElement(startTransition, startTransition.getGuardedAction().getGuard()
        .toNode()));
  }

  @Override
  public void next() {

    path = new LinkedList<>();

    while (!stack.isEmpty()) {

      // Getting the first element from the stack.
      TreeElement current = stack.pop();
      EfsmTransition currentTransition = current.getTransition();

      // Pushing the element into the visited stack.
      visited.push(current);

      // Checking whether the initial state is reached.
      if (condition.conditionMet(currentTransition.getSourceState())) {

        // Constructing the resulting path.
        path.addFirst(currentTransition.getGuardedAction());
        TreeElement tmp = current.getParent();
        while (tmp != null) {
          path.addFirst(tmp.getTransition().getGuardedAction());
          tmp = tmp.getParent();
        }
        return;
      }

      // Checking whether the limit is reached.
      if (condition.limitReached(current.getLevel())) {
        return;
      }

      // Obtaining the next elements of the tree.
      Collection<TreeElement> nextTransitions = TreeElement.getSatTransitions(current, efsm);

      if (!nextTransitions.isEmpty()) {
        for (TreeElement transition : nextTransitions) {
          if (!visited.contains(transition)) {
            transition.setParent(current);
            transition.setLevel(current.getLevel() + 1);

            // Pushing the element into the stack.
            stack.push(transition);
          }
        }
      }
    }

    isOver = true;

  }

  @Override
  public boolean hasNext() {
    return !isOver;
  }

  @Override
  public boolean hasValue() {
    return path != null && !path.isEmpty();
  }

  @Override
  public List<GuardedAction> value() {
    return path;
  }

}
