/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.extractor.conflict.searcher;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmTransition;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

/**
 * This class represents the element of the {@link Efsm} search tree. It contains the
 * {@link EfsmTransition} to explore, the {@link Node} expression which represents the path, the
 * link to the parent element and the level of the element in the tree.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
final class TreeElement {

  private EfsmTransition transition; // TODO: add comment here
  private TreeElement parent; // TODO: add comment here
  private Node node; // TODO: add comment here
  private int level; // TODO: add comment here

  /**
   * Constructs a new element.
   * 
   * @param transition - the specified {@link EfsmTransition};
   * @param parent - the specified parent {@link TreeElement};
   * @param node - the specified {@link Node} expression.
   */
  public TreeElement(EfsmTransition transition, TreeElement parent, Node node) {
    this.transition = transition;
    this.parent = parent;
    this.node = node;
    this.level = 0;
  }

  /**
   * Constructs a new element.
   * 
   * @param transition - the specified {@link EfsmTransition};
   * @param node - the specified {@link Node} expression.
   */
  public TreeElement(EfsmTransition transition, Node node) {
    this.transition = transition;
    this.node = node;
    this.level = 0;
  }

  /**
   * Returns the transition associated with the element.
   * 
   * @return The transition.
   */
  public EfsmTransition getTransition() {
    return transition;
  }

  /**
   * Returns the parent element of the element.
   * 
   * @return The parent {@link TreeElement}.
   */
  public TreeElement getParent() {
    return parent;
  }

  /**
   * Sets the parent element of the element.
   * 
   * @param element - the parent {@link TreeElement}.
   */
  public void setParent(TreeElement element) {
    this.parent = element;
  }

  /**
   * Returns the {@link Node} associated with the element.
   * 
   * @return The {@link Node} expression.
   */
  public Node getNode() {
    return node;
  }

  /**
   * Returns the level of the element.
   * 
   * @return The level.
   */
  public int getLevel() {
    return level;
  }

  /**
   * Sets the level of the element.
   * 
   * @param level - the specified level.
   */
  public void setLevel(int level) {
    this.level = level;
  }

  /**
   * Searches for previous transitions that lead to the same state where the specified transition
   * begins. Performs a backward substitution of previous assignments into specified transition
   * guard and checks its satisfiability.
   * 
   * @param element - {@link TreeElement} containing the specified transition and its guard.
   * @param efsm - {@link Efsm} containing the specified transition.
   * @return A {@link List} of {@link TreeElement} containing transitions that have been proved SAT
   *         after backward substitutions.
   */
  public static List<TreeElement> getSatTransitions(TreeElement element, Efsm efsm) {

    EfsmTransition transition = element.getTransition();
    Node expr = element.getNode();

    List<TreeElement> resultingList = new LinkedList<>();

    for (EfsmTransition item : efsm.getIncomingTransitions(transition.getSourceState())) {
      Node tmpExpr = expr;

      if (item.getGuardedAction().getAction() != null) {
        Action action = item.getGuardedAction().getAction();

        HashMap<NodeVariable, NodeVariable> inputVars = new HashMap<>();

        // Searching and labeling input variables in Uses.
        for (NodeVariable usedVar : action.getUses()) {
          VariableData data = (VariableData) usedVar.getUserData();
          if ((data.getVariableType() == VariableType.IN)
              && (!inputVars.keySet().contains(usedVar))) {
            inputVars.put(
                usedVar,
                new NodeVariable(
                    usedVar.getName() + "_" + element.getLevel(),
                    usedVar.getDataType()));
          }
        }

        for (Assignment statement : action.getAssignments()) {
          Node assignExpr = statement.getValue();

          // Substituting input variables in the assignments with labeled variables.
          for (NodeVariable var : ExprUtils.getVariables(assignExpr)) {
            if (inputVars.keySet().contains(var)) {
              assignExpr = Transformer.substitute(assignExpr, var.getName(), inputVars.get(var));
            }
          }

          // Substituting current guard variables according to the previous assignments.
          RangedVariable atomic = statement.getTarget();
          tmpExpr = Transformer.substitute(tmpExpr, atomic.getVariable().getName(), assignExpr);
        }
      }

      // Checking compatibility of current and previous guards.
      if (ExprUtils.areCompatible(tmpExpr, item.getGuardedAction().getGuard().toNode())) {

        // If SAT, add to the resulting list.
        resultingList.add(new TreeElement(item, ExprUtils.getConjunction(tmpExpr, item
            .getGuardedAction().getGuard().toNode())));
      }
    }
    return resultingList;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj == null) {
      return false;
    }
    if (getClass() != obj.getClass()) {
      return false;
    }
    TreeElement other = (TreeElement) obj;

    return this.transition == other.transition
        && this.node == other.node
        && this.parent == other.parent
        && this.level == other.level;
  }

  @Override
  public int hashCode() {
    return System.identityHashCode(this);
  }
}
