/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */
package ru.ispras.retrascope.engine.efsm.generator.test;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.Map.Entry;
import java.util.logging.Logger;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmUtils;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;
import edu.uci.ics.jung.graph.DirectedSparseMultigraph;
import edu.uci.ics.jung.graph.util.EdgeType;

/**
 * The class which performs the learning phase of the FATE algorithm.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
final class EfsmAnalyser {
  private final Logger logger;
  private final String logEntryHeader;

  private final Efsm efsm;

  private final DirectedSparseMultigraph<EfsmTransition, VariableDependency> controlDependencyGraph;
  private final DirectedSparseMultigraph<EfsmTransition, VariableDependency> dataDependencyGraph;
  private final Map<NodeVariable, Set<List<EfsmTransition>>> counters;
  private final Map<EfsmTransition, Set<List<EfsmTransition>>> definitionPaths;
  private final Map<EfsmTransition, Set<Set<List<EfsmTransition>>>> loopDefinitionPaths;

  EfsmAnalyser(Efsm efsm) {
    this.logEntryHeader = "EFSM.FATE.Analyser." + efsm.getName();
    this.logger = Log.getLogger(this.getClass());

    this.efsm = efsm;

    this.controlDependencyGraph = this.detectControlDependencies(efsm);
    this.dataDependencyGraph = this.detectDataDependencies(efsm);
    this.counters = new HashMap<>();
    this.definitionPaths = new HashMap<>();
    this.loopDefinitionPaths = new HashMap<>();
    this.detectCounters(efsm);
  }

  DirectedSparseMultigraph<EfsmTransition, VariableDependency> getControlDependencyGraph() {
    return this.controlDependencyGraph;
  }

  DirectedSparseMultigraph<EfsmTransition, VariableDependency> getDataDependencyGraph() {
    return this.dataDependencyGraph;
  }

  Map<NodeVariable, Set<List<EfsmTransition>>> getCounters() {
    return this.counters;
  }

  DirectedSparseMultigraph<EfsmTransition, VariableDependency> detectControlDependencies(Efsm efsm) {
    this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": detecting control dependencies");

    DirectedSparseMultigraph<EfsmTransition, VariableDependency> controlDependencyGraph =
        new DirectedSparseMultigraph<>();

    for (EfsmTransition transition: efsm.getTransitions()) {
      if (!controlDependencyGraph.containsVertex(transition)) {
        controlDependencyGraph.addVertex(transition);
      }

      for (NodeVariable usedVariable: transition.getGuardedAction().getUsesByGuard()) {
        String usedVariableName = usedVariable.getName();

        Object varData = usedVariable.getUserData();

        if (varData instanceof VariableData) {
          if (((VariableData)varData).getVariableType() == VariableType.REG) {
            for (EfsmTransition definingTransition:
                EfsmAnalyser.findVariableDefinition(efsm, transition, usedVariableName)) {
              if (!controlDependencyGraph.containsVertex(definingTransition)) {
                controlDependencyGraph.addVertex(definingTransition);
              }

              VariableDependency edge =
                  new VariableDependency(definingTransition, transition, usedVariable);
              controlDependencyGraph.addEdge(edge, definingTransition, transition,
                  EdgeType.DIRECTED);
            }
          }
        } else {
          throw new IllegalArgumentException(
            String.format("Incorrect data type: %s.", varData.getClass()));
        }
      }
    }

    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": control dependencies have been detected successfully");

    return controlDependencyGraph;
  }
 
  DirectedSparseMultigraph<EfsmTransition, VariableDependency> detectDataDependencies(Efsm efsm) {
    this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": detecting data dependencies");

    DirectedSparseMultigraph<EfsmTransition, VariableDependency> dataDependencyGraph =
        new DirectedSparseMultigraph<>();

    for (EfsmTransition transition : efsm.getTransitions()) {
      if (!dataDependencyGraph.containsVertex(transition)) {
        dataDependencyGraph.addVertex(transition);
      }

      for (NodeVariable usedVariable: transition.getGuardedAction().getUsesByAction()) {
        String usedVariableName = usedVariable.getName();

        Object varData = usedVariable.getUserData();

        if (varData instanceof VariableData) {
          if ((((VariableData)varData).getVariableType() == VariableType.REG)) {
            for (EfsmTransition definingTransition:
                EfsmAnalyser.findVariableDefinition(efsm, transition, usedVariableName)) {
              if (!dataDependencyGraph.containsVertex(definingTransition)) {
                dataDependencyGraph.addVertex(definingTransition);
              }

              VariableDependency edge =
                  new VariableDependency(definingTransition, transition, usedVariable);
              dataDependencyGraph.addEdge(edge, definingTransition, transition, EdgeType.DIRECTED);
            }
          }
        } else {
          throw new IllegalArgumentException(
            String.format("Incorrect data type: %s.", varData.getClass()));
        }
      }
    }

    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": data dependencies have been detected successfully");

    return dataDependencyGraph;
  }

  private static List<EfsmTransition> findVariableDefinition(Efsm efsm, EfsmTransition transition,
      String usedVariableName) {
    Set<EfsmTransition> visitedTransitions = new HashSet<>();
    Queue<EfsmTransition> queue = new LinkedList<>();
    queue.add(transition);
    List<EfsmTransition> definingTransitions = new ArrayList<>();

    while (!queue.isEmpty()) {
      EfsmTransition currentTransition = queue.poll();

      for (EfsmTransition previousTransition:
        efsm.getIncomingTransitions(currentTransition.getSourceState())) {

        if (!visitedTransitions.contains(previousTransition)) {
          boolean useDefMatching = false;

          for (NodeVariable definedVariable : previousTransition.getGuardedAction().getDefines()) {
            if (definedVariable.getName().equals(usedVariableName)) {
              useDefMatching = true;

              break;
            }
          }

          if (useDefMatching) {
            definingTransitions.add(previousTransition);
          } else {
            queue.add(previousTransition);
            visitedTransitions.add(previousTransition);
          }
        }
      }
    }

    return definingTransitions;
  }

  private void detectCounters(Efsm efsm) {
    this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": detecting counters");

    Map<NodeVariable, Set<EfsmTransition>> processedCandidates = new HashMap<>();

    for (VariableDependency dependency: this.controlDependencyGraph.getEdges()) {
      EfsmTransition definingTransition = this.controlDependencyGraph.getSource(dependency);

      if (processedCandidates.containsKey(dependency.getVariable())) {
        Set<EfsmTransition> processedTransitions = processedCandidates.get(dependency.getVariable());

        if (processedTransitions.contains(definingTransition)) {
          continue;
        } else {
          processedTransitions.add(definingTransition);
        }
      } else {
        Set<EfsmTransition> transitions = new HashSet<>();
        transitions.add(definingTransition);
        processedCandidates.put(dependency.getVariable(), transitions);
      }

      EfsmState sourceState = definingTransition.getSourceState();
      EfsmState destinationState =
          this.controlDependencyGraph.getDest(dependency).getDestinationState();
      Set<NodeVariable> definingVariables =
          EfsmUtils.getResultingDependencies(definingTransition.getGuardedAction().getAction())
              .get(dependency.getVariable());

      for (Iterator<NodeVariable> iterator = definingVariables.iterator(); iterator.hasNext();) {
        if (efsm.containsInput(iterator.next().getName())) {
          iterator.remove();
        }
      }

      if (!definingVariables.isEmpty()) {
        if (efsm.isReachable(destinationState, sourceState)) {
          Set<NodeVariable> variables = new HashSet<>(1);
          variables.add(dependency.getVariable());

          for (VariableDependency dataDependency:
              this.dataDependencyGraph.getOutEdges(definingTransition)) {
            if (dataDependency.getVariable().equals(dependency.getVariable())) {
              EfsmTransition transition = this.dataDependencyGraph.getDest(dataDependency);
              this.findCounterLoop(dataDependency.getVariable(), transition,
                  definingVariables, variables, Arrays.asList(definingTransition));
            }
          }
        }
      }
    }

    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": counters have been detected successfully");
  }

  //target transition is the first in the returned list
  private void findCounterLoop(NodeVariable potentialCounter, EfsmTransition currentTransition,
      Set<NodeVariable> targetVariables, Set<NodeVariable> currentVariables,
      List<EfsmTransition> currentPath) {
    if (currentPath.contains(currentTransition)) {
      if (currentTransition.equals(currentPath.get(0))) {
        Set<NodeVariable> intersection = new HashSet<>(targetVariables);
        intersection.retainAll(currentVariables);

        if (!intersection.isEmpty()) {
          Set<List<EfsmTransition>> counterData =
              this.counters.get(potentialCounter);

          if (counterData == null) {
            counterData = new HashSet<List<EfsmTransition>>(1);
            this.counters.put(potentialCounter, counterData);
          }

          counterData.add(currentPath);

          this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": the variable " +
              potentialCounter.getName() + " defined by the transition " + currentPath.get(0) +
              " is a counter in the loop " + currentPath);
        }
      }
    } else {
      Set<NodeVariable> newCurrentVariables = new HashSet<>();

      for (Entry<NodeVariable, Set<NodeVariable>> dependency:
          EfsmUtils.getResultingDependencies(currentTransition.getGuardedAction().getAction())
              .entrySet()) {
        NodeVariable variable = dependency.getKey();
        Set<NodeVariable> intersection = new HashSet<>(currentVariables);
        intersection.retainAll(dependency.getValue());

        if (!intersection.isEmpty()) {
          newCurrentVariables.add(variable);
        }
      }

      if (!newCurrentVariables.isEmpty()) {
        List<EfsmTransition> newCurrentPath = new ArrayList<>(currentPath);
        newCurrentPath.add(currentTransition);

        for (VariableDependency dataDependency:
            this.dataDependencyGraph.getOutEdges(currentTransition)) {
          if (newCurrentVariables.contains(dataDependency.getVariable())) {
            EfsmTransition transition = this.dataDependencyGraph.getDest(dataDependency);
            this.findCounterLoop(potentialCounter, transition, targetVariables, newCurrentVariables,
                newCurrentPath);
          }
        }
      }
    }
  }

  public Set<List<EfsmTransition>> getDefinitionPaths(EfsmTransition transition) {
    Set<List<EfsmTransition>> cachedPaths = this.definitionPaths.get(transition);

    if (cachedPaths == null) {
      cachedPaths = new HashSet<>();
      this.definitionPaths.put(transition, cachedPaths);
      Map<EfsmTransition, Set<NodeVariable>> dependencies = new HashMap<>();

      for (VariableDependency dependency: this.controlDependencyGraph.getInEdges(transition)) {
        NodeVariable dependencyVariable = dependency.getVariable();
        EfsmTransition definingTransition = this.controlDependencyGraph.getSource(dependency);
        Set<NodeVariable> variables =
            EfsmUtils.getResultingDependencies(definingTransition.getGuardedAction().getAction())
                .get(dependencyVariable);

        Set<NodeVariable> storedVariables = dependencies.get(definingTransition);

        if (storedVariables == null) {
          storedVariables = new HashSet<>();
          dependencies.put(definingTransition, storedVariables);
        }

        storedVariables.addAll(variables);
      }

      for (Entry<EfsmTransition, Set<NodeVariable>> dependency: dependencies.entrySet()) {
        cachedPaths.addAll(this.findDefinitionPaths(Arrays.asList(dependency.getKey()),
            dependency.getValue()));
      }
    }

    return cachedPaths;
  }

  public Set<Set<List<EfsmTransition>>> getLoopDefinitionPaths(EfsmTransition transition) {
    Set<Set<List<EfsmTransition>>> cachedPaths = this.loopDefinitionPaths.get(transition);

    if (cachedPaths == null) {
      cachedPaths = new HashSet<>();
      this.loopDefinitionPaths.put(transition, cachedPaths);

      for (VariableDependency dependency: this.controlDependencyGraph.getInEdges(transition)) {
        NodeVariable dependencyVariable = dependency.getVariable();
        EfsmTransition definingTransition = this.controlDependencyGraph.getSource(dependency);
        Set<List<EfsmTransition>> paths = new HashSet<>();

        if (this.counters.containsKey(dependencyVariable)) {
          for (List<EfsmTransition> path: this.counters.get(dependencyVariable)) {
            if (path.get(0).equals(definingTransition)) {
              paths.add(path);
            }
          }

          if (!paths.isEmpty()) {
            cachedPaths.add(paths);
          }
        }
      }
    }

    return cachedPaths;
  }

  private Set<List<EfsmTransition>> findDefinitionPaths(List<EfsmTransition> currentPath,
      Set<NodeVariable> currentVariables) {
    Set<List<EfsmTransition>> foundPaths = new HashSet<>();
    Set<NodeVariable> innerDependencies = new HashSet<>();

    for (NodeVariable variable: currentVariables) {
      if (this.efsm.containsInnerVariable(variable.getName())) {
        innerDependencies.add(variable);
      }
    }

    if (innerDependencies.isEmpty()) {
      foundPaths.add(currentPath);
    } else {
      if (innerDependencies.size() < currentVariables.size()) {
        foundPaths.add(currentPath);
      }

      EfsmTransition currentTransition = currentPath.get(0);
      Map<EfsmTransition, Map<NodeVariable, Set<NodeVariable>>> dependencies = new HashMap<>();

      for (VariableDependency dependency: this.dataDependencyGraph.getInEdges(currentTransition)) {
        NodeVariable dependencyVariable = dependency.getVariable();

        if (innerDependencies.contains(dependencyVariable)) {
          EfsmTransition definingTransition = this.dataDependencyGraph.getSource(dependency);
          Set<NodeVariable> variables =
              EfsmUtils.getResultingDependencies(definingTransition.getGuardedAction().getAction())
                  .get(dependencyVariable);

          Map<NodeVariable, Set<NodeVariable>> storedVariables =
              dependencies.get(definingTransition);

          if (storedVariables == null) {
            storedVariables = new HashMap<>();
            dependencies.put(definingTransition, storedVariables);
          }

          storedVariables.put(dependencyVariable, variables);
        }
      }

      for (Entry<EfsmTransition, Map<NodeVariable,Set<NodeVariable>>> dependency
          : dependencies.entrySet()) {

        if (currentPath.contains(dependency.getKey())) {
          int limit = dependency.getKey().getGuardedAction().getAction().getAssignments().size();
          int occurrences = 0;

          for (EfsmTransition transition: currentPath) {
            if (transition.equals(dependency.getKey())) {
              occurrences++;
            }
          }

          if (occurrences >= limit) {
            continue;
          }
        }

        List<EfsmTransition> newPath = new ArrayList<EfsmTransition>(currentPath);
        Set<NodeVariable> variables = new HashSet<>(innerDependencies);
        variables.removeAll(dependency.getValue().keySet());

        for (Set<NodeVariable> newVariables: dependency.getValue().values()) {
          variables.addAll(newVariables);
        }

        newPath.add(0, dependency.getKey());
        foundPaths.addAll(this.findDefinitionPaths(newPath, variables));
      }
    }

    return foundPaths;
  }
}
