/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */
package ru.ispras.retrascope.engine.efsm.generator.test;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import org.apache.commons.cli.ParseException;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.Random;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.SolverResult.Status;
import ru.ispras.fortress.solver.constraint.ConstraintUtils;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.basis.TestGenerator;
import ru.ispras.retrascope.engine.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.result.test.Vector;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

/**
 * The test generator for EFSM-based hardware models.
 * 
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public final class EfsmTestGenerator extends TestGenerator {
  /**
   * The engine ID of all {@code EfsmFateTestGenerator} instances.
   */
  public static final String ENGINE_ID = "efsm-test-generator".intern();

  public static final Parameter LOOP_LIMIT =
      new Parameter("loop-limit", true, "Loop iteration limit");
  private static final int DEFAULT_LOOP_LIMIT = 1;

  private final Logger logger;
  private final String logEntryHeader;

  /**
   * Constructs a generator.
   */
  public EfsmTestGenerator() {
    super(EfsmTestGenerator.ENGINE_ID, EfsmModel.TYPE);
    this.addParameter(EfsmTestGenerator.LOOP_LIMIT);
    this.logEntryHeader = "EFSM.TestGenerator";
    this.logger = Log.getLogger(this.getClass());
  }

  @Override
  public Test start(Map<EntityType, Entity> inputs) {
    EfsmModel efsmModel = (EfsmModel) inputs.get(EfsmModel.TYPE);

    if (efsmModel == null) {
      throw new IllegalArgumentException("No EfsmModel objects have been found among arguments");
    }

    int loopLimit = -1;

    Arguments arguments = (Arguments) inputs.get(Arguments.TYPE);

    if (arguments != null) {
      try {
        Arguments parsedArguments = this.parseCommandLine(arguments.getCommandLine());

        loopLimit = Integer.parseInt(parsedArguments.getValue(EfsmTestGenerator.
            LOOP_LIMIT, "-1"));
      } catch (ParseException | NumberFormatException e) {
        throw new IllegalArgumentException("The exception has occurred while parsing command line "
            + "arguments", e);
      }
    }

    if (loopLimit == -1) {
      loopLimit = EfsmTestGenerator.DEFAULT_LOOP_LIMIT;

      this.logger.log(LogLevel.INFO, this.logEntryHeader +
          ": loop limit isn't specified. Setting it to the default value: " + loopLimit);
    } else {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": the specified loop limit: " +
          loopLimit);
    }

    List<Efsm> efsms = new LinkedList<>(efsmModel.getEfsmList());
    int noProgressLimit;

    if (!efsms.isEmpty()) {
      Map<Efsm, Set<List<EfsmTransition>>> traversedPaths = new HashMap<>(efsms.size());
      EfsmTransitionCoverage totalCoverage = null;

      List<RandomGenerator> generators = new LinkedList<>();

      int transitionsAmount = 0;
      int statesAmount = 0;

      for (Efsm efsm: efsms) {
        traversedPaths.put(efsm, new HashSet<List<EfsmTransition>>());

        transitionsAmount = transitionsAmount + efsm.getTransitions().size();
        statesAmount = statesAmount + efsm.getStates().size();

        RandomGenerator generator = new RandomGenerator(efsm);

        generators.add(generator);

        if (totalCoverage == null) {
          totalCoverage = new EfsmTransitionCoverage(generator.getSequenceCoverage());
        } else {
          totalCoverage.include(generator.getSequenceCoverage());
        }
      }

      noProgressLimit = transitionsAmount / statesAmount;
      int sequenceNoProgressCounter = 0;

      Comparator<RandomGenerator> comparator = new Comparator<RandomGenerator>() {
        @Override
        public int compare(RandomGenerator generator1, RandomGenerator generator2) {
          BigDecimal coverage1 = generator1.getSequenceCoverage().getPercentage();
          BigDecimal coverage2 = generator2.getSequenceCoverage().getPercentage();

          return coverage2.compareTo(coverage1);
        }
      };

      Test test = new Test();
      Map<Efsm, Iterator<Vector>> vectorIterators = new HashMap<>(efsms.size());
      Map<Sequence, EfsmTransitionCoverage> sequencesCoverage = new HashMap<>();

      while (!totalCoverage.isFull() && (sequenceNoProgressCounter <= noProgressLimit)) {
        this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": starting a new sequence");

        for (RandomGenerator generator: generators) {
          vectorIterators.put(generator.getEfsm(), generator.getNextSequenceIterator());
        }

        Sequence sequence = new Sequence();

        int vectorNoProgressCounter = 0;
        List<Vector> bufferedVectors = new ArrayList<>();
        EfsmTransitionCoverage sequenceCoverage = this.combineCoverage(generators);

        while (vectorNoProgressCounter <= noProgressLimit) {
          Collections.sort(generators, comparator);
          Vector currentVector = null;

          for (RandomGenerator generator: generators) {
            Iterator<Vector> sequenceIterator = vectorIterators.get(generator.getEfsm());
            generator.setAdditionalConstraintOnIputs(currentVector);
            Vector generatedVector = sequenceIterator.next();

            if (generatedVector != null) {
              currentVector = generatedVector;
            }
          }

          if (currentVector != null) {
            for (Efsm efsm: efsms) {
              currentVector = this.fillUnusedInputsRandomly(efsm, currentVector);
            }

            for (RandomGenerator generator: generators) {
              generator.submitVector(currentVector);
            }

            BigDecimal oldCoveragePercentage = sequenceCoverage.getPercentage();
            EfsmTransitionCoverage bufferedSequenceCoverage = this.combineCoverage(generators);
            BigDecimal newCoveragePercentage = bufferedSequenceCoverage.getPercentage();

            if (newCoveragePercentage.compareTo(oldCoveragePercentage) == 1) {
              this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
                  ": the new vector has increased current sequence coverage from " +
                oldCoveragePercentage + " to " + newCoveragePercentage + "%");

              if (!bufferedVectors.isEmpty()) {
                sequence.addVectors(bufferedVectors);
                bufferedVectors.clear();
                vectorNoProgressCounter = 0;
              }

              sequence.addVector(currentVector);
              sequenceCoverage = bufferedSequenceCoverage;

              EfsmTransitionCoverage testedTotalCoverage =
                  new EfsmTransitionCoverage(totalCoverage);
              testedTotalCoverage.include(sequenceCoverage);

              if (testedTotalCoverage.isFull()) {
                this.logger.log(LogLevel.DEBUG,
                    this.logEntryHeader + ": total coverage is full now");

                break;
              }
            } else {
              vectorNoProgressCounter++;
              bufferedVectors.add(currentVector);

              this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
                  ": there is no progress in sequence coverage. Buffering the last vector. " +
                  "No progress counter: " + vectorNoProgressCounter +
                  ". No progress limit: " + noProgressLimit);
            }
          } else {
            this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
                ": failed to generate next vector. Finishing current sequence");

            break;
          }
        }

        BigDecimal oldCoveragePercentage = totalCoverage.getPercentage();
        EfsmTransitionCoverage bufferedTotalCoverage = new EfsmTransitionCoverage(totalCoverage);
        bufferedTotalCoverage.include(sequenceCoverage);
        BigDecimal newCoveragePercentage = bufferedTotalCoverage.getPercentage();

        if (newCoveragePercentage.compareTo(oldCoveragePercentage) == 1) {
          this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
              ": the new sequence has increased total coverage from " +
              oldCoveragePercentage + " to " + newCoveragePercentage + "%");

          totalCoverage = bufferedTotalCoverage;

          for (RandomGenerator generator: generators) {
            List<EfsmTransition> path = new ArrayList<>(generator.getTraversedPath());

            for (int i = 1; i <= bufferedVectors.size(); i++) {
              path.remove(path.size() - 1);
            }

            traversedPaths.get(generator.getEfsm()).add(path);
          }

          test.addSequence(sequence);
          sequencesCoverage.put(sequence, sequenceCoverage);
        } else {
          sequenceNoProgressCounter++;

          this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
              ": there is no progress in total coverage. Skipping the last sequence. " +
              "No progress counter: " + sequenceNoProgressCounter +
              ". No progress limit: " + noProgressLimit);
        }
      }

      this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
          ": the random phase of generation has been finished successfully. Total coverage: " +
          totalCoverage);

      if (!totalCoverage.isFull()) {
        this.logger.log(LogLevel.DEBUG, 
            this.logEntryHeader + ": starting the directed phase of generation");

        for (RandomGenerator randomGenerator: generators) {
          Efsm efsm = randomGenerator.getEfsm();

          if (randomGenerator.getTotalCoverage().isFull()) {
            this.logger.log(LogLevel.DEBUG,
                this.logEntryHeader + ": the " + efsm + " EFSM is fully covered. Skipping it");
          } else {
            this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": the " + efsm +
                " EFSM has uncovered transitions. Processing it");

            DirectedGenerator directedGenerator =
                new DirectedGenerator(efsm, traversedPaths.get(efsm), loopLimit);

            for (Iterator<Vector> vectorIterator = directedGenerator.getNextSequenceIterator();
                vectorIterator != null;
                vectorIterator = directedGenerator.getNextSequenceIterator()) {
              Sequence sequence = new Sequence();

              while (vectorIterator.hasNext()) {
                Vector vector = vectorIterator.next();

                for (Efsm passiveEfsm: efsms) {
                  vector = this.fillUnusedInputsRandomly(passiveEfsm, vector);
                }

                sequence.addVector(vector);
              }

              test.addSequence(sequence);

              EfsmTransitionCoverage sequenceCoverage = directedGenerator.getSequenceCoverage();
              totalCoverage.include(sequenceCoverage);
              sequencesCoverage.put(sequence, sequenceCoverage);

              this.logger.log(LogLevel.DEBUG,
                  this.logEntryHeader + ": the directed sequence has been generated successfully");
              this.logger.log(LogLevel.DEBUG,
                  this.logEntryHeader + ": current coverage: " + totalCoverage);
            }
          }
        }

        this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
                ": the directed phase of generation has been finished successfully");
      }

      this.logger.log(LogLevel.INFO, this.logEntryHeader +
          ": generation has been finished successfully. Total coverage: " + totalCoverage);
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
          ": total amount of sequences in the generated test before minimisation: " +
          test.getSequencesAmount());
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
          ": total amount of vectors in the generated test before minimisation: " +
          test.getVectorsAmount());

      test = TestMinimiser.minimise(sequencesCoverage);

      this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
          ": total amount of sequences in the generated test after minimisation: " +
          test.getSequencesAmount());
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
          ": total amount of vectors in the generated test after minimisation: " +
          test.getVectorsAmount());

      for (EfsmTransition transition: totalCoverage.getUncoveredTransitions()) {
        this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
            ": uncovered transition: " + transition);
      }

      return test;
    } else {
      throw new IllegalArgumentException("No EFSMs are specified");
    }
  }

  private Vector fillUnusedInputsRandomly(Efsm efsm, Vector inputVector) {
    Transaction randomlyFilledTransaction = new Transaction();

    this.logger.log(LogLevel.DEBUG, this.logEntryHeader
        + ": generating random values for unused inputs");

    Collection<Event> events = inputVector.getEventList().getEvents();
    Set<String> eventNames = new HashSet<>(events.size());

    for (Event event: events) {
      eventNames.add(event.getRangedVariable().getVariableName());
    }

    for (String inputName: efsm.getInputDataTypes().keySet()) {
      if (!inputVector.containsVariable(inputName)) {
        NodeValue value;
        Node constraint = efsm.getInvariantCondition(inputName);

        if (constraint == null) {
          value = new NodeValue(Random.newValue(efsm.getInputDataType(inputName)));
        } else {
          if ((efsm.getInputDataType(inputName).getTypeId() == DataTypeId.LOGIC_INTEGER) &&
              (constraint instanceof NodeOperation) &&
              (((NodeOperation) constraint).getOperandCount() == 2) &&
              (((NodeOperation) constraint).getOperand(0) instanceof NodeOperation) &&
              (((NodeOperation) constraint).getOperand(1) instanceof NodeOperation)) {
            NodeOperation firstCondition =
                (NodeOperation) ((NodeOperation) constraint).getOperand(0);
            NodeOperation secondCondition =
                (NodeOperation) ((NodeOperation) constraint).getOperand(1);

            if (firstCondition.getOperationId().equals(StandardOperation.GREATEREQ) &&
                secondCondition.getOperationId().equals(StandardOperation.LESSEQ)) {
              int min = ((NodeValue)firstCondition.getOperand(1)).getInteger().intValue();
              int max = ((NodeValue)secondCondition.getOperand(1)).getInteger().intValue();

              NodeValue module = new NodeValue(Data.newInteger(max - min));
              NodeValue minValueNode = new NodeValue(Data.newInteger(min));
              value = new NodeValue(Random.newValue(DataType.INTEGER));
              NodeVariable targetVariable = new NodeVariable(inputName, DataType.INTEGER);

              NodeOperation moduledValueComputation =
                  new NodeOperation(StandardOperation.MOD, value, module);
              NodeOperation valueComputation =
                  new NodeOperation(StandardOperation.ADD, moduledValueComputation, minValueNode);
              NodeOperation assignment =
                  new NodeOperation(StandardOperation.EQ, targetVariable, valueComputation);
              SolverResult result =
                  ConstraintUtils.solve(ConstraintUtils.newConstraint(assignment));

              if (result.getStatus() == Status.SAT) {
                value = new NodeValue(result.getVariables().iterator().next().getData());
              } else {
                throw new RetrascopeRuntimeException("Failed to resolve invariant constraint on " +
                    "input " + inputName + " due to the errors: " + result.getErrors());
              }
            } else {
              throw new IllegalArgumentException("Such invariant is currently unsupported: " +
                  constraint);
            }
          } else {
            SolverResult result = ConstraintUtils.solve(ConstraintUtils.newConstraint(constraint));

            if (result.getStatus() == Status.SAT) {
              value = new NodeValue(result.getVariables().iterator().next().getData());
            } else {
              throw new RetrascopeRuntimeException("Failed to resolve invariant constraint on " +
                  "input " + inputName + " due to the errors: " + result.getErrors());
            }
          }
        }

        randomlyFilledTransaction.addVariable(inputName, value);
      }
    }

    if (!randomlyFilledTransaction.isEmpty()) {
      inputVector.addTransaction(randomlyFilledTransaction);

      this.logger.log(LogLevel.DEBUG, this.logEntryHeader
          + ": randomly generated transaction: " + randomlyFilledTransaction);
    } else {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader
          + ": all inputs are already set");
    }

    return inputVector;
  }

  private EfsmTransitionCoverage combineCoverage(Collection<RandomGenerator> generators) {
    EfsmTransitionCoverage coverage = null;

    for (RandomGenerator generator: generators) {
      if (coverage == null) {
        coverage = new EfsmTransitionCoverage(generator.getSequenceCoverage());
      } else {
        coverage.include(generator.getSequenceCoverage());
      }
    }

    return coverage;
  }
}
