/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */
package ru.ispras.retrascope.engine.efsm.generator.test.fate;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Logger;

import edu.uci.ics.jung.graph.DirectedSparseMultigraph;
import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.Random;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.NodeBinding.BoundVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.SolverResult.Status;
import ru.ispras.fortress.solver.constraint.Constraint;
import ru.ispras.fortress.solver.constraint.ConstraintCombiner;
import ru.ispras.fortress.solver.constraint.ConstraintUtils;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.basis.AtomicTestGenerator;
import ru.ispras.retrascope.engine.basis.TestCoverage;
import ru.ispras.retrascope.engine.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.engine.efsm.generator.test.heuristic.direction.NearestUncoveredEfsmTransitionChooser;
import ru.ispras.retrascope.engine.efsm.simulator.EfsmSimulationSnapshot;
import ru.ispras.retrascope.engine.efsm.simulator.EfsmSimulator;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.model.efsm.Efsm.TransitionWeightBasis;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.result.test.Vector;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

/**
 * The FATE test generator which deals with single EFSM and operates at the vector-level.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see EfsmFateTestGenerator
 */
public class DirectedFateGenerator extends AtomicTestGenerator {
  private final Logger logger;
  private final String logEntryHeader;

  private final Map<EfsmTransition, Set<ReachabilityInformation>> reachabilityInformation;
  private final Map<EfsmTransition, Set<ReachabilityInformation>> testedReachabilityInformation;

  private EfsmTransitionCoverage totalCoverage;

  private static final String VARIABLE_RENAMING_SUFFIX = "_renamed";

  private final EfsmSimulator simulator;

  private final DirectedSparseMultigraph<EfsmTransition, VariableDependency> controlDependencyGraph;
  private final Map<NodeVariable, Set<List<EfsmTransition>>> counters;

  private NearestUncoveredEfsmTransitionChooser directedChooser;

  private final int loopLimit;

  /**
   * Constructs a test generator for the specified EFSM with the specified parameters.
   *
   * @param efsm the EFSM
   * @param loopLimit a depth limit for loop unfolding
   */
  public DirectedFateGenerator(Efsm efsm, int loopLimit,
      Map<EfsmTransition, Set<ReachabilityInformation>> reachabilityInformation) {
    this.logEntryHeader = "EFSM.FATE.DirectedGenerator." + efsm.getName();
    this.logger = Log.getLogger(this.getClass());

    this.simulator = new EfsmSimulator(efsm);

    FateEfsmAnalyser efsmAnalyser = new FateEfsmAnalyser(efsm);
    this.controlDependencyGraph = efsmAnalyser.getControlDependencyGraph();
    this.counters = efsmAnalyser.getCounters();

    this.totalCoverage = new EfsmTransitionCoverage(efsm);

    for (EfsmTransition coveredTransition: reachabilityInformation.keySet()) {
      this.totalCoverage.markTransitionAsCovered(coveredTransition);
    }

    this.reachabilityInformation = new HashMap<>(reachabilityInformation);
    this.testedReachabilityInformation = new HashMap<>();

    this.directedChooser = null;

    this.loopLimit = loopLimit;
  }

  /**
   * Returns an EFSM which this generator is related to.
   *
   * @return the EFSM
   */
  public Efsm getEfsm() {
    return this.simulator.getStaticModel();
  }

  /**
   * Returns a totalCoverage reached during generation.
   *
   * @return the totalCoverage
   */
  @Override
  public EfsmTransitionCoverage getSequenceCoverage() {
    return new EfsmTransitionCoverage(this.simulator.getCoverage());
  }

  /**
   * Returns an iterator enabling to retrieve next sequence of vectors produced by this generator.
   *
   * @return the iterator. If this generator is already finished generation, {@code null} is
   * returned
   */
  @Override
  public Iterator<Vector> getNextSequenceIterator() {
    if (!this.totalCoverage.isFull()) {
      if (this.directedChooser == null) {
        this.directedChooser = new NearestUncoveredEfsmTransitionChooser(
            this.simulator.getStaticModel(), this.totalCoverage);
      }

      for (EfsmTransition transition = this.directedChooser.chooseNext(); transition != null;
          transition = this.directedChooser.chooseNext()) {
        Sequence sequence = this.generateSequence(transition);

        if (sequence != null) {
          this.directedChooser = new NearestUncoveredEfsmTransitionChooser(
              this.simulator.getStaticModel(), this.totalCoverage);

          return sequence.getVectors().iterator();
        }
      }
    }

    return null;
  }

  private Vector reset() {
    this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": resetting the EFSM");

    this.simulator.reset();

    Vector vector = this.enableGuardedAction(this.simulator.getStaticModel().
        getResetGuardedActions().iterator().next(), null);//TODO

    if (vector != null) {
      this.logger.log(LogLevel.DEBUG,
          this.logEntryHeader + ": the EFSM has been reset successfully");
    } else {
      this.logger.log(LogLevel.WARNING,
          this.logEntryHeader + ": failed to reset the EFSM");
    }

    return vector;
  }

  private Sequence generateSequence(EfsmTransition transition) {
    Efsm efsm = this.simulator.getStaticModel();
 
    this.logger.log(LogLevel.DEBUG, this.logEntryHeader
        + ": generating a sequence to cover the transition: " + transition);

    if (transition.getSourceState().equals(this.simulator.getStaticModel().getInitialState())) {
      Vector vector = this.reset();

      if (vector != null) {
        this.simulator.setInputValues(vector.getTransactions());
        this.simulator.processEvents(vector.getEventList());
        Sequence sequence = new Sequence();
        Map<EfsmTransition, Set<ReachabilityInformation>> reachabilityMap = new HashMap<>();
        vector = this.traverseTransition(transition, sequence, reachabilityMap);

        if (vector != null) {
          sequence.addVector(vector);
          this.unbufferReachabilityInformation(reachabilityMap);

          return sequence;
        }
      }
    } else {
      Set<EfsmTransition> coveredTransitions = this.totalCoverage.getCoveredTransitions();

      for (EfsmTransition previousTransition:
          efsm.getIncomingTransitions(transition.getSourceState())) {
        if (coveredTransitions.contains(previousTransition)) {
          for (ReachabilityInformation reachabilityInformation:
              this.reachabilityInformation.get(previousTransition)) {
            if (this.testedReachabilityInformation.containsKey(transition) &&
                this.testedReachabilityInformation.get(transition).contains(reachabilityInformation)) {
              continue;
            }

            this.logger.log(LogLevel.DEBUG, this.logEntryHeader
                + ": simulating the model to the end state of the transition: " + previousTransition);

            this.reach(reachabilityInformation, true);

            Sequence sequence = new Sequence();

            for (int i = 0; i <= reachabilityInformation.vectorNumber; i++) {
              sequence.addVector(reachabilityInformation.sequence.getVector(i));
            }

            Map<EfsmTransition, Set<ReachabilityInformation>> reachabilityMap = new HashMap<>();
            Vector vector = this.traverseTransition(transition, sequence, reachabilityMap);

            if (vector != null) {
              sequence.addVector(vector);
              this.unbufferReachabilityInformation(reachabilityMap);

              if (this.testedReachabilityInformation.containsKey(transition)) {
                this.testedReachabilityInformation.remove(transition);
              }

              return sequence;
            }
          }
        }
      }
    }

    return this.coverByBackjumping(transition);
  }

  private Sequence coverByBackjumping(EfsmTransition transition) {
    this.logger.log(LogLevel.DEBUG, this.logEntryHeader
        + ": using backjumping to cover the transition: " + transition);
 
    Guard guard = transition.getGuardedAction().getGuard();
    Efsm efsm = this.simulator.getStaticModel();
 
    for (VariableDependency dependency: this.controlDependencyGraph.getInEdges(transition)) {
      EfsmTransition definingTransition = this.controlDependencyGraph.getSource(dependency);
 
      if (this.totalCoverage.getCoveredTransitions().contains(definingTransition)) {
        for (ReachabilityInformation reachabilityInformation:
            this.reachabilityInformation.get(definingTransition)) {
          if (this.testedReachabilityInformation.containsKey(transition) &&
              this.testedReachabilityInformation.get(transition).contains(reachabilityInformation)) {
            continue;
          }

          this.logger.log(LogLevel.DEBUG, this.logEntryHeader
              + ": the transition " + definingTransition + " defining the variable " +
              dependency.getVariable().getName() + " is covered. Trying it");

          boolean isCounter = false;

          if (this.counters.containsKey(dependency.getVariable())) {
            for (List<EfsmTransition> loop: this.counters.get(dependency.getVariable())) {
              if (loop.get(0).equals(definingTransition)) {
                isCounter = true;

                break;
              }
            }
          }

          List<EfsmTransition> path = efsm.getShortestPath(definingTransition.getDestinationState(),
              transition.getSourceState(), TransitionWeightBasis.GUARD_REGISTER_AMOUNT);

          this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": the shortest path from the "
              + "destination state of the defining transition to the source state of the target "
              + "transition: " + path);

          if (isCounter) {
            this.logger.log(LogLevel.DEBUG, this.logEntryHeader
                + ": simulating the model to the destination state of the defining transition");

            this.reach(reachabilityInformation, true);
          } else {
            this.logger.log(LogLevel.DEBUG, this.logEntryHeader
                + ": simulating the model to the source state of the defining transition");

            this.reach(reachabilityInformation, false);
          }

          Set<Node> predicates = new HashSet<>();

          Action action = definingTransition.getGuardedAction().getAction();
          Node variableDefinition = this.getResultingSubstitution(action, dependency.getVariable());

          for (Node predicate: guard.getPredicates()) {
            Set<NodeVariable> predicateVariables = new HashSet<>(ExprUtils.getVariables(predicate));

            if (isCounter) {
              if (predicateVariables.contains(dependency.getVariable())) {
                predicates.add(predicate);
              }
            } else {
              Node renamedPredicate = predicate;
              boolean matching = false;
  //TODO: invariants on renamed variables
              for (NodeVariable variable: predicateVariables) {
                if (variable.equals(dependency.getVariable())) {
                  matching = true;
                  predicates.add(new NodeOperation(StandardOperation.EQ, variable, variableDefinition));
                } else {
                  String variableName = variable.getName();
                  NodeVariable renamedVariable =
                      new NodeVariable(
                          variableName + DirectedFateGenerator.VARIABLE_RENAMING_SUFFIX,
                          variable.getDataType());
                  renamedPredicate =
                      Transformer.substitute(renamedPredicate, variableName, renamedVariable);
                }
              }

              if (matching) {
                predicates.add(renamedPredicate);
              }
            }
          }

          Node additionalExpression =
              ExprUtils.getConjunction(predicates.toArray(new Node[predicates.size()]));

          boolean compatibility = false;

          Sequence sequence = new Sequence();
          Map<EfsmTransition, Set<ReachabilityInformation>> reachabilityMap = new HashMap<>();

          if (!isCounter) {
            Vector vector =
                this.generateInputsForTransition(definingTransition, additionalExpression);

            if (vector != null) {
              vector = this.fillUnusedInputsRandomly(vector);

              this.simulator.setInputValues(vector.getTransactions());
              this.simulator.processEvents(vector.getEventList());

              compatibility = true;

              for (int i = 0; i < reachabilityInformation.vectorNumber; i++) {
                sequence.addVector(reachabilityInformation.sequence.getVector(i));
              }

              sequence.addVector(vector);
            }
          } else {
            for (int i = 0; i <= reachabilityInformation.vectorNumber; i++) {
              sequence.addVector(reachabilityInformation.sequence.getVector(i));
            }

            sequence = this.traverseLoop(dependency.getVariable(), definingTransition,
                additionalExpression, sequence, reachabilityMap);

            if (sequence != null) {
              compatibility = true;
            }
          }

          if (compatibility) {
            this.logger.log(LogLevel.DEBUG, this.logEntryHeader
                + ": the transitions are consistent. Traversing a path between them");

            boolean success = true;

            for (EfsmTransition nextTransition: path) {
              Vector vector = this.traverseTransition(nextTransition, sequence, reachabilityMap);

              if (vector != null) {
                sequence.addVector(vector);
              } else {
                success = false;

                this.logger.log(LogLevel.DEBUG, this.logEntryHeader
                    + ": the path is untraversable");

                break;
              }
            }

            if (success) {
              Vector vector = this.traverseTransition(transition, sequence, reachabilityMap);

              if (vector != null) {
                sequence.addVector(vector);

                if (this.testedReachabilityInformation.containsKey(transition)) {
                  this.testedReachabilityInformation.remove(transition);
                }

                this.unbufferReachabilityInformation(reachabilityMap);

                return sequence;
              }

              this.logger.log(LogLevel.DEBUG, this.logEntryHeader
                  + ": failed to generate inputs for the target transition");
            }
          } else {
            this.logger.log(LogLevel.DEBUG, this.logEntryHeader
                + ": the transitions aren't consistent");
          }

          Set<ReachabilityInformation> testedRi = this.testedReachabilityInformation.get(transition);

          if (testedRi == null) {
            testedRi = new HashSet<>();
            this.testedReachabilityInformation.put(transition, testedRi);
          }

          testedRi.add(reachabilityInformation);
        }
      } else {
        this.logger.log(LogLevel.DEBUG, this.logEntryHeader
            + ": the transition " + definingTransition + " defining the variable " +
            dependency.getVariable().getName() + " isn't covered. Skipping it");
      }
    }

    return null;
  }

  private void reach(ReachabilityInformation reachabilityInformation, boolean withLastTransition) {
    if (withLastTransition) {
      this.simulator.loadSnapshot(reachabilityInformation.destinationStateSnapshot);
    } else {
      this.simulator.loadSnapshot(reachabilityInformation.sourceStateSnapshot);
    }
  }

  private Vector traverseTransition(EfsmTransition transition, Sequence sequence,
      Map<EfsmTransition, Set<ReachabilityInformation>> reachabilityMap) {
    this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
        ": trying to traverse the transition: " + transition);

    Vector inputVector = this.generateInputsForTransition(transition, null);

    if (inputVector != null) {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader
          + ": the transition has been traversed successfully: " + transition);
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader
          + ": input vector: " + inputVector);

      inputVector = this.fillUnusedInputsRandomly(inputVector);

      EfsmSimulationSnapshot sourceStateSnapshot = this.simulator.makeSimulationSnapshot();

      this.simulator.setInputValues(inputVector.getTransactions());
      this.simulator.processEvents(inputVector.getEventList());

      Set<ReachabilityInformation> transitionReachabilityInformation =
          reachabilityMap.get(transition);

      if (transitionReachabilityInformation == null) {
        transitionReachabilityInformation = new HashSet<>();
        reachabilityMap.put(transition, transitionReachabilityInformation);
      }

      transitionReachabilityInformation.add(new ReachabilityInformation(sequence,
          sequence.getLength(), sourceStateSnapshot, this.simulator.makeSimulationSnapshot()));

      if (!this.simulator.getCurrentState().equals(transition.getDestinationState())) {
        throw new RetrascopeRuntimeException("Unexpected simulator state: " +
            this.simulator.getCurrentState());
      }

      this.logger.log(LogLevel.DEBUG, this.logEntryHeader
          + ": current totalCoverage: " + this.totalCoverage);
    }

    return inputVector;
  }

  private Vector generateInputsForTransition(EfsmTransition transition, Node additionalExpression) {
    if (!this.simulator.getCurrentState().equals(transition.getSourceState())) {
      throw new IllegalArgumentException(
          "The current state of the simulator doesn't match the transition's source state: " +
          this.simulator.getCurrentState() + " vs " + transition.getSourceState());
    }

    Vector vector = this.enableGuardedAction(transition.getGuardedAction(), additionalExpression);

    return vector;
  }

  private Vector enableGuardedAction(GuardedAction guardedAction, Node additionalExpression) {
    Node guardExpression = guardedAction.getGuard().toNode();
    Set<NodeVariable> usedVariables = new HashSet<>(ExprUtils.getVariables(guardExpression));
    List<BoundVariable> bindingList = new ArrayList<>(usedVariables.size());
    Set<Node> inputConstraints = new HashSet<>();
    Efsm efsm = this.simulator.getStaticModel();

    for (NodeVariable usedVariable: usedVariables) {
      String variableName = usedVariable.getName();

      if (efsm.containsInput(variableName)) {
        Node variableConstraint = efsm.getInvariantCondition(variableName);

        if (variableConstraint != null) {
          inputConstraints.add(variableConstraint);
        }
      } else {
        Map<String, NodeValue> innerVariables = this.simulator.getInnerValues();

        if (innerVariables.containsKey(variableName)) {
          bindingList.add(NodeBinding.bindVariable(usedVariable, innerVariables.get(variableName)));
        }
      }
    }

    guardExpression =
        Transformer.substituteAllBindings(new NodeBinding(guardExpression, bindingList));

    Constraint targetConstraint = ConstraintUtils.newConstraint(guardExpression);

    StringBuilder logMessageBuilder = new StringBuilder(this.logEntryHeader);
    logMessageBuilder.append(": trying to generate an input vector for the guard expression ");
    logMessageBuilder.append(guardExpression);

    if (additionalExpression != null)
    {
      logMessageBuilder.append("\n").append("Additional constraint: ").append(additionalExpression);

      Constraint additionalConstraint = ConstraintUtils.newConstraint(additionalExpression);
      targetConstraint = ConstraintCombiner.makeConjunction(targetConstraint, additionalConstraint);

      for (NodeVariable usedVariable: ExprUtils.getVariables(additionalExpression)) {
        String variableName = usedVariable.getName();

        if (efsm.containsInput(variableName)) {
          Node variableConstraint = efsm.getInvariantCondition(variableName);

          if (variableConstraint != null) {
            inputConstraints.add(variableConstraint);
          }
        }
      }
    }

    if (!inputConstraints.isEmpty()) {
      Node inputConstraint =
          ExprUtils.getConjunction(inputConstraints.toArray(new Node[inputConstraints.size()]));

      logMessageBuilder.append("\n").append("Constraint on inputs: ").append(inputConstraint);

      targetConstraint = ConstraintCombiner.makeConjunction(targetConstraint,
          ConstraintUtils.newConstraint(inputConstraint));
    }

    this.logger.log(LogLevel.DEBUG, logMessageBuilder.toString());

    SolverResult solverResult = ConstraintUtils.solve(targetConstraint);

    if (solverResult.getStatus().equals(Status.SAT)) {
        Vector inputVector = new Vector();
        Transaction transaction = new Transaction();

        for (Variable variable : solverResult.getVariables()) {
          String variableName = variable.getName();

          if (efsm.getInputDataTypes().containsKey(variableName)) {
            transaction.addVariable(variableName, new NodeValue(variable.getData()));
          }
        }

        inputVector.addEvents(guardedAction.getClocks());

        if (!transaction.isEmpty()) {
          inputVector.addTransaction(transaction);
        }

        this.logger.log(LogLevel.DEBUG, this.logEntryHeader
            + ": the input vector has been generated successfully: " + inputVector);

        return inputVector;
    } else {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": failed to generate an input vector");

      return null;
    }
  }

  private Sequence traverseLoop(NodeVariable counter, EfsmTransition definingTransition,
      Node stopCondition, Sequence sequence,
      Map<EfsmTransition, Set<ReachabilityInformation>> reachabilityMap) {
    Set<List<EfsmTransition>> loops = this.counters.get(counter);
    Set<List<EfsmTransition>> completeLoops = new HashSet<>(loops.size());

    for (List<EfsmTransition> loop: loops) {
      if (loop.get(0).equals(definingTransition)) {
        LinkedList<EfsmTransition> completeLoop = new LinkedList<>(loop);

        for (ListIterator<EfsmTransition> iterator = completeLoop.listIterator(); iterator.hasNext();) {
          EfsmTransition currentTransition = iterator.next();

          EfsmState sourceState = currentTransition.getDestinationState();
          EfsmState destinationState;

          if (iterator.hasNext()) {
            EfsmTransition nextTransition = iterator.next();
            destinationState = nextTransition.getSourceState();
            iterator.previous();
          } else {
            destinationState = completeLoop.getFirst().getSourceState();
          }

          for (EfsmTransition transition:
              this.simulator.getStaticModel().getShortestPath(sourceState, destinationState,
                  TransitionWeightBasis.GUARD_REGISTER_AMOUNT)) {
            iterator.add(transition);
          }
        }

        if (completeLoop.size() > 1) {
          completeLoop.addLast(completeLoop.removeFirst());
        }

        completeLoops.add(completeLoop);
      }
    }

    return this.traverseMixedLoop(completeLoops, sequence, stopCondition, 0, reachabilityMap);
  }

  private Sequence traverseMixedLoop(Set<List<EfsmTransition>> loops, Sequence sequence,
      Node stopCondition, int currentDepth,
      Map<EfsmTransition, Set<ReachabilityInformation>> reachabilityMap) {
    Set<NodeVariable> usedVariables = new HashSet<>(ExprUtils.getVariables(stopCondition));
    List<BoundVariable> bindingList = new ArrayList<>(usedVariables.size());

    for (NodeVariable usedVariable: usedVariables) {
      String variableName = usedVariable.getName();
      Map<String, NodeValue> innerVariables = this.simulator.getInnerValues();

      if (innerVariables.containsKey(variableName)) {
        bindingList.add(NodeBinding.bindVariable(usedVariable, innerVariables.get(variableName)));
      }
    }

    Node currentStopCondition =
        Transformer.substituteAllBindings(new NodeBinding(stopCondition, bindingList));

    Constraint stopConstraint = ConstraintUtils.newConstraint(currentStopCondition);
    SolverResult solverResult = ConstraintUtils.solve(stopConstraint);

    this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": iteration " + currentDepth +
        ". The current form of stop condition: " + currentStopCondition);

    if (solverResult.getStatus().equals(Status.SAT)) {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
          ": loop traverse has been finished successfully at iteration " + currentDepth);

      return sequence;
    }

    if (currentDepth > this.loopLimit) {
      return null;
    }

    EfsmSimulationSnapshot preLoopSnapshot = this.simulator.makeSimulationSnapshot();

    for (List<EfsmTransition> loop: loops) {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
          ": trying to proceed with the loop: " + loop);

      this.simulator.loadSnapshot(preLoopSnapshot);

      Sequence bufferedSequence = new Sequence();
      bufferedSequence.addVectors(sequence.getVectors());

      Map<EfsmTransition, Set<ReachabilityInformation>> bufferedReachabilityMap = new HashMap<>();

      List<Vector> iterationVectors =
          this.performLoopIteration(loop, bufferedSequence, bufferedReachabilityMap);

      if (iterationVectors != null) {
        bufferedSequence.addVectors(iterationVectors);
        bufferedSequence = this.traverseMixedLoop(loops, bufferedSequence, stopCondition,
            currentDepth + 1, bufferedReachabilityMap);

        if (bufferedSequence != null) {
          for (Entry<EfsmTransition, Set<ReachabilityInformation>> entry:
            bufferedReachabilityMap.entrySet()) {
            EfsmTransition transition = entry.getKey();

            if (reachabilityMap.containsKey(transition)) {
              reachabilityMap.get(transition).addAll(entry.getValue());
            } else {
              reachabilityMap.put(transition, entry.getValue());
            }
          }
//TODO
          return bufferedSequence;
        }
      } else {
        continue;
      }
    }

    return null;
  }

  private List<Vector> performLoopIteration(List<EfsmTransition> loop, Sequence sequence,
      Map<EfsmTransition, Set<ReachabilityInformation>> reachabilityMap) {
    List<Vector> vectors = new ArrayList<>(loop.size());

    for (EfsmTransition transition: loop) {
      Vector vector = this.traverseTransition(transition, sequence, reachabilityMap);

      if (vector != null) {
        vectors.add(vector);
      } else {
        this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
            ": there is untraversable transition in the loop: " + transition);

        return null;
      }
    }

    return vectors;
  }

  private Node getResultingSubstitution(Action action, NodeVariable variable) {
    List<Assignment> assignments = action.getAssignments();
    LinkedHashMap<NodeVariable, Node> bufferedSubstitutions =
        new LinkedHashMap<>(assignments.size());

    for (Assignment assignment: assignments) {
      RangedVariable rangedVariable = assignment.getTarget();
      Node expression = assignment.getValue();

      if ((rangedVariable == null) && (expression == null)) {
        continue;
      }

      InvariantChecks.checkNotNull(rangedVariable);
      InvariantChecks.checkNotNull(expression);

      NodeVariable nodeVariable = rangedVariable.getVariable();
      DataType dataType = nodeVariable.getDataType();

      if (rangedVariable.isRanged()) {
        if (!dataType.getTypeId().equals(DataTypeId.BIT_VECTOR)) {
          throw new IllegalArgumentException("Unsupported data type of ranged variable: " + dataType);
        }

        Range range = rangedVariable.getRange();

        NodeValue leftBound = (NodeValue) range.getYoung();
        NodeValue rightBound = (NodeValue) range.getOld();

        int leftBoundInteger = leftBound.getInteger().intValue();
        int rightBoundInteger = rightBound.getInteger().intValue();

        int variableSize = dataType.getSize();

        if (leftBoundInteger < rightBoundInteger) {
          throw new IllegalArgumentException("This type of subvectors is currently unsupported");
        }

        NodeOperation mostSignificantPart = null;
        NodeOperation lessSignificantPart = null;

        if (leftBoundInteger < variableSize - 1) {
          NodeValue leftIndex =
              new NodeValue(DataType.INTEGER.valueOf(String.valueOf(variableSize - 1), 10));
          NodeValue rightIndex =
              new NodeValue(DataType.INTEGER.valueOf(String.valueOf(leftBoundInteger + 1), 10));

          mostSignificantPart = new NodeOperation(StandardOperation.BVEXTRACT, leftIndex, rightIndex,
              nodeVariable);
        }

        if (rightBoundInteger > 0) {
          NodeValue leftIndex =
              new NodeValue(DataType.INTEGER.valueOf(String.valueOf(rightBoundInteger - 1), 10));
          NodeValue rightIndex = new NodeValue(DataType.INTEGER.valueOf("0", 10));

          lessSignificantPart = new NodeOperation(StandardOperation.BVEXTRACT, leftIndex, rightIndex,
              nodeVariable);
        }

        if ((mostSignificantPart != null) && (lessSignificantPart != null)) {
          expression = new NodeOperation(StandardOperation.BVCONCAT, mostSignificantPart, expression,
              lessSignificantPart);
        } else {
          if (mostSignificantPart != null) {
            expression =
                new NodeOperation(StandardOperation.BVCONCAT, mostSignificantPart, expression);
          } else {
            if (lessSignificantPart != null) {
              expression =
                  new NodeOperation(StandardOperation.BVCONCAT, expression, lessSignificantPart);
            }
          }
        }
      }

      Set<NodeVariable> usedVariables = new HashSet<>(ExprUtils.getVariables(expression));
      List<BoundVariable> bindingList = new ArrayList<>(usedVariables.size());

      for (Entry<NodeVariable, Node> substitution: bufferedSubstitutions.entrySet()) {
        if (usedVariables.contains(substitution.getKey())) {
          bindingList.add(NodeBinding.bindVariable(substitution.getKey(), substitution.getValue()));
        }
      }

      expression = Transformer.substituteAllBindings(new NodeBinding(expression, bindingList));

      bufferedSubstitutions.remove(nodeVariable);
      bufferedSubstitutions.put(nodeVariable, expression);
    }

    return bufferedSubstitutions.get(variable);
  }

  private Vector fillUnusedInputsRandomly(Vector inputVector) {
    Efsm efsm = this.simulator.getStaticModel();
    Transaction randomlyFilledTransaction = new Transaction();

    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": generating random values for unused inputs");

    Collection<Event> events = inputVector.getEventList().getEvents();
    Set<String> eventNames = new HashSet<>(events.size());

    for (Event event: events) {
      eventNames.add(event.getRangedVariable().getVariableName());
    }

    for (String inputName: efsm.getInputDataTypes().keySet()) {
      if (!inputVector.containsVariable(inputName)) {
        NodeValue value;
        Node constraint = efsm.getInvariantCondition(inputName);

        if (constraint == null) {
          value = new NodeValue(Random.newValue(efsm.getInputDataType(inputName)));
        } else {
          if ((efsm.getInputDataType(inputName).getTypeId() == DataTypeId.LOGIC_INTEGER) &&
              (constraint instanceof NodeOperation) &&
              (((NodeOperation) constraint).getOperandCount() == 2) &&
              (((NodeOperation) constraint).getOperand(0) instanceof NodeOperation) &&
              (((NodeOperation) constraint).getOperand(1) instanceof NodeOperation)) {
            NodeOperation firstCondition =
                (NodeOperation) ((NodeOperation) constraint).getOperand(0);
            NodeOperation secondCondition =
                (NodeOperation) ((NodeOperation) constraint).getOperand(1);

            if (firstCondition.getOperationId().equals(StandardOperation.GREATEREQ) &&
                secondCondition.getOperationId().equals(StandardOperation.LESSEQ)) {
              int min = ((NodeValue)firstCondition.getOperand(1)).getInteger().intValue();
              int max = ((NodeValue)secondCondition.getOperand(1)).getInteger().intValue();

              NodeValue module = new NodeValue(Data.newInteger(max - min));
              NodeValue minValueNode = new NodeValue(Data.newInteger(min));
              value = new NodeValue(Random.newValue(DataType.INTEGER));
              NodeVariable targetVariable = new NodeVariable(inputName, DataType.INTEGER);

              NodeOperation moduledValueComputation =
                  new NodeOperation(StandardOperation.MOD, value, module);
              NodeOperation valueComputation =
                  new NodeOperation(StandardOperation.ADD, moduledValueComputation, minValueNode);
              NodeOperation assignment =
                  new NodeOperation(StandardOperation.EQ, targetVariable, valueComputation);
              SolverResult result =
                  ConstraintUtils.solve(ConstraintUtils.newConstraint(assignment));

              if (result.getStatus() == Status.SAT) {
                value = new NodeValue(result.getVariables().iterator().next().getData());
              } else {
                throw new RetrascopeRuntimeException("Failed to resolve invariant constraint on " +
                    "input " + inputName + " due to the errors: " + result.getErrors());
              }
            } else {
              throw new IllegalArgumentException("Such invariant is currently unsupported: " +
                  constraint);
            }
          } else {
            SolverResult result = ConstraintUtils.solve(ConstraintUtils.newConstraint(constraint));

            if (result.getStatus() == Status.SAT) {
              value = new NodeValue(result.getVariables().iterator().next().getData());
            } else {
              throw new RetrascopeRuntimeException("Failed to resolve invariant constraint on " +
                  "input " + inputName + " due to the errors: " + result.getErrors());
            }
          }
        }

        randomlyFilledTransaction.addVariable(inputName, value);
      }
    }

    if (!randomlyFilledTransaction.isEmpty()) {
      inputVector.addTransaction(randomlyFilledTransaction);

      this.logger.log(LogLevel.DEBUG, this.logEntryHeader
          + ": randomly generated transaction: " + randomlyFilledTransaction);
    } else {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": all inputs are already set");
    }

    return inputVector;
  }

  private void unbufferReachabilityInformation(Map<EfsmTransition,
      Set<ReachabilityInformation>> reachabilityMap) {
    for (Entry<EfsmTransition, Set<ReachabilityInformation>> entry: reachabilityMap.entrySet()) {
      EfsmTransition transition = entry.getKey();

      if (this.reachabilityInformation.containsKey(transition)) {
        this.reachabilityInformation.get(transition).addAll(entry.getValue());
      } else {
        this.totalCoverage.markTransitionAsCovered(transition);
        this.reachabilityInformation.put(transition, entry.getValue());
      }
    }
  }

  @Override
  public TestCoverage getTotalCoverage() {
    return this.totalCoverage;
  }
}
