/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */
 
package ru.ispras.retrascope.engine.efsm.generator.test.fate;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.logging.Logger;

import org.apache.commons.cli.ParseException;

import ru.ispras.fortress.data.Random;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.basis.TestGenerator;
import ru.ispras.retrascope.engine.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.engine.efsm.generator.test.TestMinimiser;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.test.Vector;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

/**
 * The test generator for EFSM-based hardware models based on the algorithm described in the article
 * <a href="http://link.springer.com/article/10.1007%2Fs10836-011-5209-8">Efficient
 * Generation of Stimuli for Functional Verification by Backjumping Across Extended FSMs</a>.
 * <br>
 * FATE is the acronym for "functional ATPG (automated test pattern generator) to traverse EFSMs".
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class EfsmFatePlusTestGenerator extends TestGenerator {
  private final Logger logger;
  private final String logEntryHeader;

  /**
   * The engine ID of all {@code EfsmFateTestGenerator} instances.
   */
  public static final String ENGINE_ID = "efsm-fate-plus-test-generator".intern();

  public static final Parameter SEQUENCE_LENGTH_PARAMETER =
      new Parameter("sequence-length", true,
          "Amount of vectors in one randomly generated sequence");
  public static final Parameter SEQUENCES_NUMBER_PARAMETER =
      new Parameter("sequences-number", true,
          "Amount of sequences in a randomly generated fragment of a test");
  public static final Parameter LOOP_LIMIT =
      new Parameter("loop-limit", true, "Loop iteration limit");

  private static final int DEFAULT_LOOP_LIMIT = 1;

  /**
   * Constructs a generator.
   */
  public EfsmFatePlusTestGenerator() {
    super(EfsmFatePlusTestGenerator.ENGINE_ID, EfsmModel.TYPE);

    this.addParameter(EfsmFatePlusTestGenerator.SEQUENCE_LENGTH_PARAMETER);
    this.addParameter(EfsmFatePlusTestGenerator.SEQUENCES_NUMBER_PARAMETER);
    this.addParameter(EfsmFatePlusTestGenerator.LOOP_LIMIT);

    this.logEntryHeader = "EFSM.FATE+";
    this.logger = Log.getLogger(this.getClass());
  }
 
  @Override
  public Test start(Map<EntityType, Entity> inputs) {
    EfsmModel efsmModel = (EfsmModel) inputs.get(EfsmModel.TYPE);
 
    if (efsmModel == null) {
      throw new IllegalArgumentException("No EfsmModel objects have been found among arguments");
    }

    LinkedList<Efsm> efsms = new LinkedList<>(efsmModel.getEfsmList());

    int sequenceLength = -1;
    int sequencesNumber = -1;
    int loopLimit = -1;

    Arguments arguments = (Arguments) inputs.get(Arguments.TYPE);

    if (arguments != null) {
      try {
        Arguments parsedArguments = this.parseCommandLine(arguments.getCommandLine());

        sequenceLength = Integer.parseInt(parsedArguments.getValue(EfsmFatePlusTestGenerator.
            SEQUENCE_LENGTH_PARAMETER, "-1"));
        sequencesNumber = Integer.parseInt(parsedArguments.getValue(EfsmFatePlusTestGenerator.
            SEQUENCES_NUMBER_PARAMETER, "-1"));
        loopLimit =
            Integer.parseInt(parsedArguments.getValue(EfsmFatePlusTestGenerator.LOOP_LIMIT, "-1"));
      } catch (ParseException | NumberFormatException e) {
        throw new RetrascopeRuntimeException(
            "The exception has occurred while parsing command line arguments:", e);
      }
    }

    if (sequenceLength == -1) {
      int statesAmount = 0;

      for (Efsm efsm: efsms) {
        statesAmount = statesAmount + efsm.getStates().size();
      }

      sequenceLength = statesAmount;

      this.logger.log(LogLevel.INFO, this.logEntryHeader +
          ": sequence length isn't specified. Setting it to the default value: " + sequenceLength);
    } else {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": the specified sequence length: " +
          sequenceLength);
    }

    if (sequencesNumber == -1) {
      int transitionsAmount = 0;
      int statesAmount = 0;

      for (Efsm efsm: efsms) {
        transitionsAmount = transitionsAmount + efsm.getTransitions().size();
        statesAmount = statesAmount + efsm.getStates().size();
      }

      sequencesNumber = transitionsAmount / statesAmount;

      this.logger.log(LogLevel.INFO, this.logEntryHeader +
          ": amount of sequences isn't specified. Setting it to the default value: " +
          sequencesNumber);
    } else {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": the specified amount of sequences: " +
          sequencesNumber);
    }

    if (loopLimit == -1) {
      loopLimit = EfsmFatePlusTestGenerator.DEFAULT_LOOP_LIMIT;

      this.logger.log(LogLevel.INFO, this.logEntryHeader +
          ": loop limit isn't specified. Setting it to the default value: " + loopLimit);
    } else {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": the specified loop limit: " +
          loopLimit);
    }

    if (!efsms.isEmpty()) {
      EfsmTransitionCoverage totalCoverage = new EfsmTransitionCoverage();

      final Map<Efsm, Double> fCoefficients = new HashMap<>(efsms.size());
      final Map<Efsm, Double> aCoefficients = new HashMap<>(efsms.size());
      double aIncrement = Double.MAX_VALUE;

      Map<Efsm, RandomFatePlusGenerator> generators = new HashMap<>(efsms.size());

      for (Efsm efsm: efsms) {
        double f = 1d / (double) efsm.getInputDataTypes().size();

        if (f < aIncrement) {
          aIncrement = f;
        }

        fCoefficients.put(efsm, Double.valueOf(f));
        aCoefficients.put(efsm, Double.valueOf(0d));

        RandomFatePlusGenerator generator = new RandomFatePlusGenerator(efsm);

        generators.put(efsm, generator);

        totalCoverage.include(generator.getSequenceCoverage());
      }

      Random.setSeed(totalCoverage.getUncoveredTransitions().size());

      Comparator<Efsm> comparator = new Comparator<Efsm>() {
        @Override
        public int compare(Efsm efsm1, Efsm efsm2) {
          if (efsm1.isSynchronous() && !efsm2.isSynchronous()) {
            return -1;
          }

          if (!efsm1.isSynchronous() && efsm2.isSynchronous()) {
            return 1;
          }

          double efsm1Coefficient =
              fCoefficients.get(efsm1).doubleValue() + aCoefficients.get(efsm1).doubleValue();
          double efsm2Coefficient =
              fCoefficients.get(efsm2).doubleValue() + aCoefficients.get(efsm2).doubleValue();

          if (efsm1Coefficient == efsm2Coefficient) {
            return 0;
          }

          if (efsm1Coefficient > efsm2Coefficient) {
            return -1;
          } else {
            return 1;
          }
        }
      };

      Test test = new Test();
      Map<Sequence, EfsmTransitionCoverage> sequencesCoverage = new HashMap<>();
      Map<Efsm, Iterator<Vector>> vectorIterators = new HashMap<>(efsms.size());

      for (int i = 0; i < sequencesNumber; i++) {
        this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": generating a new sequence (" +
            (i + 1) + " of " + sequencesNumber + ")");

        Sequence sequence = new Sequence();
        EfsmTransitionCoverage sequenceCoverage = new EfsmTransitionCoverage();

        for (Efsm efsm: efsms) {
          vectorIterators.put(efsm, generators.get(efsm).getNextSequenceIterator());
        }

        for (int j = 0; j < sequenceLength; j++) {
          this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": generating a new vector " +
              (j + 1) + " of " + sequenceLength + ")");

          if (efsms.size() > 1) {
            Collections.sort(efsms, comparator);

            this.logger.log(LogLevel.DEBUG,
                this.logEntryHeader + ": current order of EFSMs: " + efsms);
            this.logger.log(LogLevel.DEBUG,
                this.logEntryHeader + ": current A coefficients: " + aCoefficients);
            this.logger.log(LogLevel.DEBUG,
                this.logEntryHeader + ": current F coefficients: " + fCoefficients);

            aCoefficients.put(efsms.getFirst(), Double.valueOf(0d));

            for (Efsm efsm: efsms) {
              if (!efsm.isSynchronous()) {
                aCoefficients.put(efsm, Double.valueOf(0d));

                break;
              }
            }
          }

          Vector vector = new Vector();

          for (Efsm efsm: efsms) {
            this.logger.log(LogLevel.DEBUG,
                this.logEntryHeader + ": processing the EFSM " + efsm.getName());

            RandomFatePlusGenerator generator = generators.get(efsm);
            generator.fixInputs(vector);
            vector = vectorIterators.get(efsm).next();

            EfsmTransitionCoverage efsmSequenceCoverage = generator.getSequenceCoverage();
            totalCoverage.include(efsmSequenceCoverage);
            sequenceCoverage.include(efsmSequenceCoverage);

            if (generator.wasForcedToChangeDirection()) {
              double newCoefficient = aCoefficients.get(efsm).doubleValue() + aIncrement;

              this.logger.log(LogLevel.DEBUG,
                  this.logEntryHeader + ": the generator was forced to change target transition."
                      + " Incrementing the A coefficient. New value: " + newCoefficient);

              aCoefficients.put(efsm, Double.valueOf(newCoefficient));
            }
          }

          sequence.addVector(vector);

          this.logger.log(LogLevel.DEBUG,
              this.logEntryHeader + ": the vector has been generated successfully: " + vector);
          this.logger.log(LogLevel.DEBUG,
              this.logEntryHeader + ": current coverage: " + totalCoverage);
        }

        test.addSequence(sequence);
        sequencesCoverage.put(sequence, sequenceCoverage);
      }

      this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
          ": the random phase of generation has been finished successfully. Total coverage: " +
          totalCoverage);

      if (!totalCoverage.isFull()) {
        this.logger.log(LogLevel.DEBUG,
            this.logEntryHeader + ": starting the directed phase of generation");

        for (Efsm efsm: efsms) {
          RandomFatePlusGenerator randomGenerator = generators.get(efsm);

          if (randomGenerator.getTotalCoverage().isFull()) {
            this.logger.log(LogLevel.DEBUG,
                this.logEntryHeader + ": the " + efsm + " EFSM is fully covered. Skipping it");
          } else {
            this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": the " + efsm +
                " EFSM has uncovered transitions. Processing it");

            DirectedFatePlusGenerator directedGenerator = new DirectedFatePlusGenerator(efsm, loopLimit,
                randomGenerator.getReachabilityInformation());

            for (Iterator<Vector> vectorIterator = directedGenerator.getNextSequenceIterator();
                vectorIterator != null; vectorIterator = directedGenerator.getNextSequenceIterator()) {
              Sequence sequence = new Sequence();

              while (vectorIterator.hasNext()) {
                sequence.addVector(vectorIterator.next());
              }

              test.addSequence(sequence);

              EfsmTransitionCoverage sequenceCoverage = directedGenerator.getSequenceCoverage();
              sequencesCoverage.put(sequence, sequenceCoverage);
              totalCoverage.include(sequenceCoverage);

              this.logger.log(LogLevel.DEBUG,
                  this.logEntryHeader + ": the directed sequence has been generated successfully");
              this.logger.log(LogLevel.DEBUG,
                  this.logEntryHeader + ": current coverage: " + totalCoverage);
            }
          }
        }

        this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
            ": the directed phase of generation has been finished successfully");
      }

      this.logger.log(LogLevel.INFO, this.logEntryHeader +
          ": generation has been finished successfully. Total coverage: " + totalCoverage);
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
          ": total amount of sequences in the generated test before minimisation: " +
          test.getSequencesAmount());
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
          ": total amount of vectors in the generated test before minimisation: " +
          test.getVectorsAmount());

      test = TestMinimiser.minimise(sequencesCoverage);

      this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
          ": total amount of sequences in the generated test after minimisation: " +
          test.getSequencesAmount());
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
          ": total amount of vectors in the generated test after minimisation: " +
          test.getVectorsAmount());

      for (EfsmTransition transition: totalCoverage.getUncoveredTransitions()) {
        this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
            ": uncovered transition: " + transition);
      }

      return test;
    } else {
      throw new IllegalArgumentException("No EFSMs are specified");
    }
  }
}
