/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */
package ru.ispras.retrascope.engine.efsm.generator.test.fate;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.Random;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.NodeBinding.BoundVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.SolverResult.Status;
import ru.ispras.fortress.solver.constraint.Constraint;
import ru.ispras.fortress.solver.constraint.ConstraintCombiner;
import ru.ispras.fortress.solver.constraint.ConstraintUtils;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.basis.AtomicTestGenerator;
import ru.ispras.retrascope.engine.basis.TestCoverage;
import ru.ispras.retrascope.engine.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.engine.efsm.generator.test.heuristic.direction.SemiRandomEfsmTransitionChooser;
import ru.ispras.retrascope.engine.efsm.simulator.EfsmSimulationSnapshot;
import ru.ispras.retrascope.engine.efsm.simulator.EfsmSimulator;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.result.test.Vector;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

/**
 * The FATE test generator which deals with single EFSM and operates at the vector-level.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see EfsmFateTestGenerator
 */
final class RandomFatePlusGenerator extends AtomicTestGenerator {
  private final Logger logger;
  private final String logEntryHeader;

  private Map<EfsmTransition, Set<ReachabilityInformation>> reachabilityInformation;

  private final EfsmTransitionCoverage totalCoverage;

  private Sequence currentSequence;

  private final EfsmSimulator simulator;

  private Vector fixedVector;

  private boolean wasForcedToChangeDirection;

  private long seed;

  /**
   * Constructs a test generator for the specified EFSM.
   *
   * @param efsm the EFSM
   */
  RandomFatePlusGenerator(Efsm efsm) {
    this.logEntryHeader = "EFSM.FATE+.RandomGenerator." + efsm.getName();
    this.logger = Log.getLogger(this.getClass());

    this.simulator = new EfsmSimulator(efsm);

    this.fixedVector = null;

    this.totalCoverage = new EfsmTransitionCoverage(efsm);

    this.reachabilityInformation = new HashMap<>(efsm.getTransitions().size());

    this.seed = 0L;
  }

  Map<EfsmTransition, Set<ReachabilityInformation>> getReachabilityInformation() {
    return this.reachabilityInformation;
  }

  /**
   * Returns an EFSM which this generator is related to.
   *
   * @return the EFSM
   */
  Efsm getEfsm() {
    return this.simulator.getStaticModel();
  }

  /**
   * Returns a sequenceCoverage reached during generation.
   *
   * @return the sequenceCoverage
   */
  @Override
  public EfsmTransitionCoverage getSequenceCoverage() {
    return new EfsmTransitionCoverage(this.simulator.getCoverage());
  }

  /**
   * Fixes the specified inputs.
   *
   * @param vector a vector containing the inputs to be fixed
   */
  void fixInputs(Vector vector) {
    this.fixedVector = vector;
  }

  /**
   * Returns an iterator enabling to retrieve next sequence of vectors produced by this generator.
   *
   * @return the iterator. If this generator is already finished generation, {@code null} is
   * returned
   */
  @Override
  public Iterator<Vector> getNextSequenceIterator() {
    return new RandomFateSequenceIterator();
  }

  Vector generateInputVectorRandomly() {
    this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": generating a new input vector");

    Vector vector = null;
    this.wasForcedToChangeDirection = false;
    EfsmSimulationSnapshot sourceStateSnapshot = null;
    EfsmTransition traversedTransition = null;

    if (this.simulator.getCurrentState().equals(Efsm.PREINITIAL_STATE)) {
      vector = this.reset();
    } else {
      SemiRandomEfsmTransitionChooser chooser =
          new SemiRandomEfsmTransitionChooser(this.simulator.getStaticModel(),
              this.simulator.getCurrentState(), this.totalCoverage, this.seed);

      for (EfsmTransition transition = chooser.chooseNext(); transition != null;
          transition = chooser.chooseNext()) {
        vector = this.traverseTransition(transition);

        if (vector != null) {
          traversedTransition = transition;
          sourceStateSnapshot = this.simulator.makeSimulationSnapshot();

          break;
        }
      }
    }

    if (vector == null) {
      if (this.fixedVector != null) {
        vector = this.fixedVector;

        this.logger.log(LogLevel.DEBUG, this.logEntryHeader +
            ": failed to generate any vector. Applying the externally fixed vector: " + vector);
      } else {
        vector = new Vector();

        this.logger.log(LogLevel.WARNING, this.logEntryHeader +
            ": failed to generate any vector. Applying the empty vector: " + vector);
      }
    }

    vector = this.fillUnusedInputsRandomly(vector);

    this.simulator.setInputValues(vector.getTransactions());
    this.simulator.processEvents(vector.getEventList());

    if (traversedTransition != null) {
      ReachabilityInformation reachabilityInformation =
          new ReachabilityInformation(this.currentSequence, this.currentSequence.getLength(),
              sourceStateSnapshot, this.simulator.makeSimulationSnapshot());

      Set<ReachabilityInformation> transitionReachabilityInformation =
          this.reachabilityInformation.get(traversedTransition);

      if (transitionReachabilityInformation == null) {
        transitionReachabilityInformation = new HashSet<>();
        this.reachabilityInformation.put(traversedTransition, transitionReachabilityInformation);
      }

      transitionReachabilityInformation.add(reachabilityInformation);
      this.totalCoverage.markTransitionAsCovered(traversedTransition);
    }

    this.currentSequence.addVector(vector);

    return vector;
  }

  private Vector reset() {
    this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": resetting the EFSM");

    this.simulator.reset();

    Vector vector = this.enableGuardedAction(this.simulator.getStaticModel().
        getResetGuardedActions().iterator().next(), null);//TODO

    if (vector != null) {
      this.logger.log(LogLevel.DEBUG,
          this.logEntryHeader + ": the EFSM has been reset successfully");
    } else {
      this.logger.log(LogLevel.WARNING,
          this.logEntryHeader + ": failed to reset the EFSM");
    }

    return vector;
  }

  private Vector traverseTransition(EfsmTransition transition) {
    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": traversing the transition: " + transition);

    Vector inputVector = this.generateInputsForTransition(transition, null);

    if (inputVector != null) {
      this.logger.log(LogLevel.DEBUG,
          this.logEntryHeader + ": the transition has been traversed successfully");
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader
          + ": input vector: " + inputVector);
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader
          + ": current sequenceCoverage: " + this.simulator.getCoverage());
    } else {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": failed to traverse the transition");
    }

    return inputVector;
  }

  private Vector generateInputsForTransition(EfsmTransition transition, Node additionalExpression) {
    if (!this.simulator.getCurrentState().equals(transition.getSourceState())) {
      throw new IllegalArgumentException(
          "The current state of the simulator doesn't match the transition's source state: " +
          this.simulator.getCurrentState() + " vs " + transition.getSourceState());
    }

    return this.enableGuardedAction(transition.getGuardedAction(), additionalExpression);
  }

  private Vector enableGuardedAction(GuardedAction guardedAction, Node additionalExpression) {
    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": enabling the guarded action: " + guardedAction);

    if (this.fixedVector != null) {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": fixed inputs: " + this.fixedVector);

      EventList gaEventList = new EventList();
      gaEventList.addEvents(guardedAction.getClocks());

      if (!gaEventList.compatibleWith(this.fixedVector.getEventList())) {
        this.logger.log(LogLevel.DEBUG,
            this.logEntryHeader + ": the transition isn't compatible with the fixed events");

        this.wasForcedToChangeDirection = true;

        return null;
      }
    }

    Node guardExpression = guardedAction.getGuard().toNode();
    Collection<NodeVariable> usedVariables = ExprUtils.getVariables(guardExpression);
    List<BoundVariable> bindingList = new ArrayList<>(usedVariables.size());
    boolean intersectionWithFixedVector = false;
    Set<Node> inputConstraints = new HashSet<>();
    Efsm efsm = this.simulator.getStaticModel();

    for (NodeVariable usedVariable: usedVariables) {
      String variableName = usedVariable.getName();

      if (efsm.containsInput(variableName)) {
        Node variableConstraint = efsm.getInvariantCondition(variableName);

        if (variableConstraint != null) {
          inputConstraints.add(variableConstraint);
        }
      } else {
        Map<String, NodeValue> innerVariables = this.simulator.getInnerValues();

        if (innerVariables.containsKey(variableName)) {
          bindingList.add(NodeBinding.bindVariable(usedVariable, innerVariables.get(variableName)));
        } else {
          if ((this.fixedVector != null) && this.fixedVector.containsVariable(variableName)) {
            intersectionWithFixedVector = true;
            bindingList.add(NodeBinding.bindVariable(usedVariable,
                this.fixedVector.getValue(variableName)));
          }
        }
      }
    }

    guardExpression =
        Transformer.substituteAllBindings(new NodeBinding(guardExpression, bindingList));

    Constraint targetConstraint = ConstraintUtils.newConstraint(guardExpression);

    StringBuilder logMessageBuilder = new StringBuilder(this.logEntryHeader);
    logMessageBuilder.append(": trying to generate an input vector for the guard expression ");
    logMessageBuilder.append(guardExpression);

    if (additionalExpression != null)
    {
      logMessageBuilder.append("\n").append("Additional constraint: ").append(additionalExpression);

      Constraint additionalConstraint = ConstraintUtils.newConstraint(additionalExpression);
      targetConstraint = ConstraintCombiner.makeConjunction(targetConstraint, additionalConstraint);

      for (NodeVariable usedVariable: ExprUtils.getVariables(additionalExpression)) {
        String variableName = usedVariable.getName();

        if (efsm.containsInput(variableName)) {
          Node variableConstraint = efsm.getInvariantCondition(variableName);

          if (variableConstraint != null) {
            inputConstraints.add(variableConstraint);
          }
        }
      }
    }

    if (!inputConstraints.isEmpty()) {
      Node inputConstraint =
          ExprUtils.getConjunction(inputConstraints.toArray(new Node[inputConstraints.size()]));

      logMessageBuilder.append("\n").append("Constraint on inputs: ").append(inputConstraint);

      targetConstraint = ConstraintCombiner.makeConjunction(targetConstraint,
          ConstraintUtils.newConstraint(inputConstraint));
    }

    this.logger.log(LogLevel.DEBUG, logMessageBuilder.toString());

    SolverResult solverResult = ConstraintUtils.solve(targetConstraint);

    if (solverResult.getStatus().equals(Status.SAT)) {
        Vector inputVector = new Vector();
        Transaction transaction = new Transaction();

        for (Variable variable : solverResult.getVariables()) {
          String variableName = variable.getName();

          if (efsm.getInputDataTypes().containsKey(variableName)) {
            transaction.addVariable(variableName, new NodeValue(variable.getData()));
          }
        }

        inputVector.addEvents(guardedAction.getClocks());

        if (!transaction.isEmpty()) {
          inputVector.addTransaction(transaction);
        }

        this.logger.log(LogLevel.DEBUG, this.logEntryHeader
            + ": the input vector has been generated successfully: " + inputVector);

        if (this.fixedVector != null) {
          inputVector.addEvents(this.fixedVector.getEventList().getEvents());

          for (Transaction fixedTransaction: this.fixedVector.getTransactions()) {
            inputVector.addTransaction(fixedTransaction);
          }
        }

        return inputVector;
    } else {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": failed to generate an input vector");
      
      this.wasForcedToChangeDirection = intersectionWithFixedVector;

      return null;
    }
  }

  private Vector fillUnusedInputsRandomly(Vector inputVector) {
    Efsm efsm = this.simulator.getStaticModel();
    Transaction randomlyFilledTransaction = new Transaction();

    this.logger.log(LogLevel.DEBUG,
        this.logEntryHeader + ": generating random values for unused inputs");

    Collection<Event> events = inputVector.getEventList().getEvents();
    Set<String> eventNames = new HashSet<>(events.size());

    for (Event event: events) {
      eventNames.add(event.getRangedVariable().getVariableName());
    }

    for (String inputName: efsm.getInputDataTypes().keySet()) {
      if (!inputVector.containsVariable(inputName)) {
        NodeValue value;
        
        Node constraint = efsm.getInvariantCondition(inputName);

        if (constraint == null) {
          value = new NodeValue(Random.newValue(efsm.getInputDataType(inputName)));
        } else {
          if ((efsm.getInputDataType(inputName).getTypeId() == DataTypeId.LOGIC_INTEGER) &&
              (constraint instanceof NodeOperation) &&
              (((NodeOperation) constraint).getOperandCount() == 2) &&
              (((NodeOperation) constraint).getOperand(0) instanceof NodeOperation) &&
              (((NodeOperation) constraint).getOperand(1) instanceof NodeOperation)) {
            NodeOperation firstCondition =
                (NodeOperation) ((NodeOperation) constraint).getOperand(0);
            NodeOperation secondCondition =
                (NodeOperation) ((NodeOperation) constraint).getOperand(1);

            if (firstCondition.getOperationId().equals(StandardOperation.GREATEREQ) &&
                secondCondition.getOperationId().equals(StandardOperation.LESSEQ)) {
              int min = ((NodeValue)firstCondition.getOperand(1)).getInteger().intValue();
              int max = ((NodeValue)secondCondition.getOperand(1)).getInteger().intValue();

              NodeValue module = new NodeValue(Data.newInteger(max - min));
              NodeValue minValueNode = new NodeValue(Data.newInteger(min));
              value = new NodeValue(Random.newValue(DataType.INTEGER));
              NodeVariable targetVariable = new NodeVariable(inputName, DataType.INTEGER);

              NodeOperation moduledValueComputation =
                  new NodeOperation(StandardOperation.MOD, value, module);
              NodeOperation valueComputation =
                  new NodeOperation(StandardOperation.ADD, moduledValueComputation, minValueNode);
              NodeOperation assignment =
                  new NodeOperation(StandardOperation.EQ, targetVariable, valueComputation);
              SolverResult result =
                  ConstraintUtils.solve(ConstraintUtils.newConstraint(assignment));

              if (result.getStatus() == Status.SAT) {
                value = new NodeValue(result.getVariables().iterator().next().getData());
              } else {
                throw new RetrascopeRuntimeException("Failed to resolve invariant constraint on " +
                    "input " + inputName + " due to the errors: " + result.getErrors());
              }
            } else {
              throw new IllegalArgumentException("Such invariant is currently unsupported: " +
                  constraint);
            }
          } else {
            SolverResult result = ConstraintUtils.solve(ConstraintUtils.newConstraint(constraint));

            if (result.getStatus() == Status.SAT) {
              value = new NodeValue(result.getVariables().iterator().next().getData());
            } else {
              throw new RetrascopeRuntimeException("Failed to resolve invariant constraint on " +
                  "input " + inputName + " due to the errors: " + result.getErrors());
            }
          }
        }

        randomlyFilledTransaction.addVariable(inputName, value);
      }
    }

    if (!randomlyFilledTransaction.isEmpty()) {
      inputVector.addTransaction(randomlyFilledTransaction);

      this.logger.log(LogLevel.DEBUG, this.logEntryHeader
          + ": randomly generated transaction: " + randomlyFilledTransaction);
    } else {
      this.logger.log(LogLevel.DEBUG, this.logEntryHeader + ": all inputs are already set");
    }

    return inputVector;
  }

  /**
   * Checks if during last vector generation this generator was forced to change target transition
   * because of an incompatibility of its GA with a fixed vector.
   *
   * @return {@code true} if this generator was forced to change direction, otherwise returns
   * {@code false}
   */
  boolean wasForcedToChangeDirection() {
    return this.wasForcedToChangeDirection;
  }

  private class RandomFateSequenceIterator implements Iterator<Vector> {
    private RandomFateSequenceIterator() {
      RandomFatePlusGenerator.this.simulator.reset();
      RandomFatePlusGenerator.this.seed++;
      RandomFatePlusGenerator.this.currentSequence = new Sequence();
    }

    @Override
    public boolean hasNext() {
      return true;
    }

    @Override
    public Vector next() {
      if (this.hasNext()) {
        return RandomFatePlusGenerator.this.generateInputVectorRandomly();
      } else {
        return null;
      }
    }

    @Override
    public void remove() {
      throw new UnsupportedOperationException("remove");
    }
  }

  @Override
  public TestCoverage getTotalCoverage() {
    return this.totalCoverage;
  }
}
