/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.printer.graphml;

import ru.ispras.fortress.expression.Node;
import ru.ispras.retrascope.basis.FileCreator;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmModelVisitor;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.Guard;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.result.graphml.GraphMlUtils;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Map;

/**
 * EFSM visitor for GraphML printer.
 * 
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class EfsmGraphMlVisitor implements EfsmModelVisitor {

  /* Writer to GraphML output file. */
  private PrintWriter printWriter;

  /* Map of identifiers for transitions. */
  private Map<EfsmTransition, Integer> transitionMap;

  /* Map of identifiers for states. */
  private Map<EfsmState, Integer> stateMap;

  /* Model identifier. */
  private int modelId;

  /**
   * Create GraphML file with the specified name.
   *
   * @param fileName file name
   */
  public void initFile(final String fileName) {
    try {
      /* Create GraphML file. */
      final File graphMlFile = FileCreator.newFile(fileName);

      modelId = 0;
      final FileWriter fileWriter = new FileWriter(graphMlFile.getAbsoluteFile());
      printWriter = new PrintWriter(new BufferedWriter(fileWriter));
    } catch (IOException e) {
      throw new RetrascopeRuntimeException(e);
    }
  }

  @Override
  public void onEfsmBegin(final Efsm efsm) {
    GraphMlUtils.onGraphBegin(printWriter, Integer.toString(modelId));
    stateMap = new HashMap<>();
    transitionMap = new HashMap<>();
    modelId++;  
  }

  @Override
  public void onEfsmEnd(final Efsm efsm) {
    GraphMlUtils.onGraphEnd(printWriter);
  }

  @Override
  public void onStateBegin(final EfsmState state) {
    final int idState = getStateId(state);

    GraphMlUtils.printNode(printWriter, 
        Integer.toHexString(idState) + Integer.toString(modelId), state.getId(),
        GraphMlUtils.FORM_ELLIPSE, GraphMlUtils.COLOR_YELLOW);
  }

  private int getStateId(final EfsmState state) {
    final Integer idState = stateMap.get(state);

    if (idState == null) {
      final int sizeMap = stateMap.size();
      stateMap.put(state, sizeMap);
      return sizeMap;
    }

    return idState;
  }

  @Override
  public void onStateEnd(final EfsmState state) {
    /* Empty. */
  }

  @Override
  public void onTransitionBegin(final EfsmTransition transition) {
    Integer idTransition = transitionMap.get(transition);

    if (idTransition == null) {
      final int sizeMap = transitionMap.size();
      transitionMap.put(transition, sizeMap);
      idTransition = sizeMap;
    }

    /* Get child & parent nodes. */
    final int sourceState = getStateId(transition.getSourceState());
    final GuardedAction guardedAction = transition.getGuardedAction();

    final String guardId = "Guard" + Integer.toHexString(idTransition);
    GraphMlUtils.printNode(printWriter, guardId, guardedAction.getGuard().getId(),
        GraphMlUtils.FORM_DIAMOND, GraphMlUtils.COLOR_BLUE, GraphMlUtils.CROPPING_TEXT);
    
    GraphMlUtils.printEdge(
        printWriter,
        Integer.toHexString(sourceState)  + Integer.toString(modelId), guardId);

    final String actionId = "Action" + Integer.toHexString(idTransition);
    GraphMlUtils.printNode(printWriter, actionId, guardedAction.getAction().getId(),
        GraphMlUtils.FORM_RECTANGLE, GraphMlUtils.COLOR_GREEN, GraphMlUtils.CROPPING_TEXT);
    GraphMlUtils.printEdge(printWriter, guardId, actionId);

    final int destinationState = getStateId(transition.getDestinationState());
    
    GraphMlUtils.printEdge(
        printWriter, actionId,
        Integer.toHexString(destinationState)  + Integer.toString(modelId));
  }

  @Override
  public void onTransitionEnd(final EfsmTransition transition) {
    /* Empty. */
  }

  @Override
  public void onGuardedActionBegin(final GuardedAction guardedAction) {
    /* Empty. */
  }

  @Override
  public void onGuardedActionEnd(final GuardedAction guardedAction) {
    /* Empty. */
  }

  @Override
  public void onEvent(final Event event) {
    /* Empty. */
  }

  @Override
  public void onGuardBegin(final Guard guard) {
    /* Empty. */
  }

  @Override
  public void onGuardEnd(final Guard guard) {
    /* Empty. */
  }

  @Override
  public void onAssignmentActionBegin(final Action action) {
    /* Empty. */
  }

  @Override
  public void onAssignmentActionEnd(final Action action) {
    /* Empty. */
  }

  @Override
  public void onAssignmentBegin(final Assignment statement) {
    /* Empty. */
  }

  @Override
  public void onAssignmentEnd(final Assignment statement) {
    /* Empty. */
  }

  @Override
  public void onExpression(final Node expression) {
    /* Empty. */
  }

  @Override
  public void onEfsmModelBegin(final EfsmModel efsmModel) {
    GraphMlUtils.onGraphMlBegin(printWriter);
  }

  @Override
  public void onEfsmModelEnd(final EfsmModel efsmModel) {
    GraphMlUtils.onGraphMlEnd(printWriter);

    GraphMlUtils.closeGraphMlFile(printWriter);
  }
}
