/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.test.printer.testbench.vhdl;

import org.apache.commons.cli.ParseException;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.test.printer.testbench.DeciderLikeTestPatternPrinter;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.result.cfginterface.CfgInterface;
import ru.ispras.retrascope.result.cfginterface.CfgSignalInterface;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.testbench.VhdlTestbench;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

/**
 * The engine that makes VHDL testbenches from tests.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see Test
 * @see VhdlTestbench
 */
public class TestVhdlTestbenchPrinter extends Engine {

  /**
   * The engine ID of all {@code TestVhdlTestbenchPrinter} instances.
   */
  public static final String ENGINE_ID = "test-vhdl-testbench-printer".intern();

  public static final Parameter TESTBENCH_DIRECTORY =
      new Parameter("testbench-dir", true, "A directory where testbenches are saved");

  /**
   * The default name of testbench directory.
   */
  public static final String TESTBENCH_DIR_NAME = "testbenches";

  private static final String DEFAULT_TESTBENCH_DIRECTORY =
      TESTBENCH_DIR_NAME + File.separator + TestVhdlTestbenchPrinter.MODULE_NAME_PLACEHOLDER;

  private static final String TEST_FILE_EXTENSION = ".tst";

  private static final Path TESTBENCH_PATTERN_FILE =
      Paths.get("src", "main", "templates", "vhdl", "testbench");
  private static final Path DECIDER_PARSER_FILE =
      Paths.get("src", "main", "templates", "vhdl", "tst_parser");

  private static final String DEFAULT_CLOCK_PERIOD_VALUE = "10";
  private static final String DEFAULT_CLOCK_PERIOD_UNIT = "ns";

  private static final String MODULE_NAME_PLACEHOLDER = "%MODULE_NAME%";
  private static final String CLOCK_PERIOD_VALUE_PLACEHOLDER = "%PERIOD_VALUE%";
  private static final String CLOCK_PERIOD_UNIT_PLACEHOLDER = "%PERIOD_UNIT%";
  private static final String TEST_FILE_NAME_PLACEHOLDER = "%TEST_FILE_NAME%";
  private static final String PORT_DECLARATION_PLACEHOLDER = "%PORT_DECLARATION%";
  private static final String SIGNAL_DECLARATION_PLACEHOLDER = "%SIGNAL_DECLARATION%";
  private static final String BUFFER_VARIABLE_DECLARATION_PLACEHOLDER =
      "%BUFFER_VARIABLE_DECLARATION%";
  private static final String PORT_MAP_DECLARATION_PLACEHOLDER = "%PORT_MAP_DECLARATION%";
  private static final String TABULATION_PLACEHOLDER = "%TABULATION%";
  private static final String PORT_NAME_PLACEHOLDER = "%PORT_NAME%";
  private static final String PORT_KIND_PLACEHOLDER = "%PORT_KIND%";
  private static final String PORT_TYPE_PLACEHOLDER = "%PORT_TYPE%";
  private static final String SIGNAL_NAME_PLACEHOLDER = "%SIGNAL_NAME%";
  private static final String SIGNAL_TYPE_PLACEHOLDER = "%SIGNAL_TYPE%";
  private static final String MAX_LINE_LENGTH_PLACEHOLDER = "%MAX_LINE_LENGTH%";
  private static final String MAX_VALUE_LENGTH_PLACEHOLDER = "%MAX_VALUE_LENGTH%";
  private static final String SIGNAL_ASSIGNMENT_PLACEHOLDER = "%SIGNAL_ASSIGNMENT%";
  private static final String BUFFER_VARIABLE_ASSIGNMENT_PLACEHOLDER =
      "%BUFFER_VARIABLE_ASSIGNMENT%";
  private static final String CONVERSION_FUNCTION_PLACEHOLDER = "%CONVERSION_FUNCTION%";
  private static final String BUFFER_VARIABLE_NAME_POSTFIX = "_buffer";

  private static final String PORT_DECLARATION = TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER +
      TestVhdlTestbenchPrinter.PORT_NAME_PLACEHOLDER + ": " +
      TestVhdlTestbenchPrinter.PORT_KIND_PLACEHOLDER + " " +
      TestVhdlTestbenchPrinter.PORT_TYPE_PLACEHOLDER;
  private static final String SIGNAL_DECLARATION = TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER +
      "signal " + TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER + ": " +
      TestVhdlTestbenchPrinter.SIGNAL_TYPE_PLACEHOLDER + ";";
  private static final String BUFFER_VARIABLE_DECLARATION =
      TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER +
      "variable " + TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER +
      TestVhdlTestbenchPrinter.BUFFER_VARIABLE_NAME_POSTFIX + ": " +
      TestVhdlTestbenchPrinter.SIGNAL_TYPE_PLACEHOLDER + ";";
  private static final String PORT_MAP_DECLARATION =
      TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER +
      TestVhdlTestbenchPrinter.PORT_NAME_PLACEHOLDER + " => " +
      TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER;

  private static final String SIGNAL_ASSIGNMENT = TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER +
      TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER + " <= " +
      TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER +
      TestVhdlTestbenchPrinter.BUFFER_VARIABLE_NAME_POSTFIX + ";";

  private static final String BUFFER_VARIABLE_ASSIGNMENT =
      TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER +
      TestVhdlTestbenchPrinter.CONVERSION_FUNCTION_PLACEHOLDER + "(test_line_string, " +
      TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER +
      TestVhdlTestbenchPrinter.BUFFER_VARIABLE_NAME_POSTFIX + ");";

  private static final String INPUT_PORT_KIND = "in";
  private static final String OUTPUT_PORT_KIND = "out";

  private static final String STD_LOGIC_VECTOR_PATTERN =
      "std_logic_vector (%LOWER_BOUND% to %UPPER_BOUND%)";
  private static final String UPPER_BOUND_PLACEHOLDER = "%UPPER_BOUND%";
  private static final String LOWER_BOUND_PLACEHOLDER = "%LOWER_BOUND%";

  private static final Map<DataType, String> DATA_TYPES_MAP = new HashMap<>();

  static {
    TestVhdlTestbenchPrinter.DATA_TYPES_MAP.put(DataType.BIT_VECTOR(1), "std_logic");
    TestVhdlTestbenchPrinter.DATA_TYPES_MAP.put(DataType.INTEGER, "integer");
    TestVhdlTestbenchPrinter.DATA_TYPES_MAP.put(DataType.BOOLEAN, "bit");
    TestVhdlTestbenchPrinter.DATA_TYPES_MAP.put(DataType.REAL, "real");
  }

  private static final Map<String, String> CONVERSION_FUNCTIONS_MAP =
      new TreeMap<>(String.CASE_INSENSITIVE_ORDER);

  static {
    TestVhdlTestbenchPrinter.CONVERSION_FUNCTIONS_MAP.put("std_logic", "string_to_std_logic");
    TestVhdlTestbenchPrinter.CONVERSION_FUNCTIONS_MAP.put("std_logic_vector",
    		"string_to_std_logic_vector");
    TestVhdlTestbenchPrinter.CONVERSION_FUNCTIONS_MAP.put("integer", "string_to_integer");
    TestVhdlTestbenchPrinter.CONVERSION_FUNCTIONS_MAP.put("bit", "string_to_bit");
    TestVhdlTestbenchPrinter.CONVERSION_FUNCTIONS_MAP.put("bit_vector", "string_to_bit_vector");
    TestVhdlTestbenchPrinter.CONVERSION_FUNCTIONS_MAP.put("real", "string_to_real");
  }

  /**
   * Creates a new {@code TestVhdlTestbenchPrinter} instance.
   */
  public TestVhdlTestbenchPrinter() {
    super(TestVhdlTestbenchPrinter.ENGINE_ID, EntityType.get(Test.ENTITY_TYPE_ID),
        EntityType.get(VhdlTestbench.ENTITY_TYPE_ID));
    this.addInputType(CfgInterface.TYPE);
    this.addParameter(TestVhdlTestbenchPrinter.TESTBENCH_DIRECTORY);
  }

  @Override
  public VhdlTestbench start(Map<EntityType, Entity> inputs) {
    Test test = (Test) inputs.get(EntityType.get(Test.ENTITY_TYPE_ID));
    CfgInterface cfgInterface = (CfgInterface) inputs.get(CfgInterface.TYPE);

    if (test == null) {
      throw new IllegalArgumentException("No test entity has been found among inputs");
    }

    if (cfgInterface == null) {
      throw new IllegalArgumentException("No interface entity has been found among inputs");
    }

    if (Files.notExists(TestVhdlTestbenchPrinter.TESTBENCH_PATTERN_FILE) ||
        !Files.isRegularFile(TestVhdlTestbenchPrinter.TESTBENCH_PATTERN_FILE)) {
      throw new RetrascopeRuntimeException("No testbench pattern file " +
          TestVhdlTestbenchPrinter.TESTBENCH_PATTERN_FILE + " has been found");
    }

    if (Files.notExists(TestVhdlTestbenchPrinter.DECIDER_PARSER_FILE) ||
        !Files.isRegularFile(TestVhdlTestbenchPrinter.DECIDER_PARSER_FILE)) {
      throw new RetrascopeRuntimeException("No decider parser library " +
          TestVhdlTestbenchPrinter.DECIDER_PARSER_FILE + " has been found");
    }

    Map<String, CfgSignalInterface> interfaceMapping = cfgInterface.getMapping();

    if (cfgInterface.getMapping().isEmpty()) {
      throw new IllegalArgumentException("No interface information has been provided");
    }

    if (cfgInterface.getMapping().size() > 1) {
      throw new IllegalArgumentException("CfgInterface containing multiple modules is unsupported");
    }

    String moduleName = interfaceMapping.keySet().iterator().next();
    String normalisedModuleName = this.normaliseModuleName(moduleName);
    CfgSignalInterface externalSignals = interfaceMapping.get(moduleName);
    Collection<NodeVariable> inputSignals = externalSignals.getInputs();
    Set<String> inputSignalNames = new LinkedHashSet<>(inputSignals.size());

    for (NodeVariable input: inputSignals) {
      inputSignalNames.add(input.getName());
    }

    Arguments arguments = (Arguments) inputs.get(Arguments.TYPE);

    String testbenchDirectoryString = TestVhdlTestbenchPrinter.DEFAULT_TESTBENCH_DIRECTORY;

    if (arguments != null) {
      try {
        Arguments parsedArguments = this.parseCommandLine(arguments.getCommandLine());

        testbenchDirectoryString =
            parsedArguments.getValue(TestVhdlTestbenchPrinter.TESTBENCH_DIRECTORY,
                testbenchDirectoryString);
      } catch (ParseException | NumberFormatException e) {
        throw new IllegalArgumentException("The exception has occurred while parsing command line "
            + "arguments", e);
      }
    }

    Path testbenchDirectory =
        Paths.get(testbenchDirectoryString.replace(TestVhdlTestbenchPrinter.MODULE_NAME_PLACEHOLDER,
            normalisedModuleName));

    if (!Files.exists(testbenchDirectory)) {
      try {
        Files.createDirectories(testbenchDirectory);
      } catch (IOException exception) {
        throw new RetrascopeRuntimeException(
            "The exception has occurred while creating the target directory", exception);
      }
    }

    Path testPatternFile = testbenchDirectory.resolve(normalisedModuleName +
        TestVhdlTestbenchPrinter.TEST_FILE_EXTENSION);

    try {
      DeciderLikeTestPatternPrinter.print(test, testPatternFile, inputSignalNames);
    } catch (IOException e) {
      throw new RetrascopeRuntimeException(
          "The exception has occurred while printing test pattern file", e);
    }

    int maxLineLength = 0;
    int maxValueLength = 0;

    try (BufferedReader testReader =
        Files.newBufferedReader(testPatternFile, Charset.defaultCharset())) {
      for (String line = testReader.readLine(); line != null; line = testReader.readLine()) {
        if (line.length() > maxLineLength) {
          maxLineLength = line.length();
        }

        String[] values = line.split(" ");

        for (String value: values) {
          if (value.length() > maxValueLength) {
            maxValueLength = value.length();
          }
        }
      }
    } catch (IOException e) {
      throw new RetrascopeRuntimeException(
          "The exception has occurred while reading test pattern file", e);
    }

    Path deciderParserFile = testbenchDirectory.resolve(TestVhdlTestbenchPrinter.
        DECIDER_PARSER_FILE.getFileName() + ".vhd");

    Path testbenchFile = testbenchDirectory.resolve(normalisedModuleName + "_testbench" + ".vhd");
    Set<Path> additionalFiles = new HashSet<Path>(2);
    additionalFiles.add(testPatternFile);
    additionalFiles.add(deciderParserFile);
    VhdlTestbench testbench = new VhdlTestbench(testbenchFile, additionalFiles);

    try (BufferedReader deciderParserReader =
        Files.newBufferedReader(TestVhdlTestbenchPrinter.DECIDER_PARSER_FILE,
            Charset.defaultCharset());
        BufferedWriter deciderParserWriter = Files.newBufferedWriter(deciderParserFile,
            Charset.defaultCharset(), StandardOpenOption.CREATE,
            StandardOpenOption.TRUNCATE_EXISTING)) {
      for (String line = deciderParserReader.readLine(); line != null; 
          line = deciderParserReader.readLine()) {
        line = line.replace(TestVhdlTestbenchPrinter.MAX_VALUE_LENGTH_PLACEHOLDER,
            String.valueOf(maxValueLength));
        deciderParserWriter.write(line);
        deciderParserWriter.newLine();
      }
    } catch (IOException e1) {
      throw new RetrascopeRuntimeException(
          "The exception occurred while copying decider parser library: " + e1);
    }

    try (BufferedReader templateReader =
        Files.newBufferedReader(TestVhdlTestbenchPrinter.TESTBENCH_PATTERN_FILE,
            Charset.defaultCharset());
        BufferedWriter testbenchWriter =
            Files.newBufferedWriter(testbench.getTestbenchFile(), Charset.defaultCharset())) {
      Map<String, VhdlSignal> signalMapping = this.mapExternalSignals(externalSignals);

      for (String templateLine = templateReader.readLine(); templateLine != null;
          templateLine = templateReader.readLine()) {

        String testbenchLine = templateLine.
            replace(TestVhdlTestbenchPrinter.MODULE_NAME_PLACEHOLDER, normalisedModuleName).
            replace(TestVhdlTestbenchPrinter.TEST_FILE_NAME_PLACEHOLDER,
                testPatternFile.getFileName().toString()).
            replace(TestVhdlTestbenchPrinter.CLOCK_PERIOD_UNIT_PLACEHOLDER,
                TestVhdlTestbenchPrinter.DEFAULT_CLOCK_PERIOD_UNIT).
            replace(TestVhdlTestbenchPrinter.CLOCK_PERIOD_VALUE_PLACEHOLDER,
                TestVhdlTestbenchPrinter.DEFAULT_CLOCK_PERIOD_VALUE).
            replace(TestVhdlTestbenchPrinter.MAX_LINE_LENGTH_PLACEHOLDER,
                String.valueOf(maxLineLength)).
            replace(TestVhdlTestbenchPrinter.MAX_VALUE_LENGTH_PLACEHOLDER,
                String.valueOf(maxValueLength));

        if (testbenchLine.contains(TestVhdlTestbenchPrinter.PORT_DECLARATION_PLACEHOLDER)) {
          String tabulation = testbenchLine.substring(0,
              testbenchLine.indexOf(TestVhdlTestbenchPrinter.PORT_DECLARATION_PLACEHOLDER));

          for (String portDeclaration: this.buildPortDeclaration(tabulation, externalSignals)) {
            testbenchWriter.append(portDeclaration);
            testbenchWriter.newLine();
          }

          continue;
        }

        if (testbenchLine.contains(TestVhdlTestbenchPrinter.SIGNAL_DECLARATION_PLACEHOLDER)) {
          String tabulation = testbenchLine.substring(0,
              testbenchLine.indexOf(TestVhdlTestbenchPrinter.SIGNAL_DECLARATION_PLACEHOLDER));

          for (String signalDeclaration:
              this.buildSignalDeclaration(tabulation, signalMapping.values())) {
            testbenchWriter.append(signalDeclaration);
            testbenchWriter.newLine();
          }

          continue;
        }

        if (testbenchLine.contains(TestVhdlTestbenchPrinter.
            BUFFER_VARIABLE_DECLARATION_PLACEHOLDER)) {
          String tabulation = testbenchLine.substring(0,
              testbenchLine.indexOf(TestVhdlTestbenchPrinter.
                  BUFFER_VARIABLE_DECLARATION_PLACEHOLDER));

          for (String signalDeclaration:
              this.buildBufferVariableDeclaration(tabulation, signalMapping.values())) {
            testbenchWriter.append(signalDeclaration);
            testbenchWriter.newLine();
          }

          continue;
        }

        if (testbenchLine.contains(TestVhdlTestbenchPrinter.PORT_MAP_DECLARATION_PLACEHOLDER)) {
          String tabulation = testbenchLine.substring(0,
              testbenchLine.indexOf(TestVhdlTestbenchPrinter.PORT_MAP_DECLARATION_PLACEHOLDER));

          for (String signalDeclaration: this.buildPortMapDeclaration(tabulation, signalMapping)) {
            testbenchWriter.append(signalDeclaration);
            testbenchWriter.newLine();
          }

          continue;
        }

        if (testbenchLine.contains(TestVhdlTestbenchPrinter.SIGNAL_ASSIGNMENT_PLACEHOLDER)) {
          String tabulation = testbenchLine.substring(0,
              testbenchLine.indexOf(TestVhdlTestbenchPrinter.SIGNAL_ASSIGNMENT_PLACEHOLDER));

          for (String inputSignalName: inputSignalNames) {
            testbenchWriter.append(this.buildSignalAssignemnt(tabulation,
                signalMapping.get(inputSignalName)));
            testbenchWriter.newLine();
          }

          continue;
        }

        if (testbenchLine.contains(TestVhdlTestbenchPrinter.
            BUFFER_VARIABLE_ASSIGNMENT_PLACEHOLDER)) {
          String tabulation = testbenchLine.substring(0,
              testbenchLine.indexOf(TestVhdlTestbenchPrinter.
                  BUFFER_VARIABLE_ASSIGNMENT_PLACEHOLDER));

          for (String inputSignalName: inputSignalNames) {
            testbenchWriter.append(this.buildBufferVariableAssignemnt(tabulation,
                signalMapping.get(inputSignalName)));
            testbenchWriter.newLine();
          }

          continue;
        }

        testbenchWriter.append(testbenchLine);
        testbenchWriter.newLine();
      }

      testbenchWriter.flush();
    } catch (IOException e) {
      throw new RetrascopeRuntimeException(
          "The exception has occurred while processing testbench file", e);
    }

    return testbench;
  }

  private List<String> buildPortDeclaration(String tabulation, CfgSignalInterface externalSignals) {
    List<String> portDeclaration = new ArrayList<>();

    for (NodeVariable input: externalSignals.getInputs()) {
      if (!portDeclaration.isEmpty()) {
        int lastIndex = portDeclaration.size() - 1;
        portDeclaration.set(lastIndex, portDeclaration.get(lastIndex) + ";");
      }

      portDeclaration.add(TestVhdlTestbenchPrinter.PORT_DECLARATION.
          replaceAll(TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER, tabulation).
          replaceAll(TestVhdlTestbenchPrinter.PORT_NAME_PLACEHOLDER, input.getName()).
          replaceAll(TestVhdlTestbenchPrinter.PORT_KIND_PLACEHOLDER,
              TestVhdlTestbenchPrinter.INPUT_PORT_KIND).
          replaceAll(TestVhdlTestbenchPrinter.PORT_TYPE_PLACEHOLDER,
              this.mapDataType(input)));
    }

    for (NodeVariable output: externalSignals.getOutputs()) {
      if (!portDeclaration.isEmpty()) {
        int lastIndex = portDeclaration.size() - 1;
        portDeclaration.set(lastIndex, portDeclaration.get(lastIndex) + ";");
      }

      portDeclaration.add(TestVhdlTestbenchPrinter.PORT_DECLARATION.
          replaceAll(TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER, tabulation).
          replaceAll(TestVhdlTestbenchPrinter.PORT_NAME_PLACEHOLDER, output.getName()).
          replaceAll(TestVhdlTestbenchPrinter.PORT_KIND_PLACEHOLDER,
              TestVhdlTestbenchPrinter.OUTPUT_PORT_KIND).
          replaceAll(TestVhdlTestbenchPrinter.PORT_TYPE_PLACEHOLDER,
                  this.mapDataType(output)));
    }

    return portDeclaration;
  }

  private List<String> buildSignalDeclaration(String tabulation, Collection<VhdlSignal> signals) {
    List<String> signalDeclaration = new ArrayList<>();

    for (VhdlSignal signal: signals) {
      signalDeclaration.add(TestVhdlTestbenchPrinter.SIGNAL_DECLARATION.
          replaceAll(TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER, tabulation).
          replaceAll(TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER, signal.name).
          replaceAll(TestVhdlTestbenchPrinter.SIGNAL_TYPE_PLACEHOLDER, signal.type));
    }

    return signalDeclaration;
  }

  private List<String> buildBufferVariableDeclaration(String tabulation,
      Collection<VhdlSignal> signals) {
    List<String> signalDeclaration = new ArrayList<>();

    for (VhdlSignal signal: signals) {
      signalDeclaration.add(TestVhdlTestbenchPrinter.BUFFER_VARIABLE_DECLARATION.
          replaceAll(TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER, tabulation).
          replaceAll(TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER, signal.name).
          replaceAll(TestVhdlTestbenchPrinter.SIGNAL_TYPE_PLACEHOLDER, signal.type));
    }

    return signalDeclaration;
  }

  private List<String> buildPortMapDeclaration(String tabulation, Map<String, VhdlSignal> portMap) {
    List<String> portMapDeclaration = new ArrayList<>();

    for (Map.Entry<String, VhdlSignal> mapping: portMap.entrySet()) {
      if (!portMapDeclaration.isEmpty()) {
        int lastIndex = portMapDeclaration.size() - 1;
        portMapDeclaration.set(lastIndex, portMapDeclaration.get(lastIndex) + ",");
      }

      portMapDeclaration.add(TestVhdlTestbenchPrinter.PORT_MAP_DECLARATION.
          replaceAll(TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER, tabulation).
          replaceAll(TestVhdlTestbenchPrinter.PORT_NAME_PLACEHOLDER, mapping.getKey()).
          replaceAll(TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER, mapping.getValue().name));
    }

    return portMapDeclaration;
  }

  private String mapDataType(NodeVariable variable) {
    VariableData variableData = (VariableData) variable.getUserData();

    if (variableData != null) {
      String vhdlDataType = variableData.getMetaInfo(MetaInfoType.HDL_TYPE_STR.name());

      if (vhdlDataType != null) {
        return vhdlDataType;
      }
    }

    DataType dataType = variable.getDataType();

    if (TestVhdlTestbenchPrinter.DATA_TYPES_MAP.containsKey(dataType)) {
      return TestVhdlTestbenchPrinter.DATA_TYPES_MAP.get(dataType);
    } else {
      if (dataType.getTypeId().equals(DataTypeId.BIT_VECTOR)) {
        return TestVhdlTestbenchPrinter.STD_LOGIC_VECTOR_PATTERN.
            replaceAll(TestVhdlTestbenchPrinter.UPPER_BOUND_PLACEHOLDER,
                String.valueOf(dataType.getSize() - 1)).
            replaceAll(TestVhdlTestbenchPrinter.LOWER_BOUND_PLACEHOLDER, "0");
      } else {
        throw new IllegalArgumentException("Unsupported data type: " + dataType);
      }
    }
  }

  private Map<String, VhdlSignal> mapExternalSignals(CfgSignalInterface externalSignals) {
    Collection<NodeVariable> inputSignals = externalSignals.getInputs();
    Collection<NodeVariable> outputSignals = externalSignals.getOutputs();

    Map<String, VhdlSignal> mapping = new HashMap<>(inputSignals.size() + outputSignals.size());

    for (NodeVariable input: inputSignals) {
      mapping.put(input.getName(), new VhdlSignal(input.getName(), this.mapDataType(input)));
    }

    for (NodeVariable output: outputSignals) {
      mapping.put(output.getName(), new VhdlSignal(output.getName(), this.mapDataType(output)));
    }

    return mapping;
  }

  private String buildBufferVariableAssignemnt(String tabulation, VhdlSignal signal) {
    String assignment = TestVhdlTestbenchPrinter.BUFFER_VARIABLE_ASSIGNMENT.
        replaceAll(TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER, tabulation).
        replaceAll(TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER, signal.name);

    String signalCoreType = signal.type.split(" |\\(")[0];

    if (TestVhdlTestbenchPrinter.CONVERSION_FUNCTIONS_MAP.containsKey(signalCoreType)) {
      assignment = assignment.replace(TestVhdlTestbenchPrinter.CONVERSION_FUNCTION_PLACEHOLDER,
          TestVhdlTestbenchPrinter.CONVERSION_FUNCTIONS_MAP.get(signalCoreType));
    } else {
      throw new RetrascopeRuntimeException("Unsupported signal type: " + signal.type);
    }

    return assignment;
  }

  private String buildSignalAssignemnt(String tabulation, VhdlSignal signal) {
    return TestVhdlTestbenchPrinter.SIGNAL_ASSIGNMENT.
        replaceAll(TestVhdlTestbenchPrinter.TABULATION_PLACEHOLDER, tabulation).
        replaceAll(TestVhdlTestbenchPrinter.SIGNAL_NAME_PLACEHOLDER, signal.name);
  }

  private class VhdlSignal {
    public final String name;
    public final String type;

    public VhdlSignal(String name, String type) {
      this.name = name;
      this.type = type;
    }
  }

  private String normaliseModuleName(String moduleName) {
    InvariantChecks.checkNotNull(moduleName);

    String[] splittedName = moduleName.split("\\.");

    if (splittedName.length == 1) {
      return moduleName;
    } else {
      return splittedName[splittedName.length - 1].replaceAll("\\(.*\\)", "");
    }
  }
}
