/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */
package ru.ispras.retrascope.engine.test.printer.xml;

import java.nio.file.Paths;
import java.util.Map;
import java.util.logging.Logger;

import javax.xml.bind.JAXB;

import org.apache.commons.cli.ParseException;

import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.parser.test.TestFileNames;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

/**
 * The class for marshalling tests to XML files.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see ru.ispras.retrascope.result.test.Test
 */
public class TestXmlPrinter extends Engine {
  /**
   * The engine ID of all {@code TestXmlPrinter} instances.
   */
  public static final String ENGINE_ID = "test-xml-printer".intern();

  private static final Parameter FILE_NAME_PARAMETER =
      new Parameter("file-name", true, "Output XML file name");

  private static final String DEFAULT_FILE_NAME = "test.xml";

  private final Logger logger;
  private final String logEntryHeader;

  /**
   * Constructs an {@code TestXmlPrinter} instance.
   */
  public TestXmlPrinter() {
    super(TestXmlPrinter.ENGINE_ID, EntityType.get(Test.ENTITY_TYPE_ID),
        EntityType.get(TestFileNames.ENTITY_TYPE_ID));

    this.addParameter(TestXmlPrinter.FILE_NAME_PARAMETER);

    this.logEntryHeader = "Test XML printer";
    this.logger = Log.getLogger(this.getClass());
  }

  @Override
  public TestFileNames start(Map<EntityType, Entity> inputs) {
    Test test = null;
    String fileName = null;

    for (Entity entity: inputs.values()) {
      if (entity instanceof Test) {
        test = (Test) entity;
      } else {
        if (entity instanceof Arguments) {
          try {
            Arguments arguments = this.parseCommandLine(((Arguments) entity).getCommandLine());
            fileName = arguments.getValue(TestXmlPrinter.FILE_NAME_PARAMETER);
          } catch (ParseException | NumberFormatException e) {
            this.logger.log(LogLevel.ERROR, this.logEntryHeader +
                ": the exception has occurred while parsing command line arguments:", e);
          }
        }
      }
    }

    if (test != null) {
      if (fileName == null) {
        fileName = TestXmlPrinter.DEFAULT_FILE_NAME;

        this.logger.log(LogLevel.INFO, this.logEntryHeader +
            ": an output file name is not specified. Using the default value: " + fileName);
      }

      this.logger.log(LogLevel.INFO, this.logEntryHeader + ": starting to marshal the test: " +
          test);

      JAXB.marshal(test, Paths.get(fileName).toFile());

      this.logger.log(LogLevel.INFO,
          this.logEntryHeader + ": the test has been marshalled successfully");

      return new TestFileNames(fileName);
    } else {
      throw new IllegalArgumentException("No Test instances have been found among input arguments");
    }
  }
}
