/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Assignment statement.
 *
 * <p>Consists of defined {@link RangedVariable} target
 * variable and used value expression.</p>
 *
 * @see RangedVariable
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 * 
 */
public class Assignment extends MetaInfo implements UseDef {

  private static final String STRING_SEPARATOR = ":=";

  private RangedVariable target;

  private Node value;

  /**
   * Default constructor.
   */
  public Assignment() {
    super();
  }

  /**
   * Constructs object with the specified target variable and value expression.
   *
   * @param target defined variable
   * @param value used expression
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  public Assignment(final RangedVariable target, final Node value) {
    super();
    InvariantChecks.checkNotNull(target);
    InvariantChecks.checkNotNull(value);

    this.target = target;
    this.value = value;
  }

  /**
   * Constructs object with the specified target non-ranged variable and value expression.
   *
   * @param var defined variable
   * @param expr used expression
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  public Assignment(final NodeVariable var, final Node expr) {
    super();
    InvariantChecks.checkNotNull(var);
    InvariantChecks.checkNotNull(expr);

    this.target = new RangedVariable(var);
    this.value = expr;
  }

  /**
   * Creates a deep copy of the current object.
   *
   * @return full copy of the current object
   */
  public Assignment deepCopy() {
    final Assignment deepCopy = new Assignment(getTarget().deepCopy(), getValue().deepCopy());
    deepCopy.addMetaInfo(getMetaInfo());
    return deepCopy;
  }

  /**
   * Returns defined variable.
   *
   * @return variable
   */
  public RangedVariable getTarget() {
    return this.target;
  }

  /**
   * Sets the specified variable as objects' target.
   *
   * @param variable new target variable
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void setVariable(final RangedVariable variable) {
    InvariantChecks.checkNotNull(variable);

    this.target = variable;
  }

  /**
   * Returns value expression.
   *
   * @return used expression
   */
  public Node getValue() {
    return this.value;
  }

  /**
   * Sets new value expression.
   *
   * @param value new value expression
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void setValue(final Node value) {
    InvariantChecks.checkNotNull(value);

    this.value = value;
  }

  /**
   * Returns names of variables that are used in this object.
   *
   * @return set of used variables' names.
   */
  public Set<String> getUsedVariableNames() {
    final Collection<NodeVariable> usedVariables = getUses();
    final Set<String> usedVariableNames = new LinkedHashSet<>();

    for (final NodeVariable variable: usedVariables) {
      usedVariableNames.add(variable.getName());
    }
    return usedVariableNames;
  }

  /**
   * Returns a description of object that is stored in format of the specified printer.
   *
   * @param printer expression printer
   * @param separator assignment separator between defined and used variables
   *
   * @return description
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public String getDescription(final ExprTreePrinter printer, final String separator) {
    InvariantChecks.checkNotNull(printer);
    InvariantChecks.checkNotNull(separator);

    return this.target.toString() + separator + printer.toString(this.value);
  }

  @Override
  public Set<NodeVariable> getUses() {
    Set<NodeVariable> uses;
    if (this.value == null) {
      uses = new LinkedHashSet<>();
    } else {
      uses = new LinkedHashSet<>(ExprUtils.getVariables(this.value));
    }
    return uses;
  }

  @Override
  public Set<NodeVariable> getDefines() {
    final Set<NodeVariable> defines = new LinkedHashSet<>();

    if (getTarget() != null) {
      defines.add(getTarget().getVariable());
    }

    return defines;
  }

  @Override
  public int hashCode() {
    int prime = 31;
    int result = 1;
    result = prime * result + ((this.value == null) ? 0 : this.value.hashCode());
    result = prime * result + this.target.hashCode();

    return result;
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }

    if (obj == null || !(obj instanceof Assignment)) {
      return false;
    }

    final Assignment other = (Assignment) obj;

    if (this.value == null) {
      if (other.value != null) {
        return false;
      }
    } else if (!this.value.equals(other.value)) {
      return false;
    }

    return this.target.equals(other.target);
  }

  @Override
  public String toString() {
    return this.target + " " + STRING_SEPARATOR + " " + this.value;
  }
}
