/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Meta-information storage.
 *
 * <p>This object stores some meta-information. Meta-information is presented in pairs
 * like ({@link MetaInfoType} key, {@link MetaInfoValue} value).</p>
 *
 * @see MetaInfoType
 * @see MetaInfoValue
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class MetaInfo {

  private final Map<String, String> metaInfoMap;

  /**
   * Default constructor.
   */
  public MetaInfo() {
    this.metaInfoMap = new LinkedHashMap<>();
  }

  /**
   * Checks rather object contains meta-information of the specified predefined type.
   *
   * @return type key
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public boolean hasMetaInfo(final Enum<?> type) {
    InvariantChecks.checkNotNull(type);

    return this.metaInfoMap.containsKey(type.name());
  }

  /**
   * Checks rather object contains meta-information for the specified user-defined type.
   *
   * @return type key
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public boolean hasMetaInfo(final String type) {
    InvariantChecks.checkNotNull(type);

    return this.metaInfoMap.containsKey(type);
  }

  /**
   * Adds meta-information as a pair of the specified information type and information value.
   *
   * <p>It is supposed that this method is used for the predefined types and values.</p>
   *
   * @param type information type
   * @param value information value
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public void addMetaInfo(final Enum<?> type, final Enum<?> value) {
    InvariantChecks.checkNotNull(type);
    InvariantChecks.checkNotNull(value);

    this.metaInfoMap.put(type.name(), value.name());
  }

  /**
   * Adds meta-information as a pair of the specified information type and information value.
   *
   * <p>It is supposed that this method is used for the predefined types.</p>
   *
   * @param type information type
   * @param value information value
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public void addMetaInfo(final Enum<?> type, final String value) {
    InvariantChecks.checkNotNull(type);
    InvariantChecks.checkNotNull(value);

    this.metaInfoMap.put(type.name(), value);
  }

  /**
   * Adds meta-information as a pair of the specified information type and information value.
   *
   * @param type information type
   * @param value information value
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public void addMetaInfo(final String type, final String value) {
    InvariantChecks.checkNotNull(type);
    InvariantChecks.checkNotNull(value);

    this.metaInfoMap.put(type, value);
  }

  /**
   * Adds the specified meta-information mapping.
   *
   * @param map meta-information mapping
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addMetaInfo(final Map<String, String> map) {
    InvariantChecks.checkNotNull(map);

    this.metaInfoMap.putAll(map);
  }

  /**
   * Removes information of the specified type.
   *
   * @param type information type
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void removeMetaInfo(final Enum<?> type) {
    InvariantChecks.checkNotNull(type);

    this.metaInfoMap.remove(type.name());
  }

  /**
   * Returns meta-information of the specified type.
   *
   * @param type information type
   * @return information if an object contains it, {@code null} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public String getMetaInfo(final Enum<?> type) {
    InvariantChecks.checkNotNull(type);

    return this.metaInfoMap.get(type.name());
  }

  /**
   * Returns meta-information of the specified type.
   *
   * @param key information type
   * @return information if an object contains it, {@code null} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public String getMetaInfo(final String key) {
    InvariantChecks.checkNotNull(key);

    return this.metaInfoMap.get(key);
  }

  /**
   * Returns all the meta-information.
   *
   * @return meta-information map
   */
  public Map<String, String> getMetaInfo() {
    return this.metaInfoMap;
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }

    final MetaInfo metaInfo = (MetaInfo) obj;

    return metaInfoMap.equals(metaInfo.metaInfoMap);
  }

  @Override
  public int hashCode() {
    return metaInfoMap.hashCode();
  }
}
