/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;

import javax.xml.bind.annotation.XmlAttribute;

/**
 * Pair of values.
 *
 * <p>Object can be treated as a sub-list of indices.
 * One (young) value is less or equal than another (old) value.</p>
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 * 
 */
public class Range {

  private static final String DELIMITER = ":";

  @XmlAttribute private final Node young;

  @XmlAttribute private final Node old;

  /**
   * Default constructor.
   *
   * @param young young index
   * @param old old index
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when any of the arguments is {@code null};
   *         (2) when the second argument is always less than first argument.
   */
  public Range(final Node young, final Node old) {
    InvariantChecks.checkNotNull(young);
    InvariantChecks.checkNotNull(old);

    Enum<?> opId = ExprUtils.isType(DataTypeId.BIT_VECTOR, old, young)
        ? StandardOperation.BVUGE : StandardOperation.GREATEREQ;

    if (!ExprUtils.isSAT(new NodeOperation(opId, old, young))) {
      throw new IllegalArgumentException(
          String.format("%s < %s.", old.toString(), young.toString()));
    }

    this.young = young;
    this.old = old;
  }

  /**
   * Creates a deep copy of the current object.
   *
   * @return a full copy of the current object
   */
  public Range deepCopy() {
    return new Range(this.young.deepCopy(), this.old.deepCopy());
  }

  /**
   * Returns young index.
   *
   * @return young index
   */
  public Node getYoung() {
    return this.young;
  }

  /**
   * Returns old index.
   *
   * @return old index
   */
  public Node getOld() {
    return this.old;
  }

  @Override
  public String toString() {
    return this.young + DELIMITER + this.old;
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }

    final Range range = (Range) obj;

    return this.young == range.young && this.old == range.old;
  }

  @Override
  public int hashCode() {
    int result = this.young.hashCode();
    result = 31 * result + this.old.hashCode();
    return result;
  }
}
