/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.data.types.bitvector.BitVectorAlgorithm;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;

/**
 * Variable or it's sub-part restricted by the specified range.
 *
 * <p>When the range is not {@code null} it means that not all variable is in use
 * but only it's sub-array.</p>
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class RangedVariable {

  private final NodeVariable variable;

  private Range range;

  /**
   * Constructs object with the specified variable.
   *
   * @param variable variable
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public RangedVariable(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    this.variable = variable;
    this.range = null;
  }

  /**
   * Constructs object with the specified variable and it's range.
   *
   * <p>The ranged variable MUST have either bit vector data type
   * or map data type.</p>
   *
   * @param variable variable
   * @param range range
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when any of the arguments is {@code null};
   *         (2) when variable argument is neither of
   *         {@link ru.ispras.fortress.data.types.bitvector.BitVector} data type
   *         nor of {@link ru.ispras.fortress.data.types.datamap.DataMap} data type;
   *         (3) when variable is of {@link ru.ispras.fortress.data.types.bitvector.BitVector} type
   *         but range old/young bit numbers are out of
   *         {@link ru.ispras.fortress.data.types.bitvector.BitVector} size;
   *         (4) when variable is of {@link ru.ispras.fortress.data.types.datamap.DataMap} type
   *         but range old/young bit numbers are not equal.
   */
  public RangedVariable(final NodeVariable variable, final Range range) {
    InvariantChecks.checkNotNull(variable);
    InvariantChecks.checkNotNull(range);

    checkParams(variable, range);

    this.variable = variable;
    this.range = range;
  }

  /**
   * Creates a deep copy of the current object.
   *
   * @return object's copy
   */
  public RangedVariable deepCopy() {
    final RangedVariable deepCopy = new RangedVariable((NodeVariable) this.variable.deepCopy());
    if (this.isRanged()) {
      deepCopy.setRange(this.range.deepCopy());
    }
    return deepCopy;
  }

  /**
   * Returns variable.
   *
   * @return variable
   */
  public NodeVariable getVariable() {
    return this.variable;
  }

  /**
   * Returns variable name.
   *
   * @return variable name
   */
  public String getVariableName() {
    return this.getVariable().getName();
  }

  /**
   * Returns range.
   *
   * @return range
   */
  public Range getRange() {
    return this.range;
  }

  /**
   * Sets new range.
   *
   * @param range range
   *
   *  @throws IllegalArgumentException in the following situations:
   *         (1) when any argument is {@code null};
   *         (2) when this object's variable is neither of
   *         {@link ru.ispras.fortress.data.types.bitvector.BitVector}
   *         nor of {@link ru.ispras.fortress.data.types.datamap.DataMap} data types;
   *         (3) when this object's variable
   *         is of {@link ru.ispras.fortress.data.types.bitvector.BitVector} type
   *         but range old/young bit numbers are out of
   *         {@link ru.ispras.fortress.data.types.bitvector.BitVector} size;
   *         (4) when variable is of {@link ru.ispras.fortress.data.types.datamap.DataMap} type
   *         but range old/young bit numbers are not equal.
   */
  public void setRange(final Range range) {
    InvariantChecks.checkNotNull(range);
    checkParams(getVariable(), range);

    this.range = range;
  }

  /**
   * Checks rather this is ranged.
   *
   * @return {@code true} if this contains non-{@code null} range, {@code false} otherwise
   */
  public boolean isRanged() {
    return this.range != null;
  }

  private static void checkParams(
      final NodeVariable variable,
      final Range range) {

    if (!variable.isType(DataTypeId.BIT_VECTOR)
        && !variable.isType(DataTypeId.MAP)) {
      throw new IllegalArgumentException(
          "Illegal ranged variable type id: " + variable.getDataTypeId() + ".");
    }

    final Node young = range.getYoung();
    final Node old = range.getOld();

    if (variable.isType(DataTypeId.BIT_VECTOR)) {

      /* Check rather range does not spread outside the 0..size-1 bit numbers. */

      if (ExprUtils.isType(DataTypeId.BIT_VECTOR, old, young)) {

        final DataType dataType = old.getDataType();
        final int size = dataType.getSize();

        final BitVector zero = BitVector.newEmpty(size);
        BitVectorAlgorithm.fill(zero, (byte) 0);
        final NodeValue zeroValue = NodeValue.newBitVector(zero);

        final BitVector max = BitVector.newEmpty(size);
        BitVectorAlgorithm.fill(max, (byte) 1);
        final NodeValue maxValue = NodeValue.newBitVector(max);

        final Node inRange =
            ExprUtils.getConjunction(
                new NodeOperation(StandardOperation.BVUGE, young, zeroValue),
                new NodeOperation(StandardOperation.BVULE, old, maxValue));

        if (!ExprUtils.isSAT(inRange)) {

          throw new IllegalArgumentException(
              String.format("Incorrect range: %s < %s or %s > %s.", young, zero, old, max));
        }
      } else if (ExprUtils.isType(DataType.INTEGER, old, young)) {

        final NodeValue zero = NodeValue.newInteger(0);
        final NodeValue max = NodeValue.newInteger(variable.getDataType().getSize());

        final Node inRange =
            ExprUtils.getConjunction(
                new NodeOperation(StandardOperation.GREATEREQ, young, zero),
                new NodeOperation(StandardOperation.LESSEQ, old, max));

        if (!ExprUtils.isSAT(inRange)) {

          throw new IllegalArgumentException(
            String.format("Incorrect range: %s < %s or %s > %s.", young, zero, old, max));
        }
      }
    }

    if (variable.isType(DataTypeId.MAP)
        && !ExprUtils.isSAT(new NodeOperation(StandardOperation.EQ, young, old))) {
      throw new IllegalArgumentException(
          String.format("Incorrect range: %s != %s for map", old, young));
    }
  }

  @Override
  public String toString() {
    final StringBuilder builder = new StringBuilder();
    builder.append(this.variable.toString());
    if (this.range != null) {
      builder.append('[');
      builder.append(this.range.toString());
      builder.append(']');
    }
    return builder.toString();
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }

    boolean result;
    final RangedVariable that = (RangedVariable) obj;

    if (this.range == null || that.getRange() == null) {
      result = this.variable.equals(that.variable);
    } else {
      result = this.variable.equals(that.variable) && this.range.equals(that.range);
    }
    return result;
  }

  @Override
  public int hashCode() {
    int result = this.variable.hashCode();
    if (this.range != null) {
      result = 31 * result + this.range.hashCode();
    }
    return result;
  }
}
