/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis.memory;

import ru.ispras.fortress.expression.NodeValue;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

/**
 * The hash map-based store of variables.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class HashMapMemory implements Memory {

  private static final String LEFT_STRING_BOUND = "{";
  private static final String VARIABLE_STRING_DELIMITER = "; ";
  private static final String VARIABLE_NAME_VALUE_STRING_DELIMITER = " = ";
  private static final String RIGHT_STRING_BOUND = "}";

  private final Map<String, NodeValue> store;

  /**
   * Constructs an empty (values of variables are equal to {@code null}) hash map-based store of
   * variables.
   *
   * @param variableNames a collection of names of variables to be stored
   */
  public HashMapMemory(final Collection<String> variableNames) {

    this.store = new HashMap<>(variableNames.size());

    for (final String variableName : variableNames) {
      this.store.put(variableName, null);
    }
  }

  /**
   * Constructs hash map-based store of variables as a copy of the specified {@link HashMapMemory}
   * instance.
   *
   * @param memory the memory to be copied
   */
  public HashMapMemory(final HashMapMemory memory) {
    this.store = new HashMap<>(memory.getValues());
  }

  @Override
  public void setValue(
      final String name,
      final NodeValue value) {

    this.store.put(name, value);
  }

  @Override
  public void setValues(final Map<String, NodeValue> values) {
    this.store.putAll(values);
  }

  @Override
  public boolean containsVariable(final String name) {
    return this.store.containsKey(name);
  }

  @Override
  public boolean containsVariables(final Collection<String> names) {

    for (final String name : names) {
      if (!this.store.containsKey(name)) {
        return false;
      }
    }

    return true;
  }

  @Override
  public NodeValue getValue(final String name) {
    if (this.store.containsKey(name)) {
      return this.store.get(name);
    } else {
      throw new IllegalArgumentException("The variable " + name
          + " is not found among the stored variables");
    }
  }

  @Override
  public Map<String, NodeValue> getValues(final Collection<String> names) {

    final Map<String, NodeValue> result = new HashMap<>(names.size());

    for (final String name : names) {
      if (this.store.containsKey(name)) {
        result.put(name, this.store.get(name));
      } else {
        throw new IllegalArgumentException("The variable " + name
            + " is not found among the stored variables");
      }
    }

    return result;
  }

  @Override
  public Map<String, NodeValue> getValues() {
    return this.store;
  }

  @Override
  public String toString() {

    final StringBuilder stringBuilder = new StringBuilder(HashMapMemory.LEFT_STRING_BOUND);

    final int initialStringBuilderLength = stringBuilder.length();

    for (final String variableName: this.store.keySet()) {
      if (stringBuilder.length() > initialStringBuilderLength) {
        stringBuilder.append(HashMapMemory.VARIABLE_STRING_DELIMITER);
      }

      stringBuilder.append(variableName);
      stringBuilder.append(HashMapMemory.VARIABLE_NAME_VALUE_STRING_DELIMITER);
      stringBuilder.append(this.store.get(variableName));
    }

    stringBuilder.append(HashMapMemory.RIGHT_STRING_BOUND);

    return stringBuilder.toString();
  }
}
