/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprPrinter;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.AssignmentContainer;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.UseDef;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * Basic block node of process' control flow graph.
 *
 * <p>Basic blocks can be either of concurrent (non-blocking) type
 * or of sequential (blocking) type. The concurrent type means
 * that all the assignments should be executed simultaneously
 * in such a concurrent mode. The sequential type means that
 * all the assignments should be executed sequentially in the same order
 * they have in it's collection.</p>
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 * 
 */
public class BasicBlock extends CfgNode implements AssignmentContainer,UseDef {

  /* The assignments. */
  private final List<Assignment> assignments;

  /* The concurrent type flag. */
  private boolean concurrent;

  /* The assignment separators depending on HDL and concurrent type. */

  private static final String VHDL_CON = "<=";
  private static final String VHDL_NON_CON = ":=";
  private static final String VERILOG_CON = "<=";
  private static final String VERILOG_NON_CON = "=";

  /**
   * Constructs object of the specified type flag.
   *
   * @param concurrent block type flag
   */
  public BasicBlock(final boolean concurrent) {
    super();
    this.assignments = new ArrayList<>();
    this.concurrent = concurrent;
  }

  /**
   * Constructs object with the specified assignments and the type flag.
   *
   * @param assignments assignments
   * @param concurrent block type flag
   *
   * @throws IllegalArgumentException when first argument is {@code null}.
   */
  public BasicBlock(final List<Assignment> assignments, final boolean concurrent) {
    super();
    InvariantChecks.checkNotNull(assignments);

    this.assignments = assignments;
    this.concurrent = concurrent;
  }

  /**
   * Constructs object with the specified collection of assignments.
   *
   * <p>The default value of concurrent flag is {@code false}</p>
   *
   * @param assignments assignments
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public BasicBlock(final List<Assignment> assignments) {
    this(assignments, false);
  }

  /**
   * Constructs object with the specified assignment and the flag type.
   *
   * @param assignment assignment
   * @param concurrent block type flag
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public BasicBlock(final Assignment assignment, final boolean concurrent) {
    super();
    InvariantChecks.checkNotNull(assignment);

    this.assignments = new ArrayList<>();
    this.assignments.add(assignment);
    this.concurrent = concurrent;
  }

  /**
   * Constructs object with the specified assignment and of non-concurrent type.
   *
   * @param assignment assignment
   */
  public BasicBlock(final Assignment assignment) {
    this(assignment, false);
  }

  /**
   * Constructs object that contains one assignment with the specified target and value.
   *
   * @see Assignment
   *
   * @param target target variable
   * @param value expression value
   */
  public BasicBlock(final RangedVariable target, final Node value) {
    this(new Assignment(target, value));
  }

  /**
   * Constructs object that contains one assignment with the specified target and value.
   *
   * @see Assignment
   *
   * @param target target variable
   * @param value expression value
   */
  public BasicBlock(final NodeVariable target, final Node value) {
    this(new Assignment(target, value));
  }

  /**
   * Adds assignment statement.
   *
   * <p>Assignment will be added with on 1 greater number
   * than other assignments have.</p>
   *
   * @param stmt assignment to be added
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  @Override
  public void addAssignment(final Assignment stmt) {
    InvariantChecks.checkNotNull(stmt);

    this.assignments.add(stmt);
  }

  /**
   * Appends assignments list to the last assignment of basic block.
   *
   * @param statements list of concurrent assignments
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  @Override
  public void addAssignments(final List<Assignment> statements) {
    InvariantChecks.checkNotNull(statements);

    this.assignments.addAll(statements);
  }

  /**
   * Returns all block assignments.
   *
   * @return assignments list
   */
  @Override
  public List<Assignment> getAssignments() {
    return this.assignments;
  }

  /**
   * Removes all block assignments.
   */
  @Override
  public void removeAssignments() {
    this.assignments.clear();
  }

  /**
   * Checks rather object contains only concurrent assignments.
   *
   * @return {@code true} if contains only concurrent assignments,
   *         {@code false} otherwise.
   */
  public boolean isConcurrent() {
    return this.concurrent;
  }

  /**
   * Sets the concurrent flag to the specified value.
   *
   * @param concurrent concurrent flag
   */
  public void setConcurrent(final boolean concurrent) {
    this.concurrent = concurrent;
  }

  @Override
  public CfgModelNode deepCopy() {
    return new BasicBlock(this.assignments, this.concurrent);
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {
    StringBuilder buffer = new StringBuilder();

    final Iterator<Assignment> iterator = this.assignments.iterator();
    while (iterator.hasNext()) {
      final Assignment stmt = iterator.next();

      final String kind = getKind(printer);

      if (kind == null) {
        throw new IllegalArgumentException("Unknown language.");
      } else {
        buffer.append(stmt.getDescription(printer, kind));
      }

      if (iterator.hasNext()) {
        buffer.append("\n");
      }
    }
    return buffer.toString();
  }

  private String getKind(final ExprTreePrinter printer) {
    String kind = null;
    if (printer instanceof ExprPrinter) {
      if (((ExprPrinter) printer).name().equals(ExprPrinter.VERILOG.name())) {

        if (isConcurrent()) {
          kind = VERILOG_CON;
        } else {
          kind = VERILOG_NON_CON;
        }
      } else if (((ExprPrinter) printer).name().equals(ExprPrinter.VHDL.name())) {
        if (isConcurrent()) {
          kind = VHDL_CON;
        } else {
          kind = VHDL_NON_CON;
        }
      }
    }
    return kind;
  }

  @Override
  public CfgNodeType getType() {
    return CfgNodeType.BASIC_BLOCK;
  }

  @Override
  public boolean hasOnlyParent() {
    return this.parentNode != null;
  }

  @Override
  public boolean hasOnlyChild() {
    return this.childNode != null;
  }

  @Override
  public CfgModelNode getOnlyParent() {
    return this.parentNode;
  }

  @Override
  public CfgModelNode getOnlyChild() {
    return this.childNode;
  }

  @Override
  public Set<NodeVariable> getUses() {
    final Set<NodeVariable> uses = new LinkedHashSet<>();

    for (final Assignment assignment : this.assignments) {
      final Collection<NodeVariable> assignmentUses = assignment.getUses();

      for (final NodeVariable assignmentUse : assignmentUses) {
        if (!uses.contains(assignmentUse)) {
          uses.add(assignmentUse);
        }
      }
    }
    return uses;
  }

  @Override
  public Set<NodeVariable> getDefines() {
    final Set<NodeVariable> defines = new LinkedHashSet<>();

    for (final Assignment assignment : this.assignments) {
      final Collection<NodeVariable> assignmentDefines = assignment.getDefines();

      for (final NodeVariable assignmentDefine : assignmentDefines) {
        if (!defines.contains(assignmentDefine)) {
          defines.add(assignmentDefine);
        }
      }
    }
    return defines;
  }
}
