/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Case node of control flow graph.
 *
 * <p>This object can be either of default kind (and does not contain values)
 * or of non-default kind (and does contain values). If it is non-default,
 * this object stores values of the parent {@link Switch} node's expression. So
 * these values MUST be of the same data type. If this object stores more than
 * one value, it means, that parent {@link Switch} node can be equal to one of them.</p>
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class Case extends CfgNode {

  /* The parent node. */
  private Switch parent;

  /* The values that parent node's expression can be equal to. */
  private final Collection<NodeValue> values;

  /* The default node flag: such object does not have any values. */
  private final boolean defaultNode;

  /* The default node string description. */
  private static final String DEFAULT = "default";

  /**
   * Default constructor.
   *
   * <p>Construct object with the specified default flag.</p>
   *
   * @param defaultNode default node type
   *
   */
  public Case(final boolean defaultNode) {
    super();
    this.values = new LinkedHashSet<>();
    this.defaultNode = defaultNode;
  }

  /**
   * Constructs node with the specified possible value.
   *
   * @param value value
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when argument is {@code null};
   *         (2) when argument's data type differs from
   *         {@link Switch} parent node's condition data type (if parent node exists).
   */
  public Case(final NodeValue value) {
    super();
    InvariantChecks.checkNotNull(value);
    checkValueCompatibility(value);

    this.values = new LinkedHashSet<>();
    this.values.add(value);
    this.defaultNode = false;
  }

  /**
   * Constructs node with the specified possible values.
   *
   * @param values values
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when argument is {@code null};
   *         (2) when at least one element of the argument has
   *         {@link DataType} data type that differs from one of the parent {@link Switch} node's
   *         condition;
   *         (3) when elements of the argument do not have
   *         the same {@link DataType} data type.
   */
  public Case(final Collection<NodeValue> values) {
    super();
    InvariantChecks.checkNotNull(values);
    ofSameDataType(values);
    checkParentCompatibility(values);

    this.values = values;
    this.defaultNode = false;
  }

  /**
   * Constructs node with the specified values and default flag.
   *
   * <p>If values collection is non-empty,
   * than the default flag must be set to {@code false}.
   * Otherwise the default flag must be set to {@code true}.</p>
   *
   * @param values values
   * @param defaultNode default flag
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when values argument is {@code null};
   *         (2) when the contract noted above is not satisfied;
   *         (3) when values does not have the same {@link DataType} data type;
   *         (4) when values data type is not equal to parent
   *         {@link Switch} node's condition data type.
   */
  public Case(final Collection<NodeValue> values, final boolean defaultNode) {
    super();
    InvariantChecks.checkNotNull(values);
    checkArguments(values, defaultNode);
    ofSameDataType(values);
    checkParentCompatibility(values);

    this.values = values;
    this.defaultNode = defaultNode;
  }

  /**
   * Checks rather case node is default.
   *
   * @return {@code true} if node is default, {@code false} otherwise
   */
  public boolean isDefault() {
    return this.defaultNode;
  }

  /**
   * Adds the specified value.
   *
   * <p>The argument's data type MUST be equal to either {@link Switch} parent node's
   * condition data type (if parent node exists)
   * or to all existing values data types (if exist).</p>
   *
   * @param value value
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when argument is {@code null};
   *         (2) when argument has data type that differs from either
   *         parent node's condition data type or from existing values data types.
   * @throws IllegalStateException when the object is default.
   */
  public void addValue(final NodeValue value) {
    InvariantChecks.checkNotNull(value);

    if (this.defaultNode) {
      throw new IllegalStateException("Cannot add value to default node: " + getType() + ".");
    }

    if (hasParents()) {
      final DataType conditionType = ((Switch) getOnlyParent()).getCondition().getDataType();

      if (!conditionType.equals(value.getDataType())) {
        throw new IllegalArgumentException(
            "Argument has data type that differs from: " + conditionType + ".");
      }
    } else if (!getValues().isEmpty()) {

      /* Get the first element's data type (it does not matter which element to get). */
      final DataType firstDataType = getValues().iterator().next().getDataType();

      if (!firstDataType.equals(value.getDataType())) {
        throw new IllegalArgumentException(
            "Argument has data type that differs from: " + firstDataType + ".");
      }
    }

    this.values.add(value);
  }

  /**
   * Returns node values.
   *
   * @return values
   */
  public Collection<NodeValue> getValues() {
    return this.values;
  }

  @Override
  public CfgModelNode deepCopy() {
    return new Case(this.values, this.defaultNode);
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {
    if (this.values.isEmpty()) {
      return Case.DEFAULT;
    } else {
      final StringBuilder builder = new StringBuilder();
      final Iterator<NodeValue> iterator = this.values.iterator();

      while (iterator.hasNext()) {
        final NodeValue value = iterator.next();
        builder.append(value.toString());
        if (iterator.hasNext()) {
          builder.append(',');
        }
      }
      return builder.toString();
    }
  }

  @Override
  public CfgNodeType getType() {
    return CfgNodeType.CASE;
  }

  /**
   * Adds parent node.
   *
   * @param node parent node
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when argument is {@code null};
   *         (2) when the specified node
   *         is not of {@link ru.ispras.retrascope.model.cfg.Switch} type.
   * @throws IllegalStateException when this object already
   *         has non-null parent node.
   */
  @Override
  public void addParent(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (this.parent != null) {
      throw new IllegalStateException(
          String.format(
              "Parent node for %s re-assignment with: %s.", this.getType(), node.getType()));
    }

    if (node instanceof Switch) {
      this.parent = (Switch) node;
    } else {
      throw new IllegalArgumentException("Unsupported type for parent: " + node.getType() + ".");
    }
  }

  /**
   * Removes parent node.
   *
   * @param node parent node
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when argument is {@code null};
   *         (2) when there is no such parent node in this object.
   */
  @Override
  public void removeParent(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (this.parent.equals(node)) {
      this.parent = null;
      node.removeChild(this);
    } else {
      throw new IllegalArgumentException("No such parent node: " + node.getId() + ".");
    }
  }

  @Override
  public boolean hasParents() {
    return hasParentNode();
  }

  @Override
  public boolean hasOnlyParent() {
    return this.parent != null;
  }

  @Override
  public boolean hasOnlyChild() {
    return this.childNode != null;
  }

  private boolean hasParentNode() {
    return this.parent != null;
  }

  @Override
  public CfgModelNode getOnlyParent() {
    return this.parent;
  }

  @Override
  public Collection<CfgModelNode> getParents() {
    return Collections.singleton((CfgModelNode) this.parent);
  }

  @Override
  public CfgModelNode getOnlyChild() {
    return this.childNode;
  }

  @Override
  public Set<NodeVariable> getUses() {
    return new LinkedHashSet<>();
  }

  @Override
  public Set<NodeVariable> getDefines() {
    return new LinkedHashSet<>();
  }

  private static void ofSameDataType(final Collection<NodeValue> values) {
    DataType type = null;
    for (final NodeValue value : values) {
      if (type == null) {
        type = value.getDataType();
      } else if (!type.equals(value.getDataType())) {
        throw new IllegalArgumentException(
            String.format("%s != %s.", type, value.getDataType()));
      }
    }
  }

  private void checkParentCompatibility(final Collection<NodeValue> values) {
    for (final NodeValue value : values) {
      checkValueCompatibility(value);
    }
  }

  private void checkValueCompatibility(final NodeValue value) {
    if (hasParentNode()
        && !this.parent.getCondition().getDataType().equals(value.getDataType())) {
      throw new IllegalArgumentException(
          String.format(
              "Incorrect data type %s; %s should be used.",
              value.getDataTypeId().name(),
              this.parent.getCondition().getDataTypeId().name()));
    }
  }

  private static void checkArguments(final Collection<NodeValue> values, final boolean isDefault) {
    if (values.isEmpty() && !isDefault) {
      throw new IllegalArgumentException("Can't create node with empty and false arguments.");
    }

    if (!values.isEmpty() && isDefault) {
      throw new IllegalArgumentException("Can't create node with non-empty and <true> arguments.");
    }
  }
}
