/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;

import java.util.Collection;
import java.util.LinkedHashSet;

/**
 * Control flow graph model.
 *
 * <p>Control flow graph model is a collection
 * of top-level {@link Module} modules.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgModel extends Entity {

  /**
   * The entity type.
   */
  public static final EntityType TYPE = EntityType.get("cfg");

  private final Collection<Module> modules;

  /**
   * Default constructor.
   */
  public CfgModel() {
    super(CfgModel.TYPE);

    this.modules = new LinkedHashSet<>();
  }

  /**
   * Adds the specified entity.
   *
   * <p>Able to add entity of the {@link CfgModel} type.</p>
   *
   * @param entity entity to be added
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when argument is {@code null};
   *         (2) when the specified entity
   *         is not of {@link ru.ispras.retrascope.model.cfg.CfgModel} type.
   */
  @Override
  public void add(final Entity entity) {
    InvariantChecks.checkNotNull(entity);

    if (entity instanceof CfgModel) {
      final Collection<Module> entityModules = ((CfgModel) entity).getModules();
      addModules(entityModules);
    } else {
      throw new IllegalArgumentException("Incorrect type: " + entity.getEntityType().getId() + ".");
    }
  }

  /**
   * Returns internal modules.
   *
   * @return internal modules collection
   */
  public Collection<Module> getModules() {
    return this.modules;
  }

  /**
   * Adds modules collection.
   *
   * @param modules modules collection
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addModules(final Collection<Module> modules) {
    InvariantChecks.checkNotNull(modules);

    this.modules.addAll(modules);
  }

  /**
   * Adds the specified module to the model.
   *
   * @param module module
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addModule(final Module module) {
    InvariantChecks.checkNotNull(module);

    this.modules.add(module);
  }
}
