/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.printer.ExprTreePrinter;

import java.util.Collection;

/**
 * Node of control flow graph model.
 *
 * <p>The node of {@link CfgModel} control flow graph model.</p>
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public interface CfgModelNode {

  /**
   * Returns a copy of this node.
   *
   * <p>The copy does not keep the same links to parent or child nodes
   * as this has.</p>
   *
   * @return deep copy of this node
   */
  CfgModelNode deepCopy();

  /**
   * Returns unique node identifier in string form.
   *
   * @return node identifier
   */
  String getId();

  /**
   * Returns node string description in the format of the specified printer.
   *
   * @param printer printer
   * @return description
   */
  String getDescription(ExprTreePrinter printer);

  /**
   * Returns node type.
   *
   * <p>By default, node has {@code CfgNodeType.UNKNOWN} type.</p>
   *
   * @return node type
   */
  CfgNodeType getType();

  /**
   * Checks rather node has parent nodes.
   *
   * @return {@code true} if node has at least one parent, {@code false} otherwise.
   */
  boolean hasParents();

  /**
   * Checks rather node has exactly one parent node.
   *
   * @return {@code true} if node has exactly one parent at the moment, {@code false} otherwise.
   */
  boolean hasOnlyParent();

  /**
   * Checks rather node has child nodes.
   *
   * @return {@code true} if node has at least one child, {@code false} otherwise.
   */
  boolean hasChildren();

  /**
   * Checks rather node has exactly one child node.
   *
   * @return {@code true} if node has exactly one child at the moment, {@code false} otherwise.
   */
  boolean hasOnlyChild();

  /**
   * Returns the only parent node.
   *
   * @return only parent node
   */
  CfgModelNode getOnlyParent();

  /**
   * Returns parent nodes.
   *
   * @return parent nodes collection
   */
  Collection<CfgModelNode> getParents();

  /**
   * Returns the only child node.
   *
   * @return only child node
   */
  CfgModelNode getOnlyChild();

  /**
   * Return child nodes.
   *
   * @return child nodes collection
   */
  Collection<CfgModelNode> getChildren();

  /**
   * Adds child node.
   *
   * @param node child node
   */
  void addChild(CfgModelNode node);

  /**
   * Adds parent node.
   *
   * @param node parent node
   */
  void addParent(CfgModelNode node);

  /**
   * Removes child node.
   *
   * @param node child node
   */
  void removeChild(CfgModelNode node);

  /**
   * Removes parent node.
   *
   * @param node parent node
   */
  void removeParent(CfgModelNode node);
}
