/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.MetaInfo;
import ru.ispras.retrascope.model.basis.UseDef;

import java.util.Collection;
import java.util.Collections;

/**
 * Control flow graph node.
 *
 * <p>Node of control flow graph of some {@link Process} node.
 * Has only one parent node and only one child node.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class CfgNode extends MetaInfo implements CfgModelNode,UseDef {

  protected CfgNode parentNode;

  protected CfgNode childNode;

  /* Exception messages. */
  private static final String TYPE_REQUIRED = " type is required.";
  private static final String CANT_REASSIGN = "Unable to reassign with: ";
  private static final String NO_SUCH = "No such node: ";

  @Override
  public String getId() {
    return this.toString();
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {
    return this.getType().name().toLowerCase();
  }

  @Override
  public boolean hasParents() {
    return this.parentNode != null;
  }

  @Override
  public boolean hasChildren() {
    return this.childNode != null;
  }

  /**
   * Adds child node.
   *
   * <p>The specified node will be added only when it is {@link CfgNode}
   * and this object does not have such child node.</p>
   *
   * @param node child node
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when argument is {@code null};
   *         (2) when the specified node
   *         is not of {@link ru.ispras.retrascope.model.cfg.CfgNode} type.
   * @throws IllegalStateException when this object already
   *         has non-null child node.
   */
  @Override
  public void addChild(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof CfgNode) {
      if (this.childNode == null) {

        this.childNode = (CfgNode) node;

        if (!node.hasParents() || !node.getParents().contains(this)) {
          node.addParent(this);
        }
      } else {
        throw new IllegalStateException(CANT_REASSIGN + node.getId());
      }
    } else {
      throw new IllegalArgumentException(CfgNode.class + TYPE_REQUIRED);
    }
  }

  /**
   * Adds parent node.
   *
   * <p>The specified node will be added only when it is {@link CfgNode}
   * and this object has not such parent node.</p>
   *
   * @param node parent node
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when argument is {@code null};
   *         (2) when the specified node
   *         is not of {@link ru.ispras.retrascope.model.cfg.CfgNode} type.
   * @throws IllegalStateException when this object
   *         already has non-null parent node.
   */
  @Override
  public void addParent(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof CfgNode) {

      if (this.parentNode == null) {
        this.parentNode = (CfgNode) node;

        if (!node.hasChildren() || !node.getChildren().contains(this)) {
          node.addChild(this);
        }
      } else {
        throw new IllegalStateException(CANT_REASSIGN + node.getId());
      }
    } else {
      throw new IllegalArgumentException(CfgNode.class + TYPE_REQUIRED);
    }
  }

  /**
   * Removes child node.
   *
   * <p>The specified node will be removed only when it is {@link CfgNode}
   * and this object has such child node.</p>
   *
   * @param node child node
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when argument is {@code null};
   *         (2) when the specified node
   *         is not of {@link ru.ispras.retrascope.model.cfg.CfgNode} type.
   * @throws IllegalStateException when this object
   *         does not have such child node.
   */
  @Override
  public void removeChild(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof CfgNode) {
      if (this.childNode.equals(node)) {

        this.childNode = null;

        if (node.hasParents() && node.getParents().contains(this)) {
          node.removeParent(this);
        }
      } else {
        throw new IllegalStateException(NO_SUCH + node.getId());
      }
    } else {
      throw new IllegalArgumentException(CfgNode.class + TYPE_REQUIRED);
    }
  }

  /**
   * Removes parent node.
   *
   * <p>The specified node will be removed only when it is {@link CfgNode}
   * and this object has such parent node.</p>
   *
   * @param node parent node
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when argument is {@code null};
   *         (2) when the specified node
   *         is not of {@link ru.ispras.retrascope.model.cfg.CfgNode} type.
   * @throws IllegalStateException when this object
   *         does not have such parent node.
   */
  @Override
  public void removeParent(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof CfgNode) {
      if (this.parentNode.equals(node)) {

        this.parentNode = null;

        if (node.hasChildren() && node.getChildren().contains(this)) {
          node.removeChild(this);
        }
      } else {
        throw new IllegalStateException(NO_SUCH + node.getId());
      }
    } else {
      throw new IllegalArgumentException(CfgNode.class + TYPE_REQUIRED);
    }
  }

  @Override
  public Collection<CfgModelNode> getParents() {
    return Collections.singleton((CfgModelNode) this.parentNode);
  }

  @Override
  public Collection<CfgModelNode> getChildren() {
    return Collections.singleton((CfgModelNode) this.childNode);
  }

  @Override
  public boolean equals(final Object obj) {
    return this == obj;
  }

  @Override
  public int hashCode() {
    return System.identityHashCode(this);
  }
}
