/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.transformer.ReduceOptions;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableContainer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * Utility methods for control flow graph model and it's components.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgUtils {

  private CfgUtils() {
    /* Empty. */
  }

  /**
   * Returns source node if it exists in the specified collection.
   *
   * @param nodes nodes
   * @return {@link Source} node if the specified collection contains it,
   *         {@code null} otherwise.
   */
  public static Source getSource(final Collection<CfgNode> nodes) {
    return (Source) CfgUtils.getNode(nodes, CfgNodeType.SOURCE);
  }

  /**
   * Returns sink node if it exists in the specified collection.
   *
   * @param nodes nodes
   * @return {@link Sink} node if the specified collection contains it,
   *         {@code null} otherwise.
   */
  public static Sink getSink(final Collection<CfgNode> nodes) {
    return (Sink) CfgUtils.getNode(nodes, CfgNodeType.SINK);
  }

  /**
   * Returns node of the specified type that is in the specified collection.
   *
   * @param collection nodes collection
   * @param type node type
   * @param <T> type of collection elements
   * @return node of the specified type if it exists in the specified collection
   *         and of the type of collection elements extends {@link CfgModelNode},
   *         {@code null} otherwise
   */
  public static <T> T getNode(
      final Collection<T> collection,
      final CfgNodeType type) {

    T node = null;

    for (final T elem : collection) {
      if (elem instanceof CfgModelNode && ((CfgModelNode) elem).getType().equals(type)) {
        node = elem;
        break;
      }
    }
    return node;
  }

  /**
   * Returns first node in the specified collection with the specified identifier.
   *
   * <p>Returns node with the specified identifier
   * and with minimal index in the specified collection.</p>
   *
   * @param nodes collection of nodes
   * @param nodeId node identifier
   *
   * @return node with the specified identifier if it is found,
   *         {@code null} otherwise.
   */
  public static CfgNode getFirstNode(
      final Collection<CfgNode> nodes,
      final String nodeId) {

    CfgNode first = null;

    for (final CfgNode node : nodes) {
      if (node.getId().equals(nodeId)) {
        first = node;
        break;
      }
    }
    return first;
  }

  /**
   * Returns last node in the specified list with the specified type.
   *
   * <p>Returns node with the specified identifier
   * and with maximal index in the specified list.</p>
   *
   * @param list nodes
   * @param type type of target node
   *
   * @return control flow graph node or {@code null} if there is no such node.
   */
  public static CfgNode getLastNode(
      final List<CfgNode> list,
      final CfgNodeType type) {

    int lastNodeIndex = -1;
    CfgNode last = null;

    for (final CfgNode node : list) {
      final int nodeIndex = list.indexOf(node);
      if (node.getType() == type && nodeIndex > lastNodeIndex) {
        lastNodeIndex = nodeIndex;
      }
    }

    if (lastNodeIndex >= 0) {
      last = list.get(lastNodeIndex);
    }
    return last;
  }

  /**
   * Substitutes internal source node by the specified node and adds it to the specified collection.
   *
   * <p>Previous {@link Source} node (it MUST exist) will be removed
   * from the specified collection. Method does not add the specified
   * node if the specified collection already contains it.</p>
   *
   * @param nodes collection that is to be modified
   * @param node node to be added as the new source
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws IllegalStateException when this object does not contain {@link Source} node.
   */
  public static void addAsSource(
      final Collection<CfgNode> nodes,
      final CfgNode node) {
    InvariantChecks.checkNotNull(node);

    final Source src = (Source) CfgUtils.getNode(nodes, CfgNodeType.SOURCE);
    if (src == null) {
      throw new IllegalStateException("Unable to find node of type: " + CfgNodeType.SOURCE + ".");
    }

    final CfgModelNode srcChild = src.getOnlyChild();
    src.removeChild(srcChild);
    nodes.remove(src);
    node.addChild(srcChild);

    if (!nodes.contains(node)) {
      nodes.add(node);
    }
  }

  /**
   * Substitutes internal sink node by the specified node and adds it to the specified collection.
   *
   * <p>Previous {@link Sink} node (it MUST exist) will be removed from the collection.
   * Method does not add the specified node if the specified collection already contains it.</p>
   *
   * @param nodes collection that is to be modified
   * @param node node to be added as the new sink
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws IllegalStateException when the specified collection
   *         does not contains {@link Sink} node.
   */
  public static void addAsSink(
      final Collection<CfgNode> nodes,
      final CfgNode node) {
    InvariantChecks.checkNotNull(node);

    final Sink sink = (Sink) CfgUtils.getNode(nodes, CfgNodeType.SINK);
    if (sink == null) {
      throw new IllegalStateException("Unable to find node of type: " + CfgNodeType.SINK + ".");
    }

    final Collection<CfgModelNode> sinkParents = sink.getParents();
    final Merge merge = new Merge();
    nodes.add(merge);
    for (final CfgModelNode parent : sinkParents) {
      parent.removeChild(sink);
      merge.addParent(parent);
    }
    nodes.remove(sink);
    merge.addChild(node);

    if (!nodes.contains(node)) {
      nodes.add(node);
    }
  }

  /**
   * Constructs a sequence of connected nodes
   * between the specified begin and end nodes.
   * <p>
   * Constructs a nodes sequence, to which control flow goes through
   * the specified start node and from which it goes through the specified end node.
   * The path does not contain copies of the specified begin and end nodes.
   * </p>
   * @param begin the before-first node of the path
   * @param end the after-last node of the path
   *
   * @return sub-graph of connected node copies
   */
  public static List<CfgNode> getPath(
      final CfgNode begin,
      final CfgNode end) {

    /* Utility class for keeping nodes that lie between begin/end nodes. */
    class CfgNodeCollector extends CfgDefaultVisitor {

      private Status status;

      private List<CfgNode> nodes;

      public CfgNodeCollector() {
        this.nodes = new ArrayList<>();
        this.status = Status.OK;
      }

      public void setStatus(Status status) {
        this.status = status;
      }

      @Override
      public Status getStatus() {
        return this.status;
      }

      @Override
      public void onBasicBlockBegin(BasicBlock basicBlock) {
        tryAdd(basicBlock);
      }

      @Override
      public void onBasicBlockEnd(BasicBlock basicBlock) {
        resetStatus();
      }

      @Override
      public void onCaseBegin(Case caseNode) {
        tryAdd(caseNode);
      }

      @Override
      public void onCaseEnd(Case caseNode) {
        resetStatus();
      }

      @Override
      public void onSwitchBegin(Switch switchNode) {
        tryAdd(switchNode);
      }

      @Override
      public void onSwitchEnd(Switch switchNode) {
        resetStatus();
      }

      @Override
      public void onMergeBegin(Merge merge) {
        tryAdd(merge);
      }

      @Override
      public void onMergeEnd(Merge merge) {
        resetStatus();
      }

      private void resetStatus() {
        if (getStatus().equals(Status.SKIP)) {
          setStatus(Status.OK);
        }
      }

      private void tryAdd(CfgNode node) {
        if (node.equals(end)) {
          setStatus(Status.ABORT);
        }

        if (!this.nodes.contains(node) && !node.equals(begin) && !node.equals(end)) {
          this.nodes.add(node);
        }
      }
    }

    /* Collect nodes. */
    final CfgNodeCollector collector = new CfgNodeCollector();
    final CfgWalker walker = new CfgWalker(collector, CfgWalkerMode.DFS);

    walker.visitNode(begin);

    /* Copy nodes to the returned path. */
    final Map<CfgNode, CfgNode> oldNewMap = new LinkedHashMap<>();

    final List<CfgNode> newNodes = new ArrayList<>();

    for (CfgNode node : collector.nodes) {
      CfgNode copy = (CfgNode) node.deepCopy();
      newNodes.add(copy);
      oldNewMap.put(node, copy);
    }

    /* Copy links between them. */

    for (int i = 0; i < collector.nodes.size(); i++) {

      CfgNode one = collector.nodes.get(i);

      for (int j = 0; j < collector.nodes.size(); j++) {
        CfgNode another = collector.nodes.get(j);

        if (one.getChildren().contains(another)) {
          oldNewMap.get(one).addChild(oldNewMap.get(another));
        }
      }
    }

    return newNodes;
  }

  /**
   * Substitutes usages of variable with the specified name
   * by the specified expression in all the elements of the specified collection.
   * <p>
   * Substitutes usages of variable with the specified {@link java.lang.String} name
   * by the specified {@link ru.ispras.fortress.expression.Node} expression
   * in all the elements of the specified collection. The resulting collection
   * consists of node copies of the specified collection. In which node the every
   * usage of the variable with the specified name will be substituted by the
   * specified expression.
   * </p>
   *
   * @param list collection which nodes will be copied and transformed
   * @param name variable name
   * @param node expression
   *
   * @return collection of modified nodes
   */
  public static List<CfgNode> substitute(
      final List<CfgNode> list,
      final String name,
      final Node node) {

    final int listSize = list.size();
    final List<CfgNode> result = new ArrayList<>(listSize);
    final Map<CfgNode, CfgNode> map = new LinkedHashMap<>(listSize);

    /* Substitute variable with the specified name by the specified node. */
    for (final CfgNode elem : list) {
      CfgNode copy;

      if (elem instanceof Switch) {
        copy = new Switch(Transformer.substitute(((Switch) elem).getCondition(), name, node));

      } else if (elem instanceof BasicBlock) {

        final List<Assignment> elemAssignments = ((BasicBlock) elem).getAssignments();
        final List<Assignment> assignments = new ArrayList<>(elemAssignments.size());

        for (Assignment elemAssign : elemAssignments) {
          final Node value = elemAssign.getValue();
          final RangedVariable target = elemAssign.getTarget();
          Assignment copyAssignment;
          final Node copyValue = Transformer.substitute(value, name, node);
          if (target.isRanged()) {
            final Node old = target.getRange().getOld();
            final Node young = target.getRange().getYoung();

            final Range copyRange =
                new Range(
                    Transformer.reduce(
                        ReduceOptions.NEW_INSTANCE,
                        Transformer.substitute(young, name, node)),
                    Transformer.reduce(
                        ReduceOptions.NEW_INSTANCE,
                        Transformer.substitute(old, name, node)));
            copyAssignment =
                new Assignment(new RangedVariable(target.getVariable(), copyRange), copyValue);
          } else {
            copyAssignment = new Assignment(target.getVariable(), copyValue);
          }
          assignments.add(copyAssignment);
        }

        copy = new BasicBlock(assignments, ((BasicBlock) elem).isConcurrent());
      } else {
        copy = (CfgNode) elem.deepCopy();
      }
      map.put(elem, copy);
      result.add(copy);
    }

    /* Establish parent-child links between nodes. */
    for (int i = 0; i < listSize; i++) {
      final CfgNode one = list.get(i);
      for (CfgNode another : list) {
        if (one.hasChildren() && one.getChildren().contains(another)) {
          map.get(one).addChild(map.get(another));
        }
      }
    }

    return result;
  }

  /**
   * Checks whether the specified nodes are of the specified type.
   *
   * @param type node type
   * @param nodes nodes to be checked
   *
   * @return {@code true} if all the specified nodes are of the specified type,
   *         {@code false} otherwise.
   *
   * @throws IllegalArgumentException when either the specified type is {@code null}
   *         or the specified array of nodes is empty.
   */
  public static boolean isType(
      final CfgNodeType type,
      final CfgModelNode ... nodes) {
    InvariantChecks.checkNotNull(type);
    InvariantChecks.checkNotEmpty(nodes);

    boolean result = true;

    for (CfgModelNode node : nodes) {

      if (!node.getType().equals(type)) {
        result = false;
        break;
      }
    }

    return result;
  }

  /**
   * Returns node that contains a declaration of variable with the specified name.
   *
   * @param node root node which and which child nodes are checked
   * @param name variable name
   *
   * @return node that contains the declaration of variable with the specified name
   *         or {@code null} when the specified node and all the it's children
   *         do not contain such declaration
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  public static VariableContainer getVariableContainer(
      final CfgModelNode node,
      final String name) {
    InvariantChecks.checkNotNull(node);
    InvariantChecks.checkNotNull(name);

    if (node instanceof VariableContainer
        && ((VariableContainer) node).containsVariable(name)) {
      return (VariableContainer) node;
    } else if (node.hasChildren()) {
      final Collection<CfgModelNode> rootChildren = node.getChildren();

      for (final CfgModelNode child : rootChildren) {
        VariableContainer container = getVariableContainer(child, name);

        if (container != null) {
          return container;
        }
      }
      return null;
    }
    return null;
  }
}
