/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

/**
 * Control flow graph model visitor.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public interface CfgVisitor {

  /*
   * The enumeration describes possible statuses of the visitor. Status serves as a directive
   * for the walker to alter its behavior depending on events that may occur in the visitor.
   */
  enum Status {
    /** Continue traversing. */
    OK,

    /** Skip child nodes. */
    SKIP,

    /** Stop traversing. */
    ABORT
  }

  /**
   * Returns the current status of the visitor. The status guides further actions of the walker.
   *
   * @return Current visitor status.
   */
  Status getStatus();

  /**
   * Notifies that processing has been started.
   */
  void onRootBegin();

  /**
   * Notifies that processing has been finished.
   */
  void onRootEnd();

  /**
   * Starts visiting {@link ru.ispras.retrascope.model.cfg.CfgModel} object.
   */
  void onCfgModelBegin();

  /**
   * Finishes visiting {@link ru.ispras.retrascope.model.cfg.CfgModel} object.
   */
  void onCfgModelEnd();

  /**
   * Starts visiting {@link ru.ispras.retrascope.model.cfg.Module} node.
   *
   * @param module {@link ru.ispras.retrascope.model.cfg.Module} node
   */
  void onModuleBegin(Module module);

  /**
   * Finishes visiting {@link ru.ispras.retrascope.model.cfg.Module} node.
   *
   * @param module {@link ru.ispras.retrascope.model.cfg.Module} node
   */
  void onModuleEnd(Module module);

  /**
   * Starts visiting {@link ru.ispras.retrascope.model.cfg.Process} node.
   *
   * @param process {@link ru.ispras.retrascope.model.cfg.Process} node
   */
  void onProcessBegin(Process process);

  /**
   * Finishes visiting {@link ru.ispras.retrascope.model.cfg.Process} node.
   *
   * @param process {@link ru.ispras.retrascope.model.cfg.Process} node
   */
  void onProcessEnd(Process process);

  /**
   * Starts visiting {@link ru.ispras.retrascope.model.cfg.BasicBlock} node.
   *
   * @param basicBlock {@link ru.ispras.retrascope.model.cfg.BasicBlock} node
   */
  void onBasicBlockBegin(BasicBlock basicBlock);

  /**
   * Finishes visiting {@link ru.ispras.retrascope.model.cfg.BasicBlock} node.
   *
   * @param basicBlock {@link ru.ispras.retrascope.model.cfg.BasicBlock} node
   */
  void onBasicBlockEnd(BasicBlock basicBlock);

  /**
   * Starts visiting {@link ru.ispras.retrascope.model.cfg.Case} node.
   *
   * @param caseNode {@link ru.ispras.retrascope.model.cfg.Case} node
   */
  void onCaseBegin(Case caseNode);

  /**
   * Finishes visiting {@link ru.ispras.retrascope.model.cfg.Case} node.
   *
   * @param caseNode {@link ru.ispras.retrascope.model.cfg.Case} node
   */
  void onCaseEnd(Case caseNode);

  /**
   * Starts visiting {@link ru.ispras.retrascope.model.cfg.Switch} node.
   *
   * @param switchNode switch node
   */
  void onSwitchBegin(Switch switchNode);

  /**
   * Finishes visiting {@link ru.ispras.retrascope.model.cfg.Switch} node.
   *
   * @param switchNode switch node
   */
  void onSwitchEnd(Switch switchNode);

  /**
   * Starts visiting {@link ru.ispras.retrascope.model.cfg.Merge} node.
   *
   * @param merge merge node
   */
  void onMergeBegin(Merge merge);

  /**
   * Finishes visiting {@link ru.ispras.retrascope.model.cfg.Merge} node.
   *
   * @param merge merge node
   */
  void onMergeEnd(Merge merge);

  /**
   * Starts visiting {@link ru.ispras.retrascope.model.cfg.Source} node.
   *
   * @param source source node
   */
  void onSrcBegin(Source source);

  /**
   * Finishes visiting {@link ru.ispras.retrascope.model.cfg.Source} node.
   *
   * @param source source node
   */
  void onSrcEnd(Source source);

  /**
   * Notifies that a {@link ru.ispras.retrascope.model.cfg.Sink} node has been visited.
   *
   * @param sink sink node
   */
  void onSink(Sink sink);

  /**
   * Starts visiting {@link ru.ispras.retrascope.model.cfg.Wait} node.
   *
   * @param wait wait node
   */
  void onWaitBegin(Wait wait);

  /**
   * Finishes visiting {@link ru.ispras.retrascope.model.cfg.Wait} node.
   *
   * @param wait wait node
   */
  void onWaitEnd(Wait wait);
}
