/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Control flow graph model walker.
 *
 * <p>Provides methods for traversing nodes of {@link CfgModel} control flow graph model.</p>
 *
 * <p>
 * Actions taken in the traversal process depend on the current status of the visitor (see
 * {@link ru.ispras.retrascope.model.cfg.CfgVisitor.Status}). Possible status values are:</p>
 * <ol>
 * <li>OK - continue traversal
 * <li>SKIP - skip child nodes
 * <li>ABORT - stop traversal
 * </ol>
 *
 * <p>The status is checked after calling any visitor method. Once {@code Status.ABORT} is set,
 * all traversal methods cease the execution. If after a call to a method having the Begin suffix
 * (e.g. onModuleBegin), the {@code Status.SKIP} status is set, nested elements of the visited node
 * (child nodes or subtrees) are not traversed and a corresponding terminating method
 * (that has the End suffix eg. onModuleEnd) is called.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgWalker {

  private final CfgVisitor visitor;

  private final CfgWalkerMode mode;

  /* Collection of already-visited node identifiers. */
  private final Set<String> visitedIds;

  /**
   * Constructs object with the specified visitor.
   *
   * <p>By default, object's mode is "depth-first search without repetitions".</p>
   *
   * @param visitor visitor to be applied to control flow graph model
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public CfgWalker(final CfgVisitor visitor) {
    InvariantChecks.checkNotNull(visitor);

    this.visitor = visitor;

    this.mode = CfgWalkerMode.getDefaultMode();
    this.visitedIds = null;
  }

  /**
   * Constructs object with the specified visitor and mode.
   *
   * @param visitor visitor to be applied to control flow graph model
   * @param mode visiting and walking mode
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public CfgWalker(
      final CfgVisitor visitor,
      final CfgWalkerMode mode) {
    InvariantChecks.checkNotNull(visitor);
    InvariantChecks.checkNotNull(mode);

    this.visitor = visitor;
    this.mode = mode;

    if (mode.equals(CfgWalkerMode.DFS_NO_REPEAT)) {
      this.visitedIds = new LinkedHashSet<>();
    } else {
      this.visitedIds = null;
    }
  }

  /**
   * Returns visitor.
   *
   * @see CfgVisitor
   *
   * @return visitor
   */
  public CfgVisitor getVisitor() {
    return this.visitor;
  }

  /**
   * Starts control flow graph model walking.
   *
   * @param model control flow graph model
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void start(final CfgModel model) {
    InvariantChecks.checkNotNull(model);

    visitor.onRootBegin();

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visitCfgModel(model);
      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    visitor.onRootEnd();
  }

  /**
   * Visits a sequence of control flow graph model nodes.
   *
   * @param nodes a sequence of nodes to be visited
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void visit(final Iterable<CfgModelNode> nodes) {
    InvariantChecks.checkNotNull(nodes);

    for (final CfgModelNode tree : nodes) {
      if (this.mode == CfgWalkerMode.DFS_NO_REPEAT) {
        if (this.visitedIds.isEmpty() || !this.visitedIds.contains(tree.getId())) {
          this.visitedIds.add(tree.getId());
          visitNode(tree);

          if (isStatus(CfgVisitor.Status.ABORT)) {
            return;
          }
        }
      } else {
        visitNode(tree);

        if (isStatus(CfgVisitor.Status.ABORT)) {
          return;
        }
      }
    }
  }

  /**
   * Visits the specified node.
   *
   * @param node node to be visited
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void visitNode(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    switch (node.getType()) {
      case MODULE:
        visitModule((Module) node);
        break;
      case PROCESS:
        visitProcess((Process) node);
        break;
      case BASIC_BLOCK:
        visitBasicBlock((BasicBlock) node);
        break;
      case CASE:
        visitCase((Case) node);
        break;
      case MERGE:
        visitMerge((Merge) node);
        break;
      case SWITCH:
        visitSwitch((Switch) node);
        break;
      case SINK:
        visitSink((Sink) node);
        break;
      case SOURCE:
        visitSource((Source) node);
        break;
      case WAIT:
        visitWait((Wait) node);
        break;
      default:
        throw new IllegalArgumentException("Unknown node type: " + node.getType() + ".");
    }
  }

  private boolean isStatus(final CfgVisitor.Status status) {
    return this.visitor.getStatus() == status;
  }

  private void visitCfgModel(final CfgModel model) {

    this.visitor.onCfgModelBegin();

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {

      final Collection<Module> modules = model.getModules();

      for (final Module module : modules) {

        visitModule(module);

        if (isStatus(CfgVisitor.Status.ABORT)) {
          return;
        }
      }
    }
    this.visitor.onCfgModelEnd();
  }

  private void visitModule(final Module node) {

    this.visitor.onModuleBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {

      final Collection<Process> processes = node.getProcesses();

      for (final Process process : processes) {

        visitProcess(process);

        if (isStatus(CfgVisitor.Status.ABORT)) {
          return;
        }
      }

      final Collection<Module> modules = node.getModules();

      for (final Module module : modules) {

        visitModule(module);

        if (isStatus(CfgVisitor.Status.ABORT)) {
          return;
        }
      }
    }

    this.visitor.onModuleEnd(node);
  }

  private void visitProcess(final Process node) {

    this.visitor.onProcessBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visit(node.getChildren());

      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    this.visitor.onProcessEnd(node);
  }

  private void visitSource(final Source node) {

    this.visitor.onSrcBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visit(node.getChildren());

      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    this.visitor.onSrcEnd(node);
  }

  private void visitSink(final Sink node) {
    this.visitor.onSink(node);
  }

  private void visitSwitch(final Switch node) {

    this.visitor.onSwitchBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visit(node.getChildren());

      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    this.visitor.onSwitchEnd(node);
  }

  private void visitMerge(final Merge node) {

    this.visitor.onMergeBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visit(node.getChildren());

      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    this.visitor.onMergeEnd(node);
  }

  private void visitCase(final Case node) {

    this.visitor.onCaseBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visit(node.getChildren());

      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    this.visitor.onCaseEnd(node);
  }

  private void visitBasicBlock(final BasicBlock node) {

    this.visitor.onBasicBlockBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visit(node.getChildren());

      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    this.visitor.onBasicBlockEnd(node);
  }

  private void visitWait(final Wait node) {

    this.visitor.onWaitBegin(node);

    if (isStatus(CfgVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(CfgVisitor.Status.OK)) {
      visit(node.getChildren());

      if (isStatus(CfgVisitor.Status.ABORT)) {
        return;
      }
    }
    this.visitor.onWaitEnd(node);
  }
}
