/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Merge node of control flow graph.
 *
 * <p>It is a reversed version of switch node.
 * This node has only one child node and more than one parent node.
 * Every time no more than one parent node is active.</p>
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class Merge extends CfgNode {

  private final Collection<CfgNode> parents;

  /**
   * Default constructor.
   */
  public Merge() {
    super();
    this.parents = new LinkedHashSet<>();
  }

  @Override
  public CfgModelNode deepCopy() {
    return new Merge();
  }

  @Override
  public CfgNodeType getType() {
    return CfgNodeType.MERGE;
  }

  @Override
  public boolean hasParents() {
    return !this.parents.isEmpty();
  }

  @Override
  public boolean hasOnlyParent() {
    return this.parents.size() == 1;
  }

  @Override
  public boolean hasOnlyChild() {
    return this.childNode != null;
  }

  @Override
  public CfgModelNode getOnlyParent() {
    return getParents().iterator().next();
  }

  /**
   * Adds parent node.
   *
   * @param node parent node
   *
   * @throws IllegalArgumentException when either argument is {@code null}
   *         or the specified node is not of {@link ru.ispras.retrascope.model.cfg.CfgNode} type.
   */
  @Override
  public void addParent(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof CfgNode) {
      this.parents.add((CfgNode) node);
      if (!node.hasChildren() || !node.getChildren().contains(this)) {
        node.addChild(this);
      }
    } else {
      throw new IllegalArgumentException(
          String.format("Incorrect node type: %s.", node.getType()));
    }
  }

  /**
   * Removes parent node.
   *
   * @param node parent node
   *
   * @throws IllegalArgumentException in the following situations: (1) argument is {@code null};
   *         (2) the specified node is not of {@link ru.ispras.retrascope.model.cfg.CfgNode} type.
   * @throws IllegalStateException when the object does not contain the specified parent node.
   */
  @Override
  public void removeParent(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof CfgNode) {
      if (this.hasParents() && this.getParents().contains(node)) {
        this.parents.remove(node);
        if (node.hasChildren() && node.getChildren().contains(this)) {
          node.removeChild(this);
        }
      } else {
        throw new IllegalStateException("No such parent: " + node.getId() + ".");
      }
    } else {
      throw new IllegalArgumentException("Incorrect node type: " + node.getType() + ".");
    }
  }

  @Override
  public Collection<CfgModelNode> getParents() {
    final Set<CfgModelNode> parentSet = new LinkedHashSet<>();
    parentSet.addAll(this.parents);

    return parentSet;
  }

  @Override
  public CfgModelNode getOnlyChild() {
    return this.childNode;
  }

  @Override
  public Collection<CfgModelNode> getChildren() {
    return Collections.singleton((CfgModelNode) this.childNode);
  }

  @Override
  public Set<NodeVariable> getUses() {
    return new LinkedHashSet<>();
  }

  @Override
  public Set<NodeVariable> getDefines() {
    return new LinkedHashSet<>();
  }
}
