/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.model.basis.VariableContainer;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Module of control flow graph model.
 *
 * <p>Object contains child {@link Process} processes and {@link Module} sub-modules.
 * When object is an instance of another module, it MUST have the instance name
 * with module name.</p>
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class Module extends VariableContainer implements CfgModelNode {

  private CfgModelNode parent;

  private final String name;

  private String instanceName;

  private final List<Process> processes;

  private final List<Module> modules;

  /**
   * Constructs object with the specified name.
   *
   * @param name module name
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public Module(final String name) {
    super();
    InvariantChecks.checkNotNull(name);

    this.name = name;
    this.processes = new ArrayList<>();
    this.modules = new ArrayList<>();
  }

  /**
   * Sets instance name.
   *
   * @param instanceName name of current instance
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void setInstanceName(final String instanceName) {
    InvariantChecks.checkNotNull(instanceName);

    this.instanceName = instanceName;
  }

  /**
   * Returns processes.
   *
   * @return processes
   */
  public Collection<Process> getProcesses() {
    return this.processes;
  }

  /**
   * Returns sub-modules.
   *
   * @return sub-modules
   */
  public List<Module> getModules() {
    return this.modules;
  }

  /**
   * Instantiates this object with the specified name and bindings.
   *
   * @param instanceName instance name
   * @param bindings instance bindings
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public void instantiate(final String instanceName, final Map<String, Node> bindings) {
    InvariantChecks.checkNotNull(instanceName);
    InvariantChecks.checkNotNull(bindings);

    setInstanceName(instanceName);
    setBindings(bindings);
    applyBindings(bindings);
  }

  /**
   * Returns name.
   *
   * @return module name
   */
  public String getName() {
    final StringBuilder builder = new StringBuilder(this.name);

    if (this.instanceName != null) {
      builder.append('.');
      builder.append(this.instanceName);
    }
    return builder.toString();
  }

  /**
   * Returns all the variables, that are of input signal type,
   * or empty collection, if there are no such.
   *
   * @return input signals
   */
  public Collection<NodeVariable> getInputs() {
    return getVariables(VariableType.IN, VariableType.INOUT);
  }

  /**
   * Returns all variables that are of output signal type,
   * or empty collection, if there are no such.
   *
   * @return output signals
   */
  public Collection<NodeVariable> getOutputs() {
    return getVariables(VariableType.INOUT, VariableType.OUT);
  }

  /**
   * Remove internal processes.
   */
  public void removeProcesses() {
    for (int i = 0; i < this.processes.size(); i++) {
      this.processes.get(i).removeParent(this);
    }
    this.processes.clear();
  }

  @Override
  public CfgModelNode deepCopy() {
    final Module copy = new Module(this.name);
    copy.setInstanceName(this.instanceName);
    return copy;
  }

  @Override public String getId() {
    return getDescription(Retrascope.getConfiguration().getStylePrinter());
  }

  /**
   * Adds child node.
   *
   * @param node child node
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when the specified node is neither of
   *         {@link ru.ispras.retrascope.model.cfg.Process}
   *         nor of {@link ru.ispras.retrascope.model.cfg.Module} type;
   *         (2) when argument is {@code null}.
   */
  @Override
  public void addChild(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof Process) {
      if (!this.processes.contains(node)) {
        this.processes.add((Process) node);
      }
    } else if (node instanceof Module) {
      if (!this.modules.contains(node)) {
        this.modules.add((Module) node);
      }
    } else {
      throw new IllegalArgumentException("Illegal child kind: " + node.getType() + ".");
    }
    if (!node.hasParents() || !node.getParents().contains(this)) {
      node.addParent(this);
    }
  }

  /**
   * Adds parent node.
   *
   * @param node parent node
   *
   * @throws IllegalArgumentException when argument is either {@code null}
   *         or not of {@link ru.ispras.retrascope.model.cfg.Module} type.
   */
  @Override
  public void addParent(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node.getType() == CfgNodeType.MODULE) {
      this.parent = node;

      if (!node.hasChildren() || !node.getChildren().contains(this)) {
        node.addChild(this);
      }
    } else {
      throw new IllegalArgumentException("Illegal parent kind: " + node.getType() + ".");
    }
  }

  /**
   * Removes child node.
   *
   * @param node child node
   *
   * @throws IllegalArgumentException in the following situations: (1) when the specified node
   *         is not of {@link ru.ispras.retrascope.model.cfg.Process} type
   *         and not of {@link ru.ispras.retrascope.model.cfg.Module} type;
   *         (2) this object does not contain the specified node as child;
   *         (3) argument is {@code null}.
   */
  @Override
  public void removeChild(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (this.hasChildren() && this.getChildren().contains(node)) {
      if (node instanceof Process) {
        this.processes.remove(node);
      } else if (node instanceof Module) {
        this.modules.remove(node);
      } else {
        throw new IllegalArgumentException("Illegal node kind: " + node.getType() + ".");
      }
    } else {
      throw new IllegalArgumentException("No such child node: " + node.getId() + ".");
    }
  }

  /**
   * Removes parent node.
   *
   * @param node parent node
   *
   * @throws IllegalArgumentException when either this object
   *         does not contain the specified node as parent node
   *         or argument is {@code null}.
   */
  @Override
  public void removeParent(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (this.parent == node) {
      this.parent = null;
      node.removeChild(this);
    } else {
      throw new IllegalArgumentException("No such parent node: " + node.getId() + ".");
    }
  }

  @Override
  public Collection<CfgModelNode> getParents() {
    return Collections.singleton(this.parent);
  }

  @Override
  public CfgModelNode getOnlyChild() {
    return getChildren().iterator().next();
  }

  @Override
  public Collection<CfgModelNode> getChildren() {
    final Set<CfgModelNode> children = new LinkedHashSet<>();
    children.addAll(this.getProcesses());
    children.addAll(this.getModules());
    return children;
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {
    return this.getType().name().toLowerCase() + " " + this.getName();
  }

  @Override
  public CfgNodeType getType() {
    return CfgNodeType.MODULE;
  }

  @Override
  public boolean hasParents() {
    return this.parent != null;
  }

  @Override
  public boolean hasOnlyParent() {
    return this.parent != null;
  }

  @Override
  public boolean hasChildren() {
    return !this.getProcesses().isEmpty() || !this.getModules().isEmpty();
  }

  @Override
  public boolean hasOnlyChild() {
    return this.modules.size() == 1 && this.processes.isEmpty()
        || this.modules.isEmpty() && this.processes.size() == 1;
  }

  @Override
  public CfgModelNode getOnlyParent() {
    return this.parent;
  }

  /**
   * Applies bindings to module's internal processes and sub-modules.
   *
   * @param bindings "variable name - expression" binding
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  private void applyBindings(final Map<String, Node> bindings) {
    InvariantChecks.checkNotNull(bindings);

    for (final Module module : this.modules) {
      module.applyBindings(bindings);
    }

    for (final Process process : this.processes) {
      process.applyBindings(bindings);
    }

    /* Change declarations. */
    for (final String varName : bindings.keySet()) {
      if (containsVariable(varName)) {

        /* Remove old declaration for the variable that is substituted by bindings. */
        removeDeclaration(varName);
      }
    }
  }

  private Collection<NodeVariable> getVariables(final VariableType ... types) {

    final Set<NodeVariable> variables = new LinkedHashSet<>();

    final Set<NodeVariable> moduleVariables = getVariablesMapping().keySet();
    for (final NodeVariable var : moduleVariables) {

      // extract and analyse variable data
      final Object varData = var.getUserData();
      if (varData instanceof VariableData) {
        final VariableType varType = ((VariableData) varData).getVariableType();
        for (final VariableType type : types) {
          if (varType.equals(type)) {
            variables.add(var);
            break;
          }
        }
      }
    }
    return variables;
  }

  @Override
  public boolean equals(final Object obj) {
    return this == obj;
  }

  @Override
  public int hashCode() {
    return System.identityHashCode(this);
  }
}
