/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableContainer;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Process of control flow graph model.
 *
 * <p>The object contains the embedded {@link CfgNode} control flow graph nodes
 * and {@link EventList} sensitivity list. Also it can be either
 * of initial or of non-initial type.</p>
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class Process extends VariableContainer implements CfgModelNode {

  private static final String WRONG_KIND = "Incorrect node kind: ";

  private Module parent;

  private Source child;

  /* The sensitivity list. */
  private final EventList eventList;

  private final List<CfgNode> nodes;

  /**
   * Default constructor.
   *
   * <p>Constructs non-initial process.</p>
   */
  public Process() {
    this(false);
  }

  /**
   * Constructs new process with the specified value of initial flag.
   *
   * @param initial initial flag
   */
  public Process(final boolean initial) {
    super();

    /* Add special meta-information for inital processes. */
    if (initial) {
      addMetaInfo(MetaInfoType.INITIAL_PROCESS, MetaInfoType.INITIAL_PROCESS);
    }
    this.eventList = new EventList();
    this.nodes = new ArrayList<>();
  }

  /**
   * Adds control flow graph node.
   *
   * <p>The argument node will only be added to control flow graph
   * without establishing any "child-parent" links with already added nodes.</p>
   *
   * @param node node to be added
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addNode(final CfgNode node) {
    InvariantChecks.checkNotNull(node);

    this.nodes.add(node);
  }

  /**
   * Adds control flow graph nodes collection.
   *
   * <p>The argument nodes will only be added to control flow graph
   * without establishing any "child-parent" links with already added nodes.</p>
   *
   * @param nodes nodes to be added
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addNodes(final Collection<CfgNode> nodes) {
    InvariantChecks.checkNotNull(nodes);

    this.nodes.addAll(nodes);
  }

  /**
   * Appends another control flow graph nodes collection.
   *
   * <p>The argument's nodes will not only be added to this nodes
   * but argument {@link Source} starting source node will be linked
   * with this {@link Sink} finishing sink node.</p>
   *
   * @param nodes nodes to be appended
   *
   */
  public void append(final List<CfgNode> nodes) {
    InvariantChecks.checkNotNull(nodes);

    final Sink sink = CfgUtils.getSink(this.nodes);

    final CfgModelNode sinkParent = sink.getOnlyParent();

    sinkParent.removeChild(sink);
    this.nodes.remove(sink);

    final Source cfgSource = (Source) CfgUtils.getNode(nodes, CfgNodeType.SOURCE);

    final CfgModelNode srcChild = cfgSource.getOnlyChild();

    srcChild.removeParent(cfgSource);
    nodes.remove(cfgSource);

    sinkParent.addChild(srcChild);

    this.nodes.addAll(nodes);
  }

  /**
   * Removes node.
   *
   * <p>The specified node will only be removed from this object
   * without removing any "child-parent" links with existing nodes.</p>
   *
   * @param node node to be removed
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void removeNode(final CfgNode node) {
    InvariantChecks.checkNotNull(node);

    this.nodes.remove(node);
  }

  /**
   * Returns all nodes.
   *
   * @return nodes collection
   */
  public List<CfgNode> getNodes() {
    return this.nodes;
  }

  /**
   * Checks rather process is of initial kind.
   *
   * @return {@code true} if process is initial, {@code false} otherwise.
   */
  public boolean isInitial() {
    return hasMetaInfo(MetaInfoType.INITIAL_PROCESS);
  }

  /**
   * Returns {@link ru.ispras.retrascope.model.basis.EventList} sensitivity list.
   *
   * @return sensitivity list
   */
  public EventList getSensitivityList() {
    return this.eventList;
  }

  /**
   * Sets new {@link ru.ispras.retrascope.model.basis.EventList} sensitivity list.
   *
   * @param eventList event list
   *
   * @throws IllegalArgumentException when either argument is {@code null}
   *         or object is initial and argument is non-empty.
   */
  public void setSensitivityList(final EventList eventList) {
    InvariantChecks.checkNotNull(eventList);

    if (isInitial() && !eventList.isEmpty()) {
      throw new IllegalStateException(
          "Cannot set to initial process a non-empty event collection.");
    }

    this.eventList.clear();
    this.eventList.addEvents(eventList.getEvents());
  }

  /** Applies bindings to process control flow graph and sensitivity list.
   *
   * @param bindings "variable name - expression" binding
   */
  public void applyBindings(final Map<String, Node> bindings) {
    applyToCfg(bindings);
    this.eventList.applyToEventList(bindings);
  }

  /**
   * Applies bindings to the internal collection of nodes.
   *
   * @param bindings "variable name - expression" binding
   *
   * @throws IllegalArgumentException when either argument is {@code null}
   *         or some argument pair has a key that is not of
   *         {@link ru.ispras.fortress.expression.NodeVariable} type.
   */
  private void applyToCfg(final Map<String, Node> bindings) {
    InvariantChecks.checkNotNull(bindings);

    /* Elaborate only nodes that contain code with variables. */
    for (final CfgNode node : this.nodes) {

      if (node instanceof BasicBlock) {
        applyToBasicBlock(bindings, (BasicBlock) node);
      } else if (node instanceof Switch) {
        applyToSwitch(bindings, (Switch) node);
      } else if (node instanceof Wait) {
        applyToWait(bindings, (Wait) node);
      }
    }
  }

  private void applyToWait(final Map<String, Node> bindings, final Wait node) {

    node.getEventList().applyToEventList(bindings);

    for (final Map.Entry<String, Node> entry : bindings.entrySet()) {
      node.setCondition(
          Transformer.substitute(node.getCondition(), entry.getKey(), entry.getValue()));
    }
  }

  private void applyToSwitch(final Map<String, Node> bindings, final Switch node) {

    final Set<NodeVariable> uses = node.getUses();

    for (final NodeVariable use : uses) {

      final String useName = use.getName();

      /* If used-in-expression variable is used-in-bindings. */
      if (bindings.containsKey(useName)) {
        node.setCondition(
            Transformer.substitute(node.getCondition(), useName, bindings.get(useName)));
      }
    }
  }

  private void applyToBasicBlock(final Map<String, Node> bindings, final BasicBlock node) {

    for (final Assignment stmt : node.getAssignments()) {

      final RangedVariable rangedVariable = stmt.getTarget();
      final String name = rangedVariable.getVariableName();

      /* If defined variable is used-in-bindings. */
      if (bindings.containsKey(name)) {

        final Node binding = bindings.get(name);

        if (bindings.get(name) instanceof NodeVariable) {

          stmt.setVariable(new RangedVariable((NodeVariable) binding));
        } else {
          throw new IllegalArgumentException("Illegal node kind: " + binding.getKind() + ".");
        }
      }

      final Set<String> usedVarNames = stmt.getUsedVariableNames();

      for (final String usedVarName : usedVarNames) {

        /* If used-in-expression variable is used-in-bindings. */
        if (bindings.containsKey(usedVarName)) {
          stmt.setValue(
              Transformer.substitute(stmt.getValue(), usedVarName, bindings.get(usedVarName)));
        }
      }
    }
  }

  @Override
  public CfgModelNode deepCopy() {
    final Process copy = new Process();
    copy.setSensitivityList(this.eventList);

    /* Make a copies of internal nodes. */
    final Map<CfgNode, CfgNode> oldNewMap = new LinkedHashMap<>();
    final List<CfgNode> copies = new ArrayList<>();

    for (final CfgNode node : this.nodes) {
      final CfgModelNode nodeCopy = node.deepCopy();
      oldNewMap.put(node, (CfgNode) nodeCopy);
      copies.add((CfgNode) nodeCopy);
    }

    /* Set links between node copies. */
    for (int i = 0; i < this.nodes.size(); i++) {
      for (final CfgNode node : this.nodes) {
        final CfgNode iNode = this.nodes.get(i);

        if (iNode.getChildren().contains(node)) {
          oldNewMap.get(iNode).addChild(oldNewMap.get(node));
        }
      }
    }
    copy.addNodes(copies);
    copy.addMetaInfo(getMetaInfo());
    return copy;
  }

  @Override public String getId() {
    return getDescription(Retrascope.getConfiguration().getStylePrinter());
  }

  @Override
  public void addChild(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (!(node instanceof Source)) {
      throw new IllegalArgumentException(WRONG_KIND + node.getType() + ".");
    }
    if (this.child == null) {
      this.child = (Source) node;
      if (!node.hasParents() || !node.getParents().contains(this)) {
        node.addParent(this);
      }
    } else {
      throw new IllegalStateException(
          String.format("Child node for %s re-assignment with: %s.",
              this.getType(), node.getType()));
    }
  }

  @Override
  public void addParent(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node.getType() != CfgNodeType.MODULE) {
      throw new IllegalArgumentException(WRONG_KIND + node.getType() + ".");
    }
    if (this.parent == null) {
      this.parent = (Module) node;
      if (!node.hasChildren() || node.getChildren().contains(this)) {
        node.addChild(this);
      }
    } else {
      throw new IllegalStateException(
          String.format("Parent node for %s re-assignment with: %s.", getType(), node.getType()));
    }
  }

  @Override
  public void removeChild(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node.getType() != CfgNodeType.SOURCE) {
      throw new IllegalArgumentException(WRONG_KIND + node.getType() + ".");
    }
    if (this.child.equals(node)) {

      this.child = null;

      if (node.hasParents() && node.getParents().contains(this)) {
        node.removeParent(this);
      }
    } else {
      throw new IllegalStateException("No such child node: " + node.getId() + ".");
    }
  }

  @Override
  public void removeParent(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node.getType() != CfgNodeType.MODULE) {
      throw new IllegalArgumentException(WRONG_KIND + node.getType() + ".");
    }

    if (this.parent.equals(node)) {
      this.parent = null;

      if (node.hasChildren() && node.getChildren().contains(this)) {
        node.removeChild(this);
      }
    } else {
      throw new IllegalStateException("No such parent node: " + node.getId() + ".");
    }
  }

  @Override
  public Collection<CfgModelNode> getParents() {
    return Collections.singleton((CfgModelNode) this.parent);
  }

  @Override
  public CfgModelNode getOnlyChild() {
    return this.child;
  }

  @Override
  public Collection<CfgModelNode> getChildren() {
    return Collections.singleton((CfgModelNode) this.child);
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {
    return this.getType().name().toLowerCase() + (this.eventList.isEmpty()
        ? "" : " " + this.eventList.toString());
  }

  @Override
  public CfgNodeType getType() {
    return CfgNodeType.PROCESS;
  }

  @Override
  public boolean hasParents() {
    return this.parent != null;
  }

  @Override
  public boolean hasOnlyParent() {
    return this.parent != null;
  }

  @Override
  public boolean hasChildren() {
    return this.child != null;
  }

  @Override
  public boolean hasOnlyChild() {
    return this.child != null;
  }

  @Override
  public CfgModelNode getOnlyParent() {
    return this.parent;
  }

  @Override
  public boolean equals(final Object obj) {
    return this == obj;
  }

  @Override
  public int hashCode() {
    return System.identityHashCode(this);
  }
}
