/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Source (first) node of control flow graph.
 *
 * <p>Source node has only one parent node that does not belong to control flow graph
 * and only one child node.</p>
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class Source extends CfgNode {

  private CfgModelNode parent;

  /**
   * Default constructor.
   */
  public Source() {
    super();
  }

  /**
   * Constructs object with the specified parent node.
   *
   * @param parent parent node
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public Source(final CfgModelNode parent) {
    super();
    InvariantChecks.checkNotNull(parent);

    this.parent = parent;
  }

  @Override
  public CfgModelNode deepCopy() {
    return new Source();
  }

  /**
   * Adds parent node.
   *
   * @param node parent node
   *
   * @throws IllegalArgumentException when argument is either {@code null}
   *         or of {@link CfgNode} type.
   * @throws IllegalStateException when parent node is not equal to {@code null} already.
   */
  @Override
  public void addParent(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof CfgNode) {

      throw new IllegalArgumentException("Incorrect node type: " + node.getType() + ".");

    } else {
      if (this.parent == null) {
        this.parent = node;

        if (!node.hasChildren() || !node.getChildren().contains(this)) {
          node.addChild(this);
        }
      } else {
        throw new IllegalStateException(
            String.format("Parent node for %s re-assignment with: %s.",
                this.getType(), node.getType()));
      }
    }
  }

  /**
   * Removes parent node.
   *
   * @param node parent node
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws IllegalStateException when the internal parent node
   *         is not equal to the specified node.
   */
  @Override
  public void removeParent(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (this.parent.equals(node)) {

      this.parent = null;

      if (node.hasChildren() && node.getChildren().contains(this)) {
        node.removeChild(this);
      }
    } else {
      throw new IllegalStateException("No such parent node: " + node.getId() + ".");
    }
  }

  @Override
  public Collection<CfgModelNode> getParents() {
    return Collections.singleton(this.parent);
  }

  @Override
  public CfgModelNode getOnlyChild() {
    return this.childNode;
  }

  @Override
  public CfgNodeType getType() {
    return CfgNodeType.SOURCE;
  }

  @Override
  public boolean hasParents() {
    return this.parent != null;
  }

  @Override
  public boolean hasOnlyParent() {
    return this.parent != null;
  }

  @Override
  public boolean hasOnlyChild() {
    return this.childNode != null;
  }

  @Override
  public CfgModelNode getOnlyParent() {
    return this.parent;
  }

  @Override
  public Set<NodeVariable> getUses() {
    return new LinkedHashSet<>();
  }

  @Override
  public Set<NodeVariable> getDefines() {
    return new LinkedHashSet<>();
  }
}
