/*
 * Copyright 2013 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.RangedVariable;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Switch node of control flow graph.
 *
 * <p>This node represents conditional block. It's {@link Case} children contain possible values
 * of the condition expression. Also this block may be of event type, when it's condition
 * is {@code true}, but it can be executed only when the specified event
 * {@link ru.ispras.retrascope.model.basis.Event} happens.</p>
 * <p>Note thar every {@link Case} child of this node MUST have {@link NodeValue} values
 * of the same data type that this object's {@link Node} condition has.</p>
 *
 * <p>When executed, only one output node will be active.</p>
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 * 
 */
public class Switch extends CfgNode {

  private Node condition;

  private final Event event;

  private final Collection<Case> children;

  /**
   * Constructs an object with the specified condition.
   *
   * @param condition switch condition
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public Switch(final Node condition) {
    super();
    InvariantChecks.checkNotNull(condition);

    this.condition = condition;
    this.event = null;
    this.children = new LinkedHashSet<>();
  }

  /**
   * Constructs an object with the specified event.
   *
   * <p>The condition is set to {@code NodeValue.newBoolean(true)}.</p>
   *
   * @param event event
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public Switch(final Event event) {
    super();
    InvariantChecks.checkNotNull(event);

    this.condition = NodeValue.newBoolean(true);
    this.event = event;
    this.children = new LinkedHashSet<>();
  }

  /**
   * Returns {@link Node} condition.
   *
   * @return condition
   */
  public Node getCondition() {
    return this.condition;
  }

  /**
   * Sets new {@link Node} condition.
   *
   * @param expr expression
   *
   * @throws IllegalArgumentException when argument is either {@code null}
   *         or not of the same {@link DataType} data type
   *         that its children {@link Case} nodes have.
   */
  public void setCondition(final Node expr) {
    InvariantChecks.checkNotNull(expr);

    if (hasChildren()) {
      for (final Case child : this.children) {
        final Collection<NodeValue> childValues = child.getValues();

        for (final NodeValue value : childValues) {
          if (!value.getDataType().equals(expr.getDataType())) {
            throw new IllegalArgumentException(
                String.format(
                    "Incorrect data type %s; %s must be used.",
                    expr.getDataType(),
                    value.getDataType()));
          }
        }
      }
    }
    this.condition = expr;
  }

  /**
   * Returns event.
   *
   * @return event
   */
  public Event getEvent() {
    return this.event;
  }

  @Override
  public CfgModelNode deepCopy() {
    final Switch copy = new Switch(this.event);
    copy.setCondition(this.condition);
    return copy;
  }

  /**
   * Returns a string description in the format of the specified printer.
   *
   * @param printer printer that provides an output string format
   *
   * @return object string description
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  @Override
  public String getDescription(final ExprTreePrinter printer) {
    InvariantChecks.checkNotNull(printer);

    final StringBuilder builder = new StringBuilder("");
    if (this.event != null) {
      builder.append('[');
      builder.append(this.event.toString());
      builder.append("] ");
    }
    builder.append(printer.toString(this.condition));
    return builder.toString();
  }

  @Override
  public CfgNodeType getType() {
    return CfgNodeType.SWITCH;
  }

  @Override
  public boolean hasOnlyParent() {
    return this.parentNode != null;
  }

  @Override
  public boolean hasChildren() {
    return !this.children.isEmpty();
  }

  @Override
  public boolean hasOnlyChild() {
    return this.children.size() == 1;
  }

  @Override
  public CfgModelNode getOnlyParent() {
    return this.parentNode;
  }

  @Override
  public CfgModelNode getOnlyChild() {
    return getChildren().iterator().next();
  }

  /**
   * Adds the specified child node.
   *
   * @param node child node
   *
   * @throws IllegalArgumentException in the following situations: (1) argument is either
   *         {@code null} or not of {@link ru.ispras.retrascope.model.cfg.Case} type;
   *         (2) when at least one {@link NodeValue} value of {@link Case} child
   *         argument has {@link DataType} which is not equal to the data type of condition.
   */
  @Override
  public void addChild(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof Case) {

      checkNodeDataType((Case)node);

      this.children.add((Case) node);
      if (!node.hasParents() || !node.getParents().contains(this)) {
        node.addParent(this);
      }
    } else {
      throw new IllegalArgumentException("Incorrect type of child node: " + node.getId() + ".");
    }
  }

  private void checkNodeDataType(final Case node) {
    final Collection<NodeValue> nodeValues = node.getValues();
    final DataType dataType = getCondition().getDataType();

    for (final NodeValue value : nodeValues) {
      if (!value.getDataType().equals(dataType)) {
        throw new IllegalArgumentException(
            String.format(
                "Incorrect data type %s; %s should be used.", value.getDataType(), dataType));
      }
    }
  }

  /**
   * Removes child node.
   *
   * @param node child node
   *
   * @throws IllegalArgumentException when argument is either {@code null}
   *         or is not of {@link ru.ispras.retrascope.model.cfg.Case} type.
   */
  @Override
  public void removeChild(final CfgModelNode node) {
    InvariantChecks.checkNotNull(node);

    if (node instanceof Case) {
      if (hasChildren() && getChildren().contains(node)) {
        this.children.remove(node);
        if (node.hasParents() && node.getParents().contains(this)) {
          node.removeParent(this);
        }
      }
    } else {
      throw new IllegalArgumentException("Incorrect type of child node: " + node.getId() + ".");
    }
  }

  @Override
  public Collection<CfgModelNode> getChildren() {
    final Collection<CfgModelNode> childrenSet = new LinkedHashSet<>();
    childrenSet.addAll(this.children);
    return childrenSet;
  }

  @Override
  public Set<NodeVariable> getUses() {
    final Set<NodeVariable> uses = new LinkedHashSet<>();

    /* Get variables that are used in expression. */
    final Node expr = getCondition();
    final Collection<NodeVariable> usedExprVariables = ExprUtils.getVariables(expr);
    uses.addAll(usedExprVariables);

    /* Get variables that are used in event. */
    if (getEvent() != null) {
      final RangedVariable rVariable = getEvent().getRangedVariable();
      final NodeVariable eventVariable = rVariable.getVariable();

      uses.add(eventVariable);

      if (rVariable.isRanged()) {
        final Node youngNum = rVariable.getRange().getYoung();
        final Node oldNum = rVariable.getRange().getOld();

        uses.addAll(ExprUtils.getVariables(youngNum));
        uses.addAll(ExprUtils.getVariables(oldNum));
      }
    }
    return uses;
  }

  @Override
  public Set<NodeVariable> getDefines() {
    return new LinkedHashSet<>();
  }
}
