/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.UseDef;
import ru.ispras.retrascope.model.cfg.BasicBlock;

import java.util.ArrayList;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

/**
 * Sequence of {@link Assignment} statements.
 * 
 * <p>Statements MUST be executed in the same order as they are presented in this class.</p>
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class Action implements UseDef {

  /* String representation constants. */
  private static final String LEFT_BOUND = "{";
  private static final String ASSIGN_HDR = "assignment: ";
  private static final String ASSIGN_DEL = "; ";
  private static final String RIGHT_BOUND = "}";

  private final List<Assignment> assignments;

  /**
   * Constructs object with the specified assignment.
   *
   * @param assignment assignment
   * 
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public Action(final Assignment assignment) {
    InvariantChecks.checkNotNull(assignment);

    this.assignments = new ArrayList<>();
    this.assignments.add(assignment);
  }

  /**
   * Constructs object with the specified list of assignments.
   *
   * @param list assignments
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public Action(final List<Assignment> list) {
    InvariantChecks.checkNotNull(list);

    this.assignments = list;
  }

  /**
   * Returns assignments.
   *
   * @return list of assignments
   */
  public List<Assignment> getAssignments() {
    return this.assignments;
  }

  /**
   * Transforms to non-concurrent basic block node.
   *
   * @return control flow graph node
   */
  public BasicBlock toBasicBlock() {
    return new BasicBlock(this.assignments);
  }

  /**
   * Returns unique string identifier of this object.
   *
   * @return identifier
   */
  public String getId() {
    final StringBuilder stringBuilder = new StringBuilder(Action.LEFT_BOUND);

    final int length = stringBuilder.length();

    for (final Assignment assignment: this.assignments) {
      if (stringBuilder.length() > length) {
        stringBuilder.append(Action.ASSIGN_DEL);
      }

      stringBuilder.append(Action.ASSIGN_HDR);
      stringBuilder.append(assignment);
    }

    stringBuilder.append(Action.RIGHT_BOUND);

    return stringBuilder.toString();
  }

  @Override
  public Set<NodeVariable> getUses() {
    Set<NodeVariable> uses = new LinkedHashSet<>();
    for (final Assignment stmt : this.assignments) {
      final Set<NodeVariable> stmtUses = stmt.getUses();
      if (!stmtUses.isEmpty()) {
        uses = CollectionUtils.uniteSets(uses, stmtUses);
      }
    }
    return uses;
  }

  @Override
  public Set<NodeVariable> getDefines() {
    Set<NodeVariable> defines = new LinkedHashSet<>();
    for (final Assignment stmt : this.assignments) {
      final Set<NodeVariable> stmtDefines = stmt.getDefines();
      if (!stmtDefines.isEmpty()) {
        defines = CollectionUtils.uniteSets(defines, stmtDefines);
      }
    }
    return defines;
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + this.assignments.hashCode();

    return result;
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }

    if (obj == null) {
      return false;
    }

    if (!(obj instanceof Action)) {
      return false;
    }

    final Action actionToCompare = (Action) obj;

    return this.assignments.equals(actionToCompare.assignments);
  }

  @Override
  public String toString() {
    return getId();
  }
}
