/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Utility methods for extended finite state machine model and it's components.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class EfsmUtils {

  private EfsmUtils() {
    /* Empty. */
  }

  /**
   * Returns mapping between modified variables and affecting variables.
   *
   * <p>Returns mappings between modified variables and variables which affects corresponding
   * modifications.</p>
   * <br>
   * <p>For example, if this action consists of two subsequent assignments {@code x = y} and
   * {@code z = x + i}, the resulting map contains mappings from {@code x} to {@code y} and from
   * {@code z} to {@code y} and {@code i}.</p>
   *
   * @return a map containing modified variables and variables which affect corresponding
   *         modifications
   */
  public static Map<NodeVariable, Set<NodeVariable>> getResultingDependencies(final Action action) {

    final List<Assignment> assignments = action.getAssignments();
    final Map<NodeVariable, Set<NodeVariable>> substitutions = new HashMap<>(assignments.size());

    for (final Assignment assignment: assignments) {
      final Collection<NodeVariable> usedVariables = ExprUtils.getVariables(assignment.getValue());

      Set<NodeVariable> affecting;

      if (usedVariables.isEmpty()) {
        affecting = Collections.emptySet();
      } else {
        affecting = new HashSet<>(usedVariables.size());

        for (final NodeVariable variable: usedVariables) {
          if (substitutions.containsKey(variable)) {
            affecting.addAll(substitutions.get(variable));
          } else {
            affecting.add(variable);
          }
        }
      }
      substitutions.put(assignment.getTarget().getVariable(), affecting);
    }
    return substitutions;
  }

  /**
   * Returns mapping between variables and corresponding substitutions.
   *
   * <p>Returns mappings between variables and corresponding substitutions
   * representing total variable modifications during execution of this action.
   * That is, execution of this action is equal to parallel evaluation of the
   * substitutions and subsequent assignment of derived values to the
   * corresponding variables.</p>
   *
   * @return a map containing modified variables and the substitutions representing these
   *         modifications
   */
  public static Map<NodeVariable, Node> getResultingSubstitutions(final Action action) {

    final List<Assignment> assignments = action.getAssignments();
    final Map<NodeVariable, Node> substitutions = new LinkedHashMap<>();

    for (final Assignment assignment: assignments) {
      Node expression = assignment.getValue();

      for (final NodeVariable variable: ExprUtils.getVariables(expression)) {
        if (substitutions.containsKey(variable)) {
          expression =
              Transformer.substitute(expression, variable.getName(), substitutions.get(variable));
        }
      }

      final RangedVariable rangedVariable = assignment.getTarget();
      final NodeVariable nodeVariable = rangedVariable.getVariable();

      if (rangedVariable.isRanged()) {
        final DataType dataType = nodeVariable.getDataType();

        if (!dataType.getTypeId().equals(DataTypeId.BIT_VECTOR)) {
          throw new IllegalArgumentException(
              "Unsupported data type of ranged variable: " + dataType);
        }

        final Range range = rangedVariable.getRange();

        if (!(range.getOld() instanceof NodeValue) || !(range.getYoung() instanceof NodeValue)) {
          throw new IllegalArgumentException("Variable range bounds are currently unsupported");
        }

        final NodeValue leftValue = (NodeValue) range.getYoung();
        final NodeValue rightValue = (NodeValue) range.getOld();

        final int leftBound = leftValue.getInteger().intValue();
        final int rightBound = rightValue.getInteger().intValue();

        if (leftBound < rightBound) {
          throw new IllegalArgumentException("This type of sub-vectors is currently unsupported");
        }

        NodeOperation mostSignPart = null;
        NodeOperation lessSignPart = null;
        final int variableSize = dataType.getSize();
        final int decRadix = 10;

        if (leftBound < variableSize - 1) {
          final NodeValue leftIndex =
              new NodeValue(DataType.INTEGER.valueOf(String.valueOf(variableSize - 1), decRadix));
          final NodeValue rightIndex =
              new NodeValue(DataType.INTEGER.valueOf(String.valueOf(leftBound + 1), decRadix));

          mostSignPart =
              new NodeOperation(StandardOperation.BVEXTRACT, leftIndex, rightIndex, nodeVariable);
        }

        if (rightBound > 0) {
          final NodeValue leftIndex =
              new NodeValue(DataType.INTEGER.valueOf(String.valueOf(rightBound - 1), decRadix));
          final NodeValue rightIndex = new NodeValue(DataType.INTEGER.valueOf("0", decRadix));

          lessSignPart =
              new NodeOperation(StandardOperation.BVEXTRACT, leftIndex, rightIndex, nodeVariable);
        }

        if (mostSignPart == null || lessSignPart == null) {
          if (mostSignPart == null && lessSignPart != null) {
            expression =
                new NodeOperation(StandardOperation.BVCONCAT, expression, lessSignPart);
          } else {
            expression =
                new NodeOperation(StandardOperation.BVCONCAT, mostSignPart, expression);
          }
        } else {
          expression =
              new NodeOperation(StandardOperation.BVCONCAT, mostSignPart, expression, lessSignPart);
        }
      }

      if (substitutions.containsKey(nodeVariable)) {
        substitutions.remove(nodeVariable);
      }

      substitutions.put(nodeVariable, expression);
    }

    return substitutions;
  }
}
