/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.expression.Node;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;

/**
 * The interface for EFSM visitors. Typically, EFSM visitors are used along with EFSM walkers.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see Efsm
 * @see EfsmWalker
 */
public interface EfsmVisitor {
  /**
   * Indicates that processing of the specified EFSM has been started.
   *
   * @param efsm the processed EFSM
   */
  public void onEfsmBegin(Efsm efsm);

  /**
   * Indicates that processing of the specified EFSM has been finished.
   *
   * @param efsm the processed EFSM
   */
  public void onEfsmEnd(Efsm efsm);

  /**
   * Indicates that processing of the specified EFSM state has been started.
   *
   * @param state the processed EFSM state
   */
  public void onStateBegin(EfsmState state);

  /**
   * Indicates that processing of the specified EFSM state has been finished.
   *
   * @param state the processed EFSM state
   */
  public void onStateEnd(EfsmState state);

  /**
   * Indicates that processing of the specified EFSM transition has been started.
   *
   * @param transition the processed EFSM transition
   */
  public void onTransitionBegin(EfsmTransition transition);

  /**
   * Indicates that processing of the specified EFSM transition has been finished.
   *
   * @param transition the processed EFSM transition
   */
  public void onTransitionEnd(EfsmTransition transition);

  /**
   * Indicates that processing of the specified guarded action has been started.
   *
   * @param guardedAction the processed guarded action
   */
  public void onGuardedActionBegin(GuardedAction guardedAction);

  /**
   * Indicates that processing of the specified guarded action has been finished.
   *
   * @param guardedAction the processed guarded action
   */
  public void onGuardedActionEnd(GuardedAction guardedAction);

  /**
   * Indicates encounter of the specified event.
   *
   * @param event the encountered event
   */
  public void onEvent(Event event);

  /**
   * Indicates that processing of the specified guard has been started.
   *
   * @param guard the processed guard
   */
  public void onGuardBegin(Guard guard);

  /**
   * Indicates that processing of the specified guard has been finished.
   *
   * @param guard the processed guard
   */
  public void onGuardEnd(Guard guard);

  /**
   * Indicates that processing of the specified assignment action has been started.
   *
   * @param action the processed assignment action
   */
  public void onAssignmentActionBegin(Action action);

  /**
   * Indicates that processing of the specified assignment action has been finished.
   *
   * @param action the processed assignment action
   */
  public void onAssignmentActionEnd(Action action);

  /**
   * Indicates that processing of the specified assign statement has been started.
   *
   * @param statement the processed assign statement
   */
  public void onAssignmentBegin(Assignment statement);

  /**
   * Indicates that processing of the specified assign statement has been finished.
   *
   * @param statement the processed assign statement
   */
  public void onAssignmentEnd(Assignment statement);

  /**
   * Indicates that processing of the specified expression has been started.
   *
   * @param expression the processed expression
   */
  public void onExpression(Node expression);
}
