/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.MetaInfo;
import ru.ispras.retrascope.model.basis.UseDef;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Boolean condition that consists of some predicates.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 * 
 */
public class Guard extends MetaInfo implements UseDef {

  /* String representation constants. */

  private static final String LEFT_BOUND = "{";
  private static final String PREDICATE_HEADER = "predicate: ";
  private static final String PREDICATE_DELIMITER = "; ";
  private static final String RIGHT_BOUND = "}";

  /* The Boolean predicates. */
  private final Collection<Node> predicates;

  /**
   * Default constructor.
   */
  public Guard() {
    super();
    this.predicates = new LinkedHashSet<>();
  }

  /**
   * Constructs object with the specified boolean predicate.
   *
   * @param node boolean guard predicate
   *
   * @throws IllegalArgumentException when argument is either {@code null}
   *         or is not of {@code DataType.BOOLEAN} type.
   */
  public Guard(final Node node) {
    super();
    InvariantChecks.checkNotNull(node);
    checkNode(node);

    this.predicates = new LinkedHashSet<>();

    this.predicates.add(node);
  }

  /**
   * Constructs object from the specified list of Boolean {@link Node} objects.
   *
   * @param list list of guard predicates
   *
   * @throws IllegalArgumentException when either argument is {@code null}
   *         or at least one collection element is not of {@code DataType.BOOLEAN} type.
   */
  public Guard(final Collection<Node> list) {
    super();
    InvariantChecks.checkNotNull(list);
    checkList(list);

    this.predicates = new LinkedHashSet<>();
    this.predicates.addAll(list);
  }

  private static void checkNode(final Node node) {
    if (!node.isType(DataType.BOOLEAN)) {
      throw new IllegalArgumentException(
          String.format(
              "Wrong type for predicate - %s; %s is correct.",
              node.getDataTypeId().name(),
              DataTypeId.LOGIC_BOOLEAN.name()));
    }
  }

  private static void checkList(final Collection<Node> list) {
    for (final Node node : list) {
      checkNode(node);
    }
  }

  /**
   * Returns collection of guard predicates.
   *
   * @return guard predicates list
   */
  public Collection<Node> getPredicates() {
    return this.predicates;
  }

  /**
   * Convert guard expression into {@link ru.ispras.fortress.expression.Node} expression.
   *
   * @return {@link ru.ispras.fortress.expression.Node} expression.
   */
  public Node toNode() {
    return ExprUtils.getConjunction(this.predicates.toArray(new Node[this.predicates.size()]));
  }

  /**
   * Returns unique string identifier of this object.
   *
   * @return identifier
   */
  public String getId() {
    final StringBuilder stringBuilder = new StringBuilder(Guard.LEFT_BOUND);

    final int initLength = stringBuilder.length();

    for (final Node predicate: this.predicates) {
      if (stringBuilder.length() > initLength) {
        stringBuilder.append(Guard.PREDICATE_DELIMITER);
      }

      stringBuilder.append(Guard.PREDICATE_HEADER);
      stringBuilder.append(predicate);
    }

    stringBuilder.append(Guard.RIGHT_BOUND);

    return stringBuilder.toString();
  }

  @Override
  public Set<NodeVariable> getUses() {
    Set<NodeVariable> union = new LinkedHashSet<>();

    for (final Node predicate : this.getPredicates()) {

      final Set<NodeVariable> prUses = new LinkedHashSet<>(ExprUtils.getVariables(predicate));

      if (union.isEmpty()) {
        union.addAll(prUses);
      } else {
        union = CollectionUtils.uniteSets(union, prUses);
      }
    }
    return union;
  }

  @Override
  public Set<NodeVariable> getDefines() {
    return new LinkedHashSet<>();
  }

  @Override
  public String toString() {
    return getId();
  }

  @Override
  public boolean equals(final Object obj) {
    return this == obj;
  }

  @Override
  public int hashCode() {
    return System.identityHashCode(this);
  }
}
