/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.Process;

/**
 * The control flow graph model visitor that transforms sequences of basic block nodes.
 *
 * <p>The component transforms sequences of {@link BasicBlock}
 * basic block nodes of the same kind of concurrency
 *  (all nodes of the sequence should be either concurrent
 *  or non-concurrent).</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgBasicBlockSequenceBackend extends CfgDefaultVisitor {

  /* The current process. */
  private Process process;

  /**
   * Default constructor.
   */
  public CfgBasicBlockSequenceBackend() {
    super();
  }

  @Override
  public void onProcessBegin(final Process process) {
    this.process = process;
  }

  @Override
  public void onBasicBlockBegin(final BasicBlock basicBlock) {
    while (hasSameConcurrentChild(basicBlock)) {
      merge(basicBlock, (BasicBlock) basicBlock.getOnlyChild());
    }
  }

  private void merge(final BasicBlock basicBlock, final BasicBlock child) {
    basicBlock.addAssignments(child.getAssignments());
    basicBlock.removeChild(child);
    if (child.hasOnlyChild()) {
      final CfgModelNode grandChild = child.getOnlyChild();
      grandChild.removeParent(child);
      basicBlock.addChild(grandChild);
    } else {
      throw new IllegalArgumentException("Do not have exactly one child: " + child.getId() + ".");
    }

    /* Remove hanging node. */
    this.process.removeNode(child);
  }

  /**
   * Checks rather the specified {@link BasicBlock} node has
   * {@link BasicBlock} child of the same concurrent kind.
   *
   * @param basicBlock target node that is checked
   *
   * @return {@code true} when the specified node has {@link BasicBlock} child
   *         node of the same concurrent kind, {@code false} otherwise.
   */
  private boolean hasSameConcurrentChild(final BasicBlock basicBlock) {
    if (basicBlock.hasOnlyChild()) {
      final CfgModelNode blockOnlyChild = basicBlock.getOnlyChild();

      return blockOnlyChild instanceof BasicBlock
          && ((BasicBlock) blockOnlyChild).isConcurrent() == basicBlock.isConcurrent();
    }
    return false;
  }
}
