/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.Module;

import java.util.ArrayList;
import java.util.List;

/**
 * The control flow graph model visitor that transforms concurrent basic block nodes.
 *
 * <p>TThe component transforms concurrent (non-blocking)
 * {@link BasicBlock} basic block nodes into non-concurrent
 * (blocking) ones.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgConcurrentBasicBlockBackend extends CfgDefaultVisitor {

  /* The prefix for new variables names (for those that are defined at blocking assignments). */
  private static final String LINE = "_";
  private static final String BLOCK = "BLOCK" + LINE;

  /* The current module. */
  private Module module;

  /**
   * Default constructor.
   */
  public CfgConcurrentBasicBlockBackend() {
    super();
  }

  @Override
  public void onModuleBegin(final Module module) {
    this.module = module;
  }

  @Override
  public void onBasicBlockBegin(final BasicBlock basicBlock) {

    /* Transform concurrent block into sequential (blocking) one. */
    if (basicBlock.isConcurrent()) {
      if (hasNonConstantAssigns(basicBlock)) {
        final List<Assignment> sequential = makeBlocking(basicBlock.getAssignments());
        basicBlock.removeAssignments();
        basicBlock.addAssignments(sequential);
      }
      basicBlock.setConcurrent(false);
    }
  }

  private boolean hasNonConstantAssigns(final BasicBlock basicBlock) {
    for (final Assignment assignment : basicBlock.getAssignments()) {
      if (!ExprUtils.getVariables(assignment.getValue()).isEmpty()) {
        return true;
      }
    }
    return false;
  }

  /**
   * Transforms the specified collection of non-blocking {@link Assignment} objects
   * into the collection of blocking ones.
   *
   * @param list non-blocking {@link Assignment} objects collection
   *
   * @return blocking {@link Assignment} objects collection
   */
  private List<Assignment> makeBlocking(final List<Assignment> list) {

    /* Create a sequence of assignments like "new variable := used expression". */
    final List<Assignment> blocking = new ArrayList<>();

    /* Create a sequence of assignments like "defined variable := new variable". */
    final List<Assignment> oldToNewList = new ArrayList<>();

    for (final Assignment assignment : list) {

      final RangedVariable rVariable = assignment.getTarget();

      /* Create variable with unique name. */
      final String rangeNamePart = rVariable.isRanged()
          ? rVariable.getRange().getOld() + LINE + rVariable.getRange().getYoung() + LINE : "";
      final String varName = BLOCK
          + rVariable.getVariableName() + LINE + rangeNamePart + blocking.size();

      final NodeVariable var = new NodeVariable(varName, assignment.getValue().getDataType());
      if (!this.module.containsVariable(var)) {
        this.module.declareVariable(var);
      }
      var.setUserData(new VariableData(VariableType.REG));
      blocking.add(new Assignment(var, assignment.getValue()));

      oldToNewList.add(new Assignment(rVariable, var));
    }
    blocking.addAll(oldToNewList);
    return blocking;
  }
}
