/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.MetaInfo;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.MetaInfoValue;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableContainer;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * The control flow graph model visitor that detects loops or cycles and transforms them.
 *
 * <p>The loops that this class is able to detect and transform MUST contain nodes
 * with meta-information of the following values:</p>
 * <ul>
 * <li>
 * {@code MetaInfoValue.LOOP_INIT} - {@link BasicBlock} node
 * that contains loop counter variable initialization;
 * </li>
 * <li>
 * {@code MetaInfoValue.LOOP_ENTRY} - {@link Merge} node
 * that has loop first and last nodes as child and parent nodes respectively
 * (this node is a parent for the {@link BasicBlock} node
 * with {@code MetaInfoValue.LOOP_INIT} meta-information value);
 * </li>
 * <li>
 * {@code MetaInfoValue.LOOP_COND} - {@link Switch} node
 * that contains the condition on loop counter variable;
 * </li>
 * <li>
 * {@code MetaInfoValue.LOOP_INC} - optional {@link BasicBlock} node
 * that contains {@link Assignment} that increments loop counter variable
 * (this node does not appear at the infinite loops);
 * </li>
 * <li>
 * {@code MetaInfoValue.LOOP_FIN} - {@link CfgNode} node
 * that is a parent for {@link Merge} node with
 * {@code MetaInfoValue.LOOP_ENTRY} meta information value.
 * </li>
 * </ul>
 * 
 * <p>When the loop counter condition is {@code true}, the counter-incrementing {@link BasicBlock}
 * node is absent, because all the loop-body-sub-graph will be executed in the dependant separate
 * process. Otherwise the loop is transformed to the sequence of loop-body-sub-graphs. The amount
 * of the sequence is calculated as how many times the expression-upon-counter will stay
 * satisfiable when counter will be incremented on calling the increment-function again and again.
 * </p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgLoopBackend extends CfgDefaultVisitor {

  /* The current module. */
  private Module module;

  /* The processes that are extracted from infinite loops. */
  private List<Process> loopProcesses;

  /* The current process. */
  private Process process;

  /**
   * Default constructor.
   */
  public CfgLoopBackend() {
    super();
  }

  private List<Process> getLoopProcesses() {
    return this.loopProcesses;
  }

  private Module getModule() {
    return this.module;
  }

  private Process getProcess() {
    return this.process;
  }

  @Override
  public void onModuleBegin(final Module module) {
    this.module = module;
    this.loopProcesses = new ArrayList<>();
  }

  @Override
  public void onModuleEnd(final Module module) {
    this.module = null;

    if (!getLoopProcesses().isEmpty()) {
      for (final Process loopProcess : this.loopProcesses) {
        module.addChild(loopProcess);
      }
      this.loopProcesses = null;
    }
  }

  @Override
  public void onProcessBegin(final Process process) {
    this.process = process;
  }

  @Override
  public void onMergeBegin(final Merge merge) {

    /* If node is loop entry. */
    if (hasMetaInfoValue(merge, MetaInfoValue.LOOP_ENTRY)) {
      elaborateLoop(merge);
    }
  }

  private boolean hasMetaInfoValue(CfgNode node, MetaInfoValue value) {
    return node.hasMetaInfo(MetaInfoType.LOOP_NODE)
        && node.getMetaInfo(MetaInfoType.LOOP_NODE).contains(value.name());
  }

  private void elaborateLoop(final Merge merge) {

    final Collection<CfgModelNode> parents = merge.getParents();

    /* The loop counter initialization node. */
    BasicBlock initBlock;

    /* The loop counter incrementation node. */
    BasicBlock incrementBlock;


    /* The "to-be-removed" node. */
    Merge toBeRemoved;

    final LoopComponents loopComponents =
        new LoopComponents(parents, null, null, null, null).init();
    toBeRemoved = loopComponents.getToBeRemoved();
    incrementBlock = loopComponents.getIncrementBlock();
    initBlock = loopComponents.getInitBlock();

    /* The last of the loop nodes. */
    CfgNode loopLastNode = loopComponents.getLoopLastNode();

    /* Remove the single path merge node
     * (that has only one parent increment block and only one child node). */
    if (toBeRemoved != null && incrementBlock != null) {
      merge.removeParent(toBeRemoved);
      incrementBlock.removeChild(toBeRemoved);
      incrementBlock.addChild(merge);
      getProcess().removeNode(toBeRemoved);
    }

    final CfgNode child = (CfgNode) merge.getOnlyChild();

    Switch condition = null;

    /* If node is loop condition. */
    if (hasMetaInfoValue(child, MetaInfoValue.LOOP_COND)) {
      condition = (Switch) child;
    }

    makeNoLoopSubGraph(merge, initBlock, incrementBlock, loopLastNode, condition);
  }

  private void makeNoLoopSubGraph(
      final Merge merge,
      final BasicBlock initBlock,
      final BasicBlock incrementBlock,
      final CfgNode loopLastNode,
      final Switch condition) {

    final Node conditionExpr = condition.getCondition();

    /* Detect TRUE/FALSE case child for loop condition switch node. */
    final Collection<CfgModelNode> condChildren = condition.getChildren();
    final Iterator<CfgModelNode> iterator = condChildren.iterator();
    final Case firstCase = (Case) iterator.next();
    final Case secondCase = (Case) iterator.next();
    final Case trueCase = getValueCase(firstCase, secondCase, true);
    final Case falseCase = getValueCase(firstCase, secondCase, false);

    /* Remove old links, make merge block child-free & false-case-child parent-free. */
    merge.removeChild(condition);
    final CfgNode falseCaseChild = (CfgNode) falseCase.getOnlyChild();
    falseCase.removeChild(falseCaseChild);

    if (initBlock == null || incrementBlock == null) {

      if (ExprUtils.getVariables(conditionExpr).isEmpty()) {

        processInfiniteLoop(merge, conditionExpr, trueCase, falseCaseChild);

      } else {
        throw new IllegalArgumentException("Cannot parse loop: " + condition.toString() + ".");
      }
    } else {
      processFiniteLoop(merge, initBlock, incrementBlock, conditionExpr, trueCase, falseCaseChild);
    }

    /* No more loop merge here. */
    merge.removeMetaInfo(MetaInfoType.LOOP_NODE);

    /* Remove outdated nodes. */
    getProcess().removeNode(condition);
    getProcess().removeNode(trueCase);
    getProcess().removeNode(falseCase);

    if (incrementBlock == null
        && loopLastNode != null
        && merge.getParents().contains(loopLastNode)) {

      merge.removeParent(loopLastNode);

    } else {
      getProcess().removeNode(incrementBlock);
      merge.removeParent(incrementBlock);
    }
  }

  private void processInfiniteLoop(
      final Merge merge,
      final Node conditionExpr,
      final Case trueCase,
      final CfgNode falseCaseChild) {

    if (ExprUtils.isSAT(conditionExpr)) {

      final List<CfgNode> loopPath = CfgUtils.getPath(trueCase, merge);

      final Process newProcess = new Process();

      /* Create new control flow graph for new process, */
      final List<CfgNode> processNodes = new ArrayList<>();
      final Source src = new Source();
      final Sink sink = new Sink();
      newProcess.addChild(src);
      processNodes.add(src);
      processNodes.add(sink);
      newProcess.addNodes(processNodes);

      final String loopVariableName =
          VariableContainer.NEW_VAR_PREFIX + getModule().getDeclarations().size();
      final NodeVariable loopVariable = new NodeVariable(loopVariableName, DataType.BOOLEAN);

      getModule().declareVariable(loopVariable);

      newProcess.setSensitivityList(
          new EventList(Collections.singleton(new Event(new RangedVariable(loopVariable)))));

      /* Init loop variable in the block node & add it to the parent process. */
      final BasicBlock loopVariableBlock = new BasicBlock(loopVariable, NodeValue.newBoolean(true));

      getProcess().addNode(loopVariableBlock);
      merge.addChild(loopVariableBlock);
      loopVariableBlock.addChild(falseCaseChild);

      /* Construct the sub-process control flow graph. */

      newProcess.addNodes(loopPath);

      final CfgNode first = loopPath.get(0);
      final CfgNode last = loopPath.get(loopPath.size() - 1);

      /* If loop is not isolated. */
      if (first.hasParents() || last.hasChildren()) {
        throw new IllegalArgumentException("Wrong path has been extracted.");
      } else {
        src.addChild(first);
        last.addChild(sink);
      }
      getLoopProcesses().add(newProcess);

      /* Remove internal loop nodes. */
      for (final CfgNode loopNode : loopPath) {
        getProcess().removeNode(loopNode);
      }
    } else {

      /* Dead code has been found. */
      merge.addChild(falseCaseChild);
    }
  }

  private void processFiniteLoop(
      final Merge merge,
      final BasicBlock initBlock,
      final BasicBlock incrementBlock,
      final Node conditionExpr,
      final Case trueCase,
      final CfgNode falseCaseChild) {

    final List<CfgNode> loopPath = CfgUtils.getPath(trueCase, incrementBlock);

    /* Build a sequence of iterations. */
    CfgNode unlinked = merge;
    final Node incrementExpr = incrementBlock.getAssignments().get(0).getValue();
    final Assignment initAssignment = initBlock.getAssignments().get(0);
    final String varName = initAssignment.getTarget().getVariable().getName();

    Node increment = initAssignment.getValue();
    Node whileExpr = Transformer.substitute(conditionExpr, varName, increment);

    while (ExprUtils.isSAT(whileExpr)) {

      /* Create a copy of path. */
      final List<CfgNode> loopPathCopy = new ArrayList<>(loopPath.size());
      for (final CfgNode origNode : loopPath) {
        final CfgModelNode copy = origNode.deepCopy();
        if (!loopPathCopy.isEmpty()) {
          copy.addParent(loopPathCopy.get(loopPathCopy.size() - 1));
        }
        loopPathCopy.add((CfgNode) copy);
      }

      /* Substitute iteration variable by it's incrementation expression. */
      final List<CfgNode> substituted = CfgUtils.substitute(loopPathCopy, varName, increment);

      unlinked.addChild(substituted.get(0));

      getProcess().addNodes(substituted);
      unlinked = substituted.get(substituted.size() - 1);
      increment = Transformer.substitute(incrementExpr, varName, increment);
      whileExpr = Transformer.substitute(conditionExpr, varName, increment);
    }

    /* Remove internal loop nodes. */
    for (final CfgNode loopNode : loopPath) {
      getProcess().removeNode(loopNode);
    }

    unlinked.addChild(falseCaseChild);
  }

  private static Case getValueCase(
      final Case firstCase,
      final Case secondCase,
      final boolean value) {
    return firstCase.getValues().contains(NodeValue.newBoolean(value)) ? firstCase : secondCase;
  }

  private class LoopComponents {
    private final Collection<CfgModelNode> parents;
    private BasicBlock initBlock;
    private BasicBlock incrementBlock;
    private CfgNode loopLastNode;
    private Merge toBeRemoved;

    public LoopComponents(
        final Collection<CfgModelNode> parents,
        final BasicBlock initBlock,
        final BasicBlock incrementBlock,
        final CfgNode loopLastNode,
        final Merge toBeRemoved) {

      this.parents = parents;
      this.initBlock = initBlock;
      this.incrementBlock = incrementBlock;
      this.loopLastNode = loopLastNode;
      this.toBeRemoved = toBeRemoved;
    }

    public BasicBlock getInitBlock() {
      return initBlock;
    }

    public BasicBlock getIncrementBlock() {
      return incrementBlock;
    }

    public CfgNode getLoopLastNode() {
      return loopLastNode;
    }

    public Merge getToBeRemoved() {
      return toBeRemoved;
    }

    public LoopComponents init() {
      for (final CfgModelNode parent : parents) {
        if (((MetaInfo)parent).hasMetaInfo(MetaInfoType.LOOP_NODE)) {

          final String info = ((MetaInfo)parent).getMetaInfo(MetaInfoType.LOOP_NODE);

          if (info.equals(MetaInfoValue.LOOP_INIT.name())) {

            /* The counter initialization block has been found. */
            initBlock = (BasicBlock) parent;

          } else if (info.equals(MetaInfoValue.LOOP_INC.name())) {

            /* The counter incrementation block has been found. */
            incrementBlock = (BasicBlock) parent;
          } else if (info.equals(MetaInfoValue.LOOP_FIN.name())) {

            /* The final node of the loop iteration has been found. */
            loopLastNode = (CfgNode) parent;
          }
        }
        if (parent instanceof Merge && isSinglePathNode((Merge) parent)
            && isIncrementBlock(parent.getOnlyParent())) {

          incrementBlock = (BasicBlock) parent.getOnlyParent();
          toBeRemoved = (Merge) parent;
        }
      }
      return this;
    }

    private boolean isSinglePathNode(final Merge parent) {
      return parent.getChildren().size() == 1 && parent.getParents().size() == 1;
    }

    private boolean isIncrementBlock(final CfgModelNode node) {
      return ((MetaInfo)node).hasMetaInfo(MetaInfoType.LOOP_NODE)
          && ((MetaInfo)node).getMetaInfo(MetaInfoType.LOOP_NODE)
              .contains(MetaInfoValue.LOOP_INC.name());
    }
  }
}
