/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Process;

import java.util.Collection;

/**
 * Control flow graph model visitor that detects sequences of merge nodes and transforms them.
 *
 * <p>This backend merges {@link ru.ispras.retrascope.model.cfg.Merge} nodes,
 * following each other (i.e. sequences of merge nodes).
 * Also it removes {@link ru.ispras.retrascope.model.cfg.Merge} nodes,
 * which have exactly one parent node and exactly one child node.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgMergeSequenceBackend extends CfgDefaultVisitor {

  /* Current process. */
  private Process process;

  /* Error messages. */
  private static final String NO_CHILDREN = "Has no children: ";

  /**
   * Default constructor.
   */
  public CfgMergeSequenceBackend() {
    super();
  }

  @Override
  public void onProcessBegin(final Process process) {
    this.process = process;
  }

  @Override
  public void onMergeBegin(final Merge merge) {

    while (hasOnlyMergeChild(merge)) {
      mergeSequence(merge, (Merge) merge.getOnlyChild());
    }
  }

  private void mergeSequence(
      final Merge merge,
      final Merge child) {

    merge.removeChild(child);

    if (child.hasChildren()) {

      final Collection<? extends CfgModelNode> childParents = child.getParents();

      for (final CfgModelNode childParent : childParents) {

        if (!childParent.equals(merge)) {

          childParent.removeChild(child);
          childParent.addChild(merge);

        }
      }
      if (child.hasChildren()) {

        final Collection<CfgModelNode> grandChildren = child.getChildren();

        for (final CfgModelNode grandChild : grandChildren) {

          grandChild.removeParent(child);
          grandChild.addParent(merge);
        }
      } else {
        throw new IllegalArgumentException(NO_CHILDREN + child.getType() + ".");
      }
    }

    /* Remove "hanging" nodes. */
    if (this.process != null) {
      this.process.removeNode(child);
    }
  }

  private boolean hasOnlyMergeChild(final Merge merge) {
    if (merge.hasChildren()) {
      if (merge.hasOnlyChild() && merge.getOnlyChild() instanceof Merge) {
        return true;
      }
    } else {
      throw new IllegalStateException(NO_CHILDREN + merge + ".");
    }
    return false;
  }

  @Override
  public void onMergeEnd(final Merge merge) {
    if (merge.hasOnlyParent() && merge.hasOnlyChild()) {
      final CfgModelNode parent = merge.getOnlyParent();
      final CfgModelNode child = merge.getOnlyChild();
      parent.removeChild(merge);
      child.removeParent(merge);
      parent.addChild(child);
      this.process.removeNode(merge);
    }
  }
}
