/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNodeType;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 *Control flow graph model visitor that detects one-basic-block processes and transforms them.
 *
 * <p>This backend merges {@link ru.ispras.retrascope.model.cfg.Process} processes,
 * which contain only one {@link ru.ispras.retrascope.model.cfg.BasicBlock} basic block node
 * of concurrent type, with only one {@link ru.ispras.retrascope.model.basis.Assignment} assignment,
 * and equal {@link ru.ispras.retrascope.model.basis.EventList} event lists.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgOneAssignmentProcessBackend extends CfgDefaultVisitor {

  /**
   * Default constructor.
   */
  public CfgOneAssignmentProcessBackend() {
    super();
  }

  @Override
  public void onModuleBegin(final Module module) {

    final Set<Process> merged = new LinkedHashSet<>();
    final Collection<Process> processes = module.getProcesses();

    if (!processes.isEmpty()) {
      for (final Process process : processes) {
        if (isOneBlockCfg(process)) {

          if (merged.isEmpty()) {
            merged.add(process);

          } else {

            boolean isMerged = false;
            for (final Process mergedProcess : merged) {

              /* If the target process can be merged into "merged" process. */
              if (isOneBlockCfg(mergedProcess)
                  && mergedProcess.getSensitivityList().equals(process.getSensitivityList())) {
                mergeProcesses(mergedProcess, process);
                isMerged = true;
                break;
              }
            }
            if (!isMerged) {
              merged.add(process);
            }
          }
        } else {
          merged.add(process);
        }
      }

      /* Remove "old" processes. */
      module.removeProcesses();

      /* Add new "merged" processes. */
      for (final Process process : merged) {
        module.addChild(process);
      }
    }
  }

  private void mergeProcesses(final Process result, final Process target) {

    /* Merge sensitivity lists. */
    final EventList rList = result.getSensitivityList();
    final Collection<Event> tList = target.getSensitivityList().getEvents();

    for (final Event event : tList) {
      rList.add(event);
    }

    /* Merge control flow graphs. */

    /* Get basic bock node at the result process. */
    final Source resultSrc = CfgUtils.getSource(result.getNodes());
    final BasicBlock resultBasicBlock = (BasicBlock) resultSrc.getOnlyChild();

    /* Get basic block at the target process. */
    final Sink targetSrc = CfgUtils.getSink(target.getNodes());
    final BasicBlock targetBasicBlock = (BasicBlock) targetSrc.getOnlyChild();

    resultBasicBlock.addAssignments(targetBasicBlock.getAssignments());
  }

  private boolean isOneBlockCfg(final Process process) {
    if (process.hasChildren()) {
      final Collection<CfgModelNode> children = process.getChildren();
      return process.hasOnlyChild() && isConcurrentAssignmentBlockCfg(children.iterator().next());
    } else {
      return false;
    }
  }

  private boolean isConcurrentAssignmentBlockCfg(final CfgModelNode node) {
    if (node instanceof Source) {
      final CfgModelNode child = node.getOnlyChild();

      return child instanceof BasicBlock
          && child.getChildren().iterator().next() instanceof Sink
          && ((BasicBlock) child).isConcurrent();

    } else {
      throw new IllegalArgumentException(
          String.format(
              "The %s node MUST be the %s node's child.",
              CfgNodeType.SOURCE.name(),
              CfgNodeType.PROCESS.name()));
    }
  }
}
