/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Switch;

import java.util.Collection;

/**
 * Control flow graph model visitor that detects and transforms sequences of switch nodes.
 *
 * <p>The HDL parser backend that optimises sequences
 * of control flow graph {@link ru.ispras.retrascope.model.cfg.CfgModel }model
 * {@link ru.ispras.retrascope.model.cfg.Switch} switch nodes.
 * Switch nodes must contain conditions of the following form:
 * {@code x == const_1 || ... || x == const_n}, and only the {@code x} must be used
 * in every node of the sequences to be transformed.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgSwitchSequenceBackend extends CfgDefaultVisitor {

  /* The current process. */
  private Process process;

  /**
   * Default constructor.
   */
  public CfgSwitchSequenceBackend() {
    super();
  }

  @Override
  public void onProcessBegin(final Process process) {
    this.process = process;
  }

  @Override
  public void onSwitchBegin(final Switch switchNode) {

    Switch grandChild;
    while ((grandChild = getSwitchGrandChildWithEqualExpr(switchNode)) != null) {
      collapseSwitchSubTree(switchNode, grandChild);
    }
  }

  private void collapseSwitchSubTree(
      final Switch switchNode,
      final Switch grandChild) {

    final Collection<CfgModelNode> children = grandChild.getChildren();

    /* Remove links between next switch and it's children. */
    for (final CfgModelNode child : children) {
      grandChild.removeChild(child);
      switchNode.addChild(child);
    }

    /* Remove links from and to middle node. */
    final CfgModelNode middle = grandChild.getOnlyParent();
    middle.removeChild(grandChild);
    middle.removeParent(switchNode);

    this.process.removeNode((CfgNode) middle);
    this.process.removeNode(grandChild);
  }

  private Switch getSwitchGrandChildWithEqualExpr(final Switch switchNode) {
    if (switchNode.hasChildren()) {
      final Collection<CfgModelNode> switchChildren = switchNode.getChildren();
      for (final CfgModelNode child : switchChildren) {
        if (child.hasOnlyChild()) {

          final CfgModelNode grandChild = child.getOnlyChild();
          if (grandChild instanceof Switch) {
            final Node grandChildExpr = ((Switch) grandChild).getCondition();
            final Node kernel = getMultiOrKernel(grandChildExpr);

            if (switchNode.getCondition().equals(grandChildExpr)
                || kernel != null && switchNode.getCondition().equals(kernel)) {
              return (Switch) grandChild;
            }
          }
        } else {
          throw new IllegalArgumentException(
              "Node has not exactly one child: " + child.getId() + ".");
        }
      }
      return null;
    } else {
      throw new IllegalStateException("Node has no children: " + switchNode.getType() + ".");
    }
  }

  private Node getMultiOrKernel(final Node expr) {
    if (expr instanceof NodeOperation) {
      if (((NodeOperation) expr).getOperationId() == StandardOperation.OR) {
        Node kernel = null;
        int valIndex = -1;
        int varIndex = -1;

        for (int i = 0; i < ((NodeOperation) expr).getOperandCount(); i++) {
          final Node exprOperand = ((NodeOperation) expr).getOperand(i);
          if (exprOperand instanceof NodeOperation) {
            if (((NodeOperation) exprOperand).getOperationId() == StandardOperation.EQ) {
              for (int j = 0; j < ((NodeOperation) exprOperand).getOperandCount(); j++) {
                final Node operand = ((NodeOperation) exprOperand).getOperand(j);
                if (operand instanceof NodeVariable) {
                  varIndex = j;
                  if (kernel != null && !kernel.equals(operand)) {
                    return null;
                  }
                  kernel = operand;
                } else if (operand instanceof NodeValue) {
                  valIndex = j;
                }
              }
              if (varIndex == -1 || valIndex == -1) {
                return null;
              }
            }
          } else {
            return null;
          }
        }
        return kernel;
      } else {
        return null;
      }

    } else {
      return null;
    }
  }
}
