/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableContainer;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.CfgDefaultVisitor;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.CfgNodeType;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.cfg.Wait;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Control flow graph model visitor that detects and transforms wait nodes.
 *
 * <p>The transformation result does not contain
 * {@link ru.ispras.retrascope.model.cfg.Wait} nodes.
 * The visitor substitutes them by sub-processes
 * which are 'guarded' by predicates on new variables.
 * The situations when the same sequence of graph nodes
 * can be executed both after and not after
 * {@link ru.ispras.retrascope.model.cfg.Wait} node are also supported.</p>
 *
 * <p>There is Zamia IG feature which is taken into account in this backend.
 * The IG builder creates objects (that can be treated as
 * {@link ru.ispras.retrascope.model.cfg.Wait} objects)
 * every time when it detects some process' sensitivity list.
 * The builder puts this node just before the
 * {@link ru.ispras.retrascope.model.cfg.Sink} node. Such nodes are treated
 * by the backend as finalizing and are substituted by sensitivity list.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgWaitBackend extends CfgDefaultVisitor {

  /* The current module. */
  private Module module;

  /* The current process. */
  private Process process;

  /* The current parent wait node. */
  private Wait waitNode;

  /* The mapping for wait nodes & nodes of their processes. */
  private Map<Wait, Set<CfgNode>> map;

  /* The control flow graph nodes and their related nearest parent wait nodes. */
  private Map<CfgNode, Wait> nearestWaitMap;

  /* The wait nodes parents & corresponding to them new variables. */
  private Map<CfgNode, NodeVariable> parentNewVarMap;

  /* The wait nodes & corresponding new variables. */
  private Map<Wait, NodeVariable> waitNewVarMap;

  /* The nodes that are reachable non through wait. */
  private Set<CfgNode> waitUnReachable;

  /* The collection for wait-processes. */
  private Map<Module, List<Process>> waitProcessMap;

  /* The collection of finalizing wait nodes. */
  private Map<Process, Wait> finalizing;

  /**
   * Default constructor.
   */
  public CfgWaitBackend() {
    super();
  }

  @Override
  public void onCfgModelBegin() {
    this.waitProcessMap = new LinkedHashMap<>();
    this.finalizing = new LinkedHashMap<>();
  }

  @Override
  public void onCfgModelEnd() {

    /* Remove finalizing wait nodes from the model. */
    for (final Map.Entry<Process, Wait> entry : this.finalizing.entrySet()) {
      final Wait wait = entry.getValue();
      final CfgNode waitParent = (CfgNode) wait.getOnlyParent();
      final CfgNode waitChild = (CfgNode) wait.getOnlyChild();
      entry.getKey().removeNode(wait);
      waitParent.removeChild(wait);
      wait.removeChild(waitChild);
      waitParent.addChild(waitChild);
    }

    /* Add wait sub-processes. */
    for (final Map.Entry<Module, List<Process>> entry : this.waitProcessMap.entrySet()) {
      final List<Process> waitProcesses = entry.getValue();
      for (final Process waitProcess : waitProcesses) {
        entry.getKey().addChild(waitProcess);
      }
    }

    this.waitProcessMap.clear();
  }

  @Override
  public void onModuleBegin(final Module module) {
    this.module = module;
    this.waitProcessMap.put(module, new ArrayList<Process>());
  }

  @Override
  public void onProcessBegin(final Process process) {
    this.process = process;
    this.map = new LinkedHashMap<>();
    this.waitUnReachable = new LinkedHashSet<>();
    this.nearestWaitMap = new LinkedHashMap<>();
    this.parentNewVarMap = new LinkedHashMap<>();
    this.waitNewVarMap = new LinkedHashMap<>();
  }

  @Override
  public void onProcessEnd(final Process process) {

    this.process = null;

    List<CfgNode> cfg = process.getNodes();

    final Map<CfgModelNode, CfgModelNode> oldNewMap = new LinkedHashMap<>();

    for (final Map.Entry<Wait, Set<CfgNode>> entry : this.map.entrySet()) {

      final Wait entryWait = entry.getKey();

      /* Create new process. */
      process.removeNode(entryWait);

      final Process waitProcess = new Process();
      waitProcess.setSensitivityList(getSensitivityList(entryWait));

      /* Create & fill process's control flow graph. */
      final List<CfgNode> waitCfg = new ArrayList<>();
      final Source src = new Source();
      final Sink sink = new Sink();
      waitCfg.add(src);
      waitCfg.add(sink);
      waitProcess.addChild(src);
      waitProcess.addNodes(waitCfg);

      /* Start CFG building from creating the switch-case pair
       * that checks the related new variable. */
      final NodeVariable newVariable = this.waitNewVarMap.get(entryWait);
      final Node variableCondition =
          new NodeOperation(StandardOperation.EQ, newVariable, NodeValue.newBoolean(true));
      final Node condition = entryWait.containsCondition()
          ? ExprUtils.getConjunction(entryWait.getCondition(), variableCondition)
          : variableCondition;

      final Switch mainSwitch = new Switch(condition);
      waitCfg.add(mainSwitch);
      src.addChild(mainSwitch);

      final Case aCase = new Case(NodeValue.newBoolean(true));
      waitCfg.add(aCase);
      mainSwitch.addChild(aCase);

      /* Remove the link with wait node parent, & wait node itself. */
      final CfgNode waitNodeParent = (CfgNode) entryWait.getOnlyParent();

      if (waitNodeParent != null) {
        finishPath(cfg, entryWait, newVariable);
      }

      /* Add nodes to the process control flow graph. */
      final Set<CfgNode> nodes = entry.getValue();

      if (nodes.isEmpty()) {
        final NodeVariable waitVariable = this.parentNewVarMap.get(entryWait);
        final BasicBlock waitVarBlock = new BasicBlock(waitVariable, NodeValue.newBoolean(true));
        waitCfg.add(waitVarBlock);
        waitVarBlock.addParent(aCase);
        waitVarBlock.addChild(sink);
      } else if (hasOnlySink(nodes)) {
        aCase.addChild(sink);
      } else {
        for (final CfgNode node : nodes) {

          final Collection<CfgModelNode> nodeParents = node.getParents();

          if (this.waitUnReachable.contains(node)) {

            /* Create the copy & put it to the new process. */
            final CfgNode copy = (CfgNode)node.deepCopy();
            waitCfg.add(copy);

            oldNewMap.put(node, copy);

            /* Try to find parent nodes that are already added to the new process. */
            for (final CfgModelNode nodeParent : nodeParents) {
              if (oldNewMap.containsKey(nodeParent)) {
                copy.addParent(oldNewMap.get(nodeParent));
              } else if (nodeParent instanceof CfgNode && waitCfg.contains(nodeParent)) {
                copy.addParent(nodeParent);
              }
            }

          } else {

            /* Move node to the new wait process. */
            cfg.remove(node);

            waitCfg.add(node);

            if (node.getOnlyParent().equals(entryWait)) {
              node.removeParent(entryWait);
              node.addParent(aCase);
            }
          }

          if (this.parentNewVarMap.containsKey(node)) {

            /* This node is parent for other wait, finish related sub-path with trigger block. */
            finishPath(waitCfg, (Wait) node.getOnlyChild(), this.parentNewVarMap.get(node));
          }
        }
      }

      this.waitProcessMap.get(this.module).add(waitProcess);
      cfg = waitProcess.getNodes();
    }
    this.map.clear();
    this.waitUnReachable.clear();
    this.nearestWaitMap.clear();
    this.parentNewVarMap.clear();
    this.waitNewVarMap.clear();
  }

  private EventList getSensitivityList(final Wait waitNode) {
    if (waitNode.getEventList().getEvents().isEmpty()) {
      if (waitNode.containsCondition()) {

        final Set<Event> events = new LinkedHashSet<>();
        final Collection<NodeVariable> variables = ExprUtils.getVariables(waitNode.getCondition());

        for (final NodeVariable variable : variables) {
          events.add(new Event(new RangedVariable(variable), EventType.ANY_EDGE));
        }
        return new EventList(events);
      } else {
        throw new IllegalArgumentException("Empty node of type: " + waitNode.getClass() + ".");
      }
    } else {
      return waitNode.getEventList().deepCopy();
    }
  }

  private boolean hasOnlySink(final Set<CfgNode> nodes) {
    return nodes.size() == 1 && nodes.iterator().next() instanceof Sink;
  }

  private void finishPath(
      final List<CfgNode> cfg,
      final Wait waitNode,
      final NodeVariable newVariable) {

    final Sink cfgSink = (Sink) CfgUtils.getNode(cfg, CfgNodeType.SINK);

    final CfgNode nodeParent = (CfgNode) waitNode.getOnlyParent();

    final BasicBlock trigger = new BasicBlock(newVariable, NodeValue.newBoolean(true));
    cfg.add(trigger);

    if (nodeParent.hasChildren()) {
      nodeParent.removeChild(waitNode);
    }
    nodeParent.addChild(trigger);

    /* Link trigger block with pre-sink merge node & create it if needed. */
    Merge preSinkMerge;
    final CfgNode sinkParent = (CfgNode) cfgSink.getOnlyParent();

    if (sinkParent != null && cfgSink.getOnlyParent() instanceof Merge) {

      /* Appropriate merge node has been found. */
      preSinkMerge = (Merge) cfgSink.getOnlyParent();
    } else {

      /* Insert merge that is a parent for sink node. */
      preSinkMerge = new Merge();

      if (sinkParent != null) {
        sinkParent.removeChild(cfgSink);
        sinkParent.addChild(preSinkMerge);
      }
      preSinkMerge.addChild(cfgSink);
      trigger.addChild(preSinkMerge);
    }
    cfg.add(preSinkMerge);
  }

  @Override
  public void onBasicBlockBegin(final BasicBlock basicBlock) {
    processNodeBegin(basicBlock);
  }

  @Override
  public void onBasicBlockEnd(final BasicBlock basicBlock) {
    processNodeEnd(basicBlock);
  }

  @Override
  public void onCaseBegin(final Case caseNode) {
    processNodeBegin(caseNode);
  }

  @Override
  public void onCaseEnd(final Case caseNode) {
    processNodeEnd(caseNode);
  }

  @Override
  public void onSwitchBegin(final Switch switchNode) {
    processNodeBegin(switchNode);
  }

  @Override
  public void onSwitchEnd(final Switch switchNode) {
    processNodeEnd(switchNode);
  }

  @Override
  public void onMergeBegin(final Merge merge) {
    processNodeBegin(merge);
  }

  @Override
  public void onMergeEnd(final Merge merge) {
    processNodeEnd(merge);
  }

  @Override
  public void onSrcBegin(final Source source) {
    processNodeBegin(source);
  }

  @Override
  public void onSink(final Sink sink) {
    processNodeBegin(sink);
  }

  @Override
  public void onWaitBegin(final Wait wait) {

    if (isFinalizing(wait)) {

      /* The finalizing wait node is detected. */
      this.finalizing.put(this.process, wait);

    } else {

      /* The internal wait node is detected. */
      processNodeBegin(wait);

      this.waitNode = wait;

      this.map.put(wait, new LinkedHashSet<CfgNode>());
    }
  }

  private boolean isFinalizing(final Wait wait) {
    return !wait.containsCondition()
        && wait.getEventList().equals(this.process.getSensitivityList())
        && wait.hasChildren()
        && wait.getChildren().size() == 1
        && wait.getOnlyChild() instanceof Sink;
  }

  @Override
  public void onWaitEnd(final Wait wait) {
    this.waitNode = null;
  }

  private void processNodeBegin(final CfgNode node) {
    if (this.waitNode == null) {
      this.waitUnReachable.add(node);
    } else if (!(node instanceof Wait)) {

      /* Put to wait-related nodes all children except wait children. */
      this.map.get(this.waitNode).add(node);
    }
    this.nearestWaitMap.put(node, this.waitNode);

    if (node.hasOnlyChild()) {
      checkOneChildNode(node);
    }
  }

  private void checkOneChildNode(final CfgNode node) {
    final CfgNode nodeChild = (CfgNode) node.getOnlyChild();
    if (nodeChild instanceof Wait
        && !isFinalizing((Wait) nodeChild)
        && !this.parentNewVarMap.containsKey(node)) {

      /* Associate a new variable with this node. */

      /* Create new variable that will be the trigger
       * for process extracted from wait-children sub-graph. */
      final int varNum = this.module.getVariables().size();
      final NodeVariable variable =
          new NodeVariable(VariableContainer.NEW_VAR_PREFIX + varNum, DataType.BOOLEAN);

      variable.setUserData(new VariableData(VariableType.REG));
      this.module.declareVariable(variable);

      this.parentNewVarMap.put(node, variable);
      this.waitNewVarMap.put((Wait) nodeChild, variable);
    }
  }

  private void processNodeEnd(final CfgNode node) {
    this.waitNode = this.nearestWaitMap.get(node);
  }
}
