/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License") {} you may not use this file
 * except in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.verilog;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprTreeVisitorDefault;
import ru.ispras.fortress.expression.ExprTreeWalker;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.MetaInfoValue;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.CfgNodeType;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.cfg.Wait;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;
import ru.ispras.verilog.parser.model.Activity;
import ru.ispras.verilog.parser.model.Activity.Type;
import ru.ispras.verilog.parser.model.Assign;
import ru.ispras.verilog.parser.model.AssignStatement;
import ru.ispras.verilog.parser.model.Assignment;
import ru.ispras.verilog.parser.model.Attribute;
import ru.ispras.verilog.parser.model.BlockGenerate;
import ru.ispras.verilog.parser.model.BlockStatement;
import ru.ispras.verilog.parser.model.CaseGenerate;
import ru.ispras.verilog.parser.model.CaseGenerateItem;
import ru.ispras.verilog.parser.model.CaseStatement;
import ru.ispras.verilog.parser.model.CaseStatementItem;
import ru.ispras.verilog.parser.model.Code;
import ru.ispras.verilog.parser.model.Declaration;
import ru.ispras.verilog.parser.model.DelayedStatement;
import ru.ispras.verilog.parser.model.DisableStatement;
import ru.ispras.verilog.parser.model.Generate;
import ru.ispras.verilog.parser.model.IfGenerate;
import ru.ispras.verilog.parser.model.IfGenerateBranch;
import ru.ispras.verilog.parser.model.IfStatement;
import ru.ispras.verilog.parser.model.IfStatementBranch;
import ru.ispras.verilog.parser.model.Instantiation;
import ru.ispras.verilog.parser.model.LoopGenerate;
import ru.ispras.verilog.parser.model.LoopStatement;
import ru.ispras.verilog.parser.model.Module;
import ru.ispras.verilog.parser.model.NullStatement;
import ru.ispras.verilog.parser.model.PathDeclaration;
import ru.ispras.verilog.parser.model.Port;
import ru.ispras.verilog.parser.model.PortConnection;
import ru.ispras.verilog.parser.model.Procedure;
import ru.ispras.verilog.parser.model.PulseStyle;
import ru.ispras.verilog.parser.model.ShowCancelled;
import ru.ispras.verilog.parser.model.Specify;
import ru.ispras.verilog.parser.model.Table;
import ru.ispras.verilog.parser.model.TableEntry;
import ru.ispras.verilog.parser.model.TaskStatement;
import ru.ispras.verilog.parser.model.TriggerStatement;
import ru.ispras.verilog.parser.model.WaitStatement;
import ru.ispras.verilog.parser.model.basis.Event;
import ru.ispras.verilog.parser.model.basis.EventControl;
import ru.ispras.verilog.parser.model.basis.Expression;
import ru.ispras.verilog.parser.model.basis.Path;
import ru.ispras.verilog.parser.model.basis.Reference;
import ru.ispras.verilog.parser.walker.VerilogNodeVisitor;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.Stack;
import java.util.logging.Logger;

/**
 * This class creates CfgModel from Verilog AST provided by Verilog Translator.
 *
 * @author <a href="mailto:chupilko@ispras.ru">Mikhail Chupilko</a>
 */
public final class VerilogCfgBuilder extends VerilogNodeVisitor {

  private static final String IS_UNKNOWN = " is UNKNOWN.";
  private static final String ZERO = "0";

  /*
   * Resulting CfgModel is kept here.
   */
  private CfgModel cfgModel;

  /*
   * Current CFG process to add statements.
   */
  private Process process;

  /*
   * CFG process for initializations.
   */
  private Process initialProcess;

  /*
   * List of switch identifiers.
   */
  private Stack<SwitchId> switches;

  /*
   * List of merge identifiers.
   */
  private Stack<String> merges;

  /*
   * List of loops.
   */
  private Stack<LoopId> loops;

  /*
   * List of condition identifiers.
   */
  private Stack<String> conditions;

  /*
   * Storage of CFG modules.
   */
  private List<ru.ispras.retrascope.model.cfg.Module> modules;

  /*
   * Shows whether it is always process or initial.
   */
  private Type activityType;

  /*
   * This class allows to change variable declaration type from Verilog Translator to Retrascope.
   *
   * @author <a href="mailto:chupilko@ispras.ru">Mikhail Chupilko</a>
   */
  private static class ExprVisitor extends ExprTreeVisitorDefault {

    @Override
    public void onVariable(final NodeVariable variable) {
      final Reference reference = (Reference) variable.getUserData();

      VariableType type;
      if (reference == null || reference.getDeclaration() == null) { // TODO: to be checked/removed
        type = VariableType.REG;
      } else {
        type = getVariableType(reference.getDeclaration().getType());
      }

      final VariableData variableData = new VariableData(type);

      final String varTypeHdlString =
          reference == null || reference.getDeclaration() == null ? "reg"
          : reference.getDeclaration().getElementType().getType().toString();

      variableData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, varTypeHdlString);
      variable.setUserData(variableData);
    }
  }

  /**
   * Class to represent identifiers for loops.
   *
   * @author Mikhail Chupilko
   */
  private static class LoopId {

    /**
     * Id of case block with "false" alternative.
     */
    public String falseCaseId;

    /**
     * Id of merge block being the first in loop construction.
     */
    public String firstMergeId;

    /**
     * NodeVariable to contain auxiliary variable for repeat loop.
     */
    public NodeVariable node;

    /**
     * Constructor of storage for loop id.
     *
     * @param falseCaseId id of case block with "false" alternative.
     * @param firstMergeId id of merge block being the first in loop construction.
     */
    public LoopId(final String falseCaseId, final String firstMergeId) {
      this.falseCaseId = falseCaseId;
      this.firstMergeId = firstMergeId;
      this.node = null;
    }

    /**
     * Additional constructor supporting NodeVariable.
     *
     * @param falseCaseId  if of case block with "false" alternative.
     * @param firstMergeId id of merge block being the first in loop construction.
     * @param node NodeVariable with auxiliary variable for repeat loop.
     */
    public LoopId(final String falseCaseId, final String firstMergeId, final NodeVariable node) {
      this.falseCaseId = falseCaseId;
      this.firstMergeId = firstMergeId;
      this.node = node;
    }
  }

  /**
   * Class to represent identifiers for switches.
   *
   * @author Mikhail Chupilko
   */
  private static class SwitchId {

    /**
     * String identifier of the object.
     */
    public String id;

    /**
     * Shows whether the object has true branch.
     */
    public boolean hasTrue;

    /**
     * Shows whether the object has false branch.
     */
    public boolean hasFalse;

    /**
     * Shows whether the case is typical, not casex or casez.
     */
    public boolean isNormalCase;

    /**
     * Shows the number of bits in the resulting vector.
     */
    public DataType bitSize;

    /**
     * Constructor sets id to the given one, and all the other parameters makes false.
     *
     * @param id - id to be saved.
     */
    public SwitchId(final String id, final boolean normalCase, final DataType bitSize) {
      this.id = id;
      hasTrue = false;
      hasFalse = false;
      this.isNormalCase = normalCase;
      this.bitSize = bitSize;
    }
  }

  /**
   * Returns Retrascope variable type for the given Verilog parser data type.
   *
   * @param type Verilog parser data type.
   * @return Retrascope variable type.
   */
  private static VariableType getVariableType(final Declaration.Type type) {
    switch (type) {
      case INPUT:
        return VariableType.IN;
      case OUTPUT:
        return VariableType.OUT;
      case INOUT:
        return VariableType.INOUT;
      case EVENT:
        return VariableType.REG;
      case PARAMETER:
        return VariableType.REG; //exCONST
      case SPECPARAM:
        return VariableType.REG; //exCONST
      case LOCALPARAM:
        return VariableType.REG; //exCONST
      case DEFPARAM:
        return VariableType.REG; //exCONST
      case GENVAR:
        return VariableType.REG;
      case VARIABLE:
        return VariableType.REG;
      default:
        return VariableType.REG;
    }
  }

  /**
   * Returns constructed CfgModel.
   *
   * @return CfgModel.
   */
  public CfgModel getCfgModel() {
    return cfgModel;
  }


  /**
   * Returns the CFG model reference kept in the storage according to a number of current CFG model.
   *
   * @return Current CFG model.
   */
  private ru.ispras.retrascope.model.cfg.Module getCurrentModule() {
    return modules.get(modules.size() - 1);
  }

  /**
   * Adds new switch and merge blocks into current process.
   *
   * <p>The following instructions will be added inside it.</p>
   *
   * @param node node-condition for switch block.
   * @param isNormalCase false value means casex or casez.
   * @param isLoop shows whether it is a condition of some loop or not
   */
  private void onSwitchStartDeclaration(
      final Node node,
      final boolean isNormalCase,
      final boolean isLoop) {

    final Switch switchBlock = new Switch(node/*transformNode(expression.getNode())*/);
    if (isLoop) {
      switchBlock.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_COND);
    }
    final Process processToWork = activityType == Type.INITIAL ? initialProcess : process;

    List<CfgNode> tempCfg = processToWork.getNodes();

    if (tempCfg == null) {
      tempCfg = new ArrayList<>();
    }

    CfgNode stopNode = CfgUtils.getLastNode(tempCfg, CfgNodeType.SINK);

    if (stopNode == null) {
      throw new IllegalArgumentException(
          "Collection MUST contain " + CfgNodeType.SINK.name() + " node.");
    }

    CfgNode startNode = (CfgNode)stopNode.getParents().iterator().next();

    if (!conditions.isEmpty()) {
      if (!conditions.lastElement().equals(ZERO)) {
        startNode = CfgUtils.getFirstNode(tempCfg, conditions.lastElement());
      }

      String currentId = merges.lastElement();
      for (int i = merges.size() - 1; i > 0; i--) {
        if (!merges.get(i).equals(ZERO)) {
          currentId = merges.get(i);
          break;
        }
      }

      if (!ZERO.equals(currentId)) {
        stopNode = CfgUtils.getFirstNode(tempCfg, currentId);

        if (stopNode == null) {
          throw new IllegalStateException();
        }
      }

      if (startNode == null) {
        throw new IllegalStateException();
      }

      while (startNode.getChildren().iterator().next() != stopNode) {
        startNode = (CfgNode) startNode.getChildren().iterator().next();
      }
    }

    if (startNode.hasChildren()) {
      startNode.removeChild(stopNode);
    }

    startNode.addChild(switchBlock);
    final Merge mergeBlock = new Merge();
    stopNode.addParent(mergeBlock);

    tempCfg.add(switchBlock);
    tempCfg.add(mergeBlock);

    final SwitchId switchId = new SwitchId(switchBlock.getId(), isNormalCase, node.getDataType());
    switches.add(switchId);
    merges.add(mergeBlock.getId());

    final List<CfgNode> cfgToRemove = processToWork.getNodes();

    if (cfgToRemove != null) {
      processToWork.removeChild(CfgUtils.getSource(cfgToRemove));
    }
    processToWork.addChild(CfgUtils.getSource(tempCfg));
  }

  /**
   * Adds new switch and merge blocks into current process.
   *
   * <p>The following instructions will be added inside it.</p>
   *
   * @param expression expression containing condition for switch block.
   * @param isNormalCase false value means casex or casez.
   * @param isLoop shows whether it is a condition of some loop or not
   */
  public void onSwitchStartDeclaration(
      final Expression expression,
      final boolean isNormalCase,
      final boolean isLoop) {
    onSwitchStartDeclaration(transformNode(expression.getNode()), isNormalCase, isLoop);
  }

  /**
   * Finishes current switch and merge blocks.
   */
  public void onSwitchEndDeclaration() {
    switches.remove(switches.size() - 1);
    merges.remove(merges.size() - 1);
  }

  /**
   * Adds condition-branch to the deepest switch-block.
   */
  public void onConditionStartDeclaration(final Case caseBlock) {
    final Process processToWork = activityType == Type.INITIAL ? initialProcess : process;
    final List<CfgNode> tempCfg = processToWork.getNodes();
    assert tempCfg != null : "Process to be worked upon is empty";

    CfgNode startNode;
    CfgNode stopNode;

    if (switches.lastElement().id.equals(ZERO)) {
      startNode = CfgUtils.getLastNode(tempCfg, CfgNodeType.SWITCH);
    } else {
      startNode = CfgUtils.getFirstNode(tempCfg, switches.lastElement().id);
    }

    if (merges.lastElement().equals(ZERO)) {
      stopNode = CfgUtils.getLastNode(tempCfg, CfgNodeType.MERGE);
    } else {
      stopNode = CfgUtils.getFirstNode(tempCfg, merges.lastElement());
    }

    if (startNode == null) {
      throw new IllegalStateException();
    }
    if (stopNode == null) {
      throw new IllegalStateException();
    }

    startNode.addChild(caseBlock);
    stopNode.addParent(caseBlock);

    tempCfg.add(caseBlock);
    conditions.add(caseBlock.getId());

    final List<CfgNode> cfgToRemove = processToWork.getNodes();
    processToWork.removeChild(CfgUtils.getSource(cfgToRemove));
    processToWork.addChild(CfgUtils.getSource(tempCfg));
  }

  /**
   * Finishes condition branch.
   */
  public void onConditionEndDeclaration() {
    conditions.remove(conditions.size() - 1);
  }

  /**
   * Changes userData type kept in the node from Reference to VariableType.
   *
   * @param node - node to be updated
   * @return updated node
   */
  public Node transformNode(final Node node) {
    final Node nodeClone = node.deepCopy();
    final ExprTreeWalker walker = new ExprTreeWalker(new ExprVisitor());
    walker.visitNode(nodeClone);
    return nodeClone;
  }

  /**
   * Returns the logger of the engine.
   *
   * @return the logger of the engine.
   */
  private Logger getLogger() {
    return Log.getLogger(this.getClass());
  }

  /**
   * Processes assignments of different types.
   *
   * @param assignment The assignment expression itself.
   * @param type blocking, non-blocking, assign (deassign, force, release will not be supported).
   * @param insideProcess add the assignment to the current process or create a new one.
   */
  public BasicBlock parseAssignment(
      final Assignment assignment,
      final AssignStatement.Type type,
      final boolean insideProcess) {

    final EventList eventList = new EventList();
    if (!insideProcess) {
      final Set<NodeVariable> useVars = assignment.getUsedVariables();

      for (final NodeVariable node : useVars) {
        final RangedVariable rangedVariable = new RangedVariable((NodeVariable)transformNode(node));
        final ru.ispras.retrascope.model.basis.Event retrascopeEvent =
            new ru.ispras.retrascope.model.basis.Event(rangedVariable);
        if (getCurrentModule().getVariablesMapping().get(rangedVariable.getVariable()) == null) {
          getLogger().log(
              LogLevel.WARNING,
              String.format(
                  "Variable %s is not a variable in module %s!",
                  rangedVariable.getVariable().getName(),
                  getCurrentModule().getName()));
        }
        eventList.add(retrascopeEvent);
      }
    }

    final Node expression = assignment.getExpression().isUsual()
        ? transformNode(assignment.getExpression().getMinExpression().getNode()) : null;
    // NOTICE: for unusual expressions which are not supported
    // System.out.println(assignment.getCondition().getMinExpression());
    // System.out.println(":");
    // System.out.println(assignment.getCondition().getTypExpression());
    // System.out.println(":");
    // System.out.println(assignment.getCondition().getMaxExpression());

    if (expression == null) {
      throw new IllegalStateException();
    }

    int rightSize = expression.getDataType().getSize();
    final BasicBlock basicBlock = new BasicBlock(type == AssignStatement.Type.NON_BLOCKING);

    for (final Reference ref : assignment.getReference().getReferences()) {
      final ru.ispras.retrascope.model.basis.Assignment retraAssignment =
          new ru.ispras.retrascope.model.basis.Assignment();
      final RangedVariable rangedVariable =
          new RangedVariable((NodeVariable)transformNode(ref.getVariable()));
      DataType dataType = rangedVariable.getVariable().getDataType();
      if (getCurrentModule().containsVariable(rangedVariable.getVariable())) {
        dataType =
            getCurrentModule().getVariable(rangedVariable.getVariable().getName()).getDataType();
      }

      int leftSize = dataType.getSize();

      if (ref.hasBitsSelection()) {
        final Expression leftExpression = ref.getBitsSelection().getLeftExpression();
        Expression rightExpression = ref.getBitsSelection().getRightExpression();
        if (rightExpression == null) {
          rightExpression = leftExpression;
        }
        leftSize =
            Math.abs(leftExpression.evaluateInteger() - rightExpression.evaluateInteger()) + 1;
        final Range range = new Range(leftExpression.getNode(), rightExpression.getNode());
        rangedVariable.setRange(range);
      }

      Node localExpression = expression;
      if (assignment.getReference().size() > 1) {
        final NodeValue lowValue = new NodeValue(Data.newInteger(rightSize - leftSize));
        final NodeValue highValue = new NodeValue(Data.newInteger(rightSize - 1));
        rightSize -= leftSize;
        localExpression =
            new NodeOperation(StandardOperation.BVEXTRACT, lowValue, highValue, expression);
      }

      retraAssignment.setVariable(rangedVariable);
      retraAssignment.setValue(localExpression);

      basicBlock.addAssignment(retraAssignment);
    }
    addBlock(basicBlock, eventList, insideProcess);
    return basicBlock;
  }

  /**
   * Simplified adding of CfgNodes into Cfg.
   *
   * <p>Required are only two parameters: the CFGNode itself and flag forcing
   * to add the Node into initial or normal process.
   * The other parameters are set to the following default values:
   * empty eventList and adding into existing process (if it exists).</p>
   *
   * @param block CfgNode.
   */
  public void addBlock(final CfgNode block) {
    final EventList eventList = new EventList();
    addBlock(block, eventList, true);
  }

  /**
   * Add CfgNode (some block, e.g. assignment) to CFG.
   *
   * @param block CfgNode.
   * @param eventList should contain all the events touching the process.
   *                  Used in case of continuous assignments.
   * @param insideProcess if {@code true}, the block is added to existing process,
   *                      if {@code false}, new process is created for it.
   */
  public void addBlock(
      final CfgNode block,
      final EventList eventList,
      final boolean insideProcess) {

    final List<CfgNode> cfg = new ArrayList<>();
    final Source src = new Source();
    final Sink sink = new Sink();
    cfg.add(src);
    cfg.add(sink);
    src.addChild(sink);
    cfg.add(block);

    CfgNode startNode = src;
    CfgNode stopNode = sink;

    if (insideProcess) {

      final Process localProcess = activityType == Type.INITIAL ? initialProcess : process;
      final List<CfgNode> tempCfg = localProcess.getNodes();

      if (tempCfg == null) {
        startNode.removeChild(stopNode);
        block.addParent(startNode);
        block.addChild(stopNode);
        localProcess.addChild(CfgUtils.getSource(cfg));
      } else {

        stopNode = CfgUtils.getLastNode(tempCfg, CfgNodeType.SINK);

        if (stopNode == null) {
          throw new IllegalStateException();
        }

        if (stopNode.hasParents()) {
          startNode = (CfgNode) stopNode.getParents().iterator().next();
        } else {
          startNode = CfgUtils.getLastNode(tempCfg, CfgNodeType.SOURCE);

          if (startNode == null) {
            throw new IllegalStateException();
          }

          while (startNode.hasChildren()
              && startNode.getChildren().iterator().next() != stopNode) {

            for (final CfgModelNode cfgModelNode : startNode.getChildren()) {
              startNode = (CfgNode) cfgModelNode;
              //(CfgNode)startNode.getChildren().iterator().next();
            }
          }
        }

        if (!conditions.isEmpty()) {
          if (!conditions.lastElement().equals(ZERO)) {
            startNode = CfgUtils.getFirstNode(tempCfg, conditions.lastElement());
          }

          if (!merges.lastElement().equals(ZERO)) {
            stopNode = CfgUtils.getFirstNode(tempCfg, merges.lastElement());
          }

          if (startNode == null) {
            throw new IllegalStateException();
          }

          final Random rand = new Random();
          while (startNode.hasChildren() && startNode.getChildren().iterator().next() != stopNode) {
            final Iterator<CfgModelNode> iterator = startNode.getChildren().iterator();
            do {
              startNode = (CfgNode)iterator.next();
            } while (iterator.hasNext() && rand.nextBoolean()
              /*visitedNodes.contains((String)iterator.next().getId())*/); //TODO!!!!!!
          }
          //while (startNode.getChildren().iterator().next() != stopNode)
          //startNode = (CfgNode)startNode.getChildren().iterator().next();
        }

        if (stopNode == null) {
          throw new IllegalStateException();
        }

        if (stopNode.getType() == CfgNodeType.SINK && block.getType() == CfgNodeType.MERGE) {
          for (final CfgModelNode node : stopNode.getParents()) {
            node.removeChild(stopNode);
            node.addChild(block);
            block.addChild(stopNode);
          }
        } else {
          if (startNode.hasChildren()) {
            startNode.removeChild(stopNode);
          }
          startNode.addChild(block);
          stopNode.addParent(block);
        }

        tempCfg.add(block);

        final List<CfgNode> cfgToRemove = localProcess.getNodes();

        localProcess.removeChild(CfgUtils.getSource(cfgToRemove));
        localProcess.addChild(CfgUtils.getSource(tempCfg));
      }
    } else {
      startNode.removeChild(stopNode);
      block.addParent(startNode);
      block.addChild(stopNode);

      final Process process = new Process();
      process.addChild(CfgUtils.getSource(cfg));
      process.addNodes(cfg);
      process.setSensitivityList(eventList);
      getCurrentModule().addChild(process);
    }
  }

  //------------------------------------------------------------------------------------------------
  // Activity
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onActivityBegin(final Activity node) {
    activityType = node.getType();
    if (activityType == Type.INITIAL) { // or INITIAL or ALWAYS
      if (initialProcess == null) {
        initialProcess = new Process(true);
        createEmptyCfg(initialProcess);
      }
    } else {
      process = new Process();
      createEmptyCfg(process);
    }

    return Result.OK;
  }

  private void createEmptyCfg(final Process process) {
    final List<CfgNode> nodes = new ArrayList<>();
    final Source src = new Source();
    final Sink sink = new Sink();
    src.addChild(sink);
    nodes.add(src);
    nodes.add(sink);
    process.addNodes(nodes);
    process.addChild(src);
  }

  @Override
  public Result onActivityEnd(final Activity node) {
    if (activityType == Type.INITIAL) {
      getCurrentModule().addChild(initialProcess);
      activityType = Type.ALWAYS;
      initialProcess = null;
    } else {
      getCurrentModule().addChild(process);
    }

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Assign
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onAssignBegin(final Assign node) {
    // node.getDelay() is not used
    parseAssignment(node.getAssignment(), AssignStatement.Type.ASSIGN, false);

    return Result.OK;
  }

  @Override
  public Result onAssignEnd(final Assign node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // AssignStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onAssignStatementBegin(final AssignStatement node) {
    switch (node.getType()) {
      case BLOCKING:
      case NON_BLOCKING:
      case ASSIGN: {
        parseAssignment(node.getAssignment(), node.getType(), true);
        break;
      }
      default: {
        getLogger().log(LogLevel.WARNING, node.getType() + " is unsupported.");
        break;
      }
    }

    return Result.OK;
  }

  @Override
  public Result onAssignStatementEnd(final AssignStatement node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Assignment
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onAssignmentBegin(final Assignment node) {
    parseAssignment(node, AssignStatement.Type.ASSIGN, false);

    return Result.OK;
  }

  @Override
  public Result onAssignmentEnd(final Assignment node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Attribute
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onAttributeBegin(final Attribute node) {
    // Nothing
    return Result.OK;
  }

  @Override
  public Result onAttributeEnd(final Attribute node) {
    // Nothing
    return Result.OK;
  }

  // -------------------------------------------------------------------------
  // BlockGenerate
  // -------------------------------------------------------------------------

  @Override
  public Result onBlockGenerateBegin(final BlockGenerate node) {
    return Result.OK;
  }

  @Override
  public Result onBlockGenerateEnd(final BlockGenerate node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // BlockStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onBlockStatementBegin(final BlockStatement node) {

    if (node.getType() == BlockStatement.Type.FORK) {
      getLogger().log(LogLevel.WARNING, "fork is not supported");
    }

    return Result.OK;
  }

  @Override
  public Result onBlockStatementEnd(final BlockStatement node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CaseGenerate
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onCaseGenerateBegin(final CaseGenerate node) {
    return Result.OK;
  }

  @Override
  public Result onCaseGenerateEnd(final CaseGenerate node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CaseGenerateItem
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onCaseGenerateItemBegin(final CaseGenerateItem node) {
    return Result.OK;
  }

  @Override
  public Result onCaseGenerateItemEnd(final CaseGenerateItem node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CaseStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onCaseStatementBegin(final CaseStatement node) {
    switch (node.getType()) {
      case CASE:
        if (node.getExpression().getNode().getDataType() == DataType.UNKNOWN) {
          Log.getLogger().log(
              LogLevel.INFO,
              "[onCaseStatementBegin] DataType of " + node.getExpression().getNode() + IS_UNKNOWN);
        }
        onSwitchStartDeclaration(node.getExpression(), true, false);
        break;
      case CASEX:
      case CASEZ:
        final SwitchId switchId =
            new SwitchId("start", false, node.getExpression().getNode().getDataType());
        switches.add(switchId);
        merges.add(ZERO);
        break;
      default:
        Log.print(LogLevel.ERROR, node.getType() + " is unsupported.");
        break;
    }

    return Result.OK;
  }

  @Override
  public Result onCaseStatementEnd(final CaseStatement node) {
    while (!switches.isEmpty()) {
      if (switches.lastElement().isNormalCase || switches.lastElement().id.equals("start")) {
        break;
      } else {
        onSwitchEndDeclaration();
        onConditionEndDeclaration();
      }
    }

    if (!switches.isEmpty()) {
      onSwitchEndDeclaration();
    }

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // CaseStatementItem
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onCaseStatementItemBegin(final CaseStatementItem node) {
    Case caseBlock = new Case(false);
    for (final Expression expression : node.getExpressions()) {
      caseBlock.addValue(expression.getValue());
    }

    if (!switches.lastElement().isNormalCase) {
      if (node.getExpression().getNode().getDataType() == DataType.UNKNOWN) {
        Log.getLogger().log(
            LogLevel.ERROR,
            "[onCaseStatementItemBegin] DataType of "
                + node.getExpression().getNode() + IS_UNKNOWN);
      }
      onSwitchStartDeclaration(node.getExpression(), false, false);
      final NodeValue nodeValue = new NodeValue(Data.newBoolean(true));
      caseBlock = new Case(nodeValue);
    }

    onConditionStartDeclaration(caseBlock);

    return Result.OK;
  }

  @Override
  public Result onCaseStatementItemEnd(final CaseStatementItem node) {
    onConditionEndDeclaration();
    if (!switches.lastElement().isNormalCase) {
      final NodeValue nodeValue = new NodeValue(Data.newBoolean(false));
      final Case caseBlock = new Case(nodeValue);
      onConditionStartDeclaration(caseBlock);
    }

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Code
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onCodeBegin(final Code node) {
    // Nothing
    return Result.OK;
  }

  @Override
  public Result onCodeEnd(final Code node) {
    // Creates new empty CFG model and adds it into storage of CFG models.
    cfgModel = new CfgModel();
    if (modules == null) {
      modules = new ArrayList<>();
      final ru.ispras.retrascope.model.cfg.Module module =
          new ru.ispras.retrascope.model.cfg.Module("");
      module.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VERILOG);
      modules.add(module);
    }
    cfgModel.addModules(modules);

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Declaration
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onDeclarationBegin(final Declaration node) {
    final Set<Declaration> namesakes = node.getNamesakes();

    if (namesakes.isEmpty() || node.isOutput() || node.isInout()) {
      final DataType type = node.getElementType().getDataType();
      final NodeVariable variable = new NodeVariable(node.getName(), type);

      final VariableData variableData = new VariableData(getVariableType(node.getType()));
      variableData.addMetaInfo(MetaInfoType.HDL_TYPE_STR.name(),
          node.getElementType().getType().toString());
      variable.setUserData(variableData);
      getCurrentModule().declareVariable(variable);

      getLogger().log(LogLevel.DEBUG,
          String.format("add variable %s [%s] (%s) to module %s.",
              variable, type, variable.getData(), getCurrentModule().getName()));
    }

    // final ElementType type = node.getElementType();
    // parameter type.isSpecified() is not used

    if (node.getExpression() != null) {
      if (initialProcess == null) {
        initialProcess = new Process(true);
        createEmptyCfg(initialProcess);
      }

      final Assignment assignment = new Assignment(null);
      final Reference reference = new Reference();
      final Path path = new Path(node.getName());
      reference.setDeclaration(node);
      reference.setPath(path);
      assignment.addReference(reference);
      assignment.setExpression(node.getExpression());
      activityType = Type.INITIAL;
      parseAssignment(assignment, AssignStatement.Type.ASSIGN, true);
    }

    return Result.OK;
  }

  @Override
  public Result onDeclarationEnd(final Declaration node) {
    activityType = Type.ALWAYS;
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // DelayedStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onDelayedStatementBegin(final DelayedStatement node) {
    final EventControl eventControl = node.getEventControl();

    if (eventControl.isEvents()) { //|| eventControl.isRepeat() eventControl.isDelays() not parsed
      for (final Event event : eventControl.getEvents()) {
        EventType eventType = EventType.ANY_EDGE;
        if (event.getEdge().isPositive()) {
          eventType = EventType.POSITIVE_EDGE;
        } else if (event.getEdge().isNegative()) {
          eventType = EventType.NEGATIVE_EDGE;
        } else if (event.getEdge().isUndefined()) {
          eventType = EventType.ANY_EDGE;
        } else {
          assert false;
        }

        final EventList eventList = new EventList();
        for (final NodeVariable nodeVariable : event.getExpression().getVariables()) {
          final RangedVariable rangedVariable =
              new RangedVariable((NodeVariable)transformNode(nodeVariable));

          final ru.ispras.retrascope.model.basis.Event retrascopeEvent =
              new ru.ispras.retrascope.model.basis.Event(rangedVariable, eventType);

          if (!getCurrentModule().containsVariable(nodeVariable)) {
            getLogger().log(
                LogLevel.WARNING,
                String.format(
                    "Variable %s is not a variable in module %s!",
                    nodeVariable.getName(),
                    getCurrentModule().getName()));
          }

          final Process localProcess = activityType == Type.INITIAL ? initialProcess : process;
          if (localProcess.getSensitivityList() == null) {
            final EventList emptyEventList = new EventList();
            localProcess.setSensitivityList(emptyEventList);
          }
          if (activityType == Type.INITIAL) {
            eventList.add(retrascopeEvent);
          } else {
            localProcess.getSensitivityList().add(retrascopeEvent);
          }
        }

        if (activityType == Type.INITIAL) {
          final Wait waitBlock = new Wait(eventList
              /*, transformNode(event.getExpression().getNode())*/);
          addBlock(waitBlock);
        }
      }
    }

    return Result.OK;
  }

  @Override
  public Result onDelayedStatementEnd(final DelayedStatement node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // DisableStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onDisableStatementBegin(final DisableStatement node) {
    return Result.OK;
  }

  @Override
  public Result onDisableStatementEnd(final DisableStatement node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Generate
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onGenerateBegin(final Generate node) {
    return Result.OK;
  }

  @Override
  public Result onGenerateEnd(final Generate node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // IfGenerate
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onIfGenerateBegin(final IfGenerate node) {
    return Result.OK;
  }

  @Override
  public Result onIfGenerateEnd(final IfGenerate node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // IfGenerateBranch
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onIfGenerateBranchBegin(final IfGenerateBranch node) {
    return Result.OK;
  }

  @Override
  public Result onIfGenerateBranchEnd(final IfGenerateBranch node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // IfStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onIfStatementBegin(final IfStatement node) {
    if (node.getExpression().getElaboratedNode().getDataType() == DataType.UNKNOWN) {
      Log.getLogger().log(
          LogLevel.ERROR,
          "[onIfStatementBegin] DataType of "
          + node.getExpression().getElaboratedNode() + IS_UNKNOWN);
    }
    onSwitchStartDeclaration(node.getExpression(), true, false);
    return Result.OK;
  }

  @Override
  public Result onIfStatementEnd(final IfStatement node) {
    // If one of branches is absent, add an artificial one.
    if (switches.lastElement().hasTrue && !switches.lastElement().hasFalse) {
      final NodeValue nodeValue = getCaseValue(0, false);
      final Case caseBlock = new Case(nodeValue);
      onConditionStartDeclaration(caseBlock);
      onConditionEndDeclaration();
    } else {
      final NodeValue nodeValue = getCaseValue(1, true);
      final Case caseBlock = new Case(nodeValue);
      onConditionStartDeclaration(caseBlock);
      onConditionEndDeclaration();
    }

    onSwitchEndDeclaration();

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // IfStatementBranch
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onIfStatementBranchBegin(final IfStatementBranch node) {
    if (node.isThen()) {
      final NodeValue nodeValue = getCaseValue(1, true);
      final Case caseBlock = new Case(nodeValue);
      onConditionStartDeclaration(caseBlock);
      switches.lastElement().hasTrue = true;
    } else if (node.isElse()) {
      final NodeValue nodeValue = getCaseValue(0, false);
      final Case caseBlock = new Case(nodeValue);
      onConditionStartDeclaration(caseBlock);
      switches.lastElement().hasFalse = true;
    }

    return Result.OK;
  }

  private NodeValue getCaseValue(int bvValue, boolean boolValue) {
    return new NodeValue(switches.lastElement().bitSize.getTypeId() == DataTypeId.BIT_VECTOR
        ? Data.newBitVector(bvValue, switches.lastElement().bitSize.getSize())
        : Data.newBoolean(boolValue));
  }

  @Override
  public Result onIfStatementBranchEnd(final IfStatementBranch node) {
    onConditionEndDeclaration();
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Instantiation
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onInstantiationBegin(final Instantiation node) {
    getLogger().log(LogLevel.DEBUG,
        String.format(
            "Instantiation of %s type is module %s.", node.getModuleName(), node.getName()));
    final ru.ispras.retrascope.model.cfg.Module module =
        new ru.ispras.retrascope.model.cfg.Module(node.getModuleName());
    module.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VERILOG);
    module.setInstanceName(node.getName());

    //assert (modules.isEmpty()); TODO
    // Add newly created module as a child to that one.
    // TODO: definition of the module should be found first.
    getCurrentModule().addChild(module);

    return Result.OK;
  }

  @Override
  public Result onInstantiationEnd(final Instantiation node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // LoopGenerate
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onLoopGenerateBegin(final LoopGenerate node) {
    return Result.OK;
  }

  @Override
  public Result onLoopGenerateEnd(final LoopGenerate node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // LoopStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onLoopStatementBegin(final LoopStatement node) {
    final NodeValue nodeValueFalse = new NodeValue(Data.newBoolean(false));
    final NodeValue nodeValueTrue = new NodeValue(Data.newBoolean(true));
    final Case condBlockTrue = new Case(nodeValueTrue);
    final Case condBlockFalse = new Case(nodeValueFalse);

    final Merge mergeBlock = new Merge();
    mergeBlock.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_ENTRY);

    switch (node.getType()) {
      case FOREVER: {
        addBlock(mergeBlock);

        onSwitchStartDeclaration(nodeValueTrue, true, true);
        onConditionStartDeclaration(condBlockFalse);
        onConditionEndDeclaration();
        onConditionStartDeclaration(condBlockTrue);

        loops.add(new LoopId(condBlockFalse.getId(), mergeBlock.getId()));
        break;
      }
      case REPEAT: {
        final DataType dataType = node.getExpression().getElaboratedNode().getDataType();
        final NodeVariable nodeVariable = getCurrentModule().declareNewVariable(dataType);
        final ru.ispras.retrascope.model.basis.Assignment assignment =
            new ru.ispras.retrascope.model.basis.Assignment();
        assignment.setVariable(new RangedVariable((NodeVariable)transformNode(nodeVariable)));

        NodeValue nodeValue;
        Enum<?> operationId;
        if (dataType.getTypeId().equals(DataTypeId.LOGIC_INTEGER)) {
          nodeValue = new NodeValue(Data.newInteger(0));
          operationId = StandardOperation.LESS;
        } else if (dataType.getTypeId().equals(DataTypeId.BIT_VECTOR)) {
          nodeValue = new NodeValue(Data.newBitVector(0, dataType.getSize()));
          operationId = StandardOperation.BVULT;
        } else {
          throw new IllegalArgumentException("Cannot parse: " + node.getExpression() + ".");
        }

        assignment.setValue(nodeValue);
        final BasicBlock basicBlock = new BasicBlock(assignment);
        basicBlock.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_INIT);
        addBlock(basicBlock);

        addBlock(mergeBlock);

        final NodeOperation nodeExpr =
            new NodeOperation(operationId, nodeVariable, node.getExpression().getValue());
        onSwitchStartDeclaration(nodeExpr, true, true);
        onConditionStartDeclaration(condBlockFalse);
        onConditionEndDeclaration();
        onConditionStartDeclaration(condBlockTrue);

        loops.add(new LoopId(condBlockFalse.getId(), mergeBlock.getId(), nodeVariable));
        break;
      }
      case WHILE: {
        addBlock(mergeBlock);

        if (node.getExpression().getNode().getDataType() == DataType.UNKNOWN) {
          Log.getLogger().log(
              LogLevel.ERROR,
              "[onLoopStatementBegin] DataType of "
                  + node.getExpression().getNode() + IS_UNKNOWN);
        }

        onSwitchStartDeclaration(node.getExpression(), true, true);
        onConditionStartDeclaration(condBlockFalse);
        onConditionEndDeclaration();
        onConditionStartDeclaration(condBlockTrue);

        loops.add(new LoopId(condBlockFalse.getId(), mergeBlock.getId()));
        break;
      }
      case FOR: { //TODO
        final BasicBlock basicBlock =
            parseAssignment(node.getInitialization(), AssignStatement.Type.BLOCKING, true);
        basicBlock.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_INIT);

        addBlock(mergeBlock);

        if (node.getExpression().getNode().getDataType() == DataType.UNKNOWN) {
          Log.getLogger().log(
              LogLevel.ERROR,
              "[onLoopStatementItemBegin2] DataType of "
                  + node.getExpression().getNode() + IS_UNKNOWN);
        }

        onSwitchStartDeclaration(node.getExpression(), true, true);
        onConditionStartDeclaration(condBlockFalse);
        onConditionEndDeclaration();
        onConditionStartDeclaration(condBlockTrue);

        loops.add(new LoopId(condBlockFalse.getId(), mergeBlock.getId()));
        break;
      }
      default:
        Log.print(LogLevel.ERROR, node.getType() + " is unsupported.");
    }

    return Result.OK;
  }

  @Override
  public Result onLoopStatementEnd(final LoopStatement node) {

    if (node.getType() == LoopStatement.Type.FOR) {
      final BasicBlock basicBlock =
          parseAssignment(node.getIteration(), AssignStatement.Type.BLOCKING, true);
      basicBlock.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_INC);
    }
    if (node.getType() == LoopStatement.Type.REPEAT) {
      final ru.ispras.retrascope.model.basis.Assignment assignment =
          new ru.ispras.retrascope.model.basis.Assignment();
      assignment.setVariable(
          new RangedVariable((NodeVariable) transformNode(loops.get(loops.size() - 1).node)));

      final DataType type = node.getExpression().getElaboratedNode().getDataType();
      NodeValue nodeValue;
      Enum<?> operationId;

      if (type.getTypeId().equals(DataTypeId.LOGIC_INTEGER)) {
        nodeValue = new NodeValue(Data.newInteger(1));
        operationId = StandardOperation.ADD;
      } else if (type.getTypeId().equals(DataTypeId.BIT_VECTOR)) {
        nodeValue = new NodeValue(Data.newBitVector(1, type.getSize()));
        operationId = StandardOperation.BVADD;
      } else {
        throw new IllegalArgumentException("Cannot parse: " + node.getExpression() + ".");
      }

      final NodeOperation nodeExpr =
          new NodeOperation(operationId, loops.get(loops.size() - 1).node, nodeValue);
      assignment.setValue(nodeExpr);
      final BasicBlock basicBlock = new BasicBlock(assignment);
      basicBlock.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_INC);
      addBlock(basicBlock);
    }
    onConditionEndDeclaration();

    // Change topology: last merge in switch-merge construction should point to the first merge.
    final Process localProcess = activityType == Type.INITIAL ? initialProcess : process;
    final CfgNode lastMerge = CfgUtils.getFirstNode(localProcess.getNodes(), merges.lastElement());

    if (lastMerge == null) {
      throw new IllegalStateException();
    }

    lastMerge.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_FIN);
    if (lastMerge.hasChildren()) {
      lastMerge.removeChild(lastMerge.getChildren().iterator().next());
    }
    lastMerge.addChild(
        CfgUtils.getFirstNode(localProcess.getNodes(), loops.lastElement().firstMergeId));

    onSwitchEndDeclaration();

    /* Change topology: false alternative of cases should point
     * to merge after the last merge in switch-merge construction.
     */
    final CfgNode lastNodeFalse =
        CfgUtils.getFirstNode(localProcess.getNodes(), loops.lastElement().falseCaseId);

    if (lastNodeFalse == null) {
      throw new IllegalStateException();
    }

    if (lastNodeFalse.hasChildren()) {
      lastNodeFalse.removeChild(lastNodeFalse.getChildren().iterator().next());
    }

    loops.remove(loops.size() - 1);

    /* If it was a nested loop, the false alternative
     * should point to the entrance of the previous loop. */
    CfgNode cfgNodeToConnect;
    if (loops.size() > 0) {
      cfgNodeToConnect =
          CfgUtils.getFirstNode(localProcess.getNodes(), loops.lastElement().firstMergeId);
    } else {
      cfgNodeToConnect = CfgUtils.getSink(localProcess.getNodes());
    }
    lastNodeFalse.addChild(cfgNodeToConnect);

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Module
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onModuleBegin(final Module node) {
    if (node.isMacromodule()) {
      getLogger().log(LogLevel.ERROR, "MACROMODULE is not supported");
    } else if (node.isPrimitive()) {
      getLogger().log(LogLevel.ERROR, "PRIMITIVE is not supported");
    } else {
      final ru.ispras.retrascope.model.cfg.Module module =
          new ru.ispras.retrascope.model.cfg.Module(node.getName());
      module.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VERILOG);

      getLogger().log(LogLevel.DEBUG,
          String.format("Start observing module %s.", module.getName()));

      if (modules == null) {
        modules = new ArrayList<>();
      }
      if (conditions == null) {
        conditions = new Stack<>();
      }
      if (switches == null) {
        switches = new Stack<>();
      }
      if (merges == null) {
        merges = new Stack<>();
      }
      if (loops == null) {
        loops = new Stack<>();
      }

      modules.add(module);
    }

    return Result.OK;
  }

  @Override
  public Result onModuleEnd(final Module node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // NullStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onNullStatementBegin(final NullStatement node) {
    return Result.OK;
  }

  @Override
  public Result onNullStatementEnd(final NullStatement node) {
    // Nothing

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // PathDeclaration
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onPathDeclarationBegin(final PathDeclaration node) {
    return Result.OK;
  }

  @Override
  public Result onPathDeclarationEnd(final PathDeclaration node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Port
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onPortBegin(final Port node) {
    // Nothing
    return Result.OK;
  }

  @Override
  public Result onPortEnd(final Port node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // PortConnection
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onPortConnectionBegin(final PortConnection node) {
    final String portName = node.getName();
    final Expression portExpr = node.getExpression();

    if (portExpr.getNode().getDataType() == DataType.UNKNOWN) {
      Log.getLogger().log(
          LogLevel.ERROR,
          "[onPortConnectionBegin] DataType of " + portExpr.getNode() + " is UNKNOWN");
    }

    final List<ru.ispras.retrascope.model.cfg.Module> modules = getCurrentModule().getModules();
    if (!modules.isEmpty()) {
      final ru.ispras.retrascope.model.cfg.Module module = modules.get(modules.size() - 1);
      Map<String, Node> bindings = module.getBindings();
      if (bindings == null) {
        bindings = new HashMap<>();
      }

      // A binding is not required to be present.
      if (portName != null) {
        // TODO: find this module declaration and check variables
        bindings.put(portName, portExpr.getNode());
      }

      module.setBindings(bindings);
    }
    return Result.OK;
  }

  @Override
  public Result onPortConnectionEnd(final PortConnection node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Procedure
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onProcedureBegin(final Procedure node) {
    /*final ElementType type = node.getElementType();

    if (type.isSpecified()) { // TODO: implement functions and tasks.
      text("function");
    } else {
      text("task");
    }

    if (node.isAutomatic()) {
      text(" ");
      text("automatic");
    }

    if (type.isSpecified()) {
      text(" ");
      text(type);
    }

    text(" ");
    text(node.getName());
    endl();*/

    return Result.OK;
  }

  @Override
  public Result onProcedureEnd(final Procedure node) {
    /*final ElementType type = node.getElementType();

    if (type.isSpecified()) {
      text("endfunction");
    } else {
      text("endtask");
    }

    endl();*/

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // PulseStyle
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onPulseStyleBegin(final PulseStyle node) {
    return Result.OK;
  }

  @Override
  public Result onPulseStyleEnd(final PulseStyle node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // ShowCancelled
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onShowCancelledBegin(final ShowCancelled node) {
    return Result.OK;
  }

  @Override
  public Result onShowCancelledEnd(final ShowCancelled node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Specify
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onSpecifyBegin(final Specify node) {
    return Result.OK;
  }

  @Override
  public Result onSpecifyEnd(final Specify node) {
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // Table
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onTableBegin(final Table node) {
    /*text("table"); TODO: implement tables.
    endl();*/

    return Result.OK;
  }

  @Override
  public Result onTableEnd(final Table node) {
    /*text("endtable");
    endl();*/

    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // TableEntry
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onTableEntryBegin(final TableEntry node) {
    /*text(" ", node.getInputs1());
    text(" ");
    text(" ", node.getEdge()); TODO: implement tables.
    text(" ");
    text(" ", node.getInputs2());
    text(" : ");
    text(" ", node.getOutputs());
    text(";");
    endl();*/

    return Result.OK;
  }

  @Override
  public Result onTableEntryEnd(final TableEntry node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // TaskStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onTaskStatementBegin(final TaskStatement node) {
    /*text(node.getPath());
    text("(");

    boolean comma = false;
    for (final Expression expression : node.getArguments()) {
      if (comma) {
        text(", ");
      }
      comma = true;

      text(expression); TODO: implement task statements
    }

    text(");");
    endl();*/

    return Result.OK;
  }

  @Override
  public Result onTaskStatementEnd(final TaskStatement node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // TriggerStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onTriggerStatementBegin(final TriggerStatement node) {
    return Result.OK;
  }

  @Override
  public Result onTriggerStatementEnd(final TriggerStatement node) {
    // Nothing
    return Result.OK;
  }

  //------------------------------------------------------------------------------------------------
  // WaitStatement
  //------------------------------------------------------------------------------------------------

  @Override
  public Result onWaitStatementBegin(final WaitStatement node) {
    final NodeVariable nodeVariable = (NodeVariable)transformNode(node.getExpression().getNode());

    NodeValue nodeValue;
    if (nodeVariable.getDataTypeId().equals(DataTypeId.LOGIC_INTEGER)) {
      nodeValue = new NodeValue(Data.newInteger(0));
    } else {
      if (nodeVariable.getDataTypeId().equals(DataTypeId.BIT_VECTOR)) {
        nodeValue = new NodeValue(Data.newBitVector(0, nodeVariable.getDataType().getSize()));
      } else {
        throw new IllegalArgumentException("Cannot parse: " + node.getExpression() + ".");
      }
    }

    final NodeOperation nodeExpr =
        new NodeOperation(StandardOperation.NOTEQ, nodeVariable, nodeValue);

    final Wait waitBlock = new Wait(nodeExpr);
    addBlock(waitBlock);

    return Result.OK;
  }

  @Override
  public Result onWaitStatementEnd(final WaitStatement node) {
    return Result.OK;
  }
}
