/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.verilog;

import org.apache.commons.cli.ParseException;

import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.FileNames;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.parser.basis.HdlParser;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;
import ru.ispras.verilog.parser.VerilogBackend;
import ru.ispras.verilog.parser.VerilogTranslator;
import ru.ispras.verilog.parser.processor.VerilogGenerateProcessor;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;

/**
 * Verilog Parser class. Uses Verilog Translator to create CfgModel.
 * 
 * @author <a href="mailto:chupilko@ispras.ru">Mikhail Chupilko</a>
 */
public class VerilogParser extends HdlParser {

  /** The entity identifier. */
  public static final String ENGINE_ID = "verilog-parser";

  /** The output entity type. */
  public static final EntityType OUT_TYPE = EntityType.get("v");

  /* Include path option. */
  private static final Parameter INCLUDE_PATH =
      new Parameter("include-path", true, "Path to find included files");

  /**
   * Default constructor.
   */
  public VerilogParser() {
    super(VerilogParser.ENGINE_ID);

    addInputType(VerilogParser.OUT_TYPE);
    setOutputType(CfgModel.TYPE);
    
    addParameter(INCLUDE_PATH);
  }

  @Override
  public Entity parse(final Map<EntityType, Entity> inputs) {
    final Arguments arguments = (Arguments) inputs.get(Arguments.TYPE);
    final FileNames fileNames = (FileNames) inputs.get(VerilogParser.OUT_TYPE);

    final List<String> fileNameArray = new ArrayList<>();
    for (final String fileName : fileNames.getList()) {
      fileNameArray.add(fileName);
    }

    // Get paths from command line
    final String[] separatedPaths;
    try {
      String paths;
      final Arguments engineArgs = parseCommandLine(arguments.getCommandLine());
      if (engineArgs.contains(INCLUDE_PATH)) {
        paths = engineArgs.getValue(INCLUDE_PATH);
        separatedPaths = paths.split(File.pathSeparator);
        fileNameArray.add(String.format("--%s", INCLUDE_PATH.getLongName()));
        Collections.addAll(fileNameArray, separatedPaths);
      }
    } catch (ParseException e) {
      Log.print(LogLevel.ERROR, e.getMessage());
    }

    final VerilogTranslator translator = new VerilogTranslator();

    final VerilogCfgBuilder builder = new VerilogCfgBuilder();
    translator.add(new VerilogBackend(new VerilogGenerateProcessor(builder)));

    translator.start(fileNameArray.toArray(new String[fileNameArray.size()]));

    return builder.getCfgModel();
  }
}
