/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Mapping between logic and bit-vector operation identifiers.
 *
 * <p>Mapping for transforming operation codes for logic types
 * into operation codes for bit vector.
 * Mapping is used at VHDL expressions elaboration in
 * {@link VhdlParser} component.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class BitVectorOpCodeMapping {

  private static final Map<StandardOperation, StandardOperation> MAP;

  private BitVectorOpCodeMapping() {
    /* Empty. */
  }

  static {
    MAP = new LinkedHashMap<>();

    MAP.put(StandardOperation.EQ, StandardOperation.EQ);
    MAP.put(StandardOperation.NOTEQ, StandardOperation.NOTEQ);
    MAP.put(StandardOperation.ADD, StandardOperation.BVADD);
    MAP.put(StandardOperation.SUB, StandardOperation.BVSUB);
    MAP.put(StandardOperation.AND, StandardOperation.BVAND);
    MAP.put(StandardOperation.OR, StandardOperation.BVOR);
    MAP.put(StandardOperation.XOR, StandardOperation.BVXOR);
    MAP.put(StandardOperation.NOT, StandardOperation.BVNOT);
    MAP.put(StandardOperation.MUL, StandardOperation.BVMUL);
    MAP.put(StandardOperation.MINUS, StandardOperation.BVNEG);
    MAP.put(StandardOperation.REM, StandardOperation.BVSREM);
    MAP.put(StandardOperation.MOD, StandardOperation.BVSMOD);
  }

  /**
   * Checks rather operation is supported.
   *
   * @param operation operation identifier
   *
   * @return {@code true} if operation is supported, {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static boolean hasOperation(final StandardOperation operation) {
    InvariantChecks.checkNotNull(operation);

    return MAP.containsKey(operation);
  }

  /**
   * Gets bit vector analogue operation code of the specified operation code.
   *
   * @param operation operation code
   *
   * @return corresponding operation code
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static StandardOperation getOperation(final StandardOperation operation) {
    InvariantChecks.checkNotNull(operation);

    return MAP.get(operation);
  }
}
