/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGProcess;
import org.zamia.instgraph.IGSequenceOfStatements;
import org.zamia.instgraph.IGSequentialStatement;
import org.zamia.instgraph.IGSequentialWait;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.RangedVariable;

import java.util.List;

/**
 * Utility methods for {@link EventList} object extraction from IG representation.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class EventListFactory {

  private EventListFactory() {
    /* Empty. */
  }

  /**
   * Returns event list of the specified IG process object.
   *
   * <p>Returns {@link EventList} sensitivity list
   * of the specified IG (Instantiation Graph) process object.</p>
   *
   * @param process IG process
   * @return event list
   */
  public static EventList getProcessEventList(final IGProcess process) {

    final EventList result = new EventList();

    final IGSequenceOfStatements sos = process.getSequenceOfStatements();

    final int maxStmtNum = sos.getNumStatements();

    for (int i = 0; i < maxStmtNum; i++) {

      final IGSequentialStatement sosStmt = sos.getStatement(i);

      if (sosStmt instanceof IGSequentialWait) {
        result.addEvents(getEventList((IGSequentialWait) sosStmt).getEvents());
      }
    }
    return result;
  }

  /**
   * Returns event list of the specified IG wait object.
   *
   * <p>Returns {@link EventList} sensitivity list
   * of the specified IG (Instantiation Graph) wait node object.</p>
   *
   * @param wait IG wait node
   * @return event list
   */
  public static EventList getEventList(final IGSequentialWait wait) {

    final EventList list = new EventList();

    final List<IGOperation> sensList = wait.getSensitivityList();

    if (sensList != null) {
      for (final IGOperation current : sensList) {
        list.add(getEvent(current));
      }
      final IGOperation conditionClause = wait.getConditionClause();
      if (conditionClause != null) {
        list.add(getEvent(conditionClause));
      }
    }
    return list;
  }

  private static Event getEvent(final IGOperation current) {

    final RangedVariable rVar = RangedVariableFactory.getRangedVariable(current);
    EventType type = null;

    if (isType(current, EventType.POSITIVE_EDGE)) {
      type = EventType.POSITIVE_EDGE;
    } else if (isType(current, EventType.NEGATIVE_EDGE)) {
      type = EventType.NEGATIVE_EDGE;
    }
    return type == null ? new Event(rVar) : new Event(rVar, type);
  }

  private static boolean isType(
      final IGOperation current,
      final EventType type) {

    return current.toHRString().equalsIgnoreCase(type.name());
  }
}
