/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.ZamiaException;
import org.zamia.instgraph.IGDefaultOperationVisitor;
import org.zamia.instgraph.IGItem;
import org.zamia.instgraph.IGObject;
import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGOperationAlias;
import org.zamia.instgraph.IGOperationAllocate;
import org.zamia.instgraph.IGOperationArrayAggregate;
import org.zamia.instgraph.IGOperationAttribute;
import org.zamia.instgraph.IGOperationBinary;
import org.zamia.instgraph.IGOperationDeref;
import org.zamia.instgraph.IGOperationError;
import org.zamia.instgraph.IGOperationIndex;
import org.zamia.instgraph.IGOperationInvokeSubprogram;
import org.zamia.instgraph.IGOperationLiteral;
import org.zamia.instgraph.IGOperationObject;
import org.zamia.instgraph.IGOperationPhi;
import org.zamia.instgraph.IGOperationRange;
import org.zamia.instgraph.IGOperationRecordAggregate;
import org.zamia.instgraph.IGOperationRecordField;
import org.zamia.instgraph.IGOperationTypeConversion;
import org.zamia.instgraph.IGOperationTypeQualification;
import org.zamia.instgraph.IGOperationUnary;
import org.zamia.instgraph.IGRange;
import org.zamia.instgraph.IGStaticValue;
import org.zamia.instgraph.IGType;
import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.Radix;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.data.types.datamap.DataMap;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.ReduceOptions;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;

/**
 * IG operation objects parser.
 *
 * <p>Component that transforms IG (Instantiation Graph) objects
 * to internal representation.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class IgOperationCfgVisitor extends IGDefaultOperationVisitor {

  /* The current result of object processing. */
  private IgOperationParserResult result;

  /**
   * Default constructor.
   */
  public IgOperationCfgVisitor() {
    super();
  }

  /**
   * Returns result that is formed by object on the current argument.
   *
   * @see IgOperationParserResult
   *
   * @return current result
   */
  public IgOperationParserResult getResult() {
    return this.result;
  }

  @Override
  public void visitOperation(final IGOperation igOperation) {
    super.visitOperation(igOperation);
  }

  @Override
  public void visitStaticValue(final IGStaticValue igStaticValue) {

    NodeValue value = null;
    final IGType type = igStaticValue.getType();

    if (type.isBit()) {

      /* One-bit-size vector value. */
      value = NodeValue.newBitVector(BitVector.valueOf(igStaticValue.isTrue()));
    } else if (type.isBool()) {

      /* Boolean value. */
      value = NodeValue.newBoolean(igStaticValue.isTrue());
    } else if (type.isInteger()) {

      /* Integer value. */
      value = NodeValue.newInteger(igStaticValue.getInt());
    } else if (type.isReal()) {

      /* Real value. */
      value = NodeValue.newReal(igStaticValue.getReal().doubleValue());
    } else if (type.isLogic()) {

      //TODO: implement 9-sign VHDL logic
      value = NodeValue.newBoolean(igStaticValue.getCharLiteral() == '1');
    }

    if (value == null) {
      throw new UnsupportedOperationException(igStaticValue.toHRString());
    }

    this.result = new IgOperationParserResult(value);
  }

  @Override
  public void visitRange(final IGRange igRange) {
    /* Empty. */
  }

  @Override
  public void visitUnary(final IGOperationUnary igUnary) {
    StandardOperation opId;
    switch (igUnary.getUnaryOp()) {
      case ABS:
        opId = StandardOperation.ABS;
        break;
      case NEG:
        opId = StandardOperation.MINUS;
        break;
      case BUF:
        opId = StandardOperation.ADD;
        break;
      case NOT:
        opId = StandardOperation.NOT;
        break;
      default:
        throw new UnsupportedOperationException(igUnary.toHRString());
    }
    final Node operandElem = IgOperationCfgVisitor.getNode(igUnary.getA());
    this.result = new IgOperationParserResult(new NodeOperation(opId, operandElem));
  }

  @Override
  public void visitTypeQualification(final IGOperationTypeQualification igQualification) {
    throw new UnsupportedOperationException(igQualification.toHRString());
  }

  @Override
  public void visitTypeConversion(final IGOperationTypeConversion igConversion) {
    throw new UnsupportedOperationException(igConversion.toHRString());
  }

  @Override
  public void visitRecordField(final IGOperationRecordField igRecordField) {
    throw new UnsupportedOperationException(igRecordField.toHRString());
  }

  @Override
  public void visitRecordAggregate(final IGOperationRecordAggregate igAggregate) {
    throw new UnsupportedOperationException(igAggregate.toHRString());
  }

  @Override
  public void visitOpRange(final IGOperationRange igOperationRange) {

    final Node operand = IgOperationCfgVisitor.getNode(igOperationRange.getOperand());

    InvariantChecks.checkNotNull(operand);

    final IGOperation range = igOperationRange.getRange();

    if (range instanceof IGRange) {
      final IGOperation isAscending = ((IGRange) range).getAscending();

      final Node left = IgOperationCfgVisitor.getNode(((IGRange) range).getLeft());
      final Node right = IgOperationCfgVisitor.getNode(((IGRange) range).getRight());

      final boolean isTrue = ((IGStaticValue)isAscending).isTrue();
      final Node firstIdx = isTrue ? right : left;
      final Node secondIdx = isTrue ? left : right;

      final Node expression =
          new NodeOperation(StandardOperation.BVEXTRACT, firstIdx, secondIdx, operand);

      this.result = new IgOperationParserResult(expression);
    } else {
      throw new UnsupportedOperationException(igOperationRange.toHRString());
    }
  }

  @Override
  public void visitPhi(final IGOperationPhi igPhi) {
    final Node condition = IgOperationCfgVisitor.getNode(igPhi.getOperand(0));
    final Node firstVariant = IgOperationCfgVisitor.getNode(igPhi.getOperand(1));
    final Node secondVariant = IgOperationCfgVisitor.getNode(igPhi.getOperand(2));

    final Node expression =
        new NodeOperation(StandardOperation.ITE, condition, firstVariant, secondVariant);

    this.result = new IgOperationParserResult(expression);
  }

  @Override
  public void visitObject(final IGOperationObject igOperationObject) {

    Node node;
    final IGObject.IGObjectCat category = igOperationObject.getObject().getCat();
    final IGType opType = igOperationObject.getType();

    if (isVariable(category)) {

      /* Object keeps a variable. */

      final DataType dataType = VariableParamFactory.getDataType(opType);

      final VariableType variableType = VariableParamFactory.getVariableType(igOperationObject);

      node = new NodeVariable(igOperationObject.getObject().getId(), dataType);

      /* Set full type meta-info. */
      final VariableData data = new VariableData(variableType);
      final String fullType = VariableParamFactory.getFullType(igOperationObject.getObject());
      data.addMetaInfo(MetaInfoType.HDL_TYPE_STR, fullType);
      node.setUserData(data);

    } else if (category.equals(IGObject.IGObjectCat.CONSTANT)) {

      /* Object keeps a constant. */

      final IGOperation initValue = igOperationObject.getObject().getInitialValue();
      node = IgOperationCfgVisitor.getNode(initValue);
    } else {
      throw new IllegalArgumentException(igOperationObject.toHRString());
    }
    this.result = new IgOperationParserResult(node);
  }

  private static boolean isVariable(final IGObject.IGObjectCat cat) {
    return cat.equals(IGObject.IGObjectCat.SIGNAL) || cat.equals(IGObject.IGObjectCat.VARIABLE);
  }

  @Override
  public void visitLiteral(final IGOperationLiteral igLiteral) {

    NodeValue value;
    try {

      /* Compute static value. */
      final IGStaticValue opStaticValue = igLiteral.computeStaticValue(null);

      switch (opStaticValue.getStaticType().getCat()) {
        case INTEGER:
          value = new NodeValue(Data.newInteger(opStaticValue.getInt()));
          break;
        case REAL:
          value =  new NodeValue(Data.newReal(opStaticValue.getReal().doubleValue()));
          break;
        case ARRAY:

          // TODO: change when 'Z' values will be supported
          final String strVector = substituteUnsupported(opStaticValue.toString());

          value =
              new NodeValue(
                  Data.newBitVector(
                      strVector,
                      Radix.BIN.value(),
                      opStaticValue.toHRString().length()));
          break;
        default:
          throw new UnsupportedOperationException(igLiteral.toHRString());
      }
    } catch (ZamiaException e) {
      throw new RetrascopeRuntimeException(e);
    }
    this.result = new IgOperationParserResult(value);
  }

  @Override
  public void visitIndex(final IGOperationIndex igIndex) {

    final Node index = IgOperationCfgVisitor.getNode(igIndex.getIndex());
    final Node operand = IgOperationCfgVisitor.getNode(igIndex.getOperand());

    Node resultNode;
    if (operand.isType(DataTypeId.BIT_VECTOR)) {

      /* For parametric operations their parameters should go earlier than operands. */
      resultNode = new NodeOperation(StandardOperation.BVEXTRACT, index, index, operand);
    } else if (operand.isType(DataTypeId.MAP)) {
      resultNode = new NodeOperation(StandardOperation.SELECT, operand, index);
    } else {
      throw new IllegalArgumentException(operand.toString());
    }
    this.result = new IgOperationParserResult(resultNode);
  }

  @Override
  public void visitError(final IGOperationError igError) {
    throw new UnsupportedOperationException(igError.toHRString());
  }

  @Override
  public void visitDeref(final IGOperationDeref igDeref) {
    throw new UnsupportedOperationException(igDeref.toHRString());
  }

  @Override
  public void visitAttribute(final IGOperationAttribute igAttribute) {

    if (igAttribute.getAttrOp() == IGOperationAttribute.AttrOp.EVENT) {

      /* Object keeps event-based attribute. */

      final IGItem item = igAttribute.getItem();

      if (item instanceof IGOperation) {
        final RangedVariable var = RangedVariableFactory.getRangedVariable((IGOperation) item);

        final Event event = new Event(var);

        this.result = new IgOperationParserResult(event);
      }
    } else {
      throw new UnsupportedOperationException(igAttribute.toHRString());
    }
  }

  @Override
  public void visitAggregate(final IGOperationArrayAggregate igAggregate) {

    final IGType igType = igAggregate.getType();
    final IGType indexType = igType.getIndexType();
    final IGType elementType = igType.getElementType();
    final DataType keyType = VariableParamFactory.getDataType(indexType);
    final DataType valueType = VariableParamFactory.getDataType(elementType);
    final DataMap dataMap = new DataMap(keyType, valueType);

    /* Skip 'others' element if exists. */
    int offset = 0;
    if (igAggregate.getOperand(0) == null) {
      offset = 1;
    }

    for (int i = offset; i < igAggregate.getNumOperands(); i++) {

      final Node node = IgOperationCfgVisitor.getNode(igAggregate.getOperand(i));
      final Node value = Transformer.reduce(ReduceOptions.NEW_INSTANCE, node);

      if (value instanceof NodeValue) {
        dataMap.put(Data.newInteger(i - 1), ((NodeValue) value).getData());
      } else {
        throw new UnsupportedOperationException(value.toString());
      }
    }
    this.result = new IgOperationParserResult(new NodeValue(Data.newArray(dataMap)));
  }

  @Override
  public void visitAllocate(final IGOperationAllocate igAllocate) {
    throw new UnsupportedOperationException(igAllocate.toHRString());
  }

  @Override
  public void visitAlias(final IGOperationAlias igAlias) {
    throw new UnsupportedOperationException(igAlias.toHRString());
  }

  @Override
  public void visitSubprogram(final IGOperationInvokeSubprogram igSubProgram) {

    final String subProgramName = igSubProgram.getSub().getId();

    final int operandNum = igSubProgram.getNumOperands();

    final String nonQuotedName = getNonQuotedString(subProgramName);
    StandardOperation opCode = IgSubprogramMapping.getOperationCode(nonQuotedName);

    final Node[] nodes = new Node[operandNum];
    for (int i = 0; i < operandNum; i++) {
      final Node node = IgOperationCfgVisitor.getNode(igSubProgram.getOperand(i));
      if (node != null) {
        nodes[i] = node;
      }
    }

    /* HACK: substitute binary operation identifiers to unary if needed. */
    if (nodes.length == 1) {
      if (opCode == StandardOperation.ADD) {
        opCode = StandardOperation.PLUS;
      } else if (opCode == StandardOperation.SUB) {
        opCode = StandardOperation.MINUS;
      }
    }

    /* Substitute operation to bit-vector-analogue if at least one operand is bit vector. */
    boolean hasBvOperand = containsType(DataTypeId.BIT_VECTOR, nodes);

    if (hasBvOperand && BitVectorOpCodeMapping.hasOperation(opCode)) {
      opCode = BitVectorOpCodeMapping.getOperation(opCode);
    }

    this.result = new IgOperationParserResult(new NodeOperation(opCode, nodes));
  }

  private static boolean containsType(final DataTypeId typeId, final Node ... nodes) {
    boolean hasBvOperand = false;

    for (final Node node : nodes) {
      if (node.isType(typeId)) {
        hasBvOperand = true;
        break;
      }
    }
    return hasBvOperand;
  }

  private static String getNonQuotedString(final String opName) {
    final char dQuote = '"';
    final int dqIndex = opName.indexOf(dQuote);
    return dqIndex < 0 ? opName : opName.substring(dqIndex + 1, opName.lastIndexOf(dQuote));
  }

  @Override
  public void visitBinary(final IGOperationBinary igBinary) {
    final Enum<?> type = IgBinaryMapping.getOperationCode(igBinary);

    final Node left = IgOperationCfgVisitor.getNode(igBinary.getA());
    final Node right = IgOperationCfgVisitor.getNode(igBinary.getB());

    this.result = new IgOperationParserResult(new NodeOperation(type, left, right));
  }

  /**
   * Returns syntax node that is extracted from the specified IG object.
   *
   * @param object IG object
   * @return syntax node
   */
  public static Node getNode(final IGOperation object) {
    final IgOperationCfgVisitor visitor = getIgOperationCfgVisitor(object);
    return visitor.getResult().getNode();
  }

  /**
   * Returns data object that is extracted from th specified IG object.
   *
   * @param object IG object
   * @return data object
   */
  public static Object getData(final IGOperation object) {
    final IgOperationCfgVisitor visitor = getIgOperationCfgVisitor(object);
    return visitor.getResult().getData();
  }

  private static IgOperationCfgVisitor getIgOperationCfgVisitor(final IGOperation object) {
    final IgOperationCfgVisitor visitor = new IgOperationCfgVisitor();
    visitor.visitOperation(object);
    return visitor;
  }

  /**
   * Performs the specified object into node variable.
   *
   * @param obj object
   * @return node variable
   */
  public static NodeVariable getVariable(final IGObject obj) {

    final DataType dataType = VariableParamFactory.getDataType(obj.getType());
    final NodeVariable var = new NodeVariable(obj.getId(), dataType);
    final VariableType variableType = VariableParamFactory.getVariableType(obj);
    var.setUserData(new VariableData(variableType));
    return var;
  }

  private String substituteUnsupported(final String vectorValue) {

    /* Substitutes 'Z'-values and '-'-values by zeros. */
    final char[] array = vectorValue.toCharArray();
    for (int i = 0; i < vectorValue.length(); i++) {
      if (array[i] == '-' || array[i] == 'Z') {
        array[i] = '0';
      }
    }
    return new String(array);
  }
}
