/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.SourceLocation;
import org.zamia.ZamiaException;
import org.zamia.instgraph.IGConcurrentStatement;
import org.zamia.instgraph.IGContainer;
import org.zamia.instgraph.IGContainerItem;
import org.zamia.instgraph.IGDefaultSequentialStatementVisitor;
import org.zamia.instgraph.IGInstantiation;
import org.zamia.instgraph.IGItem;
import org.zamia.instgraph.IGManager;
import org.zamia.instgraph.IGMapping;
import org.zamia.instgraph.IGModule;
import org.zamia.instgraph.IGObject;
import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGOperationAttribute;
import org.zamia.instgraph.IGProcess;
import org.zamia.instgraph.IGRange;
import org.zamia.instgraph.IGSequenceOfStatements;
import org.zamia.instgraph.IGSequentialAssert;
import org.zamia.instgraph.IGSequentialAssignment;
import org.zamia.instgraph.IGSequentialExit;
import org.zamia.instgraph.IGSequentialIf;
import org.zamia.instgraph.IGSequentialLoop;
import org.zamia.instgraph.IGSequentialNext;
import org.zamia.instgraph.IGSequentialProcedureCall;
import org.zamia.instgraph.IGSequentialReport;
import org.zamia.instgraph.IGSequentialReturn;
import org.zamia.instgraph.IGSequentialStatement;
import org.zamia.instgraph.IGSequentialWait;
import org.zamia.instgraph.IGStaticValue;
import org.zamia.instgraph.IGStructure;
import org.zamia.instgraph.IGType;
import org.zamia.instgraph.IGTypeStatic;
import org.zamia.instgraph.interpreter.IGInterpreterRuntimeEnv;
import org.zamia.util.PathName;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.MetaInfoValue;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableContainer;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableDeclaration;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.BasicBlock;
import ru.ispras.retrascope.model.cfg.Case;
import ru.ispras.retrascope.model.cfg.CfgModelNode;
import ru.ispras.retrascope.model.cfg.CfgNode;
import ru.ispras.retrascope.model.cfg.CfgNodeType;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.model.cfg.Merge;
import ru.ispras.retrascope.model.cfg.Module;
import ru.ispras.retrascope.model.cfg.Process;
import ru.ispras.retrascope.model.cfg.Sink;
import ru.ispras.retrascope.model.cfg.Source;
import ru.ispras.retrascope.model.cfg.Switch;
import ru.ispras.retrascope.model.cfg.Wait;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * IG structure objects parser.
 *
 * <p>Component of VHDL parser that transforms IG (Instantiation Graph)
 * structure objects representing processes, modules, instances to
 * control flow graph {@link ru.ispras.retrascope.model.cfg.CfgModel} model representation.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class IgStructureCfgVisitor extends IGDefaultSequentialStatementVisitor {

  /* The IG representation manager. */
  private final IGManager manager;

  /* The IG interpreter runtime environment. */
  private final IGInterpreterRuntimeEnv runtimeEnv;

  /* The parent top-level module. */
  private final Module topLevel;

  /* The list of current structure nodes. */
  private final List<CfgNode> nodes;

  /* The current process. */
  private Process process;

  /* Constants that keep severity string values. */
  private static final String SEVERITY_NOTE = "NOTE";
  private static final String SEVERITY_WARN = "WARNING";
  private static final String SEVERITY_ERROR = "ERROR";
  private static final String SEVERITY_FAIL = "FAILURE";

  /**
   * Creates object with the specified IG manager, IG runtime environment and the top-level module.
   *
   * @param manager IG representation manager
   * @param runtimeEnv IG runtime environment
   * @param module top-level module
   *
   * @throws IllegalArgumentException when {@code module} argument is {@code null}.
   */
  public IgStructureCfgVisitor(
      final IGManager manager,
      final IGInterpreterRuntimeEnv runtimeEnv,
      final Module module) {

    super(new IgOperationCfgVisitor());

    InvariantChecks.checkNotNull(module);

    this.manager = manager;
    this.runtimeEnv = runtimeEnv;
    this.topLevel = module;
    this.nodes = new ArrayList<>();
  }

  /**
   * Returns top-level module.
   *
   * @return top-level module
   */
  public Module getTopLevelModule() {
    return this.topLevel;
  }

  /**
   * Returns control flow graph model nodes collection.
   *
   * @return collection of nodes
   */
  private List<CfgNode> getNodes() {
    return this.nodes;
  }

  private Process getProcess() {
    return this.process;
  }

  private void setProcess(final Process process) {
    this.process = process;
  }

  @Override
  public void visit(
      final IGStructure igStructure,
      final PathName pathName) {

    final Map<NodeVariable, VariableDeclaration> declarations =
        getDeclarations(igStructure.getContainer());

    for (int i = 0; i < igStructure.getNumStatements(); i++) {
      final IGConcurrentStatement stmt = igStructure.getStatement(i);

      if (stmt instanceof IGInstantiation) {
        visitInstantiation((IGInstantiation) stmt, pathName);
      } else if (stmt instanceof IGProcess) {
        visitProcess((IGProcess) stmt);
      } else if (stmt instanceof IGStructure) {
        visit((IGStructure) stmt, pathName);
      }
    }
    getTopLevelModule().declareVariables(declarations);
  }

  private void visitProcess(final IGProcess igProcess) {

    final Process newProcess = new Process();
    setProcess(newProcess);

    /* Get variable declarations. */
    final Map<NodeVariable, VariableDeclaration> declarations =
        getDeclarations(igProcess.getContainer());
    newProcess.declareVariables(declarations);

    /* Extract process sensitivity list events. */
    final EventList sensList = EventListFactory.getProcessEventList(igProcess);
    newProcess.setSensitivityList(sensList);

    /* Elaborate internal nodes. */
    final IGSequenceOfStatements sos = igProcess.getSequenceOfStatements();
    visit(sos);

    final List<CfgNode> processNodes = new ArrayList<>(this.nodes);
    newProcess.addNodes(processNodes);

    /* Link process node with parent & child nodes. */
    newProcess.addChild(CfgUtils.getSource(processNodes));
    getTopLevelModule().addChild(newProcess);

    this.nodes.clear();
    setProcess(null);
  }

  private void visitInstantiation(
      final IGInstantiation instantiation,
      final PathName pathName) {

    /* Extract the module name. */
    final IGModule igModule = this.manager.findModule(instantiation.getSignature());
    final String moduleName = igModule.getDUUID().toCompactString();

    /* Extract module. */
    final IgStructureCfgVisitor visitor =
        new IgStructureCfgVisitor(this.manager, this.runtimeEnv, new Module(moduleName));

    visitor.visit(igModule.getStructure(), pathName);

    final Module module = visitor.getTopLevelModule();

    /* Extract module's variables declarations. */
    final Map<NodeVariable, VariableDeclaration> declarations =
        getDeclarations(igModule.getContainer());
    module.declareVariables(declarations);

    /* Instantiate the extracted module. */
    final Map<String, Node> bindings = getBindings(instantiation);
    final String instanceName = instantiation.getLabel();
    module.instantiate(instanceName, bindings);

    getTopLevelModule().addChild(module);
  }

  private Map<NodeVariable, VariableDeclaration> getDeclarations(final IGContainer igContainer) {

    final Map<NodeVariable, VariableDeclaration> map = new LinkedHashMap<>();

    for (int i = 0; i < igContainer.getNumLocalItems(); i++) {

      final IGContainerItem item = igContainer.getLocalItem(i);

      /* Elaborate variables (not-a-constants). */
      if (item instanceof IGObject && !isConstant((IGObject) item)) {

        final NodeVariable variable = RangedVariableFactory.getObjectVariable((IGObject) item);
        final Node initValue = getInitialValue((IGObject) item);

        final Node invariant =
            VariableParamFactory.getInvariant(variable, ((IGObject) item).getType());

        map.put(variable, new VariableDeclaration(initValue, invariant));
      }
    }

    return map;
  }

  private Node getInitialValue(final IGObject igObject) {
    Node initValue;

    final IGOperation igValue = igObject.getInitialValue();

    /* Try to calculate LEFT expression since it is initial for VHDL variable. */
    if (igValue == null) {
      try {

        final IGType type = igObject.getType();
        final IGTypeStatic staticType = type.computeStaticType(this.runtimeEnv, null, null);
        final SourceLocation sourceLocation = igObject.computeSourceLocation();
        final IGStaticValue leftValue = IGStaticValue.generateZ(staticType, sourceLocation);
        initValue = IgOperationCfgVisitor.getNode(leftValue);

      } catch (ZamiaException e) {
        throw new RetrascopeRuntimeException(e);
      }
    } else {
      initValue = IgOperationCfgVisitor.getNode(igValue);
    }

    if (initValue == null) {
      throw new IllegalStateException("Can't calculate initial value of: " + igObject.toString());
    }
    return initValue;
  }

  private boolean isConstant(final IGObject item) {
    return item.getCat().equals(IGObject.IGObjectCat.CONSTANT);
  }

  private Map<String, Node> getBindings(final IGInstantiation igInst) {

    final Map<String, Node> bindings = new LinkedHashMap<>();

    final int maxNum = igInst.getNumMappings();

    for (int i = 0; i < maxNum; i++) {
      final IGMapping mapping = igInst.getMapping(i);

      /* Extract the binding's target name. */
      final IGOperation formal = mapping.getFormal();

      final String varName = formal.getId();

      /* Extract the binding's value expression. */
      final Node expression = IgOperationCfgVisitor.getNode(mapping.getActual());

      bindings.put(varName, expression);
    }
    return bindings;
  }

  @Override
  public void visitWait(final IGSequentialWait igWait) {

    final IGOperation clause = igWait.getConditionClause();
    final Node condition = (clause != null) ? IgOperationCfgVisitor.getNode(clause) : null;

    final EventList list = EventListFactory.getEventList(igWait);
    final Wait waitNode = (condition != null) ? new Wait(list, condition) : new Wait(list);

    addNodesToThis(new ArrayList<>(Collections.singletonList((CfgNode) waitNode)));
  }

  @Override
  public void visitReturn(final IGSequentialReturn igReturn) {
    throw new UnsupportedOperationException(igReturn.toHRString());
  }

  @Override
  public void visitReport(final IGSequentialReport igReport) {
    throw new UnsupportedOperationException(igReport.toHRString());
  }

  @Override
  public void visitCall(final IGSequentialProcedureCall igCall) {
    throw new UnsupportedOperationException(igCall.toHRString());
  }

  @Override
  public void visitNext(final IGSequentialNext igNext) {
    throw new UnsupportedOperationException(igNext.toHRString());
  }

  @Override
  public void visitLoop(final IGSequentialLoop igLoop) {

    /* IG loop node children. */
    final IGRange loopRange = (IGRange)igLoop.getChild(0);
    final IGItem loopCounter = igLoop.getChild(1);
    final IGSequenceOfStatements loopBody = (IGSequenceOfStatements) igLoop.getChild(4);

    /* The range iteration variable. */
    final NodeVariable rangeVariable = IgOperationCfgVisitor.getVariable((IGObject) loopCounter);

    /* The loop condition. */
    final Node left = IgOperationCfgVisitor.getNode(loopRange.getLeft());
    final Node right = IgOperationCfgVisitor.getNode(loopRange.getRight());
    final Node ascending = IgOperationCfgVisitor.getNode(loopRange.getAscending());
    final boolean isAscending = ascending.equals(NodeValue.newBoolean(true));

    final Node range = createRange(rangeVariable, left, right, isAscending);

    /* Create the loop sub-graph. */
    final List<CfgNode> loopCfg = new ArrayList<>();
    final Source loopCfgSrc = new Source();
    final Sink loopCfgSink = new Sink();
    loopCfg.add(loopCfgSrc);
    loopCfg.add(loopCfgSink);

    /* Create the "i := 0" initialization block. */
    final BasicBlock initBlock = new BasicBlock(rangeVariable, left);
    initBlock.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_INIT);
    loopCfgSrc.addChild(initBlock);

    /* Create the "entry" merge block. */
    final Merge merge = new Merge();
    merge.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_ENTRY);
    initBlock.addChild(merge);

    /* Create the condition switch block. */
    final Switch switchBlock = new Switch(range);
    switchBlock.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_COND);

    final Case trueCase = new Case(NodeValue.newBoolean(true));
    switchBlock.addChild(trueCase);
    switchBlock.addParent(merge);

    final List<CfgNode> subGraph = getSubGraph(loopBody);
    CfgUtils.addAsSource(subGraph, trueCase);

    /* Create the incrementing "i++" basic block. */
    checkRangeParam(rangeVariable);

    final StandardOperation addOp = StandardOperation.ADD;
    final NodeValue one = NodeValue.newInteger(1);
    final NodeOperation incrementExpression = new NodeOperation(addOp, rangeVariable, one);
    final BasicBlock incrementBlock = new BasicBlock(rangeVariable, incrementExpression);
    incrementBlock.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_INC);

    /* Create merge that is a child for increment block. */
    final Merge afterIncrement = new Merge();
    afterIncrement.addMetaInfo(MetaInfoType.LOOP_NODE, MetaInfoValue.LOOP_FIN);
    afterIncrement.addParent(incrementBlock);
    CfgUtils.addAsSink(subGraph, incrementBlock);
    afterIncrement.addChild(merge);

    final Case falseCase = new Case(NodeValue.newBoolean(false));
    switchBlock.addChild(falseCase);

    /* Create merge which children do not belong to this loop. */
    final Merge finalMerge = new Merge();
    falseCase.addChild(finalMerge);
    finalMerge.addChild(loopCfgSink);

    /* Add nodes to loop control flow sub-graph. */
    loopCfg.add(finalMerge);
    loopCfg.add(falseCase);
    loopCfg.add(afterIncrement);
    loopCfg.add(switchBlock);
    loopCfg.add(merge);
    loopCfg.addAll(subGraph);

    addNodesToThis(loopCfg);
  }

  private void checkRangeParam(final NodeVariable rangeVariable) {
    if (!rangeVariable.isType(DataTypeId.LOGIC_INTEGER)) {
      throw new UnsupportedOperationException(
          "Unsupported loop counter type: " + rangeVariable.getDataType() + ".");
    }
  }

  @Override
  public void visitIf(final IGSequentialIf igIf) {

    final List<CfgNode> ifCfg = new ArrayList<>();
    final Source source = new Source();
    final Sink sink = new Sink();
    ifCfg.add(source);
    ifCfg.add(sink);

    final Merge merge = new Merge();

    final IGOperation ifCondition = igIf.getCond();
    final IGSequenceOfStatements elseSos = igIf.getElseSOS();
    final IGSequenceOfStatements thenSos = igIf.getThenSOS();

    /* Transform operands & detect event-dependent part of "x'event && x='0'" condition .*/
    final boolean hasEvents = hasEvents(ifCondition);

    /* Extract switch block. */
    Switch switchBlock;

    if (hasEvents) {

      /* Event-dependent condition has been found. */
      switchBlock = createEventCondition(ifCondition);

      processThenBranch(ifCfg, switchBlock, merge, thenSos);

    } else {

      /* No events are found in condition. */

      final Node condition = IgOperationCfgVisitor.getNode(ifCondition);
      switchBlock = new Switch(condition);

      /* Elaborate empty 'then' branch. */
      if (isEmpty(thenSos)) {
        final Case trueCase = new Case(NodeValue.newBoolean(true));
        switchBlock.addChild(trueCase);
        trueCase.addChild(merge);
        ifCfg.add(trueCase);
      } else {

        /* Elaborate non-empty 'then' branch. */
        processThenBranch(ifCfg, switchBlock, merge, thenSos);
      }

      /* Elaborate 'else' branch. */
      processElseBranch(ifCfg, switchBlock, merge, elseSos);
    }

    /* Add link between SOURCE and top condition
     * and between merge node and SINK.
     */
    source.addChild(switchBlock);
    merge.addChild(sink);

    ifCfg.add(switchBlock);

    addNodesToThis(ifCfg);
  }

  private static boolean hasEvents(IGOperation ifCondition) {

    for (int i = 0; i < ifCondition.getNumOperands(); i++) {
      final IGOperation operand = ifCondition.getOperand(i);
      if (operand instanceof IGOperationAttribute) {
        return true;
      }
    }

    return false;
  }

  private boolean isEmpty(final IGSequenceOfStatements sos) {
    return sos == null || sos.getNumStatements() == 0;
  }

  private Switch createEventCondition(final IGOperation ifCondition) {

    RangedVariable eventVariable = null;
    NodeOperation expression = null;

    final int opNum = ifCondition.getNumOperands();

    /* Extract event variable. */
    if (opNum == 2) {

      /* Condition contains "x'event" expression. */
      for (int i = 0; i < opNum; i++) {

        final IGOperation operand = ifCondition.getOperand(i);

        if (operand instanceof IGOperationAttribute) {

          /* Attribute case. */
          final Event eventData = (Event)IgOperationCfgVisitor.getData(operand);
          eventVariable = eventData.getRangedVariable();
        } else {

          /* Expression case. */
          expression = (NodeOperation) IgOperationCfgVisitor.getNode(operand);
        }
      }
    } else {
      throw new UnsupportedOperationException(ifCondition.toHRString());
    }

    /* Extract event type. */
    EventType type = EventType.ANY_EDGE;

    assert expression != null : "Failed to extract event expression: " + ifCondition.toHRString();

    final Enum<?> opId = expression.getOperationId();

    if (opId == StandardOperation.EQ) {
      final int opCount = expression.getOperandCount();
      for (int i = 0; i < opCount; i++) {
        final Node operand = expression.getOperand(i);
        type = getEventType(ifCondition, eventVariable, operand);
      }
    } else {
      throw new UnsupportedOperationException(ifCondition.toHRString());
    }

    return new Switch(new Event(eventVariable, type));
  }

  private EventType getEventType(
      final IGOperation ifCondition,
      final RangedVariable eventVariable,
      final Node operand) {

    EventType type = null;

    if (operand instanceof NodeValue) {

      if (equalsToBitVector(operand, BitVector.FALSE)) {

        type = EventType.NEGATIVE_EDGE;

      } else if (equalsToBitVector(operand, BitVector.TRUE)) {

        type = EventType.POSITIVE_EDGE;
      }
    } else if (!equalParams(eventVariable, operand)) {
      throw new UnsupportedOperationException(ifCondition.toHRString());
    }

    return type;
  }

  private static boolean equalParams(
      final RangedVariable rangedVariable,
      final Node operand) {

    final NodeVariable variable = rangedVariable.getVariable();
    return operand instanceof NodeVariable && operand.equals(variable);
  }

  private static boolean equalsToBitVector(
      final Node operand,
      final BitVector bitVector) {

    return ExprUtils.isType(DataTypeId.BIT_VECTOR, operand)
        && operand.equals(NodeValue.newBitVector(bitVector));
  }

  @Override
  public void visitExit(final IGSequentialExit igExit) {
    throw new UnsupportedOperationException(igExit.toHRString());
  }

  @Override
  public void visitAssert(final IGSequentialAssert igAssert) {

    final IGOperation assertOp = igAssert.getOp();

    /* Create new variable. */
    final String name = getErrorVariableName(igAssert);
    final NodeVariable variable = new NodeVariable(name, DataType.BIT_VECTOR(1));
    final VariableData data = new VariableData(VariableType.REG);
    data.addMetaInfo(MetaInfoType.HDL_TYPE_STR, "BIT");
    variable.setUserData(data);

    /* Declare new variable. */
    getProcess().declareVariable(variable);

    /* Insert high-level checking to the parent process control flow graph. */
    final List<CfgNode> processNodes = getProcess().getNodes();
    final Source src = CfgUtils.getSource(processNodes);
    final Sink sink = CfgUtils.getSink(processNodes);

    /* Take the source node child. */
    final CfgModelNode srcChild = src.getOnlyChild();
    src.removeChild(srcChild);

    final NodeValue trueValue = NodeValue.newBoolean(true);
    final NodeValue falseValue = NodeValue.newBoolean(false);
    final NodeValue nonInit = variable.getData().hasValue() ? falseValue : trueValue;
    final NodeOperation errorFalse = new NodeOperation(StandardOperation.EQ, variable, falseValue);
    final Switch errorSwitch = new Switch(ExprUtils.getDisjunction(errorFalse, nonInit));

    final Case errorTrueCase = new Case(NodeValue.newBoolean(true));
    errorSwitch.addChild(errorTrueCase);
    errorTrueCase.addChild(srcChild);
    final Case errorFalseCase = new Case(NodeValue.newBoolean(false));
    errorSwitch.addChild(errorFalseCase);
    errorFalseCase.addChild(sink);

    getProcess().addNode(errorSwitch);
    getProcess().addNode(errorTrueCase);
    getProcess().addChild(errorFalseCase);

    /* Transform assert node into sub-graph. */
    final List<CfgNode> assertCfg = new ArrayList<>();
    final Source asSrc = new Source();
    final Sink asSink = new Sink();
    assertCfg.add(asSrc);
    assertCfg.add(asSink);

    final Node opExpr = IgOperationCfgVisitor.getNode(assertOp);
    final Switch opSwitch = new Switch(opExpr);
    final Case falseCase = new Case(true);

    final BasicBlock errorBlock = new BasicBlock(variable, trueValue);
    falseCase.addChild(errorBlock);

    final Case trueCase = new Case(NodeValue.newBoolean(true));

    asSrc.addChild(opSwitch);
    opSwitch.addChild(falseCase);
    opSwitch.addChild(trueCase);
    trueCase.addChild(asSink);
    errorBlock.addChild(asSink);
    assertCfg.add(opSwitch);
    assertCfg.add(falseCase);
    assertCfg.add(trueCase);

    addNodesToThis(assertCfg);
  }

  private String getErrorVariableName(final IGSequentialAssert igAssert) {

    final IGOperation assertOp = igAssert.getOp();

    return VariableContainer.NEW_VAR_PREFIX + getSeverity(igAssert) + '_' + assertOp.toHRString();
  }

  private String getSeverity(final IGSequentialAssert igAssert) {

    final String sevString = igAssert.getSeverity().toHRString();

    String varSeverity;
    if (SEVERITY_NOTE.equalsIgnoreCase(sevString)) {

      varSeverity = SEVERITY_WARN;
    } else if (SEVERITY_WARN.equalsIgnoreCase(sevString)) {

      varSeverity = SEVERITY_WARN;
    } else if (SEVERITY_ERROR.equalsIgnoreCase(sevString)) {

      varSeverity = SEVERITY_ERROR;
    } else if (SEVERITY_FAIL.equalsIgnoreCase(sevString)) {

      varSeverity = SEVERITY_ERROR;
    } else {
      throw new IllegalArgumentException("Severity level: " + sevString + ".");
    }
    return varSeverity;
  }

  @Override
  public void visitAssignment(final IGSequentialAssignment igAssignment) {

    final IGOperation target = igAssignment.getTarget();
    final VariableType targetType = VariableParamFactory.getVariableType(target);

    CfgModelNode node;
    if (targetType != null) {

      final IGOperation value = igAssignment.getValue();
      final Node valueExpr = IgOperationCfgVisitor.getNode(value);
      node = getBasicBlock(target, valueExpr);
    } else {
      throw new IllegalStateException(
          String.format("Can't extract variable type from: %s.", igAssignment.toHRString()));
    }

    addNodesToThis(new ArrayList<>(Collections.singletonList((CfgNode) node)));
  }

  private BasicBlock getBasicBlock(
      final IGOperation target,
      final Node valueExpr) {

    final RangedVariable rangedVariable = RangedVariableFactory.getRangedVariable(target);

    final Assignment assignment = new Assignment(rangedVariable, valueExpr);

    /* If the variable is of signal type, then basic block is concurrent. */
    final boolean isConcurrent = isSignal(rangedVariable);

    return new BasicBlock(assignment, isConcurrent);
  }

  private boolean isSignal(final RangedVariable rangedVariable) {

    final NodeVariable variable = rangedVariable.getVariable();
    final Object userData = variable.getUserData();

    return VariableType.isSignal(((VariableData) userData).getVariableType());
  }

  private List<CfgNode> getSubGraph(final IGSequentialStatement stmt) {
    final IgStructureCfgVisitor visitor =
        new IgStructureCfgVisitor(this.manager, this.runtimeEnv, getTopLevelModule());
    visitor.visit(stmt);
    return visitor.getNodes();
  }

  private void processThenBranch(
      final List<CfgNode> ifCfg,
      final Switch switchBlock,
      final Merge merge,
      final IGSequenceOfStatements sos) {

    final Case thenCase = new Case(NodeValue.newBoolean(true));
    switchBlock.addChild(thenCase);

    final List<CfgNode> thenCfg = getSubGraph(sos);

    CfgUtils.addAsSource(thenCfg, thenCase);
    CfgUtils.addAsSink(thenCfg, merge);

    ifCfg.addAll(thenCfg);
  }

  private void processElseBranch(
      final List<CfgNode> ifCfg,
      final Switch switchBlock,
      final Merge merge,
      final IGSequenceOfStatements sos) {

    Case elseCase;
    final List<CfgNode> elseCfg;

    if (isEmpty(sos)) {

      elseCase = new Case(true);
      elseCase.addChild(merge);

    } else {
      elseCase = new Case(NodeValue.newBoolean(false));
      elseCfg = getSubGraph(sos);

      CfgUtils.addAsSource(elseCfg, elseCase);
      CfgUtils.addAsSink(elseCfg, merge);
      ifCfg.addAll(elseCfg);
    }

    switchBlock.addChild(elseCase);
  }

  private Node createRange(
      final NodeVariable var,
      final Node left,
      final Node right,
      final boolean isAscending) {

    /* Select operation identifiers depending on data types. */
    final boolean areBitVector = ExprUtils.isType(DataTypeId.BIT_VECTOR, left, right);

    final StandardOperation greaterId =
        areBitVector ? StandardOperation.BVUGE : StandardOperation.GREATEREQ;
    final StandardOperation lessId =
        areBitVector ? StandardOperation.BVULE : StandardOperation.LESSEQ;

    Node expr;
    if (isAscending) {

      /* The result is: left <= var <= right. */
      expr =
          ExprUtils.getConjunction(
            new NodeOperation(greaterId, var, left),
            new NodeOperation(lessId, var, right));
    } else {

      /* The result is: left >= var >= right. */
      expr =
          ExprUtils.getConjunction(
              new NodeOperation(lessId, var, left),
              new NodeOperation(greaterId, var, right));
    }

    return expr;
  }

  private void addNodesToThis(final Collection<CfgNode> nodes) {

    if (getNodes().isEmpty()) {
      final Source src = new Source();
      final Sink sink = new Sink();
      src.addChild(sink);
      getNodes().add(src);
      getNodes().add(sink);
    }

    final Sink cfgSink = (Sink) CfgUtils.getNode(getNodes(), CfgNodeType.SINK);

    if (cfgSink == null) {
      throw new IllegalArgumentException(
          String.format(
              "Can't find %s node at the %s argument.",
              CfgNodeType.SINK.name(), getNodes().getClass()));
    }

    final CfgModelNode cfgSinkParent = cfgSink.getOnlyParent();

    if (nodes.size() == 1) {

      final CfgNode node = nodes.iterator().next();

      /*
       * If the extracted basic block node has basic block as parent,
       * and both of them are concurrent, then they will be merged into one node
       * without back-ends calling.
       */
      if (CfgUtils.isType(CfgNodeType.BASIC_BLOCK, cfgSinkParent, node)
          && isConcurrent((BasicBlock)node, (BasicBlock)cfgSinkParent)) {

        ((BasicBlock)cfgSinkParent).addAssignments(((BasicBlock) node).getAssignments());

      } else {
        cfgSink.removeParent(cfgSinkParent);
        cfgSinkParent.addChild(node);
        node.addChild(cfgSink);
        getNodes().add(node);
      }
    } else if (!nodes.isEmpty()) {

      final Source nodesSrc = CfgUtils.getSource(nodes);
      final CfgNode nodesSrcChild = (CfgNode) nodesSrc.getOnlyChild();

      nodes.remove(nodesSrc);
      nodesSrcChild.removeParent(nodesSrc);
      cfgSinkParent.removeChild(cfgSink);
      cfgSinkParent.addChild(nodesSrcChild);
      getNodes().remove(cfgSink);

      getNodes().addAll(nodes);
    } else {
      throw new IllegalStateException("Empty collection.");
    }
  }

  private static boolean isConcurrent(BasicBlock ... blocks) {

    boolean concurrent = blocks[0].isConcurrent();

    for (BasicBlock block : blocks) {
      if (concurrent != block.isConcurrent()) {
        concurrent = false;
        break;
      }
    }

    return concurrent;
  }
}
