/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Mapping between IG subprogram identifiers and standard operation identifiers that are supported.
 * 
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class IgSubprogramMapping {

  private static final Map<String, StandardOperation> MAP;

  private IgSubprogramMapping() {
    /* Empty. */
  }

  static {
    MAP = new LinkedHashMap<>();

    MAP.put("=", StandardOperation.EQ);
    MAP.put("<=", StandardOperation.LESSEQ);
    MAP.put("<", StandardOperation.LESS);
    MAP.put(">=", StandardOperation.GREATEREQ);
    MAP.put(">", StandardOperation.GREATER);
    MAP.put("+", StandardOperation.ADD);
    MAP.put("-", StandardOperation.SUB);
    MAP.put("AND", StandardOperation.AND);
    MAP.put("OR", StandardOperation.OR);
    MAP.put("NOT", StandardOperation.NOT);
    MAP.put("/=", StandardOperation.NOTEQ);
    MAP.put("&", StandardOperation.BVCONCAT);
    MAP.put("XOR", StandardOperation.BVXOR);
    MAP.put("*", StandardOperation.MUL);
    MAP.put("MOD", StandardOperation.MOD);
    MAP.put("/", StandardOperation.DIV);
    MAP.put("**", StandardOperation.POWER);
    MAP.put("NOR", StandardOperation.BVNOR);
    MAP.put("BV_NEGATE", StandardOperation.BVNEG);
  }

  /**
   * Checks rather operation is supported.
   *
   * @param operation operation identifier
   *
   * @return {@code true} if operation is supported, {@code true} otherwise
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static boolean hasOperation(final String operation) {
    InvariantChecks.checkNotNull(operation);

    return MAP.containsKey(operation);
  }

  /**
   * Gets operation code of the operation with the specified name.
   *
   * @param opName operation name
   *
   * @return operation code
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws UnsupportedOperationException when {@code IgSubprogramMapping}
   *         does not contain an operation with the specified string name
   */
  public static StandardOperation getOperationCode(final String opName) {
    InvariantChecks.checkNotNull(opName);

    if (IgSubprogramMapping.hasOperation(opName)) {
      return MAP.get(opName);
    }
    throw new UnsupportedOperationException(opName);
  }
}
