/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.instgraph.IGObject;
import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGOperationIndex;
import org.zamia.instgraph.IGOperationInvokeSubprogram;
import org.zamia.instgraph.IGOperationObject;
import org.zamia.instgraph.IGOperationRange;
import org.zamia.instgraph.IGRange;
import org.zamia.instgraph.IGType;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableData;
import ru.ispras.retrascope.model.basis.VariableType;

/**
 * Utility methods for ranged variable extraction from IG objects.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class RangedVariableFactory {

  private RangedVariableFactory() {
    /* Empty. */
  }

  /**
   * Extracts ranged variable from IGOperation object.
   *
   * @param operation object
   *
   * @return ranged variable
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws UnsupportedOperationException when argument type is unsupported.
   */
  public static RangedVariable getRangedVariable(final IGOperation operation) {
    InvariantChecks.checkNotNull(operation);
    RangedVariable var;

    if (operation instanceof IGOperationInvokeSubprogram) {

      var = getSubprogramVariable((IGOperationInvokeSubprogram) operation);

    } else if (operation instanceof IGOperationObject) {

      var = new RangedVariable(getObjectVariable(((IGOperationObject) operation).getObject()));

    } else if (operation instanceof IGOperationIndex) {

      var = getIndexVariable((IGOperationIndex) operation);

    } else if (operation instanceof IGOperationRange) {

      var = getRangeVariable((IGOperationRange) operation);

    } else {
      throw new UnsupportedOperationException(operation.toHRString());
    }
    return var;
  }

  private static RangedVariable getRangeVariable(final IGOperationRange opRange) {

    final IgOperationCfgVisitor visitor = new IgOperationCfgVisitor();
    visitor.visitOperation(opRange.getOperand());

    final IGOperation igOperand = opRange.getOperand();
    final Node operand = IgOperationCfgVisitor.getNode(igOperand);

    final IGOperation range = opRange.getRange();

    checkParams(operand, range);

    final String name = ((NodeVariable) operand).getName();
    final DataType dataType = ((NodeVariable) operand).getData().getType();

    final IGOperation leftIndex = range.getOperand(0);
    final IGOperation rightIndex = range.getOperand(2);

    final Node left = IgOperationCfgVisitor.getNode(leftIndex);
    final Node right = IgOperationCfgVisitor.getNode(rightIndex);

    final Range operandRange = new Range(left, right);
    final NodeVariable variable = new NodeVariable(name, dataType);
    final VariableType variableType = VariableParamFactory.getVariableType(igOperand);
    final VariableData varData = new VariableData(variableType);
    varData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, opRange.getType().toHRString());
    variable.setUserData(varData);

    return new RangedVariable(variable, operandRange);
  }

  private static void checkParams(
      final Node operand,
      final IGOperation range) {

    if (operand.getKind() != Node.Kind.VARIABLE || !(range instanceof IGRange)) {
      throw new UnsupportedOperationException(range.toHRString());
    }
  }

  private static RangedVariable getIndexVariable(final IGOperationIndex opIdx) {
    final IGOperation igOperand = opIdx.getOperand();
    final Node operand = IgOperationCfgVisitor.getNode(igOperand);

    final VariableType type = VariableParamFactory.getVariableType(opIdx);
    final VariableData operandData = new VariableData(type);
    operandData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, igOperand.getType().toHRString());
    operand.setUserData(operandData);

    final RangedVariable var = new RangedVariable((NodeVariable) operand);
    final Node index = IgOperationCfgVisitor.getNode(opIdx.getIndex());

    if (index.getKind() == Node.Kind.VARIABLE || index.getKind() == Node.Kind.VALUE) {
      var.setRange(new Range(index, index));
    } else {
      throw new UnsupportedOperationException(opIdx.toHRString());
    }

    return var;
  }

  /**
   * Extracts variable from the specified IG object.
   *
   * @param object IG object
   * @return variable
   */
  public static NodeVariable getObjectVariable(final IGObject object) {

    final IGType opType = object.getType();

    final VariableType variableType = VariableParamFactory.getVariableType(object);
    final DataType variableValueType = VariableParamFactory.getDataType(opType);
    final NodeVariable variable = new NodeVariable(object.getId(), variableValueType);
    final VariableData variableData = new VariableData(variableType);
    final String fullType = VariableParamFactory.getFullType(object);
    variableData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, fullType);
    variable.setUserData(variableData);

    return variable;
  }

  private static RangedVariable getSubprogramVariable(final IGOperationInvokeSubprogram sub) {
    RangedVariable var;
    final int opNum = sub.getNumOperands();
    if (opNum == 1) {
      var = getRangedVariable(sub.getOperand(0));
    } else {
      throw new UnsupportedOperationException(sub.toHRString());
    }
    return var;
  }
}
