/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.ZamiaException;
import org.zamia.instgraph.IGContainerItem;
import org.zamia.instgraph.IGObject;
import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGRange;
import org.zamia.instgraph.IGStaticValue;
import org.zamia.instgraph.IGType;
import org.zamia.instgraph.IGTypeStatic;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.VariableType;

import java.math.BigInteger;

/**
 * Utility methods for extraction of variables' parameters from IG objects.
 *
 * <p>Methods of the following variables' parameters extraction are presented:
 * data types, {@link VariableType} variable types, type sizes (for bit vectors etc.),
 * value invariants, etc.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class VariableParamFactory {

  private VariableParamFactory() {
    /* Empty. */
  }

  /**
   * Returns a full string representation of IG object type.
   *
   * @param object instantiation graph object
   * @return string representation of object type including base type and invariant.
   */
  public static String getFullType(final IGObject object) {

    final IGType type = object.getType();

    /* Bit vectors are treated as arrays in IG representation. */
    final boolean isBitVector = type.isArray() && type.getElementType().isBit();
    final String bitVectorTypeId = isBitVector
        ? "bit_vector(" + type.getIndexType().toHRString() + ")"
        : type.getCat().name();

    /* The type representation. */
    final String typeId = (type.getId() != null) ? type.getId() : bitVectorTypeId;

    /* Extract possible range (for integer variables only). */
    final boolean isRangedInteger = type.isInteger() && type.getRange() != null;
    final String range = isRangedInteger ? " range " +  type.getRange().toHRString() : "";

    return typeId + range;
  }

  /**
   * Extracts data type of the specified IG type object.
   *
   * @param type object
   *
   * @return data type
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws UnsupportedOperationException when argument is not of expected category
   *         (INTEGER, REAL, ENUM, ARRAY) or is not of expected sub-category
   *         (ENUM-bit, ENUM-bool, ENUM-logic, ARRAY-bit, ARRAY-array, ARRAY-integer, ARRAY-logic).
   */
  public static DataType getDataType(final IGType type) {
    InvariantChecks.checkNotNull(type);

    DataType dataType;
    switch (type.getCat()) {
      case INTEGER:
        dataType = DataType.INTEGER;
        break;
      case REAL:
        dataType = DataType.REAL;
        break;
      case ENUM:
        dataType = getEnumDataType(type);
        break;
      case ARRAY:
        dataType = getArrayDataType(type);
        break;
      default:
        throw new UnsupportedOperationException(type.toHRString());
    }
    return dataType;
  }

  private static DataType getArrayDataType(final IGType type) {

    DataType dataType;

    final IGType elemType = type.getElementType();
    final IGType indexType = type.getIndexType();

    if (elemType.isBit()) {

      /* Bit vector, calculate it's size. */
      if (indexType instanceof IGTypeStatic) {

        final int size = getTypeSize((IGTypeStatic) indexType);
        dataType = DataType.BIT_VECTOR(size);

      } else {
        throw new UnsupportedOperationException(type.toHRString());
      }

    } else if (elemType.isArray()) {

      dataType = DataType.MAP(DataType.INTEGER, getDataType(elemType));

    } else if (elemType.isInteger()) {

      dataType = DataType.MAP(DataType.INTEGER, DataType.INTEGER);

    } else if (elemType.isLogic()) {

      //TODO: add full support 9-sign VHDL logic
      dataType = DataType.MAP(DataType.INTEGER, DataType.BOOLEAN);

    } else {
      throw new UnsupportedOperationException(type.toHRString());
    }
    return dataType;
  }

  private static DataType getEnumDataType(final IGType type) {

    DataType dataType;

    if (type.isBit()) {

      dataType = DataType.BIT_VECTOR(1);

    } else if (type.isBool()) {

      dataType = DataType.BOOLEAN;

    } else if (type.isLogic()) {

      //TODO: add full support 9-sign VHDL logic
      dataType = DataType.BOOLEAN;

    } else {
      throw new UnsupportedOperationException(type.toHRString());
    }
    return dataType;
  }

  /**
   * Returns size of the specified type.
   *
   * @param indexType index type
   *
   * @return size
   *
   * @throws java.lang.UnsupportedOperationException when index left/right borders
   *         cannot be transformed into {@link ru.ispras.fortress.expression.NodeValue} objects.
   * @throws java.lang.UnsupportedOperationException when index left/right borders
   *         are not of integer types.
   */
  public static int getTypeSize(final IGTypeStatic indexType) {

    final IGStaticValue range = indexType.getStaticRange();
    final Node left = IgOperationCfgVisitor.getNode(range.getLeft());
    final Node right = IgOperationCfgVisitor.getNode(range.getRight());

    checkParams(left, right);

    final int leftValue = ((BigInteger) ((NodeValue) left).getValue()).intValue();
    final int rightValue = ((BigInteger) ((NodeValue) right).getValue()).intValue();

    return Math.abs(leftValue - rightValue + 1);
  }

  private static void checkParams(
      final Node left,
      final Node right) {

    //TODO: substitute by isKind method
    if (left.getKind() != Node.Kind.VALUE || right.getKind() != Node.Kind.VALUE) {
      throw new UnsupportedOperationException(
          String.format("Unsupported types: %s, %s - should be of %s kind.",
              left.getKind(), right.getKind(), Node.Kind.VALUE));
    }
    if (!ExprUtils.isType(DataType.INTEGER, left, right)) {
      throw new UnsupportedOperationException(
          String.format("Unsupported types: %s, %s - should be of %s type.",
              ((NodeValue) left).getData().getType().toString(),
              ((NodeValue) right).getData().getType().toString(),
              DataType.INTEGER));
    }
  }

  /**
   * Extracts variable type of the specified item.
   *
   * @param item item
   *
   * @return {@link VariableType} variable type
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws UnsupportedOperationException when argument is neither
   *         of {@code IGOperation} nor of {@code IGObject} type.
   */
  public static VariableType getVariableType(final IGContainerItem item) {
    InvariantChecks.checkNotNull(item);

    VariableType type;
    if (item instanceof IGOperation) {

      type = getOperationVariableType((IGOperation) item);

    } else if (item instanceof IGObject) {

      type = getObjectVariableType((IGObject) item);

    } else {
      throw new UnsupportedOperationException(item.getClass().toString());
    }

    return type;
  }

  private static VariableType getOperationVariableType(final IGOperation operation) {

    VariableType type;

    try {
      switch (operation.getDirection()) {
        case IN:
        case LINKAGE:
          type = VariableType.IN;
          break;
        case OUT:
          type = VariableType.OUT;
          break;
        case INOUT:
          type = VariableType.INOUT;
          break;
        case NONE:
          type = VariableType.REG;
          break;
        default:
          throw new UnsupportedOperationException(operation.getDirection().name());
      }
    } catch (ZamiaException e) {
      throw new RetrascopeRuntimeException(e);
    }
    return type;
  }

  private static VariableType getObjectVariableType(final IGObject obj) {
    final  IGObject.OIDir direction = obj.getDirection();
    VariableType type;
    switch (direction) {
      case NONE:
        type = VariableType.REG;
        break;
      case IN:
      case LINKAGE:
        type = VariableType.IN;
        break;
      case INOUT:
        type = VariableType.INOUT;
        break;
      case OUT:
        type = VariableType.OUT;
        break;
      default:
        throw new UnsupportedOperationException(obj.toString());
    }
    return type;
  }

  /**
   * Extracts invariant to the specified variable from the specified IGType object.
   *
   * @param variable variable to which invariant is to be extracted
   * @param type type object
   *
   * @return invariant
   *
   * @throws UnsupportedOperationException when argument is ranged
   *         and it's range is neither of {@code IGStaticValue}
   *         nor of {@code IGRange} type.
   */
  public static Node getInvariant(
      final NodeVariable variable,
      final IGType type) {

    Node invariant;

    if (noTypeInvariant(variable, type)) {

      invariant = null;

    } else {

      final IGOperation range = type.getRange();

      if (range instanceof IGStaticValue) {
        invariant = getStaticRangeInvariant(variable, (IGStaticValue) range);
      } else if (range instanceof IGRange) {
        invariant = getVariableRangeInvariant(variable, (IGRange) range);
      } else {
        throw new UnsupportedOperationException("Cannot parse: " + range.getClass() + ".");
      }
    }
    return invariant;
  }

  private static boolean noTypeInvariant(
      final NodeVariable variable,
      final IGType type) {

    /* One-bit-sized vectors or booleans or non-ranged variables
     * do not have any type invariant. */
    return variable.isType(DataType.BIT_VECTOR(1))
        || variable.isType(DataType.BOOLEAN)
        || type.getRange() == null;
  }

  private static Node getVariableRangeInvariant(
      final NodeVariable variable,
      final IGRange range) {

    final Node left = IgOperationCfgVisitor.getNode(range.getLeft());
    final Node right = IgOperationCfgVisitor.getNode(range.getRight());

    final Node ascending = IgOperationCfgVisitor.getNode(range.getAscending());

    final NodeValue trueNode =
        ascending.isType(DataTypeId.BIT_VECTOR)
            ? NodeValue.newBitVector(BitVector.TRUE) : NodeValue.newBoolean(true);

    final NodeOperation checkAscending =
        new NodeOperation(StandardOperation.EQ, ascending, trueNode);

    /* Determine rather ascending node is true or false. */
    final boolean isAscending = ExprUtils.isSAT(checkAscending);

    return getRangeInvariant(variable, left, right, isAscending);
  }

  private static Node getStaticRangeInvariant(
      final NodeVariable variable,
      final IGStaticValue range) {

    /* The left/right borders. */
    final Node left = IgOperationCfgVisitor.getNode(range.getLeft());
    final Node right = IgOperationCfgVisitor.getNode(range.getRight());

    /* The flag that shows rather range is ascending or descending. */
    final boolean isAscending = range.getAscending().isTrue();

    return getRangeInvariant(variable, left, right, isAscending);
  }

  private static Node getRangeInvariant(
      final NodeVariable variable,
      final Node left,
      final Node right,
      final boolean isAscending) {

    /* The identifiers for operation codes. */
    final boolean areBitVector = ExprUtils.isType(DataTypeId.BIT_VECTOR, left, right);
    final StandardOperation greaterId =
        areBitVector ? StandardOperation.BVUGE : StandardOperation.GREATEREQ;
    final StandardOperation lessId =
        areBitVector ? StandardOperation.BVULE : StandardOperation.LESSEQ;

    final Node greater = isAscending ? right : left;
    final Node lesser = isAscending ? left : right;

    return ExprUtils.getConjunction(
        new NodeOperation(greaterId, variable, lesser),
        new NodeOperation(lessId, variable, greater));
  }
}
