/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.conflict;

import java.util.ArrayList;
import java.util.List;

import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlEnumValue;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.engine.efsm.extractor.conflict.jaxb.JaxbEfsmAdapter;
import ru.ispras.retrascope.engine.efsm.extractor.conflict.jaxb.JaxbGuardedActionAdapter;
import ru.ispras.retrascope.engine.efsm.extractor.conflict.jaxb.JaxbNodeVariableAdapter;
import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.efsm.GuardedAction;

/**
 * Represents the conflict of two Models.
 * 
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
@XmlRootElement(namespace = "ru.ispras.retrascope.result.conflict")
@XmlAccessorType(XmlAccessType.FIELD)
public class Conflict {

  @XmlJavaTypeAdapter(JaxbEfsmAdapter.class)
  private Model model1;
  @XmlJavaTypeAdapter(JaxbEfsmAdapter.class)
  private Model model2;
  @XmlJavaTypeAdapter(JaxbNodeVariableAdapter.class)
  private NodeVariable conflictVariable;
  @XmlElement
  private Types conflictType;
  @XmlElement(name = "Path")
  @XmlJavaTypeAdapter(JaxbGuardedActionAdapter.class)
  private List<GuardedAction> path;

  
  
  public static enum Types {
    /**
     * Read after Write
     */
    @XmlEnumValue("RAW")
    RAW,
    /**
     * Write after Read
     */
    @XmlEnumValue("WAR")
    WAR,
    /**
     * Write after Write
     */
    @XmlEnumValue("WAW")
    WAW, @XmlEnumValue("UNDEFINED")
    UNDEFINED
  }

  /**
   * Constructs EfsmConflict using the specified values.
   * 
   * @param model1 the first {@link Model};
   * @param model2 the second {@link Model};
   * @param conflictVariable conflicting {@link NodeVariable};
   * @param conflictType conflict type ({@link Types}).
   */
  public Conflict(Model model1, Model model2, NodeVariable conflictVariable, Types conflictType,
      List<GuardedAction> path) {
    this.model1 = model1;
    this.model2 = model2;
    this.setVariable(conflictVariable);
    this.setType(conflictType);
    this.setPath(path);
  }

  public Conflict() {

  }

  /**
   * Returns the path from the conflict state to the initial state.
   * @return The path as a list of guarded actions.
   */
  public List<GuardedAction> getPath() {
    return path;
  }

  public void setPath(List<GuardedAction> path) {
    this.path = path;
  }

  /**
   * Returns the pair of models.
   * @return A {@link List} containing the pair of Models.
   */

  public List<Model> getModelPair() {
    List<Model> result = new ArrayList<>(2);
    result.add(model1);
    result.add(model2);
    return result;
  }

  /**
   * Sets a pair of models.
   * @param model1 the first Model;
   * @param model2 the second Model.
   */
  public void setModelPair(Model model1, Model model2) {
    this.model1 = model1;
    this.model2 = model2;
  }

  /**
   * Returns the conflicting variable.
   * @return A conflicting {@link NodeVariable}
   */
  public NodeVariable getVariable() {
    return conflictVariable;
  }

  /**
   * Sets a new conflicting variable.
   * @param conflictVariable - the specified {@link NodeVariable}.
   */
  public void setVariable(NodeVariable conflictVariable) {
    this.conflictVariable = conflictVariable;
  }

  /**
   * Returns the type of conflict.
   * @return the conflictType
   */
  public Types getType() {
    return conflictType;
  }

  /**
   * Sets the type of a conflict.
   * @param conflictType - the specified {@link Types}.
   */
  public void setType(Types conflictType) {
    this.conflictType = conflictType;
  }

  @Override
  public String toString() {
    StringBuilder result = new StringBuilder();
    result.append("} {Variable: ");
    result.append(conflictVariable.toString());
    result.append("} {Type: ");
    result.append(conflictType.toString());
    result.append("} ]");
    return result.toString();

  }

}
