/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.test;

import java.util.HashMap;
import java.util.Map;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;
import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.retrascope.result.test.jaxb.JaxbVariableMapXmlAdapter;

/**
 * The class representing a transaction by a set of mappings between names and values of variables.
 * Typically, all variables of a transaction relate to the same interface of a device.
 * <br>
 * Two types of variable values are permitted in transactions. A first one is a concrete value
 * represented by the {@link NodeValue} class. A second one is a constraint-based description
 * represented by the {@link NodeOperation} class.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see NodeValue
 * @see NodeOperation
 */
@XmlType(name = "transaction")
@XmlRootElement
public class Transaction {
  private static final String LEFT_STRING_BOUND = "{";
  private static final String VARIABLE_STRING_DELIMITER = "; ";
  private static final String VARIABLE_NAME_VALUE_STRING_DELIMITER = " = ";
  private static final String RIGHT_STRING_BOUND = "}";

  @XmlElement
  @XmlJavaTypeAdapter(JaxbVariableMapXmlAdapter.class)
  private Map<String, Node> variables;

  /**
   * Constructs an empty transaction.
   */
  public Transaction() {
    this.variables = new HashMap<>();
  }

  /**
   * Constructs a transaction containing the specified name-value pairs.
   * 
   * @param values  a map containing mappings between variables' names and values
   */
  public Transaction(Map<String, ? extends Node> values) {
    this.variables = new HashMap<>(values);
  }

  /**
   * Constructs an empty transaction with the specified initial capacity of an underlying collection
   * of variable name-value pairs.
   * 
   * @param size    the initial capacity
   */
  public Transaction(int size) {
    this.variables = new HashMap<>(size);
  }

  /**
   * Adds a variable with the specified name and value to this transaction.
   * 
   * @param name    the name of the variable to be added
   * @param value   the value of the variable to be added
   */
  public void addVariable(String name, NodeValue value) {
    this.variables.put(name, value);
  }

  /**
   * Adds a variable with the specified name and value to this transaction.
   * 
   * @param name    the name of the variable to be added
   * @param value   the expression determining value of the variable to be added
   */
  public void addVariable(String name, NodeOperation value) {
    this.variables.put(name, value);
  }

  /**
   * Returns a value of the specified variable from this transaction.
   * 
   * @param variableName    a name of the variable
   * 
   * @return the variable's value if the specified variable is in this transaction; otherwise
   * returns {@code null}
   */
  public Node getValue(String variableName) {
    return this.variables.get(variableName);
  }

  /**
   * Returns a map containing mappings between names and values for all variables of this transaction.
   * 
   * @return the map representing this transaction
   */
  public Map<String, Node> getValues() {
    return new HashMap<>(this.variables);
  }

  /**
   * Checks if the specified variable is in this transaction.
   * 
   * @param name    a name of the variable to be checked
   * 
   * @return {@code true} if this transaction contains the specified variable. Otherwise returns
   * {@code false}.
   */
  public boolean containsVariable(String name) {
    return this.variables.containsKey(name);
  }

  /**
   * Checks if this transaction is empty.
   * 
   * @return {@code true} if this transaction contains no variables. Otherwise returns {@code false}
   */
  public boolean isEmpty() {
    return this.variables.isEmpty();
  }

  @Override
  public int hashCode() {
    final int prime = 31;

    return prime + this.variables.hashCode();
  }

  /**
   * Compares the specified object with this transaction for equality. The object is considered as
   * equal to this transaction if and only if it is also a transaction and it contains the same
   * variables with the same values as this transaction does, or more formally:
   * <ul>
   *    <li>it is an instance of the {@code Transaction} class or its subclass;
   *    <li>it contains the same amount of variables as this transaction does;
   *    <li>for each its variable there is a one with equal name and value in this transaction.
   * </ul>
   * 
   * @param obj the object to be compared for equality with this transaction
   * 
   * @return {@code true} if the specified object is equal to this transaction. Otherwise returns
   * {@code false}
   */
  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }

    if ((obj == null) || !(obj instanceof Transaction)) {
      return false;
    }

    Transaction transactionToCompare = (Transaction) obj;

    return this.variables.equals(transactionToCompare.variables);
  }

  @Override
  public String toString() {
    StringBuilder stringBuilder = new StringBuilder(Transaction.LEFT_STRING_BOUND);

    int initialStringBuilderLength = stringBuilder.length();

    for (String variableName : this.variables.keySet()) {
      if (stringBuilder.length() > initialStringBuilderLength) {
        stringBuilder.append(Transaction.VARIABLE_STRING_DELIMITER);
      }

      stringBuilder.append(variableName);
      stringBuilder.append(Transaction.VARIABLE_NAME_VALUE_STRING_DELIMITER);

      Node variableData = this.variables.get(variableName);

      if (variableData instanceof NodeValue) {
        stringBuilder.append(((NodeValue) this.variables.get(variableName)).getValue());
      } else {
        stringBuilder.append(variableData);
      }
    }

    stringBuilder.append(Transaction.RIGHT_STRING_BOUND);

    return stringBuilder.toString();
  }
}
