/*
 * Copyright 2014 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import java.util.logging.Level;

/**
 * The class containing Retrascope logging levels.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class LogLevel extends Level {
  private static final long serialVersionUID = 8571857648413626924L;
  /**
   * The logging level indicating a serious failure. Has the same priority as Level.SEVERE
   */
  public static final Level ERROR = new LogLevel("ERROR", 1000);
  /**
   * The logging level indicating a potential problem. Has the same priority as Level.WARNING
   */
  public static final Level WARNING = new LogLevel("WARNING", 900);
  /**
   * The logging level for informational messages. Has the same priority as Level.INFO
   */
  public static final Level INFO = new LogLevel("INFO", 800);
  /**
   * The logging level for information for debugging. Has the same priority as Level.FINE
   */
  public static final Level DEBUG = new LogLevel("DEBUG", 500);

  /**
   * Create a named logging level with a given integer value.
   *
   * <p>
   * Note that this constructor is "protected" to allow subclassing. In general clients of logging
   * should use one of the constant {@code LogLevel} objects such as {@code INFO} or {@code DEBUG}.
   * However, if clients need to add new logging levels, they may subclass {@code LogLevel} and
   * define new constants.
   *
   * @param name the name of the level, for example {@code NEW_LEVEL}
   * @param value an integer value for the level
   *
   * @throws NullPointerException if the name is {@code null}
   */
  protected LogLevel(String name, int value) {
    super(name, value);
  }

  /**
   * Parse a level name string into a Level.
   *
   * <p>
   * The argument string may consist of either a level name
   * or an integer value.
   *
   * <p>
   * For example:
   * <ul>
   * <li>{@code ERROR};
   * <li>{@code 1000};
   * </ul>
   *
   * @param name string to be parsed
   *
   * @return the parsed value. Passing an integer that corresponds to a known name (e.g., 700) will
   *         return the associated name (e.g., {@code CONFIG}). Passing an integer that does not
   *         (e.g., 1) will return a new level name initialised to that value
   *
   * @throws NullPointerException if the name is {@code null}
   * @throws IllegalArgumentException if the value is not valid. Valid values are integers between
   *         {@code Integer.MIN_VALUE} and {@code Integer.MAX_VALUE}, and all known level names.
   *         Known names are:
   * <ul>
   *   <li>the levels defined by this class (for example, {@code ERROR} and {@code WARNING});
   *   <li>the levels defined by the {@link Level} class (for example, {@code SEVERE} and
   *   <li>{@code FINE}); levels created by these classes with appropriate package access;
   *   <li>new levels defined or created by their subclasses.
   * </ul>
   *
   */
  public static synchronized Level parse(String name) throws IllegalArgumentException {
    return Level.parse(name);
  }
}
